/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */



/*------------------------------------------------------------------------*/
/*--									--*/
/*--	ADM_Main.c							--*/
/*--									--*/
/*--	Main function of administrator program.				--*/
/*------------------------------------------------------------------------*/

#include "ITTC.h"

/*------------------------------------------------------------------------*/
/*--	ADM_Init()							--*/
/*------------------------------------------------------------------------*/

int CheckResponse(ADM_CTX *ctx, int *problem);

ADM_CTX *ADM_Init(int argc, char *argv[])
{
  int err;
  int i;
  int timeout;
  ADM_CTX *ctx;

  if ((ctx = malloc(sizeof(ADM_CTX))) == NULL) {
    fprintf(stderr, "ADM_Init: out of memory.\n");
    return NULL;
  }

  ctx->streams = NULL;
  ctx->comctx = NULL;
  ctx->utlctx = NULL;
  ctx->keysize = NULL;
  ctx->keyname = NULL;
  ctx->password = NULL;
  ctx->ConfigFile = NULL;

  /*--  Parse command line arguments  --*/

  err = ADM_ParseCmdLine(ctx, argc, argv);

  if (err) {
    free(ctx);
    return NULL;
  }

  /*--  Load config file  --*/

  ctx->utlctx = UTL_init_ctx();

  if (ctx->utlctx == NULL) {
    fprintf(stderr, "ADM_Init: UTL_init failed.\n");
    free(ctx);
    return NULL;
  }

  err = UTL_LoadConfigFile(ctx->utlctx, ctx->ConfigFile);

  if (err) {
    free(ctx);
    return NULL;
  }

  /*--  Create log file  --*/

  err = UTL_OpenLogFile(ctx->utlctx, COM_TYPE_ADMIN, ctx->me);

  if (err) {
    UTL_CloseLogFile();
    free(ctx);
    return NULL;
  }

  /*--  Initialize communications package  --*/

  ctx->streams = (COM_STREAM**)malloc((ctx->numServers)*sizeof(COM_STREAM*));
  if (ctx->streams == NULL) {
    UTL_CloseLogFile();
    free(ctx);
    return NULL;
  }
  for (i=0; i<ctx->numServers; i++) ctx->streams[i] = NULL;

  ctx->comctx = COM_init(ctx->utlctx, COM_TYPE_ADMIN, ctx->me);

  if (ctx->comctx == NULL) {
    fprintf(stderr, "ADM_Init: COM_init failed.\n");
    UTL_CloseLogFile();
    free(ctx->streams);
    free(ctx);
    return NULL;
  }

  timeout = UTL_GetIntConfigParam(ctx->utlctx, "Client_Timeout", 10000);
  COM_ctx_set_timeout(ctx->comctx, timeout);

  return ctx;
}


/*------------------------------------------------------------------------*/
/*--                                                                    --*/
/*--	ADM_SendHeader()						--*/
/*--                                                                    --*/
/*--    Send the header and opcode to all servers.                      --*/
/*--                                                                    --*/
/*------------------------------------------------------------------------*/

int ADM_SendHeader(ADM_CTX *ctx) {
  int i; 
  int numServers;


  /*----  Initiate connection to all share servers  ----*/

  numServers = ctx->numServers;

  for (i=0; i<numServers; i++) {
    if ( (ctx->server < 0) || (ctx->server == i) ) {

      if (ctx->streams[i]==NULL) {
	UTL_Log("ADM_SendHeader", LOG_TEST, "Connecting to server %d", i);
	ctx->streams[i] = COM_connect(ctx->comctx, COM_TYPE_SHARE_SERVER, i, ITTC_SHARE_SERVER_VPORT);
	if (ctx->streams[i] == NULL) {
	  UTL_Log("ADM_SendHeader", LOG_FATAL, "Error connecting to server %d", i);
	  return RC_ERROR;
	}
      } else {
	UTL_Log("ADM_SendHeader", LOG_FATAL, "Already connected to server %d", i);
      }

    } else {

      UTL_Log("ADM_SendHeader", LOG_TEST, "Skipping server %d", i);

    }      
  }


  /*-- Send headers --*/

  for(i=0; i< numServers; ++i) {

    if (ctx->streams[i]) {

      if (COM_write_string(ctx->streams[i], "COMMAND") < 0) {
	UTL_Log("ADM_SendHeader", LOG_ERROR, "Unable to send header \"COMMAND\".");
	return RC_ERROR;
      }

      if (COM_write_int(ctx->streams[i], ctx->opcode) < 0) {
	UTL_Log("ADM_SendHeader", LOG_FATAL, "Unable to send opcode to server %d", i);
	return RC_ERROR;
      }

      if (ctx->opcode == ADM_op_start) {       // Send password on a start command
	if (COM_write_string(ctx->streams[i], ctx->password) < 0) {
	  UTL_Log("ADM_SendHeader", LOG_FATAL, "Unable to send password to server %d", i);
	  return RC_ERROR;
	}
      }


    }
  }

  return RC_OK;
}


/*------------------------------------------------------------------------*/
/*--                                                                    --*/
/*--	ADM_Work()							--*/
/*--                                                                    --*/
/*------------------------------------------------------------------------*/

/* Note:

   When the ADM is told to refresh a key, ADM actually calls recover,
   then refresh, then recover again.  The way refresh and recover are
   designed gives this process several limitations.  Specifically, it
   is only guarenteed to refresh the key under specific conditions.

   When this sequence starts, it first finds the lowest version of the 
   key for which the threshold is met.  It then removes all other versions
   of the key.  Then it performs the refresh on this version of the key.

   Problems can occur when this is the incorrect version to use.  For 
   example, if the public shared key corresponds to a different version 
   than this lowest threshold version, and some keys are corrupted,
   the ADM will run validate to see which servers are bad, and will think
   that servers are bad simply because they have a different version 
   than the public shared key.

   Specifically, if there is a crash while the servers are performing
   a version_set, refresh might not be able to recover the key.  This 
   should be fixed in the next release. 
 
*/


int ADM_Work(ADM_CTX *ctx) {
  int err;
  int count;
  int maxCount;

  /*----  Do operation  ----*/

  switch(ctx->opcode) {
    case ADM_op_start:
      err = ADM_control(ctx);
      break;
    case ADM_op_suspend:
      err = ADM_control(ctx);
      break;
    case ADM_op_shutdown:
      err = ADM_control(ctx);
      break;
    case ADM_op_corrupt:
      err = ADM_control(ctx);
      break;
    case ADM_op_genkey:
      err = ADM_GenKey(ctx);
      break;
    case ADM_op_refresh:

      /* 1. Do a recover */
      err = RC_ERROR;    /* Make it perform at least one recover */
      maxCount = UTL_GetIntConfigParam(ctx->utlctx, "Max_Recover", 5);
      for (count = 0; (err != RC_OK) && (count <= maxCount); count++)
	err = ADM_recover(ctx, ADM_LOW);

      if (err != RC_OK) {
	UTL_Log("ADM_Work", LOG_FATAL, "Unable to synchronize keys before refresh");
	fprintf(stderr, "Unable to synchronize keys before refresh\n");
      }

      /* 2. Do a refresh */
      err = ADM_refresh(ctx);
      if (err == RC_NO_KEY) {
	UTL_Log("ADM_Work", LOG_FATAL, "The key can not be loaded to perform the refresh");
	fprintf(stderr, "The key can not be loaded\n");
	break;
      }

      /* 3. Do a recover */
      err = RC_ERROR;    /* Make it perform at least one recover */
      for (count = 0; (err != RC_OK) && (count <= maxCount); count++)
	err = ADM_recover(ctx, ADM_HIGH);
      break;
    default:
      err = RC_ERROR;
  }

  return err;
}


/*------------------------------------------------------------------------*/
/*--	ADM_Shutdown()							--*/
/*------------------------------------------------------------------------*/

int ADM_Shutdown(ADM_CTX *ctx)
{
  COM_ctx_close(ctx->comctx);
  free(ctx->streams);
  UTL_free_ctx(ctx->utlctx);
  UTL_CloseLogFile();

  if (ctx != NULL)
    free(ctx);

  return RC_OK;
}


/*------------------------------------------------------------------------*/
/*--	main()								--*/
/*------------------------------------------------------------------------*/

int main(int argc, char *argv[])
{
  int err;
  ADM_CTX *ctx;

  printf("\nPerforming ITTC administration\n");

  /* Initialize */
  ITTC_init();

  /* Process command line arguments and initialize context */
  ctx = ADM_Init(argc, argv);
  if (ctx == NULL) exit(1);

  /* Main controller loop */
  err = ADM_Work(ctx);

  /* Close files and shutdown */
  ADM_Shutdown(ctx);

  if (err == RC_OK)
    printf("Done with ITTC administration\n\n");
  else
    printf("Error during ITTC administration\n\n");

  return err;
}
