/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */


#include "ITTC.h"


/*------------------------------------------------------------------------*/
/*--									--*/
/*--	ADM_do.c							--*/
/*--									--*/
/*--	Do required operation.						--*/
/*------------------------------------------------------------------------*/


int ADM_CheckResponse(ADM_CTX *ctx, char *msg)
{
  int i, num, err;
  COM_STREAM **streams = NULL;
  COM_STREAM *tStream = NULL;
  void *vt = NULL;
  char data[256];
  int count = 0;


  streams = (COM_STREAM**)malloc(sizeof(COM_STREAM*)*ctx->numServers);
  for (i=0; i<ctx->numServers; ++i) {
    streams[i] = ctx->streams[i];
    if (streams[i] != NULL) count++;
  }

  for(i=0; i<count; ++i) {

    err = COM_select_v(ctx->comctx, &vt, streams, ctx->numServers, 
		       COM_SELECT_READ | COM_SELECT_TIMEOUT | COM_SELECT_CLOSE);

    if (err == COM_SELECTED_TIMEOUT) {
      UTL_Log("ADM_CheckResponse", LOG_NOTIFY, "Timeout.");
      free(streams);
      return RC_TIMEOUT;
    }

    if (err == COM_SELECTED_CLOSE) {
      UTL_Log("ADM_CheckResponse", LOG_NOTIFY, "Stream closed.");
      free(streams);
      return RC_CLOSE;
    }

    if (err != COM_SELECTED_READ) {
      UTL_Log("ADM_CheckResponse", LOG_FATAL, "Error in select.");
      free(streams);
      return RC_ERROR;
    }

    tStream = (COM_STREAM*)vt;

    /* Get the peer's ID number */
    if ((num = UTL_StreamID(tStream)) < 0) {
      UTL_Log("ADM_CheckResponse", LOG_FATAL, "Error getting peer's ID.");
      free(streams);
      return RC_ERROR;
    }

    /* Remember not to select on this stream again */
    streams[num] = NULL;
    
    /* Read the response from the stream */
    if (COM_read_string(tStream, data) < 0) {
      UTL_Log("ADM_CheckResponse", LOG_FATAL, "Error reading response.");
      free(streams);
      return RC_ERROR;
    }
    
    if ( (strcmp(data, msg) != 0) && (strcmp(msg, "OK") == 0) ) {
      UTL_Log("ADM_CheckResponse", LOG_FATAL, "Begin code not OK:");
      UTL_Log("ADM_CheckResponse", LOG_FATAL, "%s", msg);
    }

  }
  
  free(streams);
  return RC_OK;
}


int ADM_CheckOpcode(ADM_CTX *ctx, int opcode)
{
  int i, num, err;
  COM_STREAM **streams= NULL, *tStream = NULL;
  void *vt = NULL;
  int data;
  int count = 0;


  streams = (COM_STREAM**)malloc(sizeof(COM_STREAM*)*ctx->numServers);
  for (i=0; i<ctx->numServers; ++i) {
    streams[i] = ctx->streams[i];
    if (streams[i] != NULL) count++;
  }

  for(i=0; i<count; ++i) {

    err = COM_select_v(ctx->comctx, &vt, streams, ctx->numServers, 
		       COM_SELECT_READ | COM_SELECT_TIMEOUT | COM_SELECT_CLOSE);

    if (err == COM_SELECTED_TIMEOUT) {
      UTL_Log("ADM_CheckOpcode", LOG_NOTIFY, "Timeout.");
      free(streams);
      return RC_TIMEOUT;
    }

    if (err == COM_SELECTED_CLOSE) {
      UTL_Log("ADM_CheckOpcode", LOG_NOTIFY, "Stream closed.");
      free(streams);
      return RC_CLOSE;
    }

    if (err != COM_SELECTED_READ) {
      UTL_Log("ADM_CheckOpcode", LOG_FATAL, "Error in select.");
      free(streams);
      return RC_ERROR;
    }

    tStream = (COM_STREAM*)vt;

    /* Get the peer's ID number */
    if ((num = UTL_StreamID(tStream)) < 0) {
      UTL_Log("ADM_CheckOpcode", LOG_FATAL, "Error getting peer's ID.");
      free(streams);
      return RC_ERROR;
    }

    /* Remember not to select on this stream again */
    streams[num] = NULL;
    
    /* Read the response from the stream */
    if (COM_read_int(tStream, &data) < 0) {
      UTL_Log("ADM_CheckOpcode", LOG_FATAL, "Error reading response.");
      free(streams);
      return RC_ERROR;
    }
    
    if (data != opcode) {
      UTL_Log("ADM_CheckOpcode", LOG_FATAL, "Server %d thought it had opcode %d.", num, data);
    }
  }

  free(streams);
  return RC_OK;
}


/*------------------------------------------------------------------------*/
/*--                                                                    --*/
/*--	ADM_control()							--*/
/*--                                                                    --*/
/*------------------------------------------------------------------------*/

int ADM_control(ADM_CTX *ctx)
{


  /*-- Send header and opcode to all servers --*/
  if (ADM_SendHeader(ctx) != RC_OK) {
    UTL_Log("ADM_Control", LOG_FATAL, "Error sending header information");
    return RC_ERROR;
  }

  /*-- Get header code --*/
  if (ADM_CheckResponse(ctx, "RESPONSE") != RC_OK) {
    UTL_Log("ADM_Control", LOG_FATAL, "Unable to read header");
    return RC_ERROR;
  }

  /*-- Get opcode --*/
  if (ADM_CheckOpcode(ctx, ctx->opcode) != RC_OK) {
    UTL_Log("ADM_Control", LOG_FATAL, "Unable to read opcode");
    return RC_ERROR;
  }

  /*-- Get begin code --*/
  if (ADM_CheckResponse(ctx, "OK") != RC_OK) {
    UTL_Log("ADM_Control", LOG_FATAL, "Unable to read begin code");
    return RC_ERROR;
  }

  UTL_Log("ADM_control", LOG_NOTIFY, "Operation completed successfully.\n");
  return RC_OK;
}


/*------------------------------------------------------------------------*/
/*--                                                                    --*/
/*--	ADM_GetBN()							--*/
/*--                                                                    --*/
/*------------------------------------------------------------------------*/

/* The function gets a single BIGNUM from a stream */
int ADM_GetBN(BIGNUM *get, COM_STREAM *stream) {

  int err = 0;
  void *vt = NULL;
  COM_STREAM *tStream = NULL;
  BIGNUM *bigTemp = NULL;
  /* Initializations */
  if (get == NULL) {
    UTL_Log("ADM_GetBN", LOG_FATAL, "Passed NULL BIGNUM.");
    return RC_ERROR;
  }
  bigTemp = BN_new();
  if (bigTemp == NULL) {
    UTL_Log("ADM_GetBN", LOG_FATAL, "Error in BN_new().");
    return RC_ERROR;
  }

  /* Select for input */
  err = COM_select_v(stream->ctx, &vt, &stream, 1, 
		     COM_SELECT_READ | COM_SELECT_TIMEOUT | COM_SELECT_CLOSE);
  if (err == COM_SELECTED_TIMEOUT) {
    UTL_Log("ADM_GetBN", LOG_FATAL, "Timeout in select.");
    BN_clear_free(bigTemp);
    return RC_TIMEOUT;
  }
  if (err == COM_SELECTED_CLOSE) {
    UTL_Log("ADM_GetBN", LOG_FATAL, "Close in select.");
    BN_clear_free(bigTemp);
    return RC_CLOSE;
  }
  if (err != COM_SELECTED_READ) {
    UTL_Log("ADM_GetBN", LOG_FATAL, "Error in select.");
    BN_clear_free(bigTemp);
    return RC_ERROR;
  }
  tStream = (COM_STREAM*)vt;
  
  /* Read the BIGNUM from the stream */
  if (COM_read_bn(tStream, bigTemp) < 0) {
    UTL_Log("ADM_GetBN", LOG_FATAL, "Error reading BIGNUM.");
    BN_clear_free(bigTemp);
    return RC_ERROR;
  }

  if (BN_copy(get, bigTemp) == NULL) {
    UTL_Log("ADM_GetBN", LOG_FATAL, "Error copying BIGNUM.");
    BN_clear_free(bigTemp);
    return RC_ERROR;
  }

  BN_clear_free(bigTemp);
  return RC_OK;
}




/*------------------------------------------------------------------------*/
/*--                                                                    --*/
/*--	ADM_recover()   						--*/
/*--                                                                    --*/
/*--    Synchronize versions of a key on servers and ADM                --*/
/*--    type == ADM_HIGH: Synchronize on the highest version            --*/
/*--                      which all servers share                       --*/
/*--    type == ADM_LOW: Synchronize on the lowest version for which    --*/
/*--                     there is a threshold                           --*/
/*--                                                                    --*/
/*------------------------------------------------------------------------*/

int ADM_recover(ADM_CTX *ctx, int type) {

  int i, j;
  BN_ULONG id;
  char resp[128];
  int *minv = NULL, *maxv = NULL;
  int bigMin, smallMax;
  int finalv;
  int absMin, absMax;
  int numServers;
  int threshold;
  int count;
  int op;
  int rc = RC_ERROR;
  RSA_SHARED_KEY *key = NULL;
  char * keydir = NULL;
  char rskName[256];
  char backRskName[256];
  char *pemName;
  char backPemName[256];
  char defPemName[256];
  BIGNUM *N = NULL;
  BIGNUM *e = NULL;


  //----------------------------------------
  // Initialize
  //----------------------------------------
  
  UTL_Log("ADM_recover", LOG_DEBUG, "### STARTING RECOVER");

  keydir = UTL_GetPathAdminParam(ctx->utlctx, "Key_Dir", ctx->me, NULL);

  /* Load the key */
  if (keydir)
    sprintf(rskName, "%s/%s.rsk", keydir, ctx->keyname);
  else
    sprintf(rskName, "%s.rsk", ctx->keyname);
  key = ITTC_load_RSA_key(rskName);  

  if (key == NULL) {
    UTL_Log("ADM_recover", LOG_FATAL, "Error reading public key '%s' from disk", rskName);
    return RC_NO_KEY;
  }
  id = (int)KEY_RSA_get_ID((RSA*)key);

  N = BN_dup(KEY_RSA_shared_key_n(key));
  e = BN_dup(KEY_RSA_shared_key_e(key));
  threshold = KEY_RSA_shared_key_threshold(key);
  numServers = KEY_RSA_shared_key_nparties(key);
  if (numServers != ctx->numServers) {
    printf("Please refresh this key with '-n %d', not '-n %d'\n\n", numServers, ctx->numServers);
    exit(1);
  }

  KEY_RSA_shared_key_free(key);
  key = NULL;

  minv = (int*)malloc(numServers*sizeof(int));
  if (minv == NULL) goto err;
  maxv = (int*)malloc(numServers*sizeof(int));
  if (maxv == NULL) goto err;
  for (i=0; i<ctx->numServers; i++) {
    minv[i] = -1;
    maxv[i] = -1;
  }


  //----------------------------------------
  // Send command to get version numbers
  //----------------------------------------

  UTL_Log("ADM_recover", LOG_DEBUG, "### STARTING GET VERSION");

  /* Close all connections, to start new ones later */
  for (i=0; i<ctx->numServers; i++) {
    if (ctx->streams[i] != NULL) {
      COM_close(ctx->streams[i]);
      ctx->streams[i] = NULL;
    }
  }

  /* Send opcode */
  ctx->opcode = ADM_op_get_version;
  if (ADM_SendHeader(ctx) != RC_OK) {
    UTL_Log("ADM_recover", LOG_FATAL, "Error sending header information");
    goto err;
  }

  /* Make sure all servers are active */
  for (i=0; i<ctx->numServers; ++i) {
    if (ctx->streams[i] == NULL) {
      UTL_Log("ADM_recover", LOG_FATAL, "All servers must be active to recover a key");
      goto err;
    }
  }

  /* Send key ID */
  for (i=0; i<ctx->numServers; i++) {
    if (COM_write_int(ctx->streams[i], id) < 0) {
      UTL_Log("ADM_recover", LOG_FATAL, "Unable to write to server %d", i);
      goto err;
    }
  }


  //----------------------------------------
  // Get version information from servers
  //----------------------------------------

  for (i=0; i<ctx->numServers; i++) {

    /* Get header */
    if (UTL_WaitRead(ctx->streams[i]) != RC_OK) goto err;
    if (COM_read_string(ctx->streams[i], resp) < 0 || strcmp(resp, "RESPONSE") != 0) {
      UTL_Log("ADM_recover", LOG_FATAL, "Error reading header");
      goto err;
    }

    /* Get opcode */
    if (UTL_WaitRead(ctx->streams[i]) != RC_OK) goto err;
    if (COM_read_int(ctx->streams[i], &op) < 0 || (op != ctx->opcode) ) {
      UTL_Log("ADM_recover", LOG_FATAL, "Error reading opcode");
      goto err;
    }

    /* Get begin code */
    if (UTL_WaitRead(ctx->streams[i]) != RC_OK) goto err;
    if (COM_read_string(ctx->streams[i], resp) < 0) {
      UTL_Log("ADM_recover", LOG_FATAL, "Error reading begin code");
      goto err;
    }
    if (strcmp(resp, "OK") != 0) {
      UTL_Log("ADM_recover", LOG_FATAL, "Begin code not OK:");
      UTL_Log("ADM_recover", LOG_FATAL, "%s", resp);
      goto err;
    }

    /* Get minimum version number */
    if (UTL_WaitRead(ctx->streams[i]) != RC_OK) goto err;
    if (COM_read_int(ctx->streams[i], &(minv[i])) < 0) {
      UTL_Log("ADM_recover", LOG_FATAL, "Opcode not OK   [%d]", op);
      goto err;
    }

    /* Get maximum version number */
    if (UTL_WaitRead(ctx->streams[i]) != RC_OK) goto err;
    if (COM_read_int(ctx->streams[i], &(maxv[i])) < 0) {
      UTL_Log("ADM_recover", LOG_FATAL, "Opcode not OK   [%d]", op);
      goto err;
    }

  }


  //----------------------------------------
  // Compute highest common version
  //----------------------------------------
  
  if (type == ADM_HIGH) {

    for (i=0; i<ctx->numServers; i++) {
      if ( (minv[i] == -1) || (maxv[i] == -1) ) {
	UTL_Log("ADM_recover", LOG_FATAL, "Server %d has no version of key '%s'", i, rskName);
	goto err;
      }
    }

    /* Find the largest minimum and smallest maximum versions */
    bigMin = minv[0];
    smallMax = maxv[0];
    for (i=0; i<ctx->numServers; i++) {
      UTL_Log("ADM_recover", LOG_FATAL, "Server %d:    min = %2d        max = %2d", i, minv[i], maxv[i]);
      if (minv[i] > bigMin) bigMin = minv[i];
      if (maxv[i] < smallMax) smallMax = maxv[i];
    }
    UTL_Log("ADM_recover", LOG_FATAL, "   Total: bigMin = %2d   smallMax = %2d", bigMin, smallMax);
    
    /* Now find the higest common version number (recover) */
    if (smallMax >= bigMin) {
      /* Everything worked correctly in the refresh */
      finalv = smallMax;
    } else {
      UTL_Log("ADM_recover", LOG_FATAL, "No version cover found");
      goto err;
    }
    
    printf("Setting key '%s' to version %d\n", rskName, finalv);

  } else {

    /* Find the minimum and maximum versions */
    i = 0;
    while (minv[i] == -1) i++;
    absMin = minv[i];
    absMax = maxv[i];
    for (i=0; i<ctx->numServers; i++) {
      UTL_Log("ADM_recover", LOG_FATAL, "Server %d:    min = %2d        max = %2d", i, minv[i], maxv[i]);
      if ( (minv[i] != -1) && (minv[i] < absMin) ) absMin = minv[i];
      if ( (maxv[i] != -1) && (maxv[i] > absMax) ) absMax = maxv[i];
    }

    /* Now find the lowest common version number with a threshold */
    finalv = -1;
    for (j=absMin; (j<=absMax) && (finalv<0) ; j++) {

      count = 0;
      for (i=0; i<ctx->numServers; i++)
	if ( (j >= minv[i]) && (j <= maxv[i]) ) count++;
      if (count >= threshold) finalv = j;

    }
      
    printf("Setting key '%s' to version %d\n", rskName, finalv);

  }


  //---------------------------------------------
  // Tell servers to use highest common version
  //---------------------------------------------

  UTL_Log("ADM_recover", LOG_DEBUG, "### STARTING VERSION SET");

  ctx->version = finalv;    /* remember this for refresh, if it's coming next */

  /* Close all connections, to start new ones later */
  for (i=0; i<ctx->numServers; i++) {
    if (ctx->streams[i] != NULL) {
      COM_close(ctx->streams[i]);
      ctx->streams[i] = NULL;
    }
  }

  /* Send new opcode */
  ctx->opcode = ADM_op_version_set;
  if (ADM_SendHeader(ctx) != RC_OK) goto err;

  /* Send key id and version */
  for (i=0; i<ctx->numServers; i++) {
    if (COM_write_int(ctx->streams[i], id) < 0) {
      UTL_Log("ADM_recover", LOG_FATAL, "Unable to write to server %d", i);
      goto err;
    }
    if (COM_write_int(ctx->streams[i], finalv) < 0) {
      UTL_Log("ADM_recover", LOG_FATAL, "Unable to write to server %d", i);
      goto err;
    }
    if (COM_write_int(ctx->streams[i], type) < 0) {
      UTL_Log("ADM_recover", LOG_FATAL, "Unable to write to server %d", i);
      goto err;
    }
  }

  /* Only get the public key if in ADM_HIGH mode */
  if (type == ADM_LOW) {


    //----------------------------------------
    // Not getting a public key
    //----------------------------------------

    for (i=0; i<ctx->numServers; i++) {

      /* Get header */
      if (UTL_WaitRead(ctx->streams[i]) != RC_OK) goto err;
      if (COM_read_string(ctx->streams[i], resp) < 0 || strcmp(resp, "RESPONSE") != 0) {
	UTL_Log("ADM_recover", LOG_FATAL, "Error reading header");
	goto err;
      }
      
      /* Get opcode */
      if (UTL_WaitRead(ctx->streams[i]) != RC_OK) goto err;
      if (COM_read_int(ctx->streams[i], &op) < 0 || (op != ctx->opcode) ) {
	UTL_Log("ADM_recover", LOG_FATAL, "Error reading opcode");
	goto err;
      }
      
      /* Get begin code */
      if (UTL_WaitRead(ctx->streams[i]) != RC_OK) goto err;
      if (COM_read_string(ctx->streams[i], resp) < 0) {
	UTL_Log("ADM_recover", LOG_FATAL, "Error reading begin code");
	goto err;
      }
      if (strcmp(resp, "OK") != 0) {
	UTL_Log("ADM_recover", LOG_FATAL, "Begin code not OK:");
	UTL_Log("ADM_recover", LOG_FATAL, "%s", resp);
	goto err;
      }

    }


  } else {


    //----------------------------------------
    // Back up public key
    //----------------------------------------
    
    /* Get the name of the default .pem file by changing*/
    /* "rsk" to "pem" at the end of the .rsk file's name */
    strcpy(defPemName, rskName);
    defPemName[strlen(rskName)-3] = 'p';
    defPemName[strlen(rskName)-2] = 'e';
    defPemName[strlen(rskName)-1] = 'm';
    
    /* Get the real .pem file name from the .rsk file */
    pemName = UTL_GetPathConfigParam(ctx->utlctx, "Shared_Key", defPemName);
    
    /* Make the backup name for the .rsk file  (".rsb") */
    strcpy(backRskName, rskName);
    backRskName[strlen(rskName)-1] = 'b';
    
    /* Make the backup name for the .pem file  (".peb") */
    strcpy(backPemName, pemName);
    backPemName[strlen(pemName)-1] = 'b';
    
    /* Backup the .rsk file */
    unlink(backRskName);
    if (rename(rskName, backRskName) < 0) {
      UTL_Log("ADM_recover", LOG_FATAL, "Unable to create backup file '%s'", backRskName);
      goto err;
    }
    
    /* Backup the .pem file */
    unlink(backPemName);
    if (rename(pemName, backPemName) < 0) {
      UTL_Log("ADM_recover", LOG_FATAL, "Unable to create backup file '%s'", backPemName);
      goto err;
    }
    
    
    //----------------------------------------
    // Set public key
    //----------------------------------------
    
    /* Get public key */
    if (ADM_GetPublicKey(ctx, N, e) != RC_OK) {
      UTL_Log("ADM_recover", LOG_FATAL, "Unable to get public key");
      
      /* restore the .pem file */
      unlink(pemName);
      rename(backPemName, pemName);
      
      /* restore the .rsk file */
      unlink(rskName);
      rename(backRskName, rskName);
      
      goto err;
    }
    
    
    //----------------------------------------
    // Remove backup key
    //----------------------------------------
    
    unlink(backRskName);
    unlink(backPemName);
    
  }


  //----------------------------------------
  // Free and return
  //----------------------------------------

  rc = RC_OK;


 done: 
  if (minv) free(minv);
  if (maxv) free(maxv);
  if (N) BN_clear_free(N);
  if (e) BN_clear_free(e);
  if (key) KEY_RSA_shared_key_free(key);
  return rc;


 err:
  rc = RC_ERROR;
  goto done;


  return RC_ERROR;

}




/*------------------------------------------------------------------------*/
/*--                                                                    --*/
/*--	ADM_GetPublicKey()						--*/
/*--                                                                    --*/
/*--    Get and save public key                                         --*/
/*--                                                                    --*/
/*--    If inN, inPubExp, or inG are non-NULL, their values as given by --*/
/*--    the servers are be checked against the values input to this     --*/
/*--    function, and if they don't match it is considered an error.    --*/
/*--    getMode decides whether trial decryptions are performed to      --*/
/*--    test the new public shared key.                                 --*/
/*--                                                                    --*/
/*------------------------------------------------------------------------*/

int ADM_GetPublicKey(ADM_CTX *ctx, BIGNUM *inN, BIGNUM *inPubExp)
{
  int i, j, num, rc = RC_ERROR;
  char resp[64];
  BIGNUM *temp = NULL, *N = NULL, *pubExp = NULL, *g = NULL;
  BIGNUM **arrayN = NULL;
  BIGNUM **arrayPubExp = NULL;
  BIGNUM **arrayg = NULL;
  int maxTests;
  RSA_SHARED_KEY *rsa = NULL;
  char * keydir = NULL;
  char * basename;
  char name[256];
  TLB_CTX * tc = NULL;
  int op;
  FILE * fp = NULL;
  void *vt;
  UTL_HEAP *heap = NULL;


  //----------------------------------------
  // Initialize
  //----------------------------------------

  UTL_Log("ADM_GetPublicKey", LOG_DEBUG, "### STARTING GET PUBLIC KEY");

  heap = UTL_InitHeap(UTL_GetIntConfigParam(ctx->utlctx, "Heap_Size", 1000));
  if (heap == NULL) {
    UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Error initializing heap");
    goto err;
  }

  temp = BN_new();
  if (temp == NULL) {
    UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Error in BN_new()");
    goto err;
  }

  arrayN = UTL_GetBNArray(ctx->numServers, heap);
  if (arrayN == NULL) {
    UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Error getting an array from the heap");
    goto err;
  }

  arrayPubExp = UTL_GetBNArray(ctx->numServers, heap);
  if (arrayN == NULL) {
    UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Error getting an array from the heap");
    goto err;
  }

  arrayg = UTL_GetBNArray(ctx->numServers, heap);
  if (arrayN == NULL) {
    UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Error getting an array from the heap");
    goto err;
  }



  //----------------------------------------
  // Get N, e, g, and then make the key
  //----------------------------------------

  for (i=0; i<ctx->numServers; i++) {


    //----------------------------------------
    // Need the OK first
    //----------------------------------------

    /* Get header */
    if (UTL_WaitRead(ctx->streams[i]) != RC_OK) goto err;
    if (COM_read_string(ctx->streams[i], resp) < 0 || strcmp(resp, "RESPONSE") != 0) {
      UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Error reading header");
      goto err;
    }

    /* Get opcode */
    if (UTL_WaitRead(ctx->streams[i]) != RC_OK) goto err;
    if (COM_read_int(ctx->streams[i], &op) < 0 || (op != ctx->opcode) ) {
      UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Error reading opcode");
      goto err;
    }

    /* Note: there is no call to UTL_WaitRead() here, because */
    /* we don't want it to timeout */
    
    /* Get begin code */
    if (COM_read_string(ctx->streams[i], resp) < 0) {
      UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Error reading begin code");
      goto err;
    }
    if (strcmp(resp, "OK") != 0) {
      UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Begin code not OK:");
      UTL_Log("ADM_GetPublicKey", LOG_FATAL, "%s", resp);
      goto err;
    }


    //----------------------------------------
    // Get N, e, g
    //----------------------------------------
    
    if (UTL_WaitRead(ctx->streams[i]) != RC_OK) goto err;
    if (COM_select_v(ctx->streams[i]->ctx, &vt, &(ctx->streams[i]), 1, 
		     COM_SELECT_READ | COM_SELECT_CLOSE) != COM_SELECTED_READ) {
      UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Problem while waiting for public key.");
      goto err;
    } 
    
    if (ADM_GetBN(arrayN[i], ctx->streams[i]) != RC_OK) {
      UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Error receiving N.");
      goto err;
    }
    
    if (ADM_GetBN(arrayPubExp[i], ctx->streams[i]) != RC_OK) {
      UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Error receiving e.");
      goto err;
    }
    
    if (ADM_GetBN(arrayg[i], ctx->streams[i]) != RC_OK) {
      UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Error receiving g.");
      goto err;
    }
    
  }
    

  //----------------------------------------
  // Make sure all servers agree on the key
  //----------------------------------------
  
  if (inN) {
    N = inN;
  } else {
    N = BN_dup(arrayN[0]);
    if (N == NULL) goto err;
  }

  if (inPubExp) {
    pubExp = inPubExp;
  } else {
    pubExp = BN_dup(arrayPubExp[0]);
    if (pubExp == NULL) goto err;
  }

  g = BN_dup(arrayg[0]);
  if (g == NULL) goto err;

  if (!BN_copy(g, arrayg[0])) goto err;

  for (i=0; i<ctx->numServers; i++) {

    if (BN_cmp(N, arrayN[i])) {
      UTL_Log("ADM_GetPublicKey", LOG_FATAL, "N's do not match.");
      goto err;
    }

    if (BN_cmp(pubExp, arrayPubExp[i])) {
      UTL_Log("ADM_GetPublicKey", LOG_FATAL, "e's do not match.");
      goto err;
    }

    if (BN_cmp(g, arrayg[i])) {
      UTL_Log("ADM_GetPublicKey", LOG_FATAL, "g's do not match.");
      goto err;
    }

  }


  //----------------------------------------
  // Create the key
  //----------------------------------------
  
  if ((rsa = KEY_RSA_shared_key_create(N, pubExp)) == NULL) {
    UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Unable to create RSA key.");
    goto err;    
  }
  KEY_RSA_shared_key_set_threshold_params(rsa, ctx->numServers, ctx->threshold);
  KEY_RSA_shared_key_set_g(rsa, g);


  //----------------------------------------
  // Get the shares
  //----------------------------------------

  for (i=0; i<ctx->numServers; i++) {


    //----------------------------------------
    // How many shares for this server?
    //----------------------------------------

    if (ADM_GetBN(temp, ctx->streams[i]) != RC_OK) {
      UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Error receiving number of shares from server %d.", i);
      goto err;
    }
    num = BN_get_word(temp);

    UTL_Log("ADM_GetPublicKey", LOG_INFO, "%d shares for server %d.", num, i);



    //----------------------------------------
    // Get all the shares
    //----------------------------------------

    for (j=0; j<num; j++) {
      if (ADM_GetBN(temp, ctx->streams[i]) != RC_OK) {
	UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Error receiving share %d from server %d.", j, i);
	goto err;
      }
      KEY_RSA_shared_key_add_gd(rsa, temp);
    }

  }

  UTL_Log("ADM_GetPublicKey", LOG_DEBUG, "Received all g^d");


  //--------------------------------------------------
  // Save shared key to file.   (saves  e,N,g^{d_i})
  //--------------------------------------------------

  keydir = UTL_GetPathAdminParam(ctx->utlctx, "Key_Dir", ctx->me, NULL);

  if (keydir)
    sprintf(name, "%s/%s.pem", keydir, ctx->keyname);
  else
    sprintf(name, "%s.pem", ctx->keyname);
  if (KEY_RSA_shared_key_write(rsa, name) < 0) {
    UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Error writing public key to disk");
    goto err;
  }


  //------------------------------------------------------------
  // Writes  .rsk  file
  //------------------------------------------------------------

  if (keydir)
    sprintf(name, "%s/%s.rsk", keydir, ctx->keyname);
  else
    sprintf(name, "%s.rsk", ctx->keyname);
  if ((fp = fopen(name, "w")) == NULL) {
    UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Error writing .rsk file to disk");
    goto err;
  }

  if (keydir) {
    basename = strrchr(ctx->ConfigFile, '/');
    if (basename) {
      ++basename;
    } else {
      basename = ctx->ConfigFile;
    }
    fprintf(fp, "#include ../../config/%s\n\nClient_ID: 0\nShared_Key: %s.pem\n",
	    basename, ctx->keyname);
  } else {
    fprintf(fp, "#include %s\n\nClient_ID: 0\nShared_Key: %s.pem\n",
	    ctx->ConfigFile, ctx->keyname);
  }
  fclose(fp);
  printf("Shared key written to '%s'\n", name);


  //--------------------------------------------------
  // Test the public key, including all the g^d's
  //--------------------------------------------------

  if ((tc = TLB_CTX_new(name)) == NULL) {    /* name  still holds the name of the rsk file */
    UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Error creating TLB_ctx");
    TLB_CTX_free(tc);
    goto err;
  }
  KEY_RSA_shared_key_set_ctx(rsa, tc);
  
  /* Perform the tests */
  maxTests = UTL_GetIntConfigParam(ctx->utlctx, "Max_Tests", 5);
  for (i=0; i<maxTests; i++) {
    if (TLB_TestServers(rsa) != RC_OK) {
      UTL_Log("ADM_GetPublicKey", LOG_FATAL, "Error testing servers");
      goto err;
    }
  }


  //----------------------------------------
  // Free memory, return
  //----------------------------------------

  rc = RC_OK;

 done:
  if (rsa) KEY_RSA_shared_key_free(rsa);
  if (temp) BN_clear_free(temp);
  if (N) BN_clear_free(N);
  if (pubExp) BN_clear_free(pubExp);
  if (g) BN_clear_free(g);
  if (heap) UTL_FreeHeap(heap);
  return rc;

 err:
  rc = RC_ERROR;
  goto done;
  return RC_ERROR;
}





/*------------------------------------------------------------------------*/
/*--                                                                    --*/
/*--	ADM_GenKey()							--*/
/*--                                                                    --*/
/*------------------------------------------------------------------------*/

int ADM_GenKey(ADM_CTX *ctx)
{
  int i;


  //----------------------------------------
  // Initialize
  //----------------------------------------

  UTL_Log("ADM_GenKey", LOG_DEBUG, "### STARTING GENERATING KEY");

  if (ADM_SendHeader(ctx) != RC_OK) {
    UTL_Log("ADM_GenKey", LOG_FATAL, "Error sending header information");
    return RC_ERROR;
  }

  for (i=0; i<ctx->numServers; ++i) {
    if (ctx->streams[i] == NULL) {
      UTL_Log("ADM_GenKey", LOG_FATAL, "All servers must be active to generate a key");
      return RC_ERROR;
    }
  }


  //------------------------------------------------------------
  // Send key strength, numServers, and threshold to servers
  //------------------------------------------------------------
  
  for (i=0; i<ctx->numServers; ++i) {
    if (COM_write_string(ctx->streams[i], ctx->keysize) < 0) {
      UTL_Log("ADM_GenKey", LOG_FATAL, "Unable to write to server %d", i);
      return RC_ERROR;
    }
    if (COM_write_int(ctx->streams[i], ctx->numServers) < 0) {
      UTL_Log("ADM_GenKey", LOG_FATAL, "Unable to write to server %d", i);
      return RC_ERROR;
    }
    if (COM_write_int(ctx->streams[i], ctx->threshold) < 0) {
      UTL_Log("ADM_GenKey", LOG_FATAL, "Unable to write to server %d", i);
      return RC_ERROR;
    }
  }

  printf("Waiting for distributed key to be generated\n");

  //----------------------------------------
  // Read public key from servers 
  //----------------------------------------

  if (ADM_GetPublicKey(ctx, NULL, NULL) != RC_OK) {
    UTL_Log("ADM_GenKey", LOG_FATAL, "Unable to get public key");
    return RC_ERROR;
  }


  printf("Distributed key generation complete\n");

  return RC_OK;
}



/*------------------------------------------------------------------------*/
/*--                                                                    --*/
/*--	ADM_refresh()   						--*/
/*--                                                                    --*/
/*------------------------------------------------------------------------*/

int ADM_refresh(ADM_CTX *ctx)
{
  int i;
  int coalition;
  char resp[128];
  int op;
  int id;
  int ret;
  int rc = RC_ERROR;
  int bits = 56;
  RSA_SHARED_KEY *key = NULL;
  BIGNUM *result = NULL, *msg = NULL;
  char name[256];  
  char * keydir = NULL;
  BIGNUM *N = NULL;
  BIGNUM *e = NULL;
  int n = 0;
  int t = 0;
  int numServers;


  //----------------------------------------
  // Initialize
  //----------------------------------------

  UTL_Log("ADM_refresh", LOG_DEBUG, "### STARTING REFRESH");

  keydir = UTL_GetPathAdminParam(ctx->utlctx, "Key_Dir", ctx->me, NULL);

  /* Load the key */
  if (keydir)
    sprintf(name, "%s/%s.rsk", keydir, ctx->keyname);
  else
    sprintf(name, "%s.rsk", ctx->keyname);

  key = ITTC_load_RSA_key(name);  
  if (key == NULL) {
    UTL_Log("ADM_refresh", LOG_FATAL, "Error reading public key from disk");
    rc = RC_NO_KEY;
    goto err;
  }
  id = (int)KEY_RSA_get_ID((RSA*)key);

  UTL_Log("ADM_refresh", LOG_NOTIFY, "Refreshing key 0x%X", id);

  /* Set the threshold */
  ctx->threshold = KEY_RSA_shared_key_threshold(key);

  /* Get values from key */
  N = BN_dup(KEY_RSA_shared_key_n(key));
  e = BN_dup(KEY_RSA_shared_key_e(key));
  n = KEY_RSA_shared_key_nparties(key);
  t = KEY_RSA_shared_key_threshold(key);
  
  numServers = KEY_RSA_shared_key_nparties(key);
  if (numServers != ctx->numServers) {
    printf("Please refresh this key with '-n %d', not '-n %d'\n\n", numServers, ctx->numServers);
    exit(1);
  }


  //------------------------------------------------------------
  // Find a good coalition to use
  //------------------------------------------------------------

  /* Do random trial decryptions to find good coalition */
  result = BN_new();
  if (result == NULL) {
    UTL_Log("ADM_refresh", LOG_FATAL, "Error in BN_new().");
    goto err;
  }

  msg = BN_new();
  if (msg == NULL) {
    UTL_Log("ADM_refresh", LOG_FATAL, "Error in BN_new().");
    goto err;
  }

  bits = BN_num_bits(KEY_RSA_shared_key_n(key)) - 1;
  if (bits < 2) bits = 2;
  if (!BN_rand(msg, bits, 0, 0)) {     /* Choose a random message */
    UTL_Log("ADM_refresh", LOG_FATAL, "Error in BN_rand().");
    goto err;
  }
  
  coalition = 0;      /* Tell TLB_ApplyKey to return the coalition it uses */
  ret = TLB_ApplyKey(result, msg, TLB_DECRYPT_MESSAGE, &coalition, key);
  if (ret != RC_OK) {
    UTL_Log("ADM_refresh", LOG_FATAL, "Unable to find acceptable coalition");
    goto err;
  }
  
  KEY_RSA_shared_key_free(key);
  key = NULL;


  //------------------------------------------------------------
  // Send id and coalition
  //------------------------------------------------------------

  /* Close all connections, to start new ones later */
  for (i=0; i<ctx->numServers; i++) {
    if (ctx->streams[i] != NULL) {
      COM_close(ctx->streams[i]);
      ctx->streams[i] = NULL;
    }
  }

  ctx->opcode = ADM_op_refresh;  
  if (ADM_SendHeader(ctx) != RC_OK) {
    UTL_Log("ADM_refresh", LOG_FATAL, "Error sending header information");
    goto err;
  }

  for (i=0; i<ctx->numServers; ++i) {
    if (ctx->streams[i] == NULL) {
      UTL_Log("ADM_refresh", LOG_FATAL, "All servers must be active to refresh a key");
      goto err;
    }
  }

  /* Tell servers which key and coalition to use, and N, e, n, and t */
  for (i=0; i<ctx->numServers; ++i) {
    /* id */
    if (COM_write_int(ctx->streams[i], id) < 0) {
      UTL_Log("ADM_refresh", LOG_FATAL, "Unable to write to server %d", i);
      goto err;
    }
    /* Coalition */
    if (COM_write_int(ctx->streams[i], coalition) < 0) {
      UTL_Log("ADM_refresh", LOG_FATAL, "Unable to write to server %d", i);
      goto err;
    }
    /* N */
    if (COM_write_bn(ctx->streams[i], N) < 0) {
      UTL_Log("ADM_refresh", LOG_FATAL, "Unable to write to server %d", i);
      goto err;
    }
    /* e */
    if (COM_write_bn(ctx->streams[i], e) < 0) {
      UTL_Log("ADM_refresh", LOG_FATAL, "Unable to write to server %d", i);
      goto err;
    }
    /* nparties */
    if (COM_write_int(ctx->streams[i], n) < 0) {
      UTL_Log("ADM_refresh", LOG_FATAL, "Unable to write to server %d", i);
      goto err;
    }
    /* threshold */
    if (COM_write_int(ctx->streams[i], t) < 0) {
      UTL_Log("ADM_refresh", LOG_FATAL, "Unable to write to server %d", i);
      goto err;
    }
    /* version */
    if (COM_write_int(ctx->streams[i], ctx->version) < 0) {
      UTL_Log("ADM_refresh", LOG_FATAL, "Unable to write to server %d", i);
      goto err;
    }
  }


  //----------------------------------------
  // Get the response
  //----------------------------------------
  
  printf("Waiting for distributed key to be refreshed\n");

  for (i=0; i<ctx->numServers; i++) {

    /* Get header */
    /* There is no UTL_WaitRead here because refresh might take long enough */
    /* to cause a timeout */
    if (COM_read_string(ctx->streams[i], resp) < 0 || strcmp(resp, "RESPONSE") != 0) {
      UTL_Log("ADM_refresh", LOG_FATAL, "Error reading header");
      goto err;
    }

    /* Get opcode */
    if (UTL_WaitRead(ctx->streams[i]) != RC_OK) {
      UTL_Log("ADM_refresh", LOG_FATAL, "Error waiting for opcode");
      goto err;
    }
    if (COM_read_int(ctx->streams[i], &op) < 0 || (op != ctx->opcode) ) {
      UTL_Log("ADM_refresh", LOG_FATAL, "Error reading opcode");
      goto err;
    }

    /* Get begin code */
    if (UTL_WaitRead(ctx->streams[i]) != RC_OK) {
      UTL_Log("ADM_refresh", LOG_FATAL, "Error waiting for begin code");
      goto err;
    }
    if (COM_read_string(ctx->streams[i], resp) < 0) {
      UTL_Log("ADM_refresh", LOG_FATAL, "Error reading begin code");
      goto err;
    }
    if (strcmp(resp, "OK") != 0) {
      UTL_Log("ADM_refresh", LOG_FATAL, "Begin code not OK:");
      UTL_Log("ADM_refresh", LOG_FATAL, "%s", resp);
      goto err;
    }

  }

  printf("Refreshed key '%s'\n", name);

  rc = RC_OK;

 done:
  if (N) BN_clear_free(N);
  if (e) BN_clear_free(e);
  if (result) BN_clear_free(result);
  if (msg) BN_clear_free(msg);
  if (key) KEY_RSA_shared_key_free(key);
  return rc;


 err:
  if (rc != RC_NO_KEY) rc = RC_ERROR;
  goto done;


  return RC_ERROR;
}


