/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/* Communications context handling - COM package */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#ifndef WIN32
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#endif

#include "COM.h"
#include "COM_priv.h"

#include "UTL.h"

PUBLIC COM_CTX *
COM_init(uctx, type, num)
     UTL_CTX * uctx;
     COM_TYPE type;
     int num;
{
  COM_CTX * ctx;
  char buf[128];
  struct sockaddr_in serv;
  int port;
  char * seq_file;

  if((ctx = malloc(sizeof(COM_CTX))) == NULL) {
    COM_handle_memerr("COM_init");
    return NULL;
  }

  ctx->utl_ctx = uctx;

  ctx->trans = COM_transport_lookup(COM_get_system_param(uctx, type, "Transport", num, ""));

  if(ctx->trans == NULL) {
    COM_handle_other("COM_init", "Unsupported transport");
    return NULL;
  }

  ctx->sequence = NULL;
  if(UTL_GetBoolConfigParam(uctx, "Sequence_Numbers", 1)) {
    seq_file = COM_get_path_system_param(uctx, type, "Sequence_File", num, "");
    if(seq_file == NULL || *seq_file == '\0') {
      COM_handle_other("COM_init", "No sequence file specified");
      return NULL;
    }
    if((ctx->sequence = COM_seq_init(seq_file)) == NULL) {
      COM_handle_other("COM_init", "Error loading sequence file");
      return NULL;
    }
  }

  if(type == COM_TYPE_SHARE_SERVER) {
    sprintf(buf, "Share_IP_Port_%d", num);
    port = UTL_GetIntConfigParam(uctx, buf, COM_DEFAULT_PORT);

    memset((char *) &serv, 0, sizeof(serv));
    serv.sin_family = AF_INET;
    serv.sin_port = htons((unsigned short)port);
    serv.sin_addr.s_addr = INADDR_ANY;

    if((ctx->accept_socket = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
      COM_handle_errno("COM_init (socket)");
      free(ctx);
      return NULL;
    }
    if(bind(ctx->accept_socket, (struct sockaddr *) &serv, sizeof(serv)) < 0) {
      fprintf(stderr, "error binding to port %d\n", port);
      COM_handle_errno("COM_init (bind)");
      free(ctx);
      return NULL;
    }
    if(listen(ctx->accept_socket, 5) < 0) {
      COM_handle_errno("COM_init (listen)");
      free(ctx);
      return NULL;
    }
  }
  else
    ctx->accept_socket = -1;

  ctx->myid = COM_MAKE_INFO(type, num, 0);
  UTL_mutex_init(&ctx->llock);	/*CHECKERR*/
  UTL_mutex_init(&ctx->slock);
  UTL_cond_init(&ctx->condnew);
  ctx->flags = 0;
  ctx->timeoutms = COM_CTX_DEFAULT_TIMEOUTMS;
  ctx->rseq = 0;
  ctx->streams = NULL;
  ctx->inputs = NULL;

  if((ctx->c_ctx = COM_new_client_transport_ctx(ctx->trans, ctx, num)) != NULL) {
    ctx->s_ctx = NULL;
    if(type == COM_TYPE_SHARE_SERVER) {
      ctx->s_ctx = COM_new_server_transport_ctx(ctx->trans, ctx, num);
#ifdef INCUBATOR
      UTL_thread_create(&ctx->incubator, incubator, ctx);
#endif
    }
    return ctx;
  }
  COM_transport_handle_error(ctx->trans, "COM_init");
  free(ctx);
  return NULL;
}

PUBLIC void
COM_ctx_set_timeout(ctx, ms)
     COM_CTX * ctx;
     unsigned long ms;
{
  ctx->timeoutms = ms;
}

PUBLIC int
COM_ctx_get_info(ctx)
     COM_CTX * ctx;
{
  return ctx->myid;
}

PUBLIC int
COM_ctx_get_flags(ctx)
     COM_CTX * ctx;
{
  return ctx->flags;
}

PUBLIC void
COM_ctx_set_flags(ctx, flags)
     COM_CTX * ctx;
     int flags;
{
  if((flags & COM_FLAG_NONBLOCK) && ctx->accept_socket >= 0)
    set_nonblock(ctx->accept_socket, 1);
  ctx->flags |= flags;
}

PUBLIC void
COM_ctx_unset_flags(ctx, flags)
     COM_CTX * ctx;
     int flags;
{
  if((flags & COM_FLAG_NONBLOCK) && ctx->accept_socket >= 0)
    set_nonblock(ctx->accept_socket, 0);
  ctx->flags &= ~flags;
}

PUBLIC void
COM_ctx_add_input(ctx, fd)
     COM_CTX * ctx;
     int fd;
{
  COM_INPUT * input;

  if((input = malloc(sizeof(COM_INPUT))) == NULL)
    return;

  input->fd = fd;
  input->next = ctx->inputs;
  ctx->inputs = input;
}

PUBLIC void
COM_ctx_close(ctx)
     COM_CTX * ctx;
{
  COM_INPUT * ip;

  if(ctx->accept_socket >= 0)
#ifdef WIN32
    closesocket(ctx->accept_socket);
#else
    close(ctx->accept_socket);
#endif
  while(ctx->streams != NULL)
    COM_close(ctx->streams);
  if(ctx->sequence)
    COM_seq_free(ctx->sequence);
  if(ctx->s_ctx)
    COM_close_transport_ctx(ctx->trans, ctx->s_ctx);
  if(ctx->c_ctx)
    COM_close_transport_ctx(ctx->trans, ctx->c_ctx);
  UTL_mutex_destroy(&ctx->llock);
  UTL_mutex_destroy(&ctx->slock);
  UTL_cond_destroy(&ctx->condnew);

  while(ctx->inputs != NULL) {
    ip = ctx->inputs;
    ctx->inputs = ip->next;
    free(ip);
  }
  free(ctx);
}
