/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/* Miscellaneous utility functions - COM package */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "COM.h"
#include "COM_priv.h"

#include "UTL.h"

PACKAGE FILE * errfp =
#ifdef WIN32
	NULL;
#else
	stderr;
#endif

PUBLIC void
COM_set_error_stream(ep)
     FILE * ep;
{
  errfp = ep;
}

PACKAGE void
COM_handle_memerr(fname)
     char * fname;
{
  UTL_Log(fname, LOG_FATAL, "Out of memory!");
}

PACKAGE void
COM_handle_errno(fname)
     char * fname;
{
#ifdef WIN32
  UTL_Log(fname, LOG_WARNING, "Error %d", WSAGetLastError());
#else
  if(errno != 0)
    UTL_Log(fname, LOG_WARNING, "%s", strerror(errno));
#endif
}

PACKAGE void
COM_handle_other(fname, mesg)
     char * fname;
     char * mesg;
{
  UTL_Log(fname, LOG_WARNING, "%s", mesg);
}

PACKAGE int
COM_parse_peer_info(str, inf)
     char * str;
     COM_INFO * inf;
{
  char buf[128];
  char * t;
  int num, type;

  t = strchr(str, '[');
  sscanf(t+1, "%32s %d", buf, &num);
  if(strcmp(buf, "SERVER") == 0)
    type = COM_TYPE_SHARE_SERVER;
  else if(strcmp(buf, "CLIENT") == 0)
    type = COM_TYPE_CLIENT;
  else if(strcmp(buf, "ADMIN") == 0)
    type = COM_TYPE_ADMIN;
  else
    return -1;
  *inf = COM_MAKE_INFO(type, num, 0);
  return 0;
}

PACKAGE int
COM_unparse_peer_info(str, inf)
     char * str;
     COM_INFO inf;
{
  switch(COM_INFO_GET_TYPE(inf)) {
  case COM_TYPE_CLIENT:
    sprintf(str, "[CLIENT %d]", COM_INFO_GET_NUM(inf));
    break;
  case COM_TYPE_SHARE_SERVER:
    sprintf(str, "[SERVER %d]", COM_INFO_GET_NUM(inf));
    break;
  case COM_TYPE_ADMIN:
    sprintf(str, "[ADMIN %d]", COM_INFO_GET_NUM(inf));
    break;
  default:
    return -1;	/* Invalid type */
  }
  return 0;
}

PRIVATE void *
COM_get_param_helper(f, ctx, type, name, number, dflt)
     void * ((*f)());
     UTL_CTX * ctx;
     COM_TYPE type;
     char * name;
     int number;
     void * dflt;
{
  char buf[128];
  void * res;

  res = (*f)(ctx, name, dflt);

  if(type == COM_TYPE_SHARE_SERVER)
    sprintf(buf, "Server_%s", name);
  else if(type == COM_TYPE_CLIENT)
    sprintf(buf, "Client_%s", name);
  else if(type == COM_TYPE_ADMIN)
    sprintf(buf, "Admin_%s", name);
  else
    return NULL;

  res = (*f)(ctx, buf, res);

  if(type == COM_TYPE_SHARE_SERVER)
    sprintf(buf, "Server_%s_%d", name, number);
  else if(type == COM_TYPE_CLIENT)
    sprintf(buf, "Client_%s_%d", name, number);
  else if(type == COM_TYPE_ADMIN)
    sprintf(buf, "Admin_%s_%d", name, number);

  return (*f)(ctx, buf, res);
}

PACKAGE char *
COM_get_system_param(ctx, type, name, number, dflt)
     UTL_CTX * ctx;
     COM_TYPE type;
     char * name;
     int number;
     char * dflt;
{
  return (char *) COM_get_param_helper(UTL_GetConfigParam, ctx,
				       type, name, number, (void *) dflt);
}

PACKAGE char *
COM_get_path_system_param(ctx, type, name, number, dflt)
     UTL_CTX * ctx;
     COM_TYPE type;
     char * name;
     int number;
     char * dflt;
{
  return (char *) COM_get_param_helper(UTL_GetPathConfigParam, ctx,
				       type, name, number, (void *) dflt);
}

PACKAGE int
COM_get_int_system_param(ctx, type, name, number, dflt)
     UTL_CTX * ctx;
     COM_TYPE type;
     char * name;
     int number;
     int dflt;
{
  return (int) COM_get_param_helper(UTL_GetIntConfigParam, ctx,
				    type, name, number, (void *) dflt);
}

PACKAGE int
COM_get_bool_system_param(ctx, type, name, number, dflt)
     UTL_CTX * ctx;
     COM_TYPE type;
     char * name;
     int number;
     int dflt;
{
  return (int) COM_get_param_helper(UTL_GetBoolConfigParam, ctx,
				    type, name, number, (void *) dflt);
}
