/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "ITTC.h"
#include "COM_priv.h"

typedef struct seq_list_st {
  dev_t rdev;
  ino_t ino;
  COM_SEQUENCE * seq;
  int refs;	/* reference count */
  struct seq_list_st * next;
} SEQUENCE_LIST;

static SEQUENCE_LIST * mlist = NULL;

PRIVATE time_t
COM_seq_get_mtime(seq)
     COM_SEQUENCE * seq;
{
  struct stat st;

  if(fstat(fileno(seq->sfp), &st) < 0)
    return 0;

  return st.st_mtime;
}

/*
 * Sequence number/file management for the COM package
 * File format is ASCII.  Switch to binary at own risk.
 */

PRIVATE COM_SEQUENCE *
COM_seq_init_fp(fp)
     FILE * fp;
{
  SEQUENCE_LIST * p;
  COM_SEQUENCE * ret;
  struct stat st;

  if(fstat(fileno(fp), &st) < 0) {
    COM_handle_errno("COM_seq_init_fp (fstat)");
    return NULL;
  }

  /* Look for same file in list */
  for(p = mlist; p != NULL; p = p->next)
    if(p->rdev == st.st_rdev && p->ino == st.st_ino) {
      ++p->refs;
      return p->seq;
    }

  /* Not found, make new one */
  if((ret = malloc(sizeof(COM_SEQUENCE))) == NULL)
    return NULL;
  ret->head = NULL;
  ret->sfp = fp;
  ret->sv_lock = 0;
  UTL_mutex_init(&ret->up_lock);
  UTL_cond_init(&ret->up_cond);
  ret->flags = 0;
  ret->nlocks = 0;
  ret->mtime = st.st_mtime;

  /* Add it to master list */
  if((p = malloc(sizeof(SEQUENCE_LIST))) != NULL) {
    p->rdev = st.st_rdev;
    p->ino = st.st_ino;
    p->seq = ret;
    p->refs = 1;
    p->next = mlist;
    mlist = p;
  }

  return ret;
}

PACKAGE COM_SEQUENCE *
COM_seq_init(name)
    char * name;
{
  FILE * fp;
  COM_SEQUENCE * ret;

  if((fp = fopen(name, "r+")) == NULL && (fp = fopen(name, "w+")) == NULL) {
    COM_handle_errno("COM_seq_init (fopen)");
    return NULL;
  }
  else {
    ret = COM_seq_init_fp(fp);
#ifdef WIN32
	/* TODO: Proper file locking under Win32 (remove W32_LOCK!) */
#ifdef W32_LOCK
	ret->hFile = CreateFile(name, GENERIC_READ | GENERIC_WRITE,
		FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL, NULL);
#endif
#endif
    return ret;
  }
}

PRIVATE int
COM_seq_lock(seq)
     COM_SEQUENCE * seq;
{
  int ret;
#ifndef WIN32
  struct flock fl;
#endif

  rewind(seq->sfp);
#ifdef WIN32
#ifdef W32_LOCK
  seq->dwTop = GetFileSize(seq->hFile, NULL);
  if(LockFile(seq->hFile, 0, 0, seq->dwTop, 0))
	  ret = 0;
  else
	  ret = -1;
#else
  ret = 0;
#endif W32_LOCK
#else
  /* ret = lockf(fileno(seq->sfp), F_LOCK, 0); */
  /* ret = flock(fileno(seq->sfp), LOCK_EX); */
  fl.l_type = F_WRLCK;
  fl.l_whence = SEEK_SET;
  fl.l_start = 0;
  fl.l_len = 0;

  if((ret = fcntl(fileno(seq->sfp), F_SETLKW, &fl)) < 0) {
    COM_handle_errno("COM_seq_lock");
  }
#endif

  return ret;
}

PRIVATE int
COM_seq_unlock(seq)
     COM_SEQUENCE * seq;
{
#ifndef WIN32
  struct flock fl;
#endif

  rewind(seq->sfp);
#ifdef WIN32
#ifdef W32_LOCK
  if(UnlockFile(seq->hFile, 0, 0, seq->dwTop, 0))
	  return 0;
  else
	  return -1;
#else
  return 0;
#endif W32_LOCK
#else
  /* ret = lockf(fileno(seq->sfp), F_ULOCK, 0); */
  /* ret = flock(fileno(seq->sfp), LOCK_UN); */
  fl.l_type = F_UNLCK;
  fl.l_whence = SEEK_SET;
  fl.l_start = 0;
  fl.l_len = 0;
  return fcntl(fileno(seq->sfp), F_SETLKW, &fl);
#endif
}

PRIVATE int
COM_seq_save(seq)
     COM_SEQUENCE * seq;
{
  int i;
  char buf[256];
  COM_SEQUENCE_ENTRY * p;

  /* TODO: use real semaphores/locks here */
  if(seq->sv_lock)
    return 0;
  seq->sv_lock = 1;

  if(seq->flags & COM_SEQ_DIRTY)
    seq->flags &= ~COM_SEQ_DIRTY;
  else {
    seq->sv_lock = 0;
    return 0;
  }

  UTL_Log("COM_seq_save", LOG_DEBUG, "Saving sequence (addr=%08x)...", seq);
  if(seq->mtime != COM_seq_get_mtime(seq)) {
    COM_handle_other("COM_seq_save", "Sequence file synchronization error");
    seq->sv_lock = 0;
    return -1;
  }

  rewind(seq->sfp);
  for(p = seq->head, i = 0; p != NULL; p = p->next, ++i) {
    COM_unparse_peer_info(buf, p->inf);
    fprintf(seq->sfp, "%s:%d\n", buf, p->seq);
  }
  fflush(seq->sfp);

  seq->mtime = COM_seq_get_mtime(seq);
  seq->sv_lock = 0;
  return 0;
}

PACKAGE void
COM_seq_free(seq)
     COM_SEQUENCE * seq;
{
  SEQUENCE_LIST ** pp;
  SEQUENCE_LIST * q;
  COM_SEQUENCE_ENTRY * p;
  COM_SEQ_WAITER * w;

  for(pp = &mlist; *pp != NULL; pp = &q->next) {
    q = *pp;
    if(q->seq == seq) {
      if(--q->refs > 0)
	return;
      else {
	*pp = q->next;
	free(q);
	break;
      }
    }
  }

  if(seq->nlocks > 0) {
    COM_seq_save(seq);
    COM_seq_unlock(seq);
  }
  for(p = seq->head; p != NULL; p = seq->head) {
    /* Delete waiters without calling them */
    for(w = p->wlist; w != NULL; w = p->wlist) {
      p->wlist = w->next;
      free(w);
    }
    seq->head = p->next;
    free(p);
  }
#ifdef WIN32
#ifdef W32_LOCK
  CloseHandle(seq->hFile);
#endif
#endif
  fclose(seq->sfp);
  UTL_mutex_destroy(&seq->up_lock);
  UTL_cond_destroy(&seq->up_cond);
  free(seq);
}

PRIVATE int
COM_seq_update(seq)
     COM_SEQUENCE * seq;
{
  COM_SEQUENCE_ENTRY * mark;
  COM_SEQUENCE_ENTRY ** sp;
  COM_INFO inf;
  char buf[256];
  char * p;
  time_t mt;
  int snum;

  if(seq->sv_lock)
    return 0;
  mt = COM_seq_get_mtime(seq);
  UTL_Log("COM_seq_update", LOG_DEBUG, "Refreshing sequence (addr=%08x)", seq);
  seq->mtime = mt;

  rewind(seq->sfp);
  sp = &seq->head;
  mark = *sp;
  while(fgets(buf, sizeof(buf), seq->sfp) != NULL) {
    if((p = strrchr(buf, ':')) == NULL)
      continue;
    *p++ = '\0';
    if(COM_parse_peer_info(buf, &inf) < 0)
      continue;
    snum = atoi(p);

    for(;;) {
      if(*sp == NULL)
	sp = &seq->head;
      else if((*sp)->inf == inf) {
	if(snum < (*sp)->seq)
	  UTL_Log("COM_seq_update", LOG_DEBUG, "Backward update rejected");
	else
	  (*sp)->seq = snum;
	break;
      }
      else
	sp = &(*sp)->next;
      if(*sp == mark) {		/* Failed - add new */
	if(*sp)
	  sp = &(*sp)->next;

	if((mark = malloc(sizeof(COM_SEQUENCE_ENTRY))) != NULL) {
	  mark->parent = seq;
	  mark->inf = inf;
	  mark->seq = snum;
	  mark->locked = 0;
	  mark->wlist = NULL;
	  mark->next = *sp;
	  *sp = mark;
	}
      }
    }
  }

  seq->flags &= ~COM_SEQ_DIRTY;
  return 0;
}

PRIVATE void
COM_seq_clear_waiters(seqent)
     COM_SEQUENCE_ENTRY * seqent;
{
  COM_SEQ_WAITER * olist;
  COM_SEQ_WAITER * w;

  /* Detach the old list so new entries don't interfere */
  olist = seqent->wlist;
  seqent->wlist = NULL;

  for(w = olist; w != NULL; w = olist) {
    olist = w->next;
    (*w->waiter)(w->arg);
    free(w);
  }
}

PACKAGE int
COM_seq_unlock_ent(seqent)
     COM_SEQUENCE_ENTRY * seqent;
{
  if(!seqent->locked) {
    COM_handle_other("COM_seq_unlock_ent", "Sequence entry not locked");
    return -1;
  }

  if(seqent->parent->nlocks <= 0) {
    COM_handle_other("COM_seq_unlock_ent", "Internal lock inconsistency!");
    return -1;
  }

  seqent->locked = 0;

  /*
   * Greedy algorithm - our process will hold the lock until it
   * clears the wait queue.  This results in more efficient I/O,
   * since we don't have to keep saving and re-updating.  To
   * switch to the more equitable (but less efficient) algorithm,
   * move the clear_waiters line after the if() block.
   */
  COM_seq_clear_waiters(seqent);
  UTL_mutex_lock(&seqent->parent->up_lock);
  if(--seqent->parent->nlocks == 0) {
    COM_seq_save(seqent->parent);
    COM_seq_unlock(seqent->parent);
  }
  UTL_mutex_unlock(&seqent->parent->up_lock);
  UTL_cond_notify_all(&seqent->parent->up_cond);
  /* Alternate location */

  return 0;
}

PRIVATE int
COM_seq_wait_ent(seqent, func, arg)
     COM_SEQUENCE_ENTRY * seqent;
     void (* func)();
     void * arg;
{
  COM_SEQ_WAITER * w;

  if((w = malloc(sizeof(COM_SEQ_WAITER))) == NULL)
    return -1;

  w->waiter = func;
  w->arg = arg;
  w->next = seqent->wlist;
  seqent->wlist = w;

  return 0;
}

PRIVATE COM_SEQUENCE_ENTRY *
COM_seq_add(seq, inf)
     COM_SEQUENCE * seq;
     COM_INFO inf;
{
  COM_SEQUENCE_ENTRY * p;

  if((p = malloc(sizeof(COM_SEQUENCE_ENTRY))) == NULL)
    return NULL;

  p->parent = seq;
  p->inf = inf;
  p->seq = 0;
  p->locked = 0;
  p->wlist = NULL;
  p->next = seq->head;
  seq->head = p;
  seq->flags |= COM_SEQ_DIRTY;
  return p;
}

PRIVATE COM_SEQUENCE_ENTRY *
COM_seq_lookup(seq, inf)
     COM_SEQUENCE * seq;
     COM_INFO inf;
{
  COM_SEQUENCE_ENTRY * p;

  for(p = seq->head; p != NULL; p = p->next)
    if(p->inf == inf)
      return p;
  return NULL;
}

PACKAGE COM_SEQUENCE_ENTRY *
COM_seq_get_locked_entry(seq, inf, func, arg)
     COM_SEQUENCE * seq;
     COM_INFO inf;
     void (* func)();
     void * arg;
{
  COM_SEQUENCE_ENTRY * ent;
  unsigned int me = UTL_thread_current_id();

  /* TJW: Critical section is slightly conservative... */
  UTL_mutex_lock(&seq->up_lock);
  while (1) {
    if(seq->nlocks == 0) {
      if(COM_seq_lock(seq) < 0) {
	UTL_mutex_unlock(&seq->up_lock);
	return NULL;
      }
      COM_seq_update(seq);
    }

    ent = COM_seq_lookup(seq, inf);
    if(ent == NULL)
      ent = COM_seq_add(seq, inf);
    else if(ent->locked) {
      if(seq->nlocks == 0)
	COM_seq_unlock(seq);
      if(func != NULL && ent->lock_owner == me)
	COM_seq_wait_ent(ent, func, arg);
      else {
	UTL_cond_wait(&seq->up_cond, &seq->up_lock);
	continue;
      }
      UTL_mutex_unlock(&seq->up_lock);
      return NULL;
    }
    ent->locked = 1;
    ent->lock_owner = me;
    ++seq->nlocks;
    break;
  }
  UTL_mutex_unlock(&seq->up_lock);
  return ent;
}

PACKAGE unsigned int
COM_seq_get_number(seqent)
     COM_SEQUENCE_ENTRY * seqent;
{
  return seqent->seq;
}

PACKAGE void
COM_seq_increment(seqent, inc)
     COM_SEQUENCE_ENTRY * seqent;
     int inc;
{
  if(!seqent->locked)
    return;

  seqent->seq += inc;
  seqent->parent->flags |= COM_SEQ_DIRTY;
}
