/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/* SSL transport layer - COM package */

#include <stdio.h>
#include <errno.h>
#include <openssl/err.h>
#include <openssl/ssl.h>

#include "ITTC.h"
#include "COM_priv.h"

extern PACKAGE FILE * errfp;

typedef struct {
  SSL * ssl;
  BIO * bio;
} COM_SSL_STREAM;

static void
COM_SSL_error(where)
     const char * where;
{
  UTL_Log("COM (SSL transport)", LOG_WARNING, "SSL error");
  ERR_print_errors_fp(errfp);
}

static TRANS_CTX *
COM_SSL_new_server_ctx(magic, num)
     TRANS_MAGIC * magic;
     int num;
{
  SSL_METHOD * meth;
  SSL_CTX * ctx;
  char * filename;

  if((meth = SSLv3_server_method()) != NULL &&
     (ctx = (SSL_CTX *) SSL_CTX_new(meth)) != NULL &&
     SSL_CTX_use_certificate_file(ctx, COM_get_path_system_param(((COM_CTX *) magic)->utl_ctx, COM_TYPE_SHARE_SERVER, "Cert", num, ""), SSL_FILETYPE_PEM) > 0 &&
     SSL_CTX_use_PrivateKey_file(ctx, COM_get_path_system_param(((COM_CTX *) magic)->utl_ctx, COM_TYPE_SHARE_SERVER, "Key", num, ""), SSL_FILETYPE_PEM) > 0) {
    if((filename = COM_get_path_system_param(((COM_CTX *) magic)->utl_ctx, COM_TYPE_SHARE_SERVER, "CA_Cert", num, NULL)) != NULL) {
      SSL_CTX_set_session_cache_mode(ctx,
	SSL_SESS_CACHE_NO_AUTO_CLEAR|SSL_SESS_CACHE_SERVER);
      SSL_CTX_load_verify_locations(ctx, filename, NULL);
      SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT, NULL);
    }
    return (TRANS_CTX *) ctx;
  }
  else
    return NULL;
}

static TRANS_CTX *
COM_SSL_new_client_ctx(magic, num)
     TRANS_MAGIC * magic;
     int num;
{
  SSL_METHOD * meth;
  SSL_CTX * ctx;
  char * filename;

  if((meth = SSLv3_client_method()) != NULL &&
     (ctx = (SSL_CTX *) SSL_CTX_new(meth)) != NULL) {
    SSL_CTX_set_session_cache_mode(ctx,
	SSL_SESS_CACHE_NO_AUTO_CLEAR|SSL_SESS_CACHE_SERVER);
    if((filename = COM_get_path_system_param(((COM_CTX *) magic)->utl_ctx, COM_INFO_GET_TYPE(((COM_CTX *) magic)->myid), "Cert", num, "")) != NULL)
      SSL_CTX_use_certificate_file(ctx, filename, SSL_FILETYPE_PEM);
    if((filename = COM_get_path_system_param(((COM_CTX *) magic)->utl_ctx, COM_INFO_GET_TYPE(((COM_CTX *) magic)->myid), "Key", num, "")) != NULL)
      SSL_CTX_use_PrivateKey_file(ctx, filename, SSL_FILETYPE_PEM);
    if((filename = COM_get_path_system_param(((COM_CTX *) magic)->utl_ctx, COM_INFO_GET_TYPE(((COM_CTX *) magic)->myid), "CA_Cert", num, NULL)) != NULL) {
      SSL_CTX_load_verify_locations(ctx, filename, NULL);
      SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT, NULL);
    }

    return (TRANS_CTX *) ctx;
  }
  else
    return NULL;
}

static void
COM_SSL_close_ctx(ctx)
     TRANS_CTX * ctx;
{
  SSL_CTX_free((SSL_CTX *) ctx);
}

static TRANS_STREAM *
COM_SSL_new_server_stream(ctx, fd)
     TRANS_CTX * ctx;
     int fd;
{
  COM_SSL_STREAM * s;

  if(ctx == NULL || (s = (COM_SSL_STREAM *) malloc(sizeof(COM_SSL_STREAM))) == NULL)
    return NULL;

  s->bio = BIO_new_socket(fd, BIO_NOCLOSE);
  if((s->ssl = SSL_new((SSL_CTX *) ctx)) != NULL) {
    SSL_set_bio(s->ssl, s->bio, s->bio);
    SSL_set_accept_state(s->ssl);
    /*SSL_accept(s->ssl);*/

    return (TRANS_STREAM *) s;
  }
  BIO_free(s->bio);
  free(s);
  return NULL;
}

static TRANS_STREAM *
COM_SSL_new_client_stream(ctx, fd)
     TRANS_CTX * ctx;
     int fd;
{
  COM_SSL_STREAM * s;

  if(ctx == NULL || (s = (COM_SSL_STREAM *) malloc(sizeof(COM_SSL_STREAM))) == NULL)
    return NULL;
  s->bio = BIO_new_socket(fd, BIO_NOCLOSE);
  if((s->ssl = SSL_new((SSL_CTX *) ctx)) != NULL) {
    SSL_set_bio(s->ssl, s->bio, s->bio);
    SSL_set_connect_state(s->ssl);
    /*SSL_connect(s->ssl);*/

    return (TRANS_STREAM *) s;
  }
  BIO_free(s->bio);
  free(s);
  return NULL;
}

static int
COM_SSL_do_handshake(stream)
     TRANS_STREAM * stream;
{
  int ret;

  ret = SSL_do_handshake(((COM_SSL_STREAM *) stream)->ssl);

  if(ret <= 0) {
    return -1;
  }
  else if(SSL_is_init_finished(((COM_SSL_STREAM *) stream)->ssl))
    return 1;
  else
    return 0;
}

static int
COM_SSL_handshake_status(stream)
     TRANS_STREAM * stream;
{
  int rv = 0;
  SSL * ssl = ((COM_SSL_STREAM *) stream)->ssl;

  if(SSL_is_init_finished(ssl)) {
    return 0;
  }

  while(1) {
    if(SSL_want_read(ssl))
      rv |= COM_SELECT_READ;
    if(SSL_want_write(ssl))
      rv |= COM_SELECT_WRITE;
    if(rv)
      return rv;
    if(SSL_state(ssl) & SSL_ST_BEFORE) {
      if(SSL_state(ssl) & SSL_ST_CONNECT)
	return COM_SELECT_WRITE | COM_SELECT_READ;
      else
	return COM_SELECT_READ;
    }
    /* Should rarely, if ever, get down here */
    COM_SSL_do_handshake(stream);
    if(SSL_is_init_finished(ssl)) {
      return 0;
    }
  }
}

static char *
COM_SSL_get_id(stream)
     TRANS_STREAM * stream;
{
  X509 * cert;

  cert = SSL_get_peer_certificate(((COM_SSL_STREAM *) stream)->ssl);
  if(cert == NULL)
    return NULL;
  else
    return cert->name;
}

static int
COM_SSL_read(stream, buf, len)
     TRANS_STREAM * stream;
     unsigned char * buf;
     int len;
{
  return SSL_read(((COM_SSL_STREAM *) stream)->ssl, buf, len);
}

static int
COM_SSL_write(stream, buf, len)
     TRANS_STREAM * stream;
     unsigned char * buf;
     int len;
{
  return SSL_write(((COM_SSL_STREAM *) stream)->ssl, buf, len);
}

static void
COM_SSL_close_stream(stream, shutdown)
     TRANS_STREAM * stream;
     int shutdown;
{
  COM_SSL_STREAM * s = (COM_SSL_STREAM *) stream;
  int fd;

  fd = SSL_get_fd(s->ssl);
  if(!shutdown)
    SSL_set_quiet_shutdown(s->ssl, 1);
  if(s->ssl) {
    SSL_shutdown(s->ssl);
#ifdef WIN32
    closesocket(fd);
#else
    close(fd);
#endif
    SSL_free(s->ssl);
  }
  free(stream);
}

static COM_TRANSPORT ssl_trans = {
  COM_SSL_new_server_ctx,
  COM_SSL_new_client_ctx,
  COM_SSL_close_ctx,
  COM_SSL_new_server_stream,
  COM_SSL_new_client_stream,
  COM_SSL_handshake_status,
  COM_SSL_do_handshake,
  COM_SSL_get_id,
  COM_SSL_read,
  COM_SSL_write,
  COM_SSL_close_stream,
  COM_SSL_error,
  NULL
};

/* Locking interface for OpenSSL */

static UTL_MUTEX lock_cs[CRYPTO_NUM_LOCKS];

static void
COM_SSL_locking_callback(mode, type, file, line)
int mode;
int type;
char * file;
int line;
{
  if(mode & CRYPTO_LOCK)
    UTL_mutex_lock(&lock_cs[type]);
  else
    UTL_mutex_unlock(&lock_cs[type]);
}

COM_TRANSPORT *
COM_SSL_transport() 
{
  int i;

  if(ssl_trans.magic == NULL) {
    ERR_load_SSL_strings();
    SSLeay_add_ssl_algorithms();

    for(i = 0; i < CRYPTO_NUM_LOCKS; ++i)
      UTL_mutex_init(&lock_cs[i]);
    CRYPTO_set_locking_callback((void (*)())COM_SSL_locking_callback);

    ssl_trans.magic = (void *) 1;	/* Just a flag */
  }
  return &ssl_trans;
}
