/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/* Secure communications package for ITTC */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#ifndef WIN32
#include <unistd.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#endif
#include <fcntl.h>
#include <errno.h>

#include "COM.h"
#include "COM_priv.h"

#include "RC.h"
#include "UTL.h"

#ifndef INADDR_NONE
#define INADDR_NONE     0xffffffff      /* should be in <netinet/in.h> */
#endif

PRIVATE void COM_mark(UTL_TIME * t);
PRIVATE int COM_elapsedms(UTL_TIME * t);

PRIVATE int COM_read_int_internal(COM_STREAM * s, int * np);
PRIVATE int COM_write_int_internal(COM_STREAM * s, int n);

PRIVATE int COM_read_string_internal(COM_STREAM * s, char * str);
PRIVATE int COM_write_string_internal(COM_STREAM * s, char * str);

PRIVATE int COM_read_bytes_internal(COM_STREAM * s, unsigned char * bytes);
PRIVATE int COM_write_bytes_internal(COM_STREAM * s, unsigned char * bytes, int len);

PRIVATE int COM_read_bn_internal(COM_STREAM * s, BIGNUM * num);
PRIVATE int COM_write_bn_internal(COM_STREAM * s, BIGNUM * num);

PRIVATE int COM_set_state(COM_STREAM * s, COM_STATUS state);
PRIVATE int COM_advance_state(COM_STREAM * s, int input);
PRIVATE int COM_flush_input(COM_STREAM * s);

PRIVATE COM_STREAM *
COM_new(ctx)
     COM_CTX * ctx;
{
  COM_STREAM * is;

  if((is = malloc(sizeof(COM_STREAM))) == NULL) {
    COM_handle_memerr("COM_new");
    return NULL;
  }
  is->ctx = ctx;
  is->trans = ctx->trans;		/* Inherit transport by default */
  is->stream = NULL;
  is->peerid = COM_INFO_INVALID_MASK;
  is->flags = ctx->flags;		/* Inherit flags from parent ctx */
  is->timeoutms = ctx->timeoutms;
  is->rseq = ctx->rseq++;
  is->owner_pid = getpid();
  is->fd = -1;
  is->connecting = 0;
  is->incr = 0;
  is->accept_state = COM_AST_NOT_SIGNALED;
  is->owned = 0;
  is->iowner = UTL_thread_current_id();
  is->wbuf = buffer_new();
  is->rbuf = buffer_new();
  is->tbuf = buffer_new();
  is->s_ent = NULL;
  is->usemarker = 1;
  UTL_mutex_init(&is->tlock);
  COM_mark(&is->marker);		/* Start the clock */
  COM_set_state(is, COM_STATUS_NOCONNECTION);
#ifdef STATS
  /* Everything set to 0 - customize if needed */
  memset(&is->stats, 0, sizeof(is->stats));
#endif

  UTL_mutex_lock(&ctx->llock);
  is->next = ctx->streams;	/* Add to linked list */
  ctx->streams = is;
  UTL_mutex_unlock(&ctx->llock);
  return is;
}

/* TODO(?): Implement separate read/write shutdowns */

PRIVATE void
COM_shutdown(s)
     COM_STREAM * s;
{
  if(s->stream) {
    if(COM_ioready(s))
      COM_flush(s);
    UTL_Log("COM_shutdown", LOG_DEBUG, "*** Closing fd %d ***", s->fd);
    COM_transport_close_stream(s->trans, s->stream,
			       (s->owner_pid == getpid()) ? 1 : 0);
  }
  s->stream = NULL;
  s->usemarker = 0;
  COM_set_state(s, COM_STATUS_NOCONNECTION);
}

PRIVATE void
COM_delete(ctx, s)
     COM_CTX * ctx;
     COM_STREAM * s;
{
  COM_STREAM ** pp;

  UTL_mutex_lock(&ctx->llock);
  pp = &ctx->streams;
  while(*pp != s)
    pp = &((*pp)->next);

  if(*pp == s)
    *pp = s->next;
  else
    COM_handle_other("COM_delete", "Delete failed");
  UTL_mutex_unlock(&ctx->llock);
}

PRIVATE COM_STREAM *
COM_next(ctx, s)
  COM_CTX * ctx;
  COM_STREAM * s;
{
  COM_STREAM * t;

  UTL_mutex_lock(&ctx->llock);
  for(t = ctx->streams; t; t = t->next) {
    if(t == s) {
      t = t->next;
      UTL_mutex_unlock(&ctx->llock);
      return t;
    }
  }
  t = ctx->streams;
  UTL_mutex_unlock(&ctx->llock);
  return t;
}

PUBLIC void
COM_print_stats(s)
     COM_STREAM * s;
{
#ifdef STATS
  printf("Statistics for connection to entity %08x:\n", s->peerid);
  printf("     Bytes read: %6ld\n   Packets read: %6ld\n",
	 s->stats.b_read, s->stats.p_read);
  printf("  Bytes written: %6ld\nPackets written: %6ld\n",
	 s->stats.b_write, s->stats.p_write);
#endif
}

#ifdef STATS
PUBLIC COM_STAT *
COM_get_stats(s)
     COM_STREAM * s;
{
  return &s->stats;
}
#endif

PUBLIC void
COM_set_timeout(s, ms)
     COM_STREAM * s;
     unsigned long ms;
{
  if(s == NULL)
    COM_handle_other("COM_set_timeout", "NULL stream");
  else
    s->timeoutms = ms;
}

PRIVATE void
COM_mark(t)
     UTL_TIME * t;
{
  UTL_TimeGetCurrent(t);
}

PRIVATE int
COM_elapsedms(t)
     UTL_TIME * t;
{
  UTL_TIME now;

  UTL_TimeGetCurrent(&now);
  return UTL_TimeSubtractMsec(&now, t);
}

#define COM_get_fd(s) ((s)->fd)

/* Returns: bitmask with COM_SELECT_READ and/or COM_SELECT_WRITE set,
   depending on the current state of s. */
PRIVATE int
COM_expect(s)
     COM_STREAM * s;
{
  int i;

  switch(s->state) {
  case COM_STATUS_NOCONNECTION:
  case COM_STATUS_CONNECTED:
    return 0;
  case COM_STATUS_INPROGRESS:
    return COM_SELECT_WRITE;
  case COM_STATUS_INHANDSHAKE:
    UTL_mutex_lock(&s->tlock);
    i = COM_transport_handshake_status(s->trans, s->stream);
    UTL_mutex_unlock(&s->tlock);
    if(i > 0)
      return i;
    else if(i == 0) {
      if(COM_advance_state(s, COM_SELECTED_DONE) > 0)
	return COM_expect(s);
    }
    else if(i < 0) {
      COM_transport_handle_error(s->trans, "COM_expect");
      COM_shutdown(s);
    }
    return 0;
  case COM_STATUS_SEQ_WAIT:
    return 0;
  case COM_STATUS_WAITID:
  case COM_STATUS_SEQ_A:
  case COM_STATUS_SEQ_B:
    return COM_SELECT_READ;
  default:
    return 0;
  }
}

PRIVATE void
COM_wake(s)
     void * s;
{
  COM_advance_state((COM_STREAM *) s, COM_SELECTED_UNKNOWN);
}

/* 1 == got the lock, 0 == rejected, -1 == error */
PRIVATE int
COM_obtain_seq_lock(s)
     COM_STREAM * s;
{
  COM_INFO inf;

  COM_get_peer_info(s, &inf);

  if((s->s_ent = COM_seq_get_locked_entry(s->ctx->sequence,
					  COM_MAKE_INFO(COM_INFO_GET_TYPE(inf),
							COM_INFO_GET_NUM(inf),
							0),
					  COM_wake, (void *) s)) == NULL) {
    UTL_Log("COM_obtain_seq_lock", LOG_DEBUG, "Lock rejected, entering wait state...\n");
    return 0;
  }
  else
    return 1;
}

PRIVATE int
COM_set_state(s, state)
     COM_STREAM * s;
     COM_STATUS state;
{
  COM_INFO inf, cinf;
  char * cpid;
  int i;
  unsigned char tb[64];

  if(state == COM_STATUS_SEQ_A) {
    if(s->ctx->sequence == NULL) {
      COM_set_state(s, COM_STATUS_CONNECTED);
      return 1;
    }

    /* Only server sends initial sequence number */
    if(!s->connecting) {
      i = COM_obtain_seq_lock(s);
      if(i < 0) {
	COM_handle_other("COM_set_state", "Error obtaining lock");
	COM_shutdown(s);
	return i;
      }
      else if(i == 0) {
	COM_set_state(s, COM_STATUS_SEQ_WAIT);
	return 0;
      }

      i = COM_seq_get_number(s->s_ent);
      UTL_Log("COM_set_state", LOG_DEBUG, "Sending server's sequence number %d to %08x", i, s->peerid);
      COM_write_int_internal(s, i);
      COM_seq_increment(s->s_ent, 1);
      ++s->incr;
    }
  }
  else if(state == COM_STATUS_SEQ_B) {
    /* Send my sequence number */
    i = COM_seq_get_number(s->s_ent);
    UTL_Log("COM_set_state", LOG_DEBUG, "Sending sequence number %d to %08x", i, s->peerid);
    COM_write_int_internal(s, i);
    if(!s->connecting)
      return COM_set_state(s, COM_STATUS_CONNECTED);
  }
  else if(state == COM_STATUS_CONNECTED) {
    if(s->s_ent) {
      s->incr = 0;
      COM_seq_unlock_ent(s->s_ent);
      s->s_ent = NULL;
    }

    /* TODO: should check cert correctness earlier */

    COM_get_peer_info(s, &inf);
    cpid = COM_transport_get_id(s->trans, s->stream);
    if(cpid == NULL) {
      UTL_Log("COM_set_state", LOG_DEBUG, "Cert peer ID NULL");
    }
    else {
      UTL_Log("COM_set_state", LOG_DEBUG, "Cert peer ID: %s", cpid);
      /* TODO: Don't do this if CA_Cert unset (certs disabled) */
      if(COM_parse_peer_info(cpid, &cinf) < 0 ||
	 COM_INFO_GET_TYPE(cinf) != COM_INFO_GET_TYPE(inf) ||
	 COM_INFO_GET_NUM(cinf) != COM_INFO_GET_NUM(inf)) {
	COM_handle_other("COM_set_state", "Certificate/peer mismatch");
	COM_shutdown(s);
	return -1;
      }
      else
	UTL_Log("COM_set_state", LOG_DEBUG, "Peer IDs match");
    }

    if(s->tbuf) {
      if(s->wbuf) {
	while((i = buffer_get(s->tbuf, tb, sizeof(tb))) > 0)
	  buffer_append(s->wbuf, tb, i);
	buffer_free(s->tbuf);
	s->tbuf = NULL;
      }
      else
	s->wbuf = s->tbuf;
    }
    s->usemarker = 0;
  }
  else if(state == COM_STATUS_NOCONNECTION) {
    if(s->s_ent) {
      if(s->incr > 0)
	COM_seq_increment(s->s_ent, -s->incr);
      COM_seq_unlock_ent(s->s_ent);
      s->s_ent = NULL;
    }
  }

  if(state == COM_STATUS_CONNECTED && !s->connecting)
    UTL_mutex_lock(&s->ctx->slock);
  s->state = state;
  if(state == COM_STATUS_CONNECTED && !s->connecting)
    UTL_mutex_unlock(&s->ctx->slock);
  return 1;
}

/* Returns: 1 if state changed, 0 if nothing happened, -1 if error */

PRIVATE int
COM_advance_state(s, input)
     COM_STREAM * s;
     int input;
{
  COM_PKT_T t;
  int i, fd;
  fd_set writefds, readfds;
  struct sockaddr_in name;
  struct timeval tm = { 0L, 0L };

  /* Maybe NOCONNECTION should be an error? */
  if(s->state == COM_STATUS_NOCONNECTION || s->state == COM_STATUS_CONNECTED)
    return 0;

  if(input != COM_SELECTED_READ && input != COM_SELECTED_WRITE &&
     input != COM_SELECTED_DONE) {
    if(COM_packet_pending(s, NULL, NULL) > 0)
      input = COM_SELECTED_READ;
    else {
      fd = COM_get_fd(s);
      FD_ZERO(&readfds);
      FD_ZERO(&writefds);
      i = COM_expect(s);
      if(i & COM_SELECT_READ)
	FD_SET(fd, &readfds);
      if(i & COM_SELECT_WRITE)
	FD_SET(fd, &writefds);

      i = select(fd + 1, &readfds, &writefds, NULL, &tm);

      if(i > 0) {
        if(FD_ISSET(fd, &readfds)) {
	  i = COM_flush_input(s);
	  if((COM_ioready(s) && i <= 0) || i < 0) {
	    COM_shutdown(s);
	    return i;
	  }
	  else if(COM_packet_pending(s, NULL, &i) > 0)
	    input = COM_SELECTED_READ;
        }
        else if(FD_ISSET(fd, &writefds))
	  input = COM_SELECTED_WRITE;
      }
      else if(i < 0)
	return i;
    }
  }

  if(s->state == COM_STATUS_INPROGRESS) {
    if(input == COM_SELECTED_WRITE) {
      i = sizeof(name);
      /* The getpeername hack - check if the connect actually succeeded */
      if(
#ifdef WIN32
		  0
#else
		  getpeername(COM_get_fd(s), (struct sockaddr *) &name, &i) < 0
#endif
		  ) {
	COM_handle_errno("COM_advance_state");
	COM_shutdown(s);
	return -1;
      }
      else {
	COM_set_state(s, COM_STATUS_INHANDSHAKE);
	return 1;
      }
    }
  }
  else if(s->state == COM_STATUS_INHANDSHAKE) {
    /* TJW: We may need to fix this to pay attention to which side of the
       connection we're on and what input we just received (read/write) */
    UTL_mutex_lock(&s->tlock);
    if(input == COM_SELECTED_READ || input == COM_SELECTED_WRITE)
      i = COM_transport_do_handshake(s->trans, s->stream);
    else
      i = 0;
    if(i >= 0 &&
       (i = COM_transport_handshake_status(s->trans, s->stream)) == 0) {
      UTL_mutex_unlock(&s->tlock);
      if(!COM_INFO_GET_VALID(s->peerid))
	return COM_set_state(s, COM_STATUS_WAITID);
      else
	return COM_set_state(s, COM_STATUS_SEQ_A);
    }
    else {
      UTL_mutex_unlock(&s->tlock);
      if(i < 0 && (s->flags & COM_FLAG_NONBLOCK) == 0) {
	COM_transport_handle_error(s->trans, "COM_advance_state");
	COM_shutdown(s);
	return -1;
      }
    }
  }
  else if(s->state == COM_STATUS_WAITID) {
    if(input == COM_SELECTED_READ && COM_packet_pending(s, &t, &i) > 0) {
      if(t != COM_PKT_INT) {
	COM_handle_other("COM_advance_state", "Peer ID missing");
	COM_shutdown(s);
	return -1;
      }
      else {
	COM_read_int_internal(s, &s->peerid);
	return COM_set_state(s, COM_STATUS_SEQ_A);
      }
    }
  }
  else if(s->state == COM_STATUS_SEQ_WAIT) {
    if(COM_set_state(s, COM_STATUS_SEQ_A) < 0)
      return -1;
    else if(s->state == COM_STATUS_SEQ_WAIT)
      return 0;
    else
      return 1;
  }
  else if(s->state == COM_STATUS_SEQ_A) {
    if(input == COM_SELECTED_READ && COM_packet_pending(s, &t, &i) > 0) {
      if(t != COM_PKT_INT) {
	COM_handle_other("COM_advance_state", "Sequence number expected");
	COM_shutdown(s);
	return -1;
      }
      else {
	/* The client didn't get the lock when entering the SEQ_A state
	   (to prevent deadlock), so it needs to get it now. */
	if(s->connecting) {
	  i = COM_obtain_seq_lock(s);
	  if(i < 0) {
	    COM_handle_other("COM_advance_state", "Error locking sequence");
	    COM_shutdown(s);
	    return i;
	  }
	  else if(i == 0) {
	    COM_set_state(s, COM_STATUS_SEQ_WAIT);
	    return 0;
	  }
	}

	COM_read_int_internal(s, &i);
	UTL_Log("COM_advance_state", LOG_DEBUG, "Expecting sequence number %d, got %d from %08x",
	       COM_seq_get_number(s->s_ent), i, s->peerid);
	if(i == COM_seq_get_number(s->s_ent) + 1 && s->connecting) {
	  /* They're ahead, catch up */
	  UTL_Log("COM_advance_state", LOG_INFO, "Recovering from difference of 1...\n");
	  COM_seq_increment(s->s_ent, 1);
	  ++s->incr;
	}
	else if(i != COM_seq_get_number(s->s_ent)) {
	  COM_handle_other("COM_advance_state", "Sequence number mismatch");
	  COM_shutdown(s);
	  return -1;
	}
	COM_seq_increment(s->s_ent, 1);
	++s->incr;
	COM_set_state(s, COM_STATUS_SEQ_B);
	return 1;
      }
    }
  }
  else if(s->state == COM_STATUS_SEQ_B) {
    if(input == COM_SELECTED_READ && COM_packet_pending(s, &t, &i) > 0) {
      if(t != COM_PKT_INT) {
	COM_handle_other("COM_advance_state", "Second sequence number expected");
	COM_shutdown(s);
	return -1;
      }
      else {
	COM_read_int_internal(s, &i);
	UTL_Log("COM_advance_state", LOG_DEBUG, "Expecting sequence number %d, got %d",
	       COM_seq_get_number(s->s_ent) + 1, i);
	if(i != COM_seq_get_number(s->s_ent) + 1) {
	  COM_handle_other("COM_advance_state", "Sequence number mismatch");
	  COM_shutdown(s);
	  return -1;
	}
	COM_seq_increment(s->s_ent, 1);
	++s->incr;
	COM_set_state(s, COM_STATUS_CONNECTED);
	return 1;
      }
    }
  }

  return 0;
}

/* nb == true: set nonblocking; else, set blocking */
PACKAGE int
set_nonblock(fd, nb)
     int fd, nb;
{
  int i;

#ifdef WIN32
  if(ioctlsocket(fd, FIONBIO, &nb) != 0) {
	COM_handle_errno("set_nonblock (ioctlsocket)");
	return -1;
  }

  return 0;
#else
  if((i = fcntl(fd, F_GETFL, 0)) < 0) {
    COM_handle_errno("set_nonblock (fcntl GET)");
    return i;
  }
  if(nb)
    i |= O_NDELAY;
  else
    i &= ~O_NDELAY;
  if(fcntl(fd, F_SETFL, i) < 0) {
    COM_handle_errno("set_nonblock (fcntl SET)");
    return -1;
  }
  return 0;
#endif
}

PUBLIC COM_STREAM *
COM_connect(ctx, type, num, vport)
     COM_CTX * ctx;
     COM_TYPE type;
     int num;
     vport_t vport;
{
  COM_TRANSPORT * trans;
  char * host;
  int port, id;
  COM_STREAM * is;
  struct sockaddr_in saddr;
  struct hostent hent;
#ifdef WIN32
  struct hostent * hptr;
#endif
  unsigned char buf[256];
  unsigned long inAddr;
  int s;
  int status;
  int err;

  trans = COM_transport_lookup(COM_get_system_param(ctx->utl_ctx, type, "Transport", num, ""));
  if(trans == NULL)
    trans = ctx->trans;

  if(type == COM_TYPE_SHARE_SERVER)
    sprintf(buf, "Share_IP_Port_%d", num);
  else if(type == COM_TYPE_CLIENT)
    sprintf(buf, "Client_IP_Port_%d", num);
  else if(type == COM_TYPE_ADMIN)	/* Usually not used */
    sprintf(buf, "Admin_IP_Port_%d", num);
  else
    return NULL;
  port = UTL_GetIntConfigParam(ctx->utl_ctx, buf, COM_DEFAULT_PORT);

  host = COM_get_system_param(ctx->utl_ctx, type, "IP_Addr", num, "localhost");

  id = COM_MAKE_INFO(type, num, vport);

  memset((char *) &saddr, 0, sizeof(saddr));
  saddr.sin_family = AF_INET;
  saddr.sin_port = htons(port);

  /**** TJW: Can we do non-blocking I/O on DNS??? ****/
  /* TODO: Take advantage of non-blocking DNS on WIN32 */
  if((inAddr = inet_addr(host)) == INADDR_NONE) {
#ifdef WIN32
    if((hptr = gethostbyname(host)) == NULL) {
      COM_handle_other("COM_connect_to_host", "Unable to resolve hostname");
      return NULL;
    }
    else
      memcpy((char *) &saddr.sin_addr, hptr->h_addr, hptr->h_length);
#else
    if(gethostbyname_r(host, &hent, buf, sizeof(buf), &err) == NULL) {
      COM_handle_other("COM_connect_to_host", "Unable to resolve hostname");
      return NULL;
    }
    else
      memcpy((char *) &saddr.sin_addr, hent.h_addr, hent.h_length);
#endif
  }
  else
    saddr.sin_addr.s_addr = inAddr;
      
  if((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    COM_handle_errno("COM_connect (socket)");
    return NULL;
  }

  if(!(ctx->flags & COM_FLAG_BLOCKING_CONNECT))
    set_nonblock(s, 1);

  /* TODO: Need to set an alarm for a blocking connect w/timeout */
  status = COM_STATUS_NOCONNECTION;
  if(connect(s, (struct sockaddr *) &saddr, sizeof(saddr)) < 0) {
    if(
#ifdef WIN32
		WSAGetLastError() == WSAEWOULDBLOCK
#else
		errno == EINPROGRESS
#endif
		)
      status = COM_STATUS_INPROGRESS;
    else {
      COM_handle_errno("COM_connect (connect)");
      return NULL;
    }
  }
  else
    status = COM_STATUS_INHANDSHAKE;
  UTL_Log("COM_connect", LOG_DEBUG, "*** Connecting fd %d ***", s);

  if((ctx->flags & COM_FLAG_NONBLOCK) && (ctx->flags & COM_FLAG_BLOCKING_CONNECT))
    set_nonblock(s, 1);
  else if(!(ctx->flags & COM_FLAG_NONBLOCK) && !(ctx->flags & COM_FLAG_BLOCKING_CONNECT))
    set_nonblock(s, 0);

  if((is = COM_new(ctx)) != NULL) {
    is->connecting = 1;
    is->fd = s;
    is->trans = trans;
    if((is->stream = COM_new_transport_client_stream(trans, ctx->c_ctx, s)) != NULL) {
      COM_set_state(is, status);
      is->peerid = id;
      UTL_Log("COM_connect_helper", LOG_DEBUG, "Peer id (outgoing): %08x", id);
      COM_write_int_internal(is, COM_MAKE_INFO(COM_INFO_GET_TYPE(ctx->myid),
					       COM_INFO_GET_NUM(ctx->myid),
					       vport));
      return is;
    }
    COM_transport_handle_error(trans, "COM_connect (transport)");
    COM_delete(ctx, is);
    free(is);
  }
  close(s);
  return NULL;
}

PUBLIC COM_STREAM *
COM_accept(ctx, vport)
     COM_CTX * ctx;
     vport_t vport;
{
  COM_STREAM * s;
  int flags;

  if(vport == COM_ACCEPT_ANY)
    flags = COM_SELECT_ACCEPT;
  else
    flags = COM_SELECT_PORT(vport);
  if(COM_select_v(ctx, (void **) &s, NULL, 0, flags) == COM_SELECTED_ACCEPT)
    return s;
  else
    return NULL;
}

PRIVATE COM_STREAM *
COM_internal_accept(ctx)
     COM_CTX * ctx;
{
  COM_STREAM * is;
  struct sockaddr_in clnt;
  int cs;
  int len;

  if(ctx->accept_socket < 0) {
    COM_handle_other("COM_internal_accept", "cannot accept within client context");
    return NULL;
  }

  /* TODO: timeout in blocking mode */
  len = sizeof(clnt);
  if((cs = accept(ctx->accept_socket, (struct sockaddr *) &clnt, &len)) >= 0) {
    if(ctx->flags & COM_FLAG_NONBLOCK)
      set_nonblock(cs, 1);

    if((is = COM_new(ctx)) != NULL) {
      is->fd = cs;
      if((is->stream = COM_new_transport_server_stream(ctx->trans, ctx->s_ctx, cs)) != NULL) {
	COM_set_state(is, COM_STATUS_INHANDSHAKE);
	return is;
      }
      COM_transport_handle_error(ctx->trans, "COM_internal_accept");
      COM_delete(ctx, is);
      free(is);
    }
    close(cs);
  }
#ifdef WIN32
  else if(WSAGetLastError() != WSAEWOULDBLOCK)
#else
  else if(errno != EWOULDBLOCK)
#endif
    COM_handle_errno("COM_internal_accept");
  return NULL;
}

PUBLIC COM_CTX *
COM_get_CTX(s)
     COM_STREAM * s;
{
  return s->ctx;
}

PUBLIC void
COM_fork(f, s, magic)
     forkfunc f;
     COM_STREAM * s;
     void * magic;
{
  int pid;

#ifndef WIN32
  if((pid = fork()) == 0) {
    s->owner_pid = getpid();
    (*f)(s, magic);
    exit(0);
  }
  else
    s->owner_pid = pid;
#endif
}

PUBLIC void
COM_grab(s)
     COM_STREAM * s;
{
  s->owner_pid = getpid();
}

PUBLIC void
COM_ungrab(s)
     COM_STREAM * s;
{
  s->owner_pid = 0;
}

PUBLIC int
COM_get_peer_info(s, inf)
     COM_STREAM * s;
     COM_INFO * inf;
{
  if(s == NULL || !COM_INFO_GET_VALID(s->peerid))
    return -1;
  *inf = s->peerid;
  return 0;
}

PUBLIC int
COM_get_flags(s)
     COM_STREAM * s;
{
  return s->flags;
}

PUBLIC void
COM_set_flags(s, flags)
     COM_STREAM * s;
     int flags;
{
  if(flags & COM_FLAG_NONBLOCK)
    set_nonblock(COM_get_fd(s), 1);
  s->flags |= flags;
}

PUBLIC void
COM_unset_flags(s, flags)
     COM_STREAM * s;
     int flags;
{
  if(flags & COM_FLAG_NONBLOCK)
    set_nonblock(COM_get_fd(s), 0);
  s->flags &= ~flags;
}

PRIVATE int
bytes2int(bytes)
     unsigned char bytes[4];
{
  return (bytes[0] << 24) | (bytes[1] << 16) | (bytes[2] << 8) | bytes[3];
}

PUBLIC int
COM_packet_pending(s, tp, lp)
     COM_STREAM * s;
     COM_PKT_T * tp;
     int * lp;
{
  int len;
  int * p;
  unsigned char c, numbuf[4];

  if(lp)
    p = lp;
  else
    p = &len;

  *p = buffer_length(s->rbuf);
  if(*p == 0)
    return 0;

  buffer_peek(s->rbuf, 0, &c, 1);
  if(c != HEADER_BYTE) {
    if(tp)
      *tp = COM_PKT_OTHER;
    *p = 1;
    return 1;
  }

  if(*p < 2)
    return 0;
  *p -= 2;

  buffer_peek(s->rbuf, 1, &c, 1);
  switch(c) {
  case COM_PKT_BYTE:
    if(*p >= 1)
      *p = 1;
    else
      return 0;
    break;
  case COM_PKT_SHORT:
    if(*p >= 2)
      *p = 2;
    else
      return 0;
    break;
  case COM_PKT_INT:
    if(*p >= 4)
      *p = 4;
    else
      return 0;
    break;
  case COM_PKT_BYTES:
  case COM_PKT_BIGNUM:
  case COM_PKT_STRING:
    if(*p < 4)
      return 0;
    buffer_peek(s->rbuf, 2, numbuf, 4);
    if(*p < bytes2int(numbuf) + 4)
      return 0;
    *p = bytes2int(numbuf);
    break;
  default:
    if(tp)
      *tp = COM_PKT_OTHER;
    return 1;
  }
  if(tp)
    *tp = c;
  return 1;
}

PRIVATE int
COM_flush_input(s)
     COM_STREAM * s;
{
  int i;
  char buf[1024];

  if(s->state == COM_STATUS_NOCONNECTION) {
    COM_handle_other("COM_flush_input", "Connection closed");
    return -1;
  }
  if(s->stream) {
    if(!COM_ioready(s)) {
      /* COM_SELECTED_DONE breaks a potential recursion,
	 because COM_advance_state() would call us otherwise */
      if((i = COM_advance_state(s, COM_SELECTED_DONE)) <= 0)
	return i;
      if(s->state == COM_STATUS_NOCONNECTION)
	return -1;
      if(!COM_ioready(s))
	return 0;
    }

    /* TODO: implement blocking mode w/timeout */
    UTL_mutex_lock(&s->tlock);
    i = COM_transport_read(s->trans, s->stream, buf, sizeof(buf));
    UTL_mutex_unlock(&s->tlock);
  }
  else
    return 0;

  if(i > 0) {
    buffer_append(s->rbuf, buf, i);
#ifdef STATS
    s->stats.b_read += i;
#endif
  }
  else if(i < 0) {
    if((s->flags & COM_FLAG_NONBLOCK) &&
       (errno == 0
#ifdef WIN32
	   || WSAGetLastError() == WSAEWOULDBLOCK
#else
	   || errno == EINPROGRESS
#endif
	   ))
      return 0;
    else {
      COM_handle_errno("COM_flush_input (read)");
      COM_transport_handle_error(s->trans, "COM_flush_input");
    }
  }
  return i;
}

#ifdef INCUBATOR

PACKAGE void
incubate_stream(a_str)
     void * a_str;
{
  int fd;
  int r, i, n;
  int mintime;
  fd_set readfds;
  fd_set writefds;
  COM_STREAM * rs;
  COM_STREAM * s = (COM_STREAM *) a_str;
  struct timeval tm;

  fd = COM_get_fd(s);
  ++(s->owned);

  while(s->state != COM_STATUS_NOCONNECTION) {
    FD_ZERO(&readfds);
    FD_ZERO(&writefds);

    n = 0;
    mintime = -1;

    rs = NULL;

    if(s->state == COM_STATUS_CONNECTED)
      break;

    if(s->usemarker &&
       ((i = s->timeoutms - COM_elapsedms(&s->marker)) < mintime || mintime < 0))
      mintime = (i < 0) ? 0 : i;

    if(s->state < COM_STATUS_CONNECTED && s->state > COM_STATUS_INHANDSHAKE) {
      while(COM_packet_pending(s, NULL, NULL) > 0 &&
	s->state != COM_STATUS_CONNECTED &&
	s->state != COM_STATUS_NOCONNECTION) {
	if((i = COM_advance_state(s, COM_SELECTED_READ)) <= 0)
	  break;
      }
    }

    if(s->state == COM_STATUS_CONNECTED)
      break;
    else if(s->state > COM_STATUS_NOCONNECTION) {
      i = COM_expect(s);
      if(i & COM_SELECT_READ)
	FD_SET(fd, &readfds);
      if(i & COM_SELECT_WRITE)
	FD_SET(fd, &writefds);
    }
    else
      break;

    if(mintime >= 0) {
      tm.tv_sec = (mintime + FUDGE_MS) / 1000;
      tm.tv_usec = 1000 * ((mintime + FUDGE_MS) % 1000);
      r = select(fd + 1, &readfds, &writefds, NULL, &tm);
    }
    else
      r = select(fd + 1, &readfds, &writefds, NULL, NULL);

    if(r < 0) {
      COM_handle_errno("COM_incubate_stream");
      break;
    }

    n = 0;

    /* Timed out? */
    if(r == 0 && !FD_ISSET(COM_get_fd(s), &readfds) && !FD_ISSET(COM_get_fd(s), &writefds) && s->usemarker && COM_elapsedms(&s->marker) >= s->timeoutms) {
      UTL_Log("COM (server thread)", LOG_NOTIFY, "Incoming connection timed out");
      COM_shutdown(s);
      break;
    }

    if(FD_ISSET(fd, &readfds)) {
      FD_CLR(fd, &readfds);
      if(COM_ioready(s)) {
	if((i = COM_flush_input(s)) == 0) {
	  /* TJW: End of file -- this is the only way
	   * to detect a read EOF in non-blocking mode. */
	  UTL_Log("COM_select_v", LOG_DEBUG, "Shutting down fd %d", COM_get_fd(s));
	  COM_shutdown(s);
	}
	else if(i < 0) {
	  fprintf(stderr, "abnormal close, %s:%d\n", __FILE__, __LINE__);
	  COM_shutdown(s);
	}
      }
      if(s->state != COM_STATUS_CONNECTED && s->state != COM_STATUS_NOCONNECTION) {
	if(COM_advance_state(s, COM_SELECTED_READ) < 0) {
	  UTL_Log("COM_select_v", LOG_INFO, "Shutting down stream");
	  COM_shutdown(s);
	}
      }
    }
    if(s->state == COM_STATUS_NOCONNECTION)
      break;
    if(FD_ISSET(fd, &writefds)) {
      if(s->state == COM_STATUS_CONNECTED) {
	COM_flush(s);
      }
      else {
	if(COM_ioready(s) && COM_write_pending(s) > 0)
	  COM_flush(s);
	if(COM_advance_state(s, COM_SELECTED_WRITE) < 0) {
	  COM_shutdown(s);
	}
      }
    }
  }

  --(s->owned);

  if(s->state == COM_STATUS_CONNECTED) {
    s->accept_state = COM_AST_SIGNALED;
    UTL_cond_notify_all(&s->ctx->condnew);
  }
  else if(s->state == COM_STATUS_NOCONNECTION)
    COM_close(s);

  return;
}

PACKAGE void
incubator(a_ctx)
     void * a_ctx;
{
  int asock;
  int r;
  fd_set readfds;
  COM_STREAM * s;
  pid_t mypid;
  UTL_THREAD lastthread;
  COM_CTX * ctx = (COM_CTX *) a_ctx;

  asock = ctx->accept_socket;

  mypid = getpid();
  while(1) {
    FD_ZERO(&readfds);
    FD_SET(asock, &readfds);

    r = select(asock + 1, &readfds, NULL, NULL, NULL);

    if(r < 0) {
      COM_handle_errno("incubator");
      break;
    }

    if(FD_ISSET(asock, &readfds)) {
      s = COM_internal_accept(ctx);	 /* Automatically adds stream to ctx */
      if(s)
	UTL_thread_create(&lastthread, incubate_stream, (void *) s);
    }
  }
}

#endif // INCUBATOR

PUBLIC int
COM_select_v(ctx, ret, streams, nstreams, flags)
     COM_CTX * ctx;
     void ** ret;
     COM_STREAM ** streams;
     int nstreams;
     int flags;
{
  unsigned int me;
  int asock, fd;
  int r, i, n, cseq = 0, rv = COM_SELECTED_DONE, isuser;
  int mintime, mintime_all = 0;
  fd_set readfds;
  fd_set writefds;
  COM_STREAM * rs;
  COM_STREAM * s;
  COM_STREAM * t;
  COM_INPUT * ip;
  struct timeval tm;
  UTL_TIME gclock;
  pid_t mypid;
  int lk1 = 0;
  int aport;

  me = UTL_thread_current_id();
  if(flags & COM_SELECT_ACCEPT) {
    asock = ctx->accept_socket;
    aport = flags & COM_ACCEPT_MASK;
  }
  else {
    asock = -1;
    aport = 0;
  }
  for(i = 0; i < nstreams; ++i)
    if(streams[i] != NULL) {
      if(streams[i]->state == COM_STATUS_NOCONNECTION)
	UTL_Log("COM_select_v", LOG_WARNING, "Selecting on closed stream");
      UTL_mutex_lock(&streams[i]->tlock);
      if(streams[i]->owned)
	UTL_Log("COM_select_v", LOG_WARNING, "Multiple selects on same stream");
      ++(streams[i]->owned);
      UTL_mutex_unlock(&streams[i]->tlock);
    }

  COM_mark(&gclock);

  mypid = getpid();
  while(1) {
    if(lk1 == 0
#ifdef INCUBATOR
       && (flags & COM_SELECT_ACCEPT)
#endif
       ) {
      /* Must enter with lock1 */
      UTL_mutex_lock(&ctx->slock);
      lk1 = 1;
    }
    FD_ZERO(&readfds);
    FD_ZERO(&writefds);
    r = 0;
    if(asock > 0) {
      r = asock + 1;
      FD_SET(asock, &readfds);
    }
    n = 0;
    mintime = -1;
    t = ctx->streams;
    rs = NULL;
    for(;;) {
      if(
#ifdef INCUBATOR
	 !(flags & COM_SELECT_ACCEPT) &&
#endif
	 n < nstreams) {
	s = streams[n++];
	if(s == NULL)
	  continue;
	isuser = 1;
      }
      else if(t) {
	s = t;
	t = COM_next(ctx, t);
#ifdef INCUBATOR
	if(!(flags & COM_SELECT_ACCEPT) && !s->connecting)
	  continue;
#endif
	if(s->state == COM_STATUS_CONNECTED && (flags & COM_SELECT_ACCEPT) &&
	   !s->connecting && s->accept_state == COM_AST_SIGNALED &&
	   (aport == COM_ACCEPT_ANY || aport == COM_INFO_GET_PORT(s->peerid))) {
	  s->accept_state = COM_AST_ACCEPTED;
	  if(ret)
	    *ret = (void *) s;
	  rv = COM_SELECTED_ACCEPT;
	  goto finish;
	}
	if(s->owned || s->iowner != me)
	  continue;
	isuser = 0;
      }
      else
	break;

      if(s->owner_pid != mypid || s->state == COM_STATUS_NOCONNECTION)
	continue;

      if(s->usemarker &&
	 ((i = s->timeoutms - COM_elapsedms(&s->marker)) < mintime || mintime < 0))
	mintime = (i < 0) ? 0 : i;

      if((flags & COM_SELECT_TIMEOUT) && s->timeoutms > 0 &&
	 (mintime_all <= 0 || s->timeoutms < mintime_all))
	mintime_all = s->timeoutms;

      fd = COM_get_fd(s);
      if(s->state < COM_STATUS_CONNECTED && s->state > COM_STATUS_INHANDSHAKE) {
	while(COM_packet_pending(s, NULL, NULL) > 0 &&
	      s->state != COM_STATUS_CONNECTED &&
	      s->state != COM_STATUS_NOCONNECTION) {
	  if((i = COM_advance_state(s, COM_SELECTED_READ)) <= 0)
	    break;
	}
      }

      if(s->state == COM_STATUS_CONNECTED) {
	if((flags & COM_SELECT_ACCEPT) && s->accept_state == COM_AST_SIGNALED && !s->connecting &&
	   (aport == COM_ACCEPT_ANY || aport == COM_INFO_GET_PORT(s->peerid))) {
	  s->accept_state = COM_AST_ACCEPTED;
	  if(ret)
	    *ret = (void *) s;
	  rv = COM_SELECTED_ACCEPT;
	  goto finish;
	}
	if(isuser && (flags & COM_SELECT_READ)) {
	  if(COM_packet_pending(s, NULL, NULL) > 0) {
	    if(rs == NULL || s->rseq < cseq) {
	      rs = (void *) s;
	      cseq = s->rseq;
	      rv = COM_SELECTED_READ;
	    }
	  }
	  FD_SET(fd, &readfds);
	  if(fd + 1 > r)
	    r = fd + 1;
	}
	if(isuser && (flags & COM_SELECT_WRITE) && COM_write_pending(s) == 0) {
	  if(rs == NULL || s->rseq < cseq) {
	    rs = (void *) s;
	    cseq = s->rseq;
	    rv = COM_SELECTED_WRITE;
	  }
	}
      }
      else if(s->state == COM_STATUS_NOCONNECTION) {
	if((flags & COM_SELECT_CLOSE) && isuser) {
	  if(ret)
	    *ret = (void *) s;
	  rv = COM_SELECTED_CLOSE;
	  goto finish;
	}
      }
      else {
	i = COM_expect(s);
	if(i & COM_SELECT_READ)
	  FD_SET(fd, &readfds);
	if(i & COM_SELECT_WRITE)
	  FD_SET(fd, &writefds);
	if(i && fd + 1> r)
	  r = fd + 1;
      }
      if(COM_ioready(s) && COM_write_pending(s) > 0) {
	FD_SET(fd, &writefds);
	if(fd + 1 > r)
	  r = fd + 1;
      }
    }

    if(rs) {
      ((COM_STREAM *) rs)->rseq = ctx->rseq++;
      if(ret)
	*ret = rs;
      goto finish;
    }

    if(flags & COM_SELECT_INPUT)
      for(ip = ctx->inputs; ip; ip = ip->next) {
	FD_SET(ip->fd, &readfds);
	if(ip->fd + 1 > r)
	  r = ip->fd + 1;
      }

    if(r == 0) {
      rv = COM_SELECTED_DONE;
      goto finish;
    }

    if(mintime >= 0 || ((flags & COM_SELECT_TIMEOUT) && mintime_all > 0)) {
      if(flags & COM_SELECT_TIMEOUT)
	i = mintime_all - COM_elapsedms(&gclock);
      else
	i = 0;
      if(i < 0)
	i = 0;
      else if(mintime >= 0 && (i == 0 || i > mintime))
	i = mintime;

      tm.tv_sec = (i + FUDGE_MS) / 1000;
      tm.tv_usec = 1000 * ((i + FUDGE_MS) % 1000);
    }
    else
      i = -1;

#ifdef INCUBATOR
    if(!(flags & COM_SELECT_ACCEPT)) {
      UTL_mutex_unlock(&ctx->slock);
      lk1 = 0;
    }
    else if(i < 0) {
      UTL_cond_wait(&ctx->condnew, &ctx->slock);
      continue;
    }
    else {
      r = UTL_cond_wait_timed(&ctx->condnew, &ctx->slock, i + FUDGE_MS);
      if(r == RC_TIMEOUT && COM_elapsedms(&gclock) >= mintime_all) {
	rv = COM_SELECTED_TIMEOUT;
	goto finish;
      }
      continue;
    }
#endif
    if(lk1) {
      UTL_mutex_unlock(&ctx->slock);
      lk1 = 0;
    }

    if(i < 0)
      r = select(r, &readfds, &writefds, NULL, NULL);
    else
      r = select(r, &readfds, &writefds, NULL, &tm);

    if(r < 0) {
      COM_handle_errno("COM_select");
      rv = COM_SELECTED_ERROR;
      goto finish;
    }
    else if(r == 0 && (flags & COM_SELECT_TIMEOUT) &&
	    COM_elapsedms(&gclock) >= mintime_all) {
      rv = COM_SELECTED_TIMEOUT;
      goto finish;
    }

    n = 0;
    t = ctx->streams;
    for(;;) {
      if(
#ifdef INCUBATOR
	 !(flags & COM_SELECT_ACCEPT) &&
#endif
	 n < nstreams) {
	s = streams[n++];
	if(s == NULL)
	  continue;
	isuser = 1;
      }
      else if(t) {
	s = t;
	t = COM_next(ctx, t);
#ifdef INCUBATOR
	if(!(flags & COM_SELECT_ACCEPT) && !s->connecting)
	  continue;
#endif
	if(s->owned || s->iowner != me)
	  continue;
	isuser = 0;
      }
      else
	break;

      if(s->owner_pid != mypid)
	continue;

      if(s->state == COM_STATUS_NOCONNECTION)
	continue;

      /* Timed out? */
      if(r == 0 && !FD_ISSET(COM_get_fd(s), &readfds) && !FD_ISSET(COM_get_fd(s), &writefds) && s->usemarker && COM_elapsedms(&s->marker) >= s->timeoutms) {
	UTL_Log("COM_select_v", LOG_INFO, "Stream Timeout");
	COM_shutdown(s);
	continue;
      }

      if(FD_ISSET(COM_get_fd(s), &readfds)) {
	FD_CLR(COM_get_fd(s), &readfds);
	if(COM_ioready(s)) {
	  if((i = COM_flush_input(s)) == 0) {
	    /* End of file -- this is the only way to detect
	     * a read EOF in non-blocking mode. */
	    UTL_Log("COM_select_v", LOG_DEBUG, "Shutting down fd %d", COM_get_fd(s));
	    COM_shutdown(s);
	  }
	  else if(i < 0) {
	    COM_shutdown(s);
	  }
	}
	if(s->state != COM_STATUS_CONNECTED) {
	  if(COM_advance_state(s, COM_SELECTED_READ) < 0) {
	    COM_shutdown(s);
	  }
	}
      }
      if(s->state == COM_STATUS_NOCONNECTION)
	continue;
      if(FD_ISSET(COM_get_fd(s), &writefds)) {
	FD_CLR(COM_get_fd(s), &writefds);
	if(s->state == COM_STATUS_CONNECTED) {
	  COM_flush(s);
	}
	else {
	  if(COM_ioready(s) && COM_write_pending(s) > 0)
	    COM_flush(s);
	  if(COM_advance_state(s, COM_SELECTED_WRITE) < 0) {
	    COM_shutdown(s);
	  }
	}
      }
    }

    if(asock > 0 && FD_ISSET(asock, &readfds))
      COM_internal_accept(ctx);	 /* Automatically adds stream to ctx */

    /* Might want to use cseq/rseq protocol here to allocate fairly */
    if(flags & COM_SELECT_INPUT)
      for(ip = ctx->inputs; ip; ip = ip->next)
	if(FD_ISSET(ip->fd, &readfds)) {
	  if(ret)
	    *ret = (void *) ip;
	  rv = COM_SELECTED_INPUT;
	  goto finish;
	}
  }

finish:
  if(lk1)
    UTL_mutex_unlock(&ctx->slock);
  for(i = 0; i < nstreams; ++i)
    if(streams[i] != NULL) {
      --(streams[i]->owned);
    }
  return rv;
}

PUBLIC int
COM_select(ctx, ret, flags)
     COM_CTX * ctx;
     void ** ret;
     int flags;
{
  unsigned int me;
  COM_STREAM * s;
  COM_STREAM ** vec;
  int i;

  me = UTL_thread_current_id();
  i = 0;
  UTL_mutex_lock(&ctx->llock);
  for(s = ctx->streams; s; s = s->next)
    if(s->state > COM_STATUS_NOCONNECTION && s->owned == 0 && s->iowner == me
#ifdef INCUBATOR
       && (flags & COM_SELECT_ACCEPT) == 0
#endif
       )
      ++i;

  if(i == 0)
    vec = NULL;
  else if((vec = malloc(i * sizeof(COM_STREAM *))) == NULL) {
    UTL_mutex_unlock(&ctx->llock);
    return COM_SELECTED_ERROR;
  }
  else {
    i = 0;
    for(s = ctx->streams; s; s = s->next)
      if(s->state > COM_STATUS_NOCONNECTION && s->owned == 0 && s->iowner == me
#ifdef INCUBATOR
       && (flags & COM_SELECT_ACCEPT) == 0
#endif
	 )
	vec[i++] = s;
  }
  UTL_mutex_unlock(&ctx->llock);

  i = COM_select_v(ctx, ret, vec, i, flags);
  if(vec)
    free(vec);
  return i;
}

PRIVATE int
COM_write(b, buf, count)
     BUFFER * b;
     char * buf;
     unsigned int count;
{
  if(b == NULL) {
    COM_handle_other("COM_write", "Buffer closed");
    return -1;
  }
  return buffer_append(b, buf, count);
}

PRIVATE int
COM_read(s, buf, count)
     COM_STREAM * s;
     char * buf;
     unsigned int count;
{
  int i;

  while(buffer_length(s->rbuf) < count)
    if((i = COM_flush_input(s)) <= 0)
      return i;

  return buffer_get(s->rbuf, buf, count);
}

PUBLIC int
COM_write_pending(s)
     COM_STREAM * s;
{
  return buffer_length(s->wbuf);
}

PUBLIC int
COM_flush(s)
     COM_STREAM * s;
{
  char buf[1024];
  int i, r, ret = 0;

  if(s->state == COM_STATUS_NOCONNECTION) {
    COM_handle_other("COM_flush", "Connection closed");
    return -1;
  }
  if(s->stream) {
    if(!COM_ioready(s)) {
      if((r = COM_advance_state(s, COM_SELECTED_UNKNOWN)) <= 0)
	return r;
      if(s->state == COM_STATUS_NOCONNECTION)
	return -1;
      if(!COM_ioready(s))
	return 0;
    }

    UTL_mutex_lock(&s->tlock);
    while((i = buffer_peek(s->wbuf, 0, buf, sizeof(buf))) > 0) {
#ifdef SLOW_NET	/* not defined - default; defined - simulate a slow network */
      int j;
      for(j = 0; j < i; ++j) {
	if((r = COM_transport_write(s->trans, s->stream, &buf[j], 1)) < 0) {
	  COM_transport_handle_error(s->trans, "COM_flush");
	  ret = -1;
	  goto done;
	}
	else if(r > 0)
	  buffer_get(s->wbuf, NULL, 1);
	sleep(1);
      }
#else
      /* TODO: timeouts on blocking writes */
      if((r = COM_transport_write(s->trans, s->stream, buf, i)) < 0) {
	if(!(s->flags & COM_FLAG_NONBLOCK) ||
	   (
#ifdef WIN32
	   WSAGetLastError() != WSAECONNABORTED
#else
	   errno != EAGAIN
#endif
#ifdef WIN32
	   && WSAGetLastError() != WSAEWOULDBLOCK
#else
	   && errno != EINPROGRESS
#endif
	   )) {
	  COM_handle_errno("COM_flush");
	  COM_transport_handle_error(s->trans, "COM_flush");
	  ret = -1;
	  goto done;
	}
	else
	{
	  ret = 0;
	  goto done;
	}
      }
      else if(r > 0) {
	buffer_get(s->wbuf, NULL, r);
#ifdef STATS
	s->stats.b_write += r;
#endif
      }
#endif
    }
    ret = 1;
  }
  else
    return 0;

done:
  UTL_mutex_unlock(&s->tlock);
  return ret;
}

PRIVATE int
COM_recv_int(s, np)
     COM_STREAM * s;
     int * np;
{
  unsigned char bytes[4];
  int i;

  if((i = COM_read(s, bytes, sizeof(bytes))) <= 0)
    return i;
  *np = bytes2int(bytes);
  return 1;
}

PRIVATE int
COM_send_int(b, n)
     BUFFER * b;
     int n;
{
  unsigned char bytes[4];

  bytes[0] = (n >> 24) & 0xff;
  bytes[1] = (n >> 16) & 0xff;
  bytes[2] = (n >> 8) & 0xff;
  bytes[3] = n & 0xff;
  return COM_write(b, bytes, 4);
}

PRIVATE int
COM_send_hdr(s, b, type, len)
     COM_STREAM * s;
     BUFFER * b;
     COM_PKT_T type;
     int len;
{
  unsigned char c;

  c = HEADER_BYTE;
  if(COM_write(b, &c, 1) < 0)
    return -1;
  c = type;
  if(COM_write(b, &c, 1) < 0)
    return -1;

  switch(type) {
  case COM_PKT_BIGNUM:
  case COM_PKT_STRING:
  case COM_PKT_BYTES:
#ifdef STATS
    ++s->stats.p_write;
#endif
    return COM_send_int(b, len);
  case COM_PKT_BYTE:
  case COM_PKT_SHORT:
  case COM_PKT_INT:
#ifdef STATS
    ++s->stats.p_write;
#endif
    return 1;
  default:
    return 0;
  }
}

PRIVATE int
COM_recv_hdr(s, tp, lenp)
     COM_STREAM * s;
     COM_PKT_T * tp;
     int * lenp;
{
  unsigned char c;
  int i;

  if((i = COM_read(s, &c, 1)) < 1)
    return i;

  if(c != HEADER_BYTE) {
    COM_handle_other("COM_recv_hdr", "Invalid header byte");
    return -1;
  }

  if((i = COM_read(s, &c, 1)) < 1)
    return i;

  *tp = c;
  switch(*tp) {
  case COM_PKT_BIGNUM:
  case COM_PKT_STRING:
  case COM_PKT_BYTES:
    if(lenp == NULL) {
      COM_handle_other("COM_recv_hdr", "Length pointer expected");
      return -1;	/* Should push back header */
    }
    else {
#ifdef STATS
      ++s->stats.p_read;
#endif
      return COM_recv_int(s, lenp);
    }
  default:
#ifdef STATS
    ++s->stats.p_read;
#endif
    if(lenp)
      *lenp = 0;
    return 1;
  }
}

PRIVATE int
COM_wait_ext_read(s)
     COM_STREAM * s;
{
  int x;

  /* Uncomment this if you want COM_read_* to use nonblocking semantics. */
  /*
  if(COM_usermode(s))
    return 1;
  */
  x = COM_select_v(s->ctx, NULL, &s, 1, COM_SELECT_READ | COM_SELECT_CLOSE);
  if(x == COM_SELECTED_READ)
    return 1;
  else if(x == COM_SELECTED_CLOSE)
    return 0;
  else
    return -1;
}

PRIVATE int
COM_read_int_internal(s, np)
     COM_STREAM * s;
     int * np;
{
  COM_PKT_T t;
  int i;

  if((i = COM_recv_hdr(s, &t, NULL)) <= 0)
    return i;
  if(t != COM_PKT_INT) {
    COM_handle_other("COM_read_int", "Integer packet expected");
    return -1;
  }
  return COM_recv_int(s, np);
}

PUBLIC int
COM_read_int(s, np)
     COM_STREAM * s;
     int * np;
{
  int x;

  x = COM_wait_ext_read(s);
  if(x > 0)
    return COM_read_int_internal(s, np);
  else
    return x;
}

PRIVATE int
COM_write_int_helper(s, b, n)
     COM_STREAM * s;
     BUFFER * b;
     int n;
{
  int i;

  if(s->state == COM_STATUS_NOCONNECTION) {
    COM_handle_other("COM_write_int", "Stream closed");
    return -1;
  }
  if((i = COM_send_hdr(s, b, COM_PKT_INT, 0)) <= 0)
    return i;

  i = COM_send_int(b, n);
  if(!(s->flags & COM_FLAG_BUFFERED) && b == s->wbuf)
    if(COM_flush(s) < 0)
      return -1;
  return i;
}

PRIVATE int
COM_write_int_internal(s, n)
     COM_STREAM * s;
     int n;
{
  return COM_write_int_helper(s, s->wbuf, n);
}

PUBLIC int
COM_write_int(s, n)
     COM_STREAM * s;
     int n;
{
  return COM_write_int_helper(s, COM_usermode(s) ? s->wbuf : s->tbuf, n);
}

PRIVATE int
COM_read_string_internal(s, str)
     COM_STREAM * s;
     char * str;
{
  COM_PKT_T t;
  int i, len;

  if((i = COM_recv_hdr(s, &t, &len)) <= 0)
    return i;
  if(t != COM_PKT_STRING) {
    COM_handle_other("COM_read_string", "String packet expected");
    return -1;
  }
  len = COM_read(s, str, len);
  if(len >= 0)
    str[len] = '\0';
  return len;
}

PUBLIC int
COM_read_string(s, str)
     COM_STREAM * s;
     char * str;
{
  int x;

  x = COM_wait_ext_read(s);
  if(x > 0)
    return COM_read_string_internal(s, str);
  else
    return x;
}

PUBLIC int
COM_write_string_helper(s, b, str)
     COM_STREAM * s;
     BUFFER * b;
     char * str;
{
  int i;

  if(s->state == COM_STATUS_NOCONNECTION) {
    COM_handle_other("COM_write_string", "Stream closed");
    return -1;
  }
  if((i = COM_send_hdr(s, b, COM_PKT_STRING, strlen(str))) < 0)
    return i;

  i = COM_write(b, str, strlen(str));
  if(!(s->flags & COM_FLAG_BUFFERED) && b == s->wbuf)
    if(COM_flush(s) < 0)
      return -1;
  return i;
}

PRIVATE int
COM_write_string_internal(s, str)
     COM_STREAM * s;
     char * str;
{
  return COM_write_string_helper(s, s->wbuf, str);
}

PUBLIC int
COM_write_string(s, str)
     COM_STREAM * s;
     char * str;
{
  return COM_write_string_helper(s, COM_usermode(s) ? s->wbuf : s->tbuf, str);
}

PRIVATE int
COM_read_bytes_internal(s, bytes)
     COM_STREAM * s;
     unsigned char * bytes;
{
  COM_PKT_T t;
  int i, len;

  if((i = COM_recv_hdr(s, &t, &len)) <= 0)
    return i;
  if(t != COM_PKT_BYTES) {
    COM_handle_other("COM_read_bytes", "Bytes packet expected");
    return -1;
  }
  return COM_read(s, bytes, len);
}

PUBLIC int
COM_read_bytes(s, bytes)
     COM_STREAM * s;
     unsigned char * bytes;
{
  int x;

  x = COM_wait_ext_read(s);
  if(x > 0)
    return COM_read_bytes_internal(s, bytes);
  else
    return x;
}

PRIVATE int
COM_write_bytes_helper(s, b, bytes, len)
     COM_STREAM * s;
     BUFFER * b;
     unsigned char * bytes;
     int len;
{
  int i;

  if(s->state == COM_STATUS_NOCONNECTION) {
    COM_handle_other("COM_write_bytes", "Stream closed");
    return -1;
  }
  if((i = COM_send_hdr(s, b, COM_PKT_BYTES, len)) <= 0)
    return i;

  i = COM_write(b, bytes, len);
  if(!(s->flags & COM_FLAG_BUFFERED) && b == s->wbuf)
    if(COM_flush(s) < 0)
      return -1;
  return i;
}

PRIVATE int
COM_write_bytes_internal(s, bytes, len)
     COM_STREAM * s;
     unsigned char * bytes;
     int len;
{
  return COM_write_bytes_helper(s, s->wbuf, bytes, len);
}

PUBLIC int
COM_write_bytes(s, bytes, len)
     COM_STREAM * s;
     unsigned char * bytes;
     int len;
{
  return COM_write_bytes_helper(s, COM_usermode(s) ? s->wbuf : s->tbuf,
				bytes, len);
}

PRIVATE int
COM_read_bn_internal(s, num)
     COM_STREAM * s;
     BIGNUM * num;
{
  COM_PKT_T t;
  int i, len;
  unsigned char * temp;

  if((i = COM_recv_hdr(s, &t, &len)) <= 0)
    return i;
  if(t != COM_PKT_BIGNUM) {
    COM_handle_other("COM_read_bn", "Bignum packet expected");
    return -1;
  }
  if((temp = malloc(len)) == NULL) {
    COM_handle_other("COM_read_bn", "Out of memory");
    return -1;
  }

  if(COM_read(s, temp, len) != len) {
    free(temp);
    COM_handle_other("COM_read_bn", "Bignum packet length incorrect");
    return -1;
  }
  BN_bin2bn(temp + 1, len - 1, num);
  if(*temp & 0x80)
    num->neg = 1;
  else
    num->neg = 0;
  free(temp);
  return 1;
}

PUBLIC int
COM_read_bn(s, num)
     COM_STREAM * s;
     BIGNUM * num;
{
  int x;

  x = COM_wait_ext_read(s);
  if(x > 0)
    return COM_read_bn_internal(s, num);
  else
    return x;
}

PRIVATE int
COM_write_bn_helper(s, b, num)
     COM_STREAM * s;
     BUFFER * b;
     BIGNUM * num;
{
  int i;
  unsigned char * temp;

  if(s->state == COM_STATUS_NOCONNECTION) {
    COM_handle_other("COM_write_bn", "Stream closed");
    return -1;
  }
  i = BN_num_bytes(num);
  if((temp = malloc(i + 1)) == NULL) {
    COM_handle_other("COM_write_bn", "Out of memory");
    return -1;
  }

  if((i = COM_send_hdr(s, b, COM_PKT_BIGNUM, i + 1)) <= 0)
    return i;

  if(i > 0 && temp) {
    i = BN_bn2bin(num, temp + 1);
    if(num->neg)
      *temp = 0x80;
    else
      *temp = 0;
    i = COM_write(b, temp, i + 1);
    free(temp);
  }
  if(!(s->flags & COM_FLAG_BUFFERED) && b == s->wbuf)
    if(COM_flush(s) < 0)
      return -1;
  return 1;
}

PRIVATE int
COM_write_bn_internal(s, num)
     COM_STREAM * s;
     BIGNUM * num;
{
  return COM_write_bn_helper(s, s->wbuf, num);
}

PUBLIC int
COM_write_bn(s, num)
     COM_STREAM * s;
     BIGNUM * num;
{
  return COM_write_bn_helper(s, COM_usermode(s) ? s->wbuf : s->tbuf, num);
}

PUBLIC void
COM_close(s)
     COM_STREAM * s;
{
  COM_shutdown(s);
  buffer_free(s->wbuf);
  buffer_free(s->rbuf);
  if(s->tbuf)
    buffer_free(s->tbuf);
  UTL_mutex_destroy(&s->tlock);
  COM_delete(s->ctx, s);
  free(s);
}
