/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/* The Ring Buffer data structure (C implementation) */

#include <stdlib.h>
#include "rbuffer.h"

BUFFER *
buffer_new()
{
  return buffer_create(DEFAULT_BUFSIZE);
}

BUFFER *
buffer_create(size)
     unsigned size;
{
  BUFFER * buf;

  if((buf = malloc(sizeof(BUFFER))) == NULL)
    return NULL;

  if((buf->data = malloc(size)) == NULL) {
    free(buf);
    return NULL;
  }

  buf->size = size;
  buf->head = buf->len = 0;
  return buf;
}

void
buffer_free(buf)
     BUFFER * buf;
{
  free(buf->data);
  free(buf);
}

unsigned
buffer_length(buf)
     BUFFER * buf;
{
  return buf->len;
}

int
buffer_peek(buf, offset, bytes, n)
     BUFFER * buf;
     unsigned offset;
     unsigned char * bytes;
     unsigned n;
{
  int i, j;

  if(offset >= buf->len)
    return 0;

  if(n + offset > buf->len)
    n = buf->len - offset;

  j = buf->head + offset;
  if(j >= buf->size)
    j -= buf->size;

  i = buf->size - j;
  if(bytes) {
    if(n > i) {
      memcpy(bytes, buf->data + j, i);
      memcpy(bytes + i, buf->data, n - i);
    }
    else
      memcpy(bytes, buf->data + j, n);
  }

  return n;
}

static int
buffer_poke(buf, offset, bytes, n)
     BUFFER * buf;
     unsigned offset;
     unsigned char * bytes;
     unsigned n;
{
  int i, j;

  j = buf->head + offset;
  if(j >= buf->size)
    j -= buf->size;

  i = buf->size - j;
  if(n > i) {
    memcpy(buf->data + j, bytes, i);
    memcpy(buf->data, bytes + i, n - i);
  }
  else
    memcpy(buf->data + j, bytes, n);

  return n;
}

static int
buffer_expand(buf, newsize)
     BUFFER * buf;
     unsigned int newsize;
{
  unsigned char * p;

  if((p = malloc(newsize)) == NULL)
    return -1;

  buffer_peek(buf, 0, p, buffer_length(buf));
  free(buf->data);
  buf->data = p;
  buf->head = 0;
  buf->size = newsize;
  return 0;
}

static int
buffer_accommodate(buf, delta)
     BUFFER * buf;
     unsigned int delta;
{
  int i = buf->size;

  if(buf->len + delta > i) {
    while(buf->len + delta > i)
      i *= 2;
    if(buffer_expand(buf, i) < 0)
      return -1;
  }
  return i;
}

int
buffer_append(buf, bytes, n)
     BUFFER * buf;
     unsigned char * bytes;
     unsigned n;
{
  int i;

  if(buffer_accommodate(buf, n) < 0)
    return 0;

  i = buffer_poke(buf, buf->len, bytes, n);
  buf->len += i;
  return i;
}

int
buffer_get(buf, bytes, n)
     BUFFER * buf;
     unsigned char * bytes;
     unsigned n;
{
  int i;

  i = buffer_peek(buf, 0, bytes, n);
  buf->len -= i;
  buf->head += i;
  if(buf->head >= buf->size)
    buf->head -= buf->size;
  return i;
}

int
buffer_unget(buf, bytes, n)
     BUFFER * buf;
     unsigned char * bytes;
     unsigned n;
{
  if(buffer_accommodate(buf, n) < 0)
    return 0;

  if(buf->head < n)
    buf->head += buf->size - n;
  else
    buf->head -= n;
  buf->len += n;

  return buffer_poke(buf, 0, bytes, n);
}
