/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */



/*------------------------------------------------------------------------*/
/*--									--*/
/*--	CTL_ConFromAdm.c						--*/
/*--									--*/
/*--	Handle Connect from Adm program.				--*/
/*------------------------------------------------------------------------*/

#include "ITTC.h"


int CTL_Respond(COM_STREAM *stream, int opcode, char *msg) {

  if (COM_write_string(stream, "RESPONSE") < 0) {
    UTL_Log("CTL_Respond", LOG_ERROR, "Unable to send header");
    return RC_ERROR;
  }

  if (COM_write_int(stream, opcode) < 0) {
    UTL_Log("CTL_Respond", LOG_ERROR, "Unable to send opcode");
    return RC_ERROR;
  }

  if (COM_write_string(stream, msg) < 0) {
    UTL_Log("CTL_Respond", LOG_ERROR, "Unable to send '%s'", msg);
    return RC_ERROR;
  }

  return RC_OK;
}




int CTL_SendPublicKey(CTL_CTX *ctx, COM_STREAM *stream, int id, int version) {
  RSA_SHARE *key = NULL;
  BIGNUM *temp = NULL;
  BN_CTX *bnCtx = NULL;
  int i;
  int k;
  int err = RC_ERROR;


  /* Load the key */
  key = KEY_CTX_getbyidver(ctx->keyctx, id, version);
  if (key == NULL) {
    UTL_Log("CTL_SendPublicKey", LOG_INFO, "Error finding private share.");
    goto error;
  }

  /* Initialize BIGNUM stuff */
  temp = BN_new();
  if (temp == NULL) {
    UTL_Log("CTL_SendPublicKey", LOG_INFO, "Error in BN_new().");
    goto error;
  }

  bnCtx = BN_CTX_new();
  if (bnCtx == NULL) {
    UTL_Log("CTL_SendPublicKey", LOG_INFO, "Error in BN_CTX_new().");
    goto error;
  }

  /* Send N, e, and g */
  if (COM_write_bn(stream, KEY_RSA_share_n(key)) <= 0) {
    UTL_Log("CTL_SendPublicKey", LOG_INFO, "Error sending N to ADM.");
    goto error;
  }
  if (COM_write_bn(stream, KEY_RSA_share_e(key)) <= 0) {
    UTL_Log("CTL_SendPublicKey", LOG_INFO, "Error sending e to ADM.");
    goto error;
  }
  if (COM_write_bn(stream, KEY_RSA_share_d(key, 0)) <= 0) {
    UTL_Log("CTL_SendPublicKey", LOG_INFO, "Error sending g to ADM.");
    goto error;
  }
  UTL_Log("CTL_SendPublicKey", LOG_INFO, "Sent N, e, and g to ADM.");

  /* Send the number of g^d's */
  k = KEY_RSA_share_k(key);   /* k is the number of g^d's */
  if (k < 0) {
    UTL_Log("CTL_SendPublicKey", LOG_INFO, "Error finding number of shares.");
    goto error;
  }
    
  if (!BN_set_word(temp, k-1)) {
    UTL_Log("CTL_SendPublicKey", LOG_INFO, "Error in BN_set_word().");
    BN_clear_free(temp);
    goto error;
  }

  if (COM_write_bn(stream, temp) < 0) {
    UTL_Log("CTL_SendPublicKey", LOG_INFO, "Unable to send number of g^d's");
    BN_clear_free(temp);
    goto error;
  }

  UTL_Log("CTL_SendPublicKey", LOG_FATAL, "Sending %d shares to ADM.", k-1);

  /* Calculate, send the g^d's */
  for (i=1; i<k; i++) {

    if (!IT_mod_exp_NEG(temp, KEY_RSA_share_d(key, 0), KEY_RSA_share_d(key, i), 
			KEY_RSA_share_n(key), bnCtx)) goto error;

    if (COM_write_bn(stream, temp) < 0) {
      UTL_Log("CTL_SendPublicKey", LOG_INFO, "Unable to send g^d number %d", i);
      goto error;
    }

    UTL_Log("CTL_SendPublicKey", LOG_FATAL, "Sent share %d", i);

  }

  err = RC_OK;

 done:

  BN_clear_free(temp);
  BN_CTX_free(bnCtx);
  return err;

 error:
  UTL_Log("CTL_SendPublicKey", LOG_FATAL, "Error");
  err = RC_ERROR;
  goto done;

  return RC_ERROR;
}

  

/*========================================================================*/
/*========================================================================*/


int CTL_ConFromAdm(CTL_CTX *ctx, COM_STREAM *stream)
{

  int opcode, keySize, numServers, threshold, err;
  int id, coalition;
  int rc = RC_ERROR;
  int version, vmin, vmax;
  int send;
  char data[128];
  char pwd[128];
  COM_STREAM *tStream = NULL;
  char txt[256];
  BIGNUM *N = NULL;
  BIGNUM *e = NULL;
  int n;
  int t;
  int remove = FALSE;


  do {

    // Wait for request from peer on current stream.
    //--------------------------------------------------
    
    err = COM_select_v(ctx->comctx, (void **) &tStream, &stream, 1, 
		       COM_SELECT_READ | COM_SELECT_TIMEOUT | COM_SELECT_CLOSE);
    stream = tStream;
    
    /* Check for a timeout or a closed stream */
    if (err == COM_SELECTED_TIMEOUT) {
      UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Timeout.");
      goto error;
    }

    /* Check for a closed stream */
    if (err == COM_SELECTED_CLOSE) {
      UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Stream closed");
      break;
    }

    /* Check for an error */
    if (err != COM_SELECTED_READ) {
      UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Error in select.");
      goto error;
    }
    

    // Read the command header
    //--------------------------------------------------

    if (COM_read_string(stream, txt) < 0) {
      UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to read header");
      CTL_Respond(stream, opcode, "ERROR: Unable to read header");
      goto error;
    }
    if (strcmp(txt, "COMMAND") != 0) {
      UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Incorrect header");
      CTL_Respond(stream, opcode, "ERROR: Incorrect header");
      goto error;
    }
    
    
    // Read the opcode.
    //--------------------------------------------------
    
    if (UTL_WaitRead(stream) != RC_OK) goto error;
    if (COM_read_int(stream, &opcode) < 0) {
      UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to read opcode from Admin.");
      CTL_Respond(stream, opcode, "ERROR: Unable to read opcode");
      goto error;
    }
    UTL_Log("CTL_ConFromAdm", LOG_TEST, "opcode = %d", opcode);
    
    
    // If suspended, can only start or shutdown.
    //--------------------------------------------------
    
    if (ctx->status == CTL_SUSPENDED) {
      if ( (opcode != ADM_op_shutdown) && (opcode != ADM_op_start) && (opcode != ADM_op_corrupt) ) {
	CTL_Respond(stream, opcode, "ERROR: Server is suspended");
	goto done;
      }
    }
    
    
    // Administrator is pretending to be a client.
    // ------------------------------------------------    
    if ( (opcode == TLB_OP_APPLY_DECRYPT) ||
	 (opcode == TLB_OP_APPLY_SIG)     ||	
	 (opcode == TLB_OP_APPLY_HASHED) ) {
      
      if (COM_write_string(stream, "RESPONSE") < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to send header");
	goto error;
      }

      if (COM_write_int(stream, opcode) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to send opcode");
	goto error;
      }

      err = NTD_serve(ctx, opcode, stream);
      if (err != RC_OK) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Error in NTD_serve");
	goto error;
      }

      continue;              // Continue the session
    } 
    
    
    // Adminisrator says: start
    // --------------------------------------------------
    
    if (opcode == ADM_op_start) {
      
      // Read admin password.
      if (UTL_WaitRead(stream) != RC_OK) goto error;
      if (COM_read_string(stream, pwd) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to read password from Admin.");
	CTL_Respond(stream, opcode, "ERROR: Unable to read password");
	goto error;
      }
      
      // Set the password for loading and saving keys
      KEY_set_passphrase(pwd);
    
      // Load key shares using admin password.	
      if (ctx->keyctx != NULL) {
	KEY_CTX_close(ctx->keyctx);
      }
      ctx->keyctx = KEY_Init(ctx->utlctx, ctx->serverID);
      
      if (ctx->keyctx == NULL) {
	CTL_Respond(stream, opcode, "ERROR: Can not load any keys");
	goto error;
      }
      
      ctx->status = CTL_ACTIVE;
      CTL_Respond(stream, opcode, "OK");
      
      UTL_Log("CTL_ConFromAdm", LOG_NOTIFY, "Server started");
      UTL_monitor(ctx->utlctx, 0, "Server started");
      
      goto done;
    }
    
    
    // Adminisrator says: suspend.
    // -------------------------------------------------------
    
    if (opcode == ADM_op_suspend) {
      
      ctx->status = CTL_SUSPENDED;
      CTL_Respond(stream, opcode, "OK");

      UTL_monitor(ctx->utlctx, 0, "Server suspended");
      
      goto done;
    }
    
    
    // Adminisrator says: corrupt.
    // -------------------------------------------------------
    
    if (opcode == ADM_op_corrupt) {
      
      ctx->status = CTL_CORRUPT;
      CTL_Respond(stream, opcode, "OK");
      
      UTL_monitor(ctx->utlctx, 0, "Server corrupted");

      goto done;
    }
    
    
    // Adminisrator says: shutdown.
    // -------------------------------------------------------
    
    if (opcode == ADM_op_shutdown) {
      
      CTL_Respond(stream, opcode, "OK");
      
      UTL_Log("CTL_ConFromAdm", LOG_NOTIFY, "Shutting down server");
      UTL_monitor(ctx->utlctx, 0, "Shutting down server");
      
      return RC_CLOSE;
    }
    
    
    // Adminisrator says: generate a new key.
    // ----------------------------------------------------------
    
    if (opcode == ADM_op_genkey) {
      
      // Determine key size
      
      if (UTL_WaitRead(stream) != RC_OK) goto error;
      if (COM_read_string(stream, data) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to read key size from Admin.");
	CTL_Respond(stream, opcode, "ERROR: Problem reading key size");
	goto error;
      }
      
      switch (data[0]) {
      case 'S':
      case 's':
	keySize = STRONG_KEY;
	break;
      case 'N':
      case 'n':
	keySize = NORMAL_KEY;
	break;
      case 'W':
      case 'w':
	keySize = WEAK_KEY;
	break;
      case 'T':
      case 't':
	keySize = TEST_KEY;
	break;
      default:
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Invalid key size to generate.");
	CTL_Respond(stream, opcode, "ERROR: Invalid key size");
	goto error;
      }
      
      if (UTL_WaitRead(stream) != RC_OK) goto error;
      if (COM_read_int(stream, &numServers) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to read number of servers from Admin.");
	CTL_Respond(stream, opcode, "ERROR: Unable to read number of servers");
	goto error;
      }
      
      if (UTL_WaitRead(stream) != RC_OK) goto error;
      if (COM_read_int(stream, &threshold) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to read threshold from Admin.");
	CTL_Respond(stream, opcode, "ERROR: Unable to read threshold");
	goto error;
      }
      
      UTL_monitor(ctx->utlctx, 0, "Generating key. Strength = %s", data);
      UTL_Log("CTL_ConFromAdm", LOG_NOTIFY, "Generating key. Strength = %s", data);
      
      if (COM_write_string(stream, "RESPONSE") < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to send header");
	return RC_ERROR;
      }
      
      if (COM_write_int(stream, opcode) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to send opcode");
	return RC_ERROR;
      }

      err = GEN_GenKey(ctx->serverID, stream, keySize, 0, numServers, threshold,
		       NULL, ctx->utlctx, ctx->keyctx, ctx->comctx);
      
      if (err != RC_OK) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Error in GenKey");
	goto error;
      }

      goto done;
    }
    
    

    // Adminisrator says: refresh
    // --------------------------------------
    
    if (opcode == ADM_op_refresh) {

      if ( (N = BN_new()) == NULL) goto error;
      if ( (e = BN_new()) == NULL) goto error;

      /* Key ID */
      if (UTL_WaitRead(stream) != RC_OK) goto error;
      if (COM_read_int(stream, &id) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to read key ID from Admin.");
	CTL_Respond(stream, opcode, "ERROR: Unable to read key ID");
	goto error;
      }
      
      /* Coalition */
      if (UTL_WaitRead(stream) != RC_OK) goto error;
      if (COM_read_int(stream, &coalition) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to read coalition from Admin.");
	CTL_Respond(stream, opcode, "ERROR: Unable to read coalition");
	goto error;
      }

      /* N */
      if (UTL_WaitRead(stream) != RC_OK) goto error;
      if (COM_read_bn(stream, N) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to read N.");
	CTL_Respond(stream, opcode, "ERROR: Unable to read N");
	goto error;
      }

      /* e */
      if (UTL_WaitRead(stream) != RC_OK) goto error;
      if (COM_read_bn(stream, e) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to read e.");
	CTL_Respond(stream, opcode, "ERROR: Unable to read e");
	goto error;
      }

      /* num parties */
      if (UTL_WaitRead(stream) != RC_OK) goto error;
      if (COM_read_int(stream, &n) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to read num parties from Admin.");
	CTL_Respond(stream, opcode, "ERROR: Unable to read num parties");
	goto error;
      }

      /* threshold */
      if (UTL_WaitRead(stream) != RC_OK) goto error;
      if (COM_read_int(stream, &t) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to read threshold from Admin.");
	CTL_Respond(stream, opcode, "ERROR: Unable to read threshold");
	goto error;
      }

      /* version */
      if (UTL_WaitRead(stream) != RC_OK) goto error;
      if (COM_read_int(stream, &version) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to read version from Admin.");
	CTL_Respond(stream, opcode, "ERROR: Unable to read version");
	goto error;
      }

      err = REF_Refresh((BN_ULONG)id, coalition, N, e, n, t, version,
			ctx->keyctx, ctx->comctx, ctx->utlctx);
      if (err != RC_OK) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Error in Refresh");
	CTL_Respond(stream, opcode, "ERROR: Problem refreshing key");
	goto error;
      }

      CTL_Respond(stream, opcode, "OK");

      UTL_monitor(ctx->utlctx, 0, "Refreshed key 0x%8X", id);

      goto done;
    }
    
    
    // Adminisrator says: get_version
    // --------------------------------------
    
    if (opcode == ADM_op_get_version) {

      if (UTL_WaitRead(stream) != RC_OK) goto error;
      if (COM_read_int(stream, &id) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to read key ID from Admin.");
	CTL_Respond(stream, opcode, "ERROR: Unable to read key ID");
	goto error;
      }

      err = KEY_CTX_getversions(ctx->keyctx, id, &vmin, &vmax);
      if (err != TRUE) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Key does not exist");
	vmin = -1;
	vmax = -1;
      }
      
      if (CTL_Respond(stream, opcode, "OK") != RC_OK) goto error;

      if (COM_write_int(stream, vmin) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to send min version");
	return RC_ERROR;
      }

      if (COM_write_int(stream, vmax) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to send max version");
	return RC_ERROR;
      }

      UTL_monitor(ctx->utlctx, 0, "Key 0x%8X goes from version %d to %d", id, vmin, vmax);

      goto done;
    }
    
    
    // Adminisrator says: version_set
    // --------------------------------------
    
    if (opcode == ADM_op_version_set) {

      if (UTL_WaitRead(stream) != RC_OK) goto error;
      if (COM_read_int(stream, &id) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to read key ID from Admin.");
	CTL_Respond(stream, opcode, "ERROR: Unable to read key ID");
	goto error;
      }

      if (UTL_WaitRead(stream) != RC_OK) goto error;
      if (COM_read_int(stream, &version) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to read key version from Admin.");
	CTL_Respond(stream, opcode, "ERROR: Unable to read key version");
	goto error;
      }

      if (UTL_WaitRead(stream) != RC_OK) goto error;
      if (COM_read_int(stream, &send) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to read 'send'.");
	CTL_Respond(stream, opcode, "ERROR: Unable to read 'send'");
	goto error;
      }

      /* Make sure the good version has been saved to disk */
      err = KEY_CTX_flushidver(ctx->keyctx, id, version);
      if (err != RC_OK) {
	UTL_Log("CTL_ConFromAdm", LOG_NOTIFY, "This will remove all versions of key 0x%8X", id);
	remove = TRUE;
      } else {
	remove = FALSE;
      }

      /* Remove all but the correct key version */
      err = KEY_CTX_collapse_idver(ctx->keyctx, id, version);
      if (err != RC_OK) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Error removing old key versions");
	CTL_Respond(stream, opcode, "ERROR: Problem removing old versions of key");
	goto error;
      }

      /* Send the headers */
      if (CTL_Respond(stream, opcode, "OK") != RC_OK) goto error;

      /* Send the correct public key back to the ADM */
      if (send == ADM_HIGH) {
	err = CTL_SendPublicKey(ctx, stream, id, version);
	if (err != RC_OK) {
	  UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Failure sending public key");
	  goto error;
	}
      }
      
      if (remove) {

	UTL_Log("CTL_ConFromAdm", LOG_NOTIFY, "Key 0x%8X removed", id);
	UTL_monitor(ctx->utlctx, 0, "Key 0x%8X removed", id);

      } else {

	UTL_Log("CTL_ConFromAdm", LOG_NOTIFY, "Version for key 0x%8X set to %d", id, version);
	UTL_monitor(ctx->utlctx, 0, "Version for key 0x%8X set to %d", id, version);

      }

      goto done;
    }
    
    
    // Adminisrator says: get_pub_key
    // --------------------------------------
    
    if (opcode == ADM_op_get_pub_key) {

      if (UTL_WaitRead(stream) != RC_OK) goto error;
      if (COM_read_int(stream, &id) < 0) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Unable to read key ID from Admin.");
	CTL_Respond(stream, opcode, "ERROR: Unable to read key ID");
	goto error;
      }

      /* Figure out which version to use */
      err = KEY_CTX_getversions(ctx->keyctx, id, &vmin, &vmax);
      if (err != TRUE) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Key does not exist");
	CTL_Respond(stream, opcode, "ERROR: Key does not exist");
	rc = RC_NO_KEY;
	goto error;
      }

      /* Send the headers */
      if (CTL_Respond(stream, opcode, "OK") != RC_OK) goto error;

      /* Send the correct public key back to the ADM */
      err = CTL_SendPublicKey(ctx, stream, id, vmax);
      if (err != RC_OK) {
	UTL_Log("CTL_ConFromAdm", LOG_ERROR, "Failure sending public key");
	goto error;
      }
      
      UTL_monitor(ctx->utlctx, 0, "Sent public key to ADM");

      goto done;
    }
    
    
    // Unknown command from administrator.
    // -----------------------------------
    
    UTL_Log("CTL_ConFromAdm", LOG_INFO, "Unknown opcode from administrator: %d", opcode);
    CTL_Respond(stream, opcode, "ERROR: Unknown opcode");
    goto error;
    
  } while (1);



 /* Handle errors, close the stream */
 error:
  UTL_Log("CTL_ConFromAdm", LOG_DEBUG, "Error-handler activated");
  if (N) BN_clear_free(N);
  if (e) BN_clear_free(e);
  if (stream != NULL) COM_close(stream);
  return RC_ERROR;



 /* Close connection correctly, return */
 done:
  if (N) BN_clear_free(N);
  if (e) BN_clear_free(e);
  if (stream != NULL) COM_close(stream);
  return RC_OK;


}

