/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */



/*------------------------------------------------------------------------*/
/*--									--*/
/*--	CTL_ConFromClient.c						--*/
/*--									--*/
/*--	Handle connect from client.					--*/
/*------------------------------------------------------------------------*/

#include "ITTC.h"

typedef struct CTL_thread_args {
  CTL_CTX *ctx;
  COM_STREAM *stream;
} CTL_THREAD_ARGS;



/* This is the new thread that is started */
void CTL_Thread(CTL_THREAD_ARGS *cta) {

  int op;
  CTL_CTX *ctx = NULL;
  COM_STREAM *stream = NULL;
  COM_STREAM *tStream = NULL;
  int err;
  char txt[256];


  /*--------------------------------------------------*/
  /* Initialize                                       */
  /*--------------------------------------------------*/

  UTL_Log("CTL_Thread", LOG_INFO, "Starting thread", cta->stream);

  /* Get data from passed argument */
  ctx = cta->ctx;
  stream = cta->stream;

  /* Prepare for session */
  COM_set_timeout(stream, ctx->timeout);


  /*--------------------------------------------------*/
  /* Start a session:                                 */
  /* Execute commands until the connection is closed  */
  /*--------------------------------------------------*/

  do {

    /* Wait for request from peer on current stream */
    err = COM_select_v(ctx->comctx, (void **) &tStream, &stream, 1, 
		       COM_SELECT_READ | COM_SELECT_TIMEOUT | COM_SELECT_CLOSE);
    if(tStream != stream) {
      UTL_Log("CTL_Thread", LOG_WARNING, "Received event from wrong stream (%08x != %08x)", tStream, stream);
    }
    stream = tStream;
    
    /* Check for a timeout */
    if (err == COM_SELECTED_TIMEOUT) {
      UTL_Log("CTL_Thread", LOG_ERROR, "Timeout");
      break;
    }

    /* Check for a closed stream */
    if ( (err == COM_SELECTED_CLOSE) || (err == COM_SELECTED_DONE) ) {
      UTL_Log("CTL_Thread", LOG_TEST, "Stream closed");
      break;
    }

    /* Check for an error */
    if (err != COM_SELECTED_READ) {
      UTL_Log("CTL_Thread", LOG_ERROR, "Error in select");
      break;
    }
    
    /* Check for the correct command header */
    if (COM_read_string(stream, txt) < 0) {
      UTL_Log("CTL_Thread", LOG_ERROR, "Unable to read header");
      break;
    }
    if (strcmp(txt, "COMMAND") != 0) {
      UTL_Log("CTL_Thread", LOG_ERROR, "Invalid header");
      break;
    }
    
    /* Read the opcode */
    if (UTL_WaitRead(stream) != RC_OK) break;
    if(COM_read_int(stream, &op) <= 0) {
      UTL_Log("CTL_Thread", LOG_ERROR, "Error reading opcode");
      CTL_Respond(stream, op, "ERROR: Unable to read opcode");
      break;
    }
   
    /* Send the header */
    if (COM_write_string(stream, "RESPONSE") < 0) {
      UTL_Log("CTL_Thread", LOG_ERROR, "Unable to send header");
      break;
    }

    /* Send the opcode */
    if (COM_write_int(stream, op) < 0) {
      UTL_Log("CTL_Thread", LOG_ERROR, "Unable to send opcode");
      break;
    }
 
    /* Do the decryption */
    err = NTD_serve(ctx, op, stream);
    if (err != RC_OK) {
      UTL_Log("CTL_Thread", LOG_ERROR, "Error in NTD_serve");
      break;
    }

  } while (1);

  /* Close the stream after the session is complete */
  if (stream != NULL) COM_close(stream);
  if (cta) free(cta);
}


/*------------------------------------------------------------*/
/* CTL_ConFromClient()                                        */
/*   Handles a connection from a client.                      */
/*   Actually starts a thread to do the work                  */
/*------------------------------------------------------------*/

int CTL_ConFromClient(CTL_CTX *ctx, COM_STREAM *stream) {
  UTL_THREAD thread;
  CTL_THREAD_ARGS *cta;

  /* Prepare to pass arguments to thread */
  cta = (CTL_THREAD_ARGS*)malloc(sizeof(CTL_THREAD_ARGS));
  cta->ctx = ctx;
  cta->stream = stream;

  UTL_Log("CTL_ConFromClient", LOG_INFO, "Starting new thread for client decryption");

  /* Cread the thread */
  UTL_thread_create(&thread, (void*)CTL_Thread, (void*)cta);
  UTL_thread_destroy(&thread);    // Only loses pid.

  return RC_OK;
}



