/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */



/*------------------------------------------------------------------------*/
/*--									--*/
/*--	CTL_Main.c							--*/
/*--									--*/
/*--	Main function of share server daemon controller.		--*/
/*------------------------------------------------------------------------*/

#include "ITTC.h"

/*------------------------------------------------------------------------*/
/*--	CTL_Init()							--*/
/*------------------------------------------------------------------------*/

CTL_CTX *CTL_Init(int argc, char *argv[]) {
  struct CTL_ArgsStruc Args;
  int err;
  CTL_CTX *ctx = NULL;

  /* Parse command line arguments */
  err = CTL_ParseArguments(argc, argv, &Args);
  if (Args.HelpFlag) return NULL;
  if (err) {
    fprintf(stderr, "Unable to parse command line.\nUse '-h' to get a list of options.\n\n");
    return NULL;
  }

  /* Create CTL Context */
  if ((ctx = malloc(sizeof(CTL_CTX))) == NULL) {
    fprintf(stderr, "CTL_Init: out of memory.\n\n");
    return NULL;
  }
  ctx->utlctx = NULL;
  ctx->comctx = NULL;
  ctx->keyctx = NULL;
  ctx->status = CTL_SUSPENDED;		/* Start server in suspended mode */

  ctx->serverID = Args.server_num;

  /* Load config file */
  ctx->utlctx = UTL_init_ctx();
  err = UTL_LoadConfigFile(ctx->utlctx, Args.ConfigFile);
  if (err) goto err;

  /* Create log file */
  err = UTL_OpenLogFile(ctx->utlctx, COM_TYPE_SHARE_SERVER, Args.server_num);
  if (err) goto err;
  UTL_MonitorInit(ctx->utlctx, COM_TYPE_SHARE_SERVER, Args.server_num);

  /* Initialize communications package */
  ctx->comctx = COM_init(ctx->utlctx, COM_TYPE_SHARE_SERVER, Args.server_num);
  if (ctx->comctx == NULL) {
    fprintf(stderr, "Unable to initialize communications.\n\n");
    goto err;
  }
  ctx->timeout = UTL_GetIntConfigParam(ctx->utlctx, "Server_Timeout", 10000);
  COM_ctx_set_timeout(ctx->comctx, ctx->timeout);

  /*COM_ctx_set_flags(ctx->comctx, COM_FLAG_BLOCKING_CONNECT);*/
  /*COM_ctx_set_flags(ctx->comctx, COM_FLAG_NONBLOCK);*/

  /* Set KEY context to NULL */
  ctx->keyctx = NULL;

  return ctx;


 err:
  if (ctx == NULL)
    return NULL;
  if (ctx->comctx)
    COM_ctx_close(ctx->comctx);
  if (ctx->utlctx) {
    UTL_CloseLogFile();
    UTL_MonitorClose(ctx->utlctx);
    UTL_free_ctx(ctx->utlctx);
  }
  free(ctx);
  return NULL;
}



/*------------------------------------------------------------------------*/
/*--	CTL_Work()							--*/
/*------------------------------------------------------------------------*/

int CTL_Work(CTL_CTX *ctx)
{
  int err;
  int vport = ITTC_SHARE_SERVER_VPORT;
  COM_INFO con_info;
  COM_STREAM *stream = NULL;


  /*----------------------------------------------*/
  /*    Wait for connection request		  */
  /*----------------------------------------------*/
  
  /* Print out status */
  UTL_Log("CTL_Work", LOG_NOTIFY, "------------------------------------------------------------");
  if (ctx->status == CTL_SUSPENDED) {
    UTL_Log("CTL_Work", LOG_NOTIFY, "Server is suspended");
    UTL_monitor(ctx->utlctx, 0, "Server is suspended");
  } else if (ctx->status == CTL_CORRUPT) {
    UTL_Log("CTL_Work", LOG_NOTIFY, "Waiting for connection (corrupt)");
    UTL_monitor(ctx->utlctx, 0, "Waiting for connection (corrupt)");
  } else {
    UTL_Log("CTL_Work", LOG_NOTIFY, "Waiting for connection");
    UTL_monitor(ctx->utlctx, 0, "Waiting for connection");
  }

  /* Wait for a connection */
  err = COM_select(ctx->comctx, (void **) &stream, COM_SELECT_ACCEPT | COM_SELECT_PORT(vport));
  UTL_Log("CTL_Work", LOG_NOTIFY, "Connection!");

  /* Handle security breach on connecting side */
  // if (err == COMRC_BAD_SEQNUM) {
  //   UTL_Log("CTL_Work", LOG_INFO, "Security breach detected!");
  //   CTL_SecurityBreach(con_type);
  //   continue;
  // }

  /* Ignore the connection if there were errors */
  if (err != COM_SELECTED_ACCEPT) {
    UTL_Log("CTL_Work", LOG_ERROR, "Ignoring connection.");
    if (stream != NULL) COM_close(stream);
    return RC_OK;
  } 

  /* Get connection type */
  err = COM_get_peer_info(stream, &con_info);
  if (err) {
    UTL_Log("CTL_Work", LOG_ERROR, "Error in COM_get_peer_info");
    if (stream != NULL) COM_close(stream);
    return RC_OK;
  }

  /* If suspended, close stream and exit */
  if ( (ctx->status == CTL_SUSPENDED) && (COM_INFO_GET_TYPE(con_info) != COM_TYPE_ADMIN) ) {
    if (stream != NULL) COM_close(stream);
    return RC_OK;
  }

  /*-------------------------------*/
  /*    Handle legal connection    */
  /*-------------------------------*/

  /* Do different things for different connection types */
  switch (COM_INFO_GET_TYPE(con_info)) {
    
    case COM_TYPE_SHARE_SERVER:   	
      UTL_Log("CTL_Work", LOG_INFO, "Connection from share server");
      CTL_ConFromServer(ctx, stream);
      break;
      
      // Connection from client can only be handled after all
      // key shares have been read from disk.  Key shares are
      // read from disk when administrator activates server.
      // Admin password is necessary for decrypting key shares stored
      // on disk.
      
    case COM_TYPE_CLIENT:		 
      UTL_Log("CTL_Work", LOG_INFO, "Connection from client");
      CTL_ConFromClient(ctx, stream);
      break;
      
    case COM_TYPE_ADMIN:
      UTL_Log("CTL_Work", LOG_INFO, "Connection from admin");
      err = CTL_ConFromAdm(ctx, stream);
      if (err == RC_CLOSE) {
	return RC_CLOSE;    // Handle the shutdown command
      }
      break;
    
    default:
      UTL_Log("CTL_Work", LOG_INFO, "Unknown connection type received.");
  }
  
  return RC_OK;
}



/*------------------------------------------------------------------------*/
/*--	CTL_Shutdown()							--*/
/*------------------------------------------------------------------------*/

int CTL_Shutdown(CTL_CTX *ctx) {
  if (ctx == NULL) return RC_ERROR;
  if (ctx->comctx) COM_ctx_close(ctx->comctx);
  if (ctx->keyctx) KEY_CTX_close(ctx->keyctx);
  if (ctx->utlctx) {
    UTL_MonitorClose(ctx->utlctx);
    UTL_free_ctx(ctx->utlctx);
    UTL_CloseLogFile();
  }
  free(ctx);
  return RC_OK;
}



/*------------------------------------------------------------------------*/
/*--	main()								--*/
/*------------------------------------------------------------------------*/

int main(int argc, char *argv[]) {
  CTL_CTX *ctx;
  int rc;

  /* Initialize */
  ITTC_init();
 
  /* Process command line arguments and initialize daemon */
  ctx = CTL_Init(argc, argv);

  if (ctx == NULL)
    exit(1);

  /* Main controller loop */
  do {
    rc = CTL_Work(ctx);
  } while (rc == RC_OK);

  /* Close files and shutdown */
  if (CTL_Shutdown(ctx) != RC_OK) exit(-1);

  return 0;
}

