/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	GEN_BWG.c    							--*/
/*--									--*/
/*--	Take a pair of bignums, x and y, and communicate with 		--*/
/*--	other servers to calculate N = (x1+x2+...+xn)*(y1+y2+...+yn)	--*/
/*--									--*/
/*------------------------------------------------------------------------*/

#include "ITTC.h"


/* A wrapper for BN_rand() */
int GEN_Rand(BIGNUM *rnd, int bits, int top, int bottom) {
  int x;

  x = BN_rand(rnd, bits, top, bottom);

  return x;
}


/* 

   F(x), G(x) and H(x) are polynomials with the following properties:
   
   F(0) = X
   G(0) = Y
   H(0) = 0;
   
   
   If share = 1, GEN_BGW returns an additive sharing of N.  Otherwise
   it returns the value of N.
   
   xType and yType are not used at this time.
   
*/     


int GEN_BGW(BIGNUM *N, BIGNUM *X, BIGNUM *Y, BIGNUM *modulus,
	    int share, int xType, int yType, GEN_CTX *genCtx) {

  int i, j;              /* Loop variables */
  int me;                /* My server number */
  int numServers;        /* Number of servers */
  int heapState;
  int bits;              /* Number of bits in random numbers */
  int bigThresh;         /* Degree of polynomial N */
  int t;                 /* Degree of polynomials F and G to make.
			    F and G each have t+1 */
  
  /* These are local versions of variables in the GEN_CTX */
  BN_CTX *bnCtx = NULL;         /* Bignum context */
  BIGNUM *temp = NULL, *tmp = NULL;    /* Temporary bignums, used to compute Fi(me), Gi(me) */
  BIGNUM *t1 = NULL, *t2 = NULL, *t3 = NULL;  /* More temporary bignums */
  BIGNUM *f = NULL, *g = NULL;         /* F and G */
  BIGNUM *one = NULL;           /* The constant 1 */
  BIGNUM *zero = NULL;          /* The constant zero */
  BIGNUM *id = NULL;            /* A server's ID number as a BIGNUM */
  BIGNUM *R = NULL;
  UTL_HEAP *heap = NULL;

  BIGNUM **Ns = NULL;           /* An array of the N(i) */
  BIGNUM **fs = NULL, **gs = NULL, **hs = NULL;  /* Hold F(x), G(x) and H(x) */
  BIGNUM **fRands = NULL;       /* Random numbers for F */
  BIGNUM **gRands = NULL;       /* Random numbers for G */
  BIGNUM **hRands = NULL;       /* Random numbers for H */
  BIGNUM **arrayF = NULL;       /* An array of BIGNUM*'s */
  BIGNUM **arrayG = NULL;       /* An array of BIGNUM*'s */


  /*-----------------------------------------------*/
  /* Initialize                                    */
  /*-----------------------------------------------*/
  
  /* Initialize variables */
  me = GEN_MyID(genCtx);
  numServers = GEN_NumServers(genCtx);
  t = GEN_Threshold(genCtx);
  bigThresh = 2*t + 1;
  zero = GEN_Zero(genCtx);
  one = GEN_One(genCtx);
  heap = GEN_Heap(genCtx);
  heapState = UTL_GetHeapState(heap);
  bits = GEN_CoeffBits(genCtx);
  bnCtx = GEN_BNContext(genCtx);

  /* Initialize BIGNUM's from heap */
  if ( (R = UTL_GetBN(heap)) == NULL) goto error;
  if (!BN_copy(R, modulus)) goto error;
  if ( (temp = UTL_GetBN(heap)) == NULL) goto error;
  if ( (tmp = UTL_GetBN(heap)) == NULL) goto error;
  if ( (t1 = UTL_GetBN(heap)) == NULL) goto error;
  if ( (t2 = UTL_GetBN(heap)) == NULL) goto error;
  if ( (t3 = UTL_GetBN(heap)) == NULL) goto error;
  if ( (id = UTL_GetBN(heap)) == NULL) goto error;
  if ( (f = UTL_GetBN(heap)) == NULL) goto error;
  if ( (g = UTL_GetBN(heap)) == NULL) goto error;

  if ( (fRands = UTL_GetBNArray(t+1, heap)) == NULL) goto error;
  if ( (gRands = UTL_GetBNArray(t+1, heap)) == NULL) goto error;
  if ( (hRands = UTL_GetBNArray(2*t+1, heap)) == NULL) goto error;
  if ( (fs = UTL_GetBNArray(numServers, heap)) == NULL) goto error;
  if ( (gs = UTL_GetBNArray(numServers, heap)) == NULL) goto error;
  if ( (hs = UTL_GetBNArray(numServers, heap)) == NULL) goto error;
  if ( (Ns = UTL_GetBNArray(numServers, heap)) == NULL) goto error;
  if ( (arrayF = UTL_GetBNArray(numServers, heap)) == NULL) goto error;
  if ( (arrayG = UTL_GetBNArray(numServers, heap)) == NULL) goto error;


  /*-----------------------------------------------*/
  /* Choose random numbers to make my polynomials  */
  /*-----------------------------------------------*/
  
  /* Set the constant coefficients of F, G and H */
  if (!BN_copy(fRands[0], X)) {
    UTL_Log("GEN_BGW", LOG_FATAL, "Error generating random numbers.");
    goto error;
  }
  if (!BN_copy(gRands[0], Y)) {
    UTL_Log("GEN_BGW", LOG_FATAL, "Error generating random numbers.");
    goto error;
  }
  if (!BN_zero(hRands[0])) {
    UTL_Log("GEN_BGW", LOG_FATAL, "Error generating random numbers.");
    goto error;
  }

  /* Generate other coefficients of F, G and H */
  for (i=1; i<=t; i++) {

    if (!BN_zero(fRands[i])) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error generating random numbers.");
      goto error;
    }
    if (!GEN_Rand(temp, bits, 0, 0)) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error generating random numbers.");
      goto error;
    }
    if (!BN_mod(fRands[i], temp, R, bnCtx)) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error generating random numbers.");
      goto error;
    }

    if (!BN_zero(gRands[i])) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error generating random numbers.");
      goto error;
    }
    if (!GEN_Rand(temp, bits, 0, 0)) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error generating random numbers.");
      goto error;
    }
    if (!BN_mod(gRands[i], temp, R, bnCtx)) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error generating random numbers.");
      goto error;
    }
  }

  for (i=1; i<=2*t; i++) {
    if (!BN_zero(hRands[i])) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error generating random numbers.");
      goto error;
    }
    if (!GEN_Rand(temp, bits, 0, 0)) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error generating random numbers.");
      goto error;
    }
    if (!BN_mod(hRands[i], temp, R, bnCtx)) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error generating random numbers.");
      goto error;
    }
  }
  
  /*-----------------------------------------------*/
  /* Calculate F_me(i) G_me(i) and H_me(i)         */
  /*-----------------------------------------------*/
  
  for(i=0; i<numServers; i++) {

    /* Save ID + 1 as a bignum */
    if (!BN_set_word(id, i+1)) {   /* use +1 to that it's 1 to numServers, not 0 to numServers-1 */
      UTL_Log("GEN_BGW", LOG_FATAL, "Error converting ID to a BIGNUM.");
      goto error;
    }    

    /*-------------------------------------*/
    /* Calculate F(i) with Horner's method */
    /*-------------------------------------*/

    if (BN_copy(temp, fRands[t]) == NULL) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error copying BIGNUM.");
      goto error;
    }        
    for (j=t-1; j>=0; j--) {
      if (!BN_mod_mul(t1, temp, id, R, bnCtx)){
	UTL_Log("GEN_BGW", LOG_FATAL, "Error calculating F(%d) - mod_mul.", i);
	goto error;
      }      
      if (!BN_add(temp, t1, fRands[j])) {
	UTL_Log("GEN_BGW", LOG_FATAL, "Error calculating F(%d) - add.", i);
	goto error;
      }
    }
    if (!BN_mod(fs[i], temp, R, bnCtx)) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error calculating F(%d) - mod.", i);
      goto error;
    }


    /*-------------------------------------*/
    /* Calculate G(i) with Horner's method */
    /*-------------------------------------*/

    if (BN_copy(temp, gRands[t]) == NULL) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error copying BIGNUM.");
      goto error;
    }    
    for (j=t-1; j>=0; j--) {
      if (!BN_mod_mul(t1, temp, id, R, bnCtx)){
	UTL_Log("GEN_BGW", LOG_FATAL, "Error calculating G(%d) - mod_mul.", i);
	goto error;
      }      
      if (!BN_add(temp, t1, gRands[j])) {
	UTL_Log("GEN_BGW", LOG_FATAL, "Error calculating G(%d) - add.", i);
	goto error;
      }
    }
    if (!BN_mod(gs[i], temp, R, bnCtx)) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error calculating G(%d) - mod.", i);
      goto error;
    }
    
    /*-------------------------------------*/
    /* Calculate H(i) with Horner's method */
    /*-------------------------------------*/

    if (BN_copy(temp, hRands[2*t]) == NULL) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error copying BIGNUM.");
      goto error;
    }    
    for (j=2*t-1; j>=0; j--) {
      if (!BN_mod_mul(t1, temp, id, R, bnCtx)){
	UTL_Log("GEN_BGW", LOG_FATAL, "Error calculating H(%d) - mod_mul.", i);
	goto error;
      }      
      if (!BN_add(temp, t1, hRands[j])) {
	UTL_Log("GEN_BGW", LOG_FATAL, "Error calculating H(%d) - add.", i);
	goto error;
      }
    }
    if (!BN_mod(hs[i], temp, R, bnCtx)) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error calculating H(%d) - mod.", i);
      goto error;
    }
    
  }


  /*--------------------------------------------------------------------*/
  /* Send F_me(i), G_me(i), and H_me(i) or store if i == me             */
  /*--------------------------------------------------------------------*/

  if (UTL_SendToAll(fs, GEN_Streams(genCtx)) != RC_OK) {
    UTL_Log("GEN_BGW", LOG_INFO, "Error sending F to all.");
    goto error;
  }
  if (UTL_SendToAll(gs, GEN_Streams(genCtx)) != RC_OK) {
    UTL_Log("GEN_BGW", LOG_INFO, "Error sending G to all.");
    goto error;
  }
  if (UTL_SendToAll(hs, GEN_Streams(genCtx)) != RC_OK) {
    UTL_Log("GEN_BGW", LOG_INFO, "Error sending H to all.");
    goto error;
  }  


  /*--------------------------------------------------------*/
  /* Receive Fi(me), Gi(me), and Fi(me) from all servers i  */
  /*--------------------------------------------------------*/

  if ((UTL_GetFromAll(fs, GEN_Streams(genCtx), GEN_Heap(genCtx), GEN_Stat(genCtx))) != RC_OK) {
    UTL_Log("GEN_BGW", LOG_INFO, "Error getting Fi from all.");
    goto error;
  }
  if ((UTL_GetFromAll(gs, GEN_Streams(genCtx), GEN_Heap(genCtx), GEN_Stat(genCtx))) != RC_OK) {
    UTL_Log("GEN_BGW", LOG_INFO, "Error getting Gi from all.");
    goto error;
  }
  if ((UTL_GetFromAll(hs, GEN_Streams(genCtx), GEN_Heap(genCtx), GEN_Stat(genCtx))) != RC_OK) {
    UTL_Log("GEN_BGW", LOG_INFO, "Error getting Hi from all.");
    goto error;
  }


  /*---------------------------------------------------------------------------------*/
  /* Find N(me) = [ F1(me)+F2(me)+... ][ G1(me)+G2(me)+... ] + [ H1(me)+H2(me)+... ] */
  /*---------------------------------------------------------------------------------*/

  /* Add up F, G and H */
  if (!BN_zero(t1)) {
    UTL_Log("GEN_BGW", LOG_FATAL, "Error with BN_zero().");
    goto error;
  }
  if (!BN_zero(t2)) {
    UTL_Log("GEN_BGW", LOG_FATAL, "Error with BN_zero().");
    goto error;
  }
  if (!BN_zero(t3)) {
    UTL_Log("GEN_BGW", LOG_FATAL, "Error with BN_zero().");
    goto error;
  }
  for (i=0; i<numServers; i++) {
    
    if (!BN_add(temp, t1, fs[i])) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error adding F, G, and H.");
      goto error;
    }
    if (BN_copy(t1, temp) == NULL) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error adding F, G, and H.");
      goto error;
    }
    
    if (!BN_add(temp, t2, gs[i])) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error adding F, G, and H.");
      goto error;
    }
    if (BN_copy(t2, temp) == NULL) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error adding F, G, and H.");
      goto error;
    }

    if (!BN_add(temp, t3, hs[i])) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error adding F, G, and H.");
      goto error;
    }
    if (BN_copy(t3, temp) == NULL) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error adding F, G, and H.");
      goto error;
    }
  }

  /* Set my value of N(me) */
  if (!BN_mod_mul(temp, t1, t2, R, bnCtx)) {
    UTL_Log("GEN_BGW", LOG_FATAL, "Error calculating N(me).");
    goto error;
  }
  if (!BN_add(tmp, temp, t3)) {
    UTL_Log("GEN_BGW", LOG_FATAL, "Error calculating N(me).");
    goto error;
  }
  if (!BN_mod(Ns[me], tmp, R, bnCtx)) {
    UTL_Log("GEN_BGW", LOG_FATAL, "Error calculating N(me).");
    goto error;
  }
  

  /*-----------------------------------------------*/
  /* Multiply by binomial coefficient              */
  /*-----------------------------------------------*/

  /* The binomials with alternating positive and negative terms are
   * actually lambda_i, the functions used in LaGrange interpolation.
   * Some calculations will show that, with i = 1, 2, ... t,
   *
   * lambda_i = (-1)^(i-1) * (k choose i)
   *
   * Because of the factor (-1)^(i-1) I call these the modified
   * binomial coefficients.  
   *
   * Since all of the N(i) have been pre-multiplied by their
   * LaGrange coefficient, interpolation here is merely adding
   * up the N(i).
   */
    
  /* Find the LaGrange interpolation coefficient mod R */
  if ( (me & 1) ) {
    if (!BN_sub(temp, R, GEN_Binomials(genCtx)[me])) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error in BN_sub().");
      goto error;
    }
  } else {
    if (!BN_copy(temp, GEN_Binomials(genCtx)[me])) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error in BN_sub().");
      goto error;
    }
  }

  /* Now multiply by the lagrange coefficient */
  if (!BN_mod_mul(tmp, Ns[me], temp, R, bnCtx)) {
    UTL_Log("GEN_BGW", LOG_FATAL, "Error in BN_mod_mul().");
    goto error;
  }

  /* At this point we have an additive sharing of 
   * the desired result.  We continue only if we 
   * desire that everyone have the result, not an 
   * additive share of the result.
   */

  if (share == BGW_INTERPOLATE) {

    /* Send N(me) to all servers */
    if (UTL_SendSameToAll(tmp, GEN_Streams(genCtx)) != RC_OK) {
      UTL_Log("GEN_BGW", LOG_INFO, "Error sending to all.");
      goto error;
    }  

    /* Receive N(i) from all servers, do LaGrange interpolation */
    if (UTL_GetSumFromAll(N, tmp, R, bnCtx, GEN_Streams(genCtx), GEN_Heap(genCtx), GEN_Stat(genCtx)) != RC_OK) {
      UTL_Log("GEN_BGW", LOG_INFO, "Error getting from all.");
      goto error;
    }
    
  } else if (share == BGW_ADDITIVE_SHARE) {

    if (!BN_copy(N, tmp)) {
      UTL_Log("GEN_BGW", LOG_FATAL, "Error in BN_copy().");
      goto error;
    }

  } else {

      UTL_Log("GEN_BGW", LOG_FATAL, "Error in share parameter.");
      goto error;

  }


  /*-----------------------------------------------*/
  /* Testing                                       */
  /*-----------------------------------------------*/

  if (GEN_Testing(genCtx) ){

    UTL_SendSameToAll(X, GEN_Streams(genCtx));
    UTL_SendSameToAll(Y, GEN_Streams(genCtx));
    UTL_GetSumFromAll(t1, X, R, bnCtx, GEN_Streams(genCtx), GEN_Heap(genCtx), GEN_Stat(genCtx));
    UTL_GetSumFromAll(t2, Y, R, bnCtx, GEN_Streams(genCtx), GEN_Heap(genCtx), GEN_Stat(genCtx));
    BN_mod_mul(t3, t1, t2, R, bnCtx);

    if (share == BGW_INTERPOLATE) {
      if (BN_cmp(N, t3)) {
	UTL_Log("GEN_BGW", LOG_FATAL, "N was computed incorrectly!");
	goto error;
      }
    } else {
      UTL_SendSameToAll(N, GEN_Streams(genCtx));
      UTL_GetSumFromAll(t1, N, R, bnCtx, GEN_Streams(genCtx), GEN_Heap(genCtx), GEN_Stat(genCtx));
      if (BN_cmp(t1, t3)) {
	UTL_Log("GEN_BGW", LOG_FATAL, "The share of N was computed incorrectly!");
	goto error;
      }
    }

  }

  /*-----------------------------------------------*/
  /* Return                                        */
  /*-----------------------------------------------*/

  return RC_OK;

 error:

  UTL_RevertHeap(heapState, heap);
  return RC_ERROR;
}





