/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*----------------------------------------------------------*/
/*--                                                      --*/
/*--  GEN_Context.c                                       --*/
/*--                                                      --*/
/*--  Functions to handle GEN_CTX data structures         --*/
/*--                                                      --*/
/*----------------------------------------------------------*/

#include "ITTC.h"


/***********************/
/* Access functions    */
/***********************/

/* Set the thread number */
int GEN_SetThread(int num, GEN_CTX* genCtx) {
  if (genCtx) {
    genCtx->thread = num;
    return RC_OK;
  }
  return RC_ERROR;
}


/* Return the thread number */
int GEN_Thread(GEN_CTX* genCtx) {
  if (genCtx) return genCtx->thread;
  return -1;
}


int GEN_SetStats(UTL_STATS* stats, GEN_CTX* genCtx) {
  if (genCtx) {
    genCtx->statistics = stats;
    return RC_OK;
  }
  return RC_ERROR;
}

/* Returns 1 if testing, 0 if running */
int GEN_Testing(GEN_CTX *genCtx) {
  if ( (genCtx) && (genCtx->testMode == TESTING) ) return 1;
  return 0;
}

UTL_STREAMS *GEN_Streams(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->streams;
  return NULL;
}

UTL_HEAP* GEN_Heap(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->heap;
  return NULL;
}

int GEN_NumServers(GEN_CTX *genCtx) {
  if (genCtx) return UTL_NumServers(GEN_Streams(genCtx));
  return -1;
}

int GEN_MyID(GEN_CTX *genCtx) {
  if (genCtx) return UTL_MyID(GEN_Streams(genCtx));
  return -1;
}

COM_STREAM* GEN_Stream(int num, GEN_CTX *genCtx) {
  if (genCtx) return UTL_Stream(num, GEN_Streams(genCtx));
  return NULL;
}

COM_CTX* GEN_COMContext(GEN_CTX *genCtx) {
  if (genCtx) return UTL_COMContext(GEN_Streams(genCtx));
  return NULL;
}

UTL_CTX* GEN_UTLContext(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->utlCtx;
  return NULL;
}


int GEN_SetStream(int num, COM_STREAM* stream, GEN_CTX *genCtx) {
  if (genCtx) return UTL_SetStream(num, stream, GEN_Streams(genCtx));
  return RC_ERROR;
}

int GEN_Security(GEN_CTX *ctx) {
  if (ctx) return ctx->security;
  return 0;
}


/* The number of bits in a word on this system */
int GEN_WordBits(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->wordBits;
  return -1;
}


/* The number of bits in R */
int GEN_RBits(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->rBits;
  return -1;
}


/* The number of bits in a BGW polynomial coefficient */
int GEN_CoeffBits(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->coeffBits;
  return -1;
}


/* The number of bits in P and Q */
int GEN_ShareBits(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->shareBits;
  return -1;
}


/* The degree of BGW polynomials */
int GEN_Threshold(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->threshold;
  return -1;
}


/* Returns R */
BIGNUM* GEN_R(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->R;
  return NULL;
}


/* Returns the sieve */
char* GEN_Sieve(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->sieve;
  return NULL;
}


/* Returns 1 if the sieve is active, 0 otherwise */
int GEN_SieveP(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->sieveP;
  return -1;
}


/* The number of primes used in sieving */
int GEN_NumSievePrimes(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->numSmallPrimes;
  return -1;
}


/* The range of the sieve */
int GEN_SieveRange(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->range;
  return -1;
}


/* Returns M */
BIGNUM* GEN_M(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->M;
  return NULL;
}


/* The upper bound for trial division primes */
int GEN_TrialDivBound(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->trialDivEnd;
  return -1;
}


/* The number of factors used in trial division */
int GEN_NumTrialDivFactors(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->numFactors;
  return -1;
}


/* returns an array of the trial division factors */
long unsigned int* GEN_TrialDivFactors(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->factors;
  return NULL;
}


/* Returns the binomials for LaGrange interpolation */
BIGNUM** GEN_Binomials(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->binomials;
  return NULL;
}

/* Return the BIGNUM representation of these numbers */
BIGNUM* GEN_Zero(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->zero;
  return NULL;
}

BIGNUM* GEN_One(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->one;
  return NULL;
}

BIGNUM* GEN_Two(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->two;
  return NULL;
}

BIGNUM* GEN_Three(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->three;
  return NULL;
}


/* Return the BIGNUM's used to set a number mod 60 */
BIGNUM* GEN_CRTa(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->CRTa;
  return NULL;
}

BIGNUM* GEN_CRTb(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->CRTb;
  return NULL;
}


/* Return sixty times M */
BIGNUM* GEN_SixtyM(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->sixtyM;
  return NULL;
}


/* Return the BIGNUM context */ 
BN_CTX* GEN_BNContext(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->bnCtx;
  return NULL;
}


/* Return the UTL statistics */
UTL_STATS* GEN_Stat(GEN_CTX *genCtx) {
  if (genCtx) return genCtx->statistics;
  return NULL;
}



/*----------------------------------------*/
/*-- Initialization and Close functions --*/
/*----------------------------------------*/

int GEN_InitSecurity(int keySize, int keyType, GEN_CTX *genCtx) {
  UTL_CTX *utlCtx = NULL;

  utlCtx = GEN_UTLContext(genCtx);
  if (utlCtx == NULL) {
    UTL_Log("InitSecurity", LOG_INFO, "No UTL context.");
    return -1;
  }

  /*  switch (keySize) {*/
  switch (keySize) {
  case STRONG_KEY:
    UTL_Log("InitSecurity", LOG_TEST, "Strong Security.");
    genCtx->rBits = UTL_GetIntConfigParam(utlCtx, "R_Bits", STRONG_R_BITS);
    genCtx->coeffBits = UTL_GetIntConfigParam(utlCtx, "Coeff_Bits", STRONG_COEFF_BITS);
    genCtx->shareBits = UTL_GetIntConfigParam(utlCtx, "Share_Bits", STRONG_SHARE_BITS);
    return 3;
  case NORMAL_KEY:
    UTL_Log("InitSecurity", LOG_TEST, "Normal Security.");
    genCtx->rBits = UTL_GetIntConfigParam(utlCtx, "R_Bits", NORMAL_R_BITS);
    genCtx->coeffBits = UTL_GetIntConfigParam(utlCtx, "Coeff_Bits", NORMAL_COEFF_BITS);
    genCtx->shareBits = UTL_GetIntConfigParam(utlCtx, "Share_Bits", NORMAL_SHARE_BITS);
    return 2;
  case WEAK_KEY:
    UTL_Log("InitSecurity", LOG_TEST, "Weak Security.");
    genCtx->rBits = UTL_GetIntConfigParam(utlCtx, "R_Bits", WEAK_R_BITS);
    genCtx->coeffBits = UTL_GetIntConfigParam(utlCtx, "Coeff_Bits", WEAK_COEFF_BITS);
    genCtx->shareBits = UTL_GetIntConfigParam(utlCtx, "Share_Bits", WEAK_SHARE_BITS);
    return 1;
  case TEST_KEY:
    UTL_Log("InitSecurity", LOG_TEST, "Test Security.");
    genCtx->rBits = UTL_GetIntConfigParam(utlCtx, "R_Bits", TEST_R_BITS);
    genCtx->coeffBits = UTL_GetIntConfigParam(utlCtx, "Coeff_Bits", TEST_COEFF_BITS);
    genCtx->shareBits = UTL_GetIntConfigParam(utlCtx, "Share_Bits", TEST_SHARE_BITS);
    return 0;
  default:
    UTL_Log("InitSecurity", LOG_TEST, "Problem with configuration file.");
    return -1;
  }

  return -1;
}


int GEN_InitializeGenCtx(COM_CTX *context, int keySize, int keyType, int numServers, 
			 UTL_CTX *utlCtx, GEN_CTX **pgenCtx) {

  int i;
  COM_INFO info;
  int me;
  GEN_CTX* genCtx = NULL;

  /* Initialize */
  genCtx = (GEN_CTX*)malloc(sizeof(GEN_CTX));
  if (genCtx == NULL) goto error;
  genCtx->streams = NULL;
  *pgenCtx = genCtx;

  genCtx->threshold = (int)floor((double)((double)numServers-1.0)/2.0);
  if ( (genCtx->bnCtx = BN_CTX_new()) == NULL) goto error;
  if ( (genCtx->CRTa = BN_new()) == NULL) goto error;
  if ( (genCtx->CRTb = BN_new()) == NULL) goto error;
  if ( (genCtx->M = BN_new()) == NULL) goto error;
  if ( (genCtx->R = BN_new()) == NULL) goto error;
  if ( (genCtx->zero = BN_new()) == NULL) goto error;
  if ( (genCtx->one = BN_new()) == NULL) goto error;
  if ( (genCtx->two = BN_new()) == NULL) goto error;
  if ( (genCtx->three = BN_new()) == NULL) goto error;
  if ( (genCtx->sixtyM = BN_new()) == NULL) goto error;
  if (!BN_set_word(genCtx->zero, 0)) goto error;
  if (!BN_set_word(genCtx->one, 1)) goto error;
  if (!BN_set_word(genCtx->two, 2)) goto error;
  if (!BN_set_word(genCtx->three, 3)) goto error;
  if ( (genCtx->binomials = (BIGNUM**)malloc(numServers*sizeof(BIGNUM*))) == NULL) goto error;
  for (i=0; i<numServers; i++) {
    if ( (genCtx->binomials[i] = BN_new()) == NULL) goto error;
  }
  genCtx->utlCtx = utlCtx;

  /* Initialize Heap */
  if ( (genCtx->heap = UTL_InitHeap(UTL_GetIntConfigParam(utlCtx, "Heap_Size", 1000))) == NULL) {
    UTL_Log("GEN_InitializeGenCtx", LOG_INFO, "Error initializing heap.");
    return RC_ERROR;
  }  
    
  /* Initialize Security */
  if ((genCtx->security = GEN_InitSecurity(keySize, keyType, genCtx)) == -1) {
    UTL_Log("GEN_InitGenCtx", LOG_INFO, "Error in InitSecurity.");
    return RC_ERROR;
  }  

  /* Get my ID number */
  info = COM_ctx_get_info(context);
  me = COM_INFO_GET_NUM(info);

  /* Find the word size */
  genCtx->wordBits = UTL_GetIntConfigParam(utlCtx, "Word_Size", 32);
  if (genCtx->wordBits == -1 ) goto error;

  /* Initialize the streams */
  genCtx->streams = UTL_InitStreams(numServers, me, context);
  if (genCtx->streams == NULL) goto error;

  /* Make R */
  if (GEN_MakeR(genCtx) != RC_OK) {
    UTL_Log("GEN_InitGenCtx", LOG_INFO, "Problem generating R.");
    goto error;
  }

  /* Set B1 and B2 */
  genCtx->sieveP = UTL_GetBoolConfigParam(utlCtx, "Sieve", 1);
  genCtx->trialDivEnd = UTL_GetIntConfigParam(utlCtx, "TrialDiv_End", 10000);

  /* Prepare for the sieve */
  if (GEN_PrepareSieve(genCtx) != RC_OK) {
    UTL_Log("GEN_InitGenCtx", LOG_INFO, "Failed to prepare sieve.");
    goto error;
  }

  /* Make factors for trial division */
  if (GEN_MakeFactors(genCtx) != RC_OK) {
    UTL_Log("GEN_InitgenCtx", LOG_INFO, "Problem generating prime factors.");
    goto error;
  }

  /* Generate the LaGrange interpolation constants */
  if (GEN_ModBin(genCtx) != RC_OK) {
    UTL_Log("GEN_InitgenCtx", LOG_INFO, "Problem generating modified binomials.");
    goto error;
  }

  /* Decide if in test mode */
  if (UTL_GetBoolConfigParam(utlCtx, "Test_Mode", FALSE)) {
    genCtx->testMode = TESTING;
  } else {
    genCtx->testMode = RUNNING;
  }

  /* Return */
  return RC_OK;

 error:
  GEN_CloseGenCtx(genCtx);
  return RC_ERROR;

}


/* Close a genCtx and free everything inside of it */
int GEN_CloseGenCtx(GEN_CTX *genCtx) {
  int i;

  if (genCtx == NULL) return RC_OK;

  /* The heap */
  if (GEN_Heap(genCtx)) UTL_FreeHeap(GEN_Heap(genCtx));

  for (i=0; i<GEN_NumServers(genCtx); i++) {
    if (GEN_Binomials(genCtx)[i]) BN_clear_free(GEN_Binomials(genCtx)[i]);
  }
  if (GEN_Binomials(genCtx)) free(GEN_Binomials(genCtx));

  if (GEN_BNContext(genCtx)) BN_CTX_free(GEN_BNContext(genCtx));

  if (GEN_CRTa(genCtx)) BN_clear_free(GEN_CRTa(genCtx));
  if (GEN_CRTb(genCtx)) BN_clear_free(GEN_CRTb(genCtx));
  if (GEN_Zero(genCtx)) BN_clear_free(GEN_Zero(genCtx));
  if (GEN_One(genCtx)) BN_clear_free(GEN_One(genCtx));
  if (GEN_Two(genCtx)) BN_clear_free(GEN_Two(genCtx));
  if (GEN_Three(genCtx))  BN_clear_free(GEN_Three(genCtx));
  if (GEN_M(genCtx)) BN_clear_free(GEN_M(genCtx));
  if (GEN_R(genCtx)) BN_clear_free(GEN_R(genCtx));
  if (GEN_SixtyM(genCtx)) BN_clear_free(GEN_SixtyM(genCtx));

  if (GEN_NumTrialDivFactors(genCtx) > 0) free(GEN_TrialDivFactors(genCtx));

  free(genCtx);
  return RC_OK;
}



