/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	GEN_GenKey.c							--*/
/*--									--*/
/*--	Generate a new shared public/private key pair			--*/
/*--									--*/
/*------------------------------------------------------------------------*/

#include "ITTC.h"


BIGNUM *global_N = NULL;
BIGNUM *global_P_share = NULL;
BIGNUM *global_Q_share = NULL;
GEN_CTX *globalGenCtx = NULL;
int genDone = 0;     /* A global variable to notify threads when a modulus is found */
UTL_MUTEX doneS;    /* A semaphore to ensure that only one thread gets */
                 /*    the critical section at a time */


typedef struct GEN_args {
  int numServers;
  int me;
  int thread;
  int keySize;
  int keyType;
  UTL_CTX *utlCtx;
  COM_CTX *comCtx;
} GEN_ARGS;



/* Called by thread */
void GEN_GM(GEN_ARGS *ga) {
  UTL_Log("GEN_GM", LOG_DEBUG, "***  Thread %d starting", ga->thread);
  if (GEN_GenerateModulus(ga->numServers, ga->me, ga->thread, ga->keySize, ga->keyType, 
			  ga->utlCtx, ga->comCtx) != RC_OK) {
    UTL_Log("GEN_GM", LOG_INFO, "*** Thread %d has died.", ga->thread);
    return;
  } else {
    UTL_Log("GEN_GM", LOG_DEBUG, "***  Thread %d done", ga->thread);
  }
}


int GEN_GenKey(int myServer, COM_STREAM *admStream, int keySize, int keyType, 
	       int numServers, int threshold, UTL_STATS** statistics, UTL_CTX *utlCtx,
	       KEY_CTX *keyCtx, COM_CTX *comCtx)
{
  
  GEN_CTX *genCtx = NULL;
  UTL_STATS *stats = NULL;

  UTL_THREAD *thr = NULL;
  int madeMutex = 0;

  GEN_ARGS *ga = NULL;
  BIGNUM *N = NULL;
  BIGNUM *P_share = NULL;
  BIGNUM *Q_share = NULL;
  BIGNUM *pubExp = NULL;
  UTL_STREAMS *streams = NULL;
  int me;
  int err = RC_OK;
  int i;
  RSA_SHARE *key = NULL;
  COM_STAT *comStat = NULL;
  int threads;


  /*--------------------------------------------------------------------*/
  /*  Initialize           			                        */
  /*--------------------------------------------------------------------*/

  UTL_monitor(utlCtx, 0, "Beginning key generation.");

  global_N = NULL;
  global_P_share = NULL;
  global_Q_share = NULL;
  globalGenCtx = NULL;

  /* Check input parameters */
  if (admStream == NULL) {
    UTL_Log("GEN_GenKey", LOG_FATAL, "No ADM stream.");
    goto error;
  }
  if (utlCtx == NULL) {
    UTL_Log("GEN_GenKey", LOG_FATAL, "No UTL context.");
    goto error;
  }
  if (keyCtx == NULL) {
    UTL_Log("GEN_GenKey", LOG_FATAL, "No KEY context.");
    goto error;
  }
  if (comCtx == NULL) {
    UTL_Log("GEN_GenKey", LOG_FATAL, "No COM context.");
    goto error;
  }


  genDone = 0;
  threads = UTL_GetIntConfigParam(utlCtx, "Threads", 6);
  
  UTL_mutex_init(&doneS);
  madeMutex = 1;
  pubExp = BN_new();
  if (pubExp == NULL) goto error;


  /*--------------------------------------------------------------------*/
  /*  Loop until RSA modulus is found.					*/
  /*--------------------------------------------------------------------*/

  if (threads > 0) {

    /* Multithread initialization */
    thr = (UTL_THREAD*)malloc(threads*sizeof(UTL_THREAD));
    if (thr == NULL) goto error;
    ga = (GEN_ARGS*)malloc(threads*sizeof(GEN_ARGS));
    if (ga == NULL) goto error;

    /* Start the threads */
    for (i=0; i<threads; i++) {
      ga[i].numServers = numServers;
      ga[i].me = myServer;
      ga[i].thread = i;
      ga[i].keySize = keySize;
      ga[i].keyType = keyType;
      ga[i].utlCtx = utlCtx;
      ga[i].comCtx = comCtx;
      UTL_thread_create(&thr[i], (void*)GEN_GM, &ga[i]);
    }
    
    /* Wait until the threads are done */
    for (i=0; i<threads; i++) {
      UTL_thread_join(&thr[i]);
    }

  } else {

    ga = (GEN_ARGS*)malloc(sizeof(GEN_ARGS));

    UTL_Log("GEN_GenKey", LOG_TEST, "*** Main process starting!!!");
    ga->numServers = numServers;
    ga->me = myServer;
    ga->thread = 0;
    ga->keySize = keySize;
    ga->keyType = keyType;
    ga->utlCtx = utlCtx;
    ga->comCtx = comCtx;
    GEN_GM(ga);
    UTL_Log("GEN_GenKey", LOG_TEST, "*** Main process ended");
      
  }


  /*--------------------------------------------------------------------*/
  /*  Prepare for the rest of the program.                   		*/
  /*--------------------------------------------------------------------*/

  if (globalGenCtx == NULL) goto error;

  genCtx = globalGenCtx;
  stats = GEN_Stat(genCtx);
  N = global_N;
  P_share = global_P_share;
  Q_share = global_Q_share;

  me = GEN_MyID(genCtx);
  streams = GEN_Streams(genCtx);


  /*--------------------------------------------------------------------*/
  /*  Construct private/public keys.					*/
  /*--------------------------------------------------------------------*/

  if (!BN_set_word(pubExp, UTL_GetIntConfigParam(utlCtx, "Public_Key", GEN_DEFAULT_PUBLIC_KEY))) {
    UTL_Log("GEN_GenKey", LOG_FATAL, "Can't set public exponent.");
    goto error;
  }

  if ((key = KEY_RSA_share_create(N, pubExp)) == NULL) {
    UTL_Log("GEN_GenKey", LOG_FATAL, "Unable to create RSA key.");
    goto error;
  }

  UTL_StartStat(4, stats);
  err = GEN_PrivateKey(key, N, P_share, Q_share, genCtx);
  UTL_FinishStat(4, stats);
  
  /* Set the threshold characteristics */
  KEY_RSA_share_set_threshold_params(key, numServers, threshold);

  if (err != RC_OK) {
    UTL_Log("GEN_GenKey", LOG_FATAL, "Unable to compute private key.");
    goto error;
  }


  /*-----------------------------------------------*/
  /* Decide upon g, generate t-out-of-k sharing    */
  /*-----------------------------------------------*/

  err = GEN_GenerateSharedKey(key, admStream, genCtx);
  if (err != RC_OK) {
    UTL_Log("GEN_GenKey", LOG_ERROR, "Error sharing key.");
    goto error;
  }


  /*--------------------------------------------------------------------*/
  /*  Save new key to disk.                                             */
  /*--------------------------------------------------------------------*/

  err = GEN_SaveKeyToDisk(keyCtx, key);
  if (err != RC_OK) {
    UTL_Log("GEN_GenKey", LOG_ERROR, "Cannot save new key to disk.");
    goto noSendError;
  }
  /*** TJW: Uncomment when KEY_CTX_add makes a copy of the key! ***/
  /*KEY_RSA_share_free(key);*/

  
  /*--------------------------------------------------------------------*/
  /*  Print out the statistics.   					*/
  /*--------------------------------------------------------------------*/
  
  if (UTL_GetBoolConfigParam(utlCtx, "Gen_Stats", FALSE)) {
    UTL_FinishStat(5, stats);
    UTL_PrintStats(stats);
    for (i=0; i<numServers; i++) if (i != me) {
      COM_print_stats(GEN_Stream(i, genCtx));
      comStat = COM_get_stats(GEN_Stream(i, genCtx));
      stats->pread += comStat->p_read;
      stats->pwrite += comStat->p_write;
      stats->bread += comStat->b_read;
      stats->bwrite += comStat->b_write;
    }
  }
  UTL_FlushLogFile();


  /*--------------------------------------------------------------------*/
  /*  Close connections to peers and free memory.			*/
  /*--------------------------------------------------------------------*/

  err = RC_OK;
  UTL_monitor(utlCtx, 0, "Done with key generation.");
  UTL_Log("GEN_GenKey", LOG_NOTIFY, "*** Distributed RSA key generated.");


 done:

  if (ga) free(ga);
  if (thr) free(thr);
  if (madeMutex) UTL_mutex_destroy(&doneS);

  if ( (genCtx) && (GEN_Streams(genCtx)) ) {
    if (UTL_CloseAllConnections(GEN_Streams(genCtx)) != RC_OK) {
      UTL_Log("GEN_GenKey", LOG_FATAL, "Failure closing connections.");
    }
  }
  if (genCtx) GEN_CloseGenCtx(genCtx);

  if (statistics) {
    *statistics = stats;    /* Return the statistics of the successful thread*/
  } else {
    if (stats) UTL_FreeStats(stats);
  }
  if (pubExp) BN_clear_free(pubExp);

  return err;


 error:

  COM_write_string(admStream, "ERROR: Unable to complete key generation");
  err = RC_ERROR;
  goto done;


 noSendError:
  err = RC_ERROR;
  goto done;


  return RC_ERROR;

}


