/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	GEN_GenerateModulus.c						--*/
/*--									--*/
/*--	Generate an RSA modulus                 			--*/
/*--									--*/
/*------------------------------------------------------------------------*/

#include "ITTC.h"


extern BIGNUM *global_N;
extern BIGNUM *global_P_share;
extern BIGNUM *global_Q_share;
extern GEN_CTX *globalGenCtx;
extern UTL_MUTEX doneS;
extern int genDone;



int GEN_SetupStats(UTL_STATS* stats) {

  if (UTL_InitStats(4, 9, stats) != RC_OK) {
    UTL_Log("GEN_GenerateModulus", LOG_INFO, "Failed to initialize GEN stats.");
    return RC_ERROR;
  }

  UTL_NameCountStat(0, "Iterations", stats);
  UTL_NameCountStat(1, "Primality Tests", stats);
  UTL_NameCountStat(2, "BGW's", stats);
  UTL_NameCountStat(3, "Select's", stats);

  UTL_NameTimeStat(0, "Pick Candidates", stats);
  UTL_NameTimeStat(1, "Generate Modulus", stats);
  UTL_NameTimeStat(2, "Trial Division", stats);
  UTL_NameTimeStat(3, "Primality Test", stats);
  UTL_NameTimeStat(4, "Generate Keys", stats);
  UTL_NameTimeStat(5, "Total", stats);
  UTL_NameTimeStat(6, "BGW", stats);
  UTL_NameTimeStat(7, "Iteration Time", stats);
  UTL_NameTimeStat(8, "Select", stats);

  return RC_OK;
}



int GEN_GenerateModulus(int numServers, int me, int thread, int keySize, int keyType, 
			UTL_CTX *utlCtx, COM_CTX *comCtx) {
  
  BIGNUM *N = NULL;
  BIGNUM *P_share = NULL;
  BIGNUM *Q_share = NULL;
  BIGNUM **Ps = NULL;        /* Used to verify result at end */
  BIGNUM **Qs = NULL;        /* Used to verify result at end */
  BIGNUM *P = NULL;          /* Used to verify result at end */
  BIGNUM *Q = NULL;          /* Used to verify result at end */
  BIGNUM *testN = NULL;      /* Used to verify result at end */
  BIGNUM *temp = NULL;       /* Used to verify result at end */
  UTL_HEAP *heap = NULL;
  UTL_STREAMS *streams = NULL;
  int i;
  int failed;
  char *txt = NULL;
  int iteration;
  int state;
  int err = RC_OK;
  UTL_STATS *stats = NULL;
  GEN_CTX *genCtx = NULL;


  /*--------------------------------------------------------------------*/
  /*  Initialization				                        */
  /*--------------------------------------------------------------------*/

  UTL_monitor(utlCtx, 0, "Thread %d - Beginning search for modulus.", thread);

  /* Make GEN context */
  genCtx = (GEN_CTX*)malloc(sizeof(GEN_CTX));
  if (genCtx == NULL) goto error;
  if (GEN_InitializeGenCtx(comCtx, keySize, keyType, numServers, utlCtx, &genCtx) != RC_OK) {
    UTL_Log("GEN_GenerateModulus", LOG_FATAL, "Failed to initialize key context.");
    goto error;
  }
  err = GEN_SetThread(thread, genCtx);
  if (err != RC_OK) goto error;

  /* Get info from GEN context */
  heap = GEN_Heap(genCtx);
  if (heap == NULL) goto error;

  streams = GEN_Streams(genCtx);
  if (streams == NULL) goto error;

  stats = (UTL_STATS*)malloc(sizeof(UTL_STATS));
  if (stats == NULL) goto error;

  iteration = thread;   /* Pretend that each thread is starting on a different iteration */

  /* Make stats */
  if (GEN_SetupStats(stats) != RC_OK) {
    UTL_Log("GEN_GenerateModulus", LOG_INFO, "Failed to initialize stats.");
    goto error;
  }
  err = GEN_SetStats(stats, genCtx);
  if (err != RC_OK) goto error;

  /* Start stats */
  UTL_StartStat(5, stats);

  /* Make connections with all other servers */
  if (UTL_SetupConnections(~0, GEN_Streams(genCtx), GEN_Heap(genCtx), thread) != RC_OK) {
    UTL_Log("GEN_GenerateModulus", LOG_FATAL, "Failed to setup connections.");
    goto error;
  }

  /* Make a few BIGNUM's */
  P_share = UTL_GetBN(heap);
  if (P_share == NULL) goto error;

  Q_share = UTL_GetBN(heap);
  if (Q_share == NULL) goto error;

  N = UTL_GetBN(heap);
  if (N == NULL) goto error;


  /*--------------------------------------------------------------------*/
  /*  Loop until RSA modulus is found					*/
  /*--------------------------------------------------------------------*/

 redo:

  do {

    /*--------------------------*/
    /* Initialize the iteration */
    /*--------------------------*/

    /* Prepare the statistics */
    err = UTL_AddCountStat(0, stats);
    if (err != RC_OK) goto error;

    err = UTL_StartStat(7, stats);
    if (err != RC_OK) goto error;

    /* Set the iteration number */
    iteration++;

    /* Prepare the heap */
    state = UTL_GetHeapState(heap); 
    if (state < 0) goto error;

    /* Exit the thread if one of the others has found a modulus */
    if (genDone) goto notFind;

    /* Output information */
    UTL_Log("GEN_GenerateModulus", LOG_TEST, "********** Attempt %d, Thread %d", UTL_CountStat(0, stats), thread);
    if (!(UTL_CountStat(0, stats)%100)) {
      fprintf(stderr, "(%d)", thread);
      UTL_monitor(utlCtx, 0, "Thread %d - Searched *%d* prospective moduli.", thread, UTL_CountStat(0, stats));
    } else if (!(UTL_CountStat(0, stats)%10)) {
      fprintf(stderr, "%d", thread);
      UTL_monitor(utlCtx, 0, "Thread %d - Searched %d prospective moduli.", thread, UTL_CountStat(0, stats));
    }


    /*------------------------------------------------------------------*/
    /*	Pick random private shares.					*/
    /*------------------------------------------------------------------*/

    UTL_StartStat(0, stats);
    if (GEN_PickShares(P_share, Q_share, genCtx) != RC_OK) {
      UTL_Log("GEN_GenerateModulus", LOG_INFO, "Failed to pick private shares.");
      goto error;
    }
    UTL_FinishStat(0, stats);


    /*------------------------------------------------------------------*/
    /*	Generate a candidate modulus					*/
    /*------------------------------------------------------------------*/    

    UTL_StartStat(1, stats);                   /* Generate Modulus stat */
    UTL_AddCountStat(2, GEN_Stat(genCtx));     /* BGW stat */
    UTL_StartStat(6, GEN_Stat(genCtx));        /* BGW stat */

    if (GEN_BGW(N, P_share, Q_share, GEN_R(genCtx), BGW_INTERPOLATE, 0, 0, genCtx) != RC_OK) {
      UTL_Log("GEN_GenerateModulus", LOG_INFO, "Modulus generation failed.");
      goto error;
    }
    
    UTL_FinishStat(6, GEN_Stat(genCtx));       /* BGW stat */
    UTL_FinishStat(1, stats);                  /* Generate Modulus stat */
    UTL_Log("GEN_GenerateModulus", LOG_TEST, "Modulus generated.");


    /*------------------------------------------------------------------*/
    /*	Do trial division on candidate	Notify peers of result.		*/
    /*------------------------------------------------------------------*/

    UTL_StartStat(2, stats);
    if (GEN_TrialDivision(&failed, N, genCtx)!= RC_OK) {
      UTL_Log("GEN_GenerateModulus", LOG_INFO, "Problem with trial division.");
      goto error;
    }
    UTL_FinishStat(2, stats);
    
    if (UTL_Notify(&failed, GEN_Streams(genCtx), GEN_Heap(genCtx)) != RC_OK) {
      UTL_Log("GEN_GenerateModulus", LOG_INFO, "Failed in notification.");
      goto error;
    }
    
    if (failed)	{		/* Candidate failed trial division. */
      UTL_FinishStat(7, stats);
      err = UTL_RevertHeap(state, heap);
      if (err != RC_OK) goto error;
      continue;
    }

    UTL_Log("GEN_GenerateModulus", LOG_TEST, "Passed trial division.");
     

    /*------------------------------------------------------------------*/
    /*	Combine results of primality test.				*/
    /*------------------------------------------------------------------*/

    UTL_AddCountStat(1, stats);    
    UTL_StartStat(3, stats);
    if (GEN_PrimeTest(&failed, N, P_share, Q_share, iteration % numServers, genCtx) != RC_OK) {
      UTL_Log("GEN_GenerateModulus", LOG_INFO, "Cannot determine primality test outcome.");
      goto error;
    }
    UTL_FinishStat(3, stats);
    UTL_FinishStat(7, stats);

    err = UTL_RevertHeap(state, heap);
    if (err != RC_OK) goto error;

  } while (failed);


  /*--------------------------------------------------------------------*/
  /*  !!! This is the critical section !!!                              */
  /*--------------------------------------------------------------------*/

  UTL_mutex_lock(&doneS);
  if (genDone != 0) {
    UTL_mutex_unlock(&doneS);
    goto notFind;
  }
  genDone = 1;
  UTL_mutex_unlock(&doneS);
  err = UTL_RevertHeap(state, heap);
  if (err != RC_OK) goto foundError;
  UTL_Log("GEN_GenerateModulus", LOG_TEST, "Thread %d found the modulus.", thread);


  /*--------------------------------------------------------------------*/
  /*  A modulus has been found!                                         */
  /*--------------------------------------------------------------------*/

  UTL_monitor(utlCtx, 0, "Found modulus!.");
  UTL_Log("GEN_GenerateModulus", LOG_TEST, "Passed primality test.");


  /*--------------------------------------------------------------------*/
  /*  Test results of modulus generation.  Only used in test mode.	*/
  /*--------------------------------------------------------------------*/

  if (GEN_Testing(genCtx) | UTL_GetBoolConfigParam(utlCtx, "Final_Test", FALSE)) {

    Ps = UTL_GetBNArray(numServers, heap);
    Qs = UTL_GetBNArray(numServers, heap);
    P = UTL_GetBN(heap);
    Q = UTL_GetBN(heap);
    testN = UTL_GetBN(heap);
    temp = UTL_GetBN(heap);
    
    /* Send shares to others, get others' shares */
    UTL_SendSameToAll(P_share, streams);
    UTL_SendSameToAll(Q_share, streams);
    UTL_GetFromAll(Ps, streams, heap, stats);
    UTL_GetFromAll(Qs, streams, heap, stats);
    BN_copy(Ps[GEN_MyID(genCtx)], P_share);
    BN_copy(Qs[GEN_MyID(genCtx)], Q_share);
    
    /* Add up shares */
    for (i=0; i<numServers; i++) {
      BN_add(temp, P, Ps[i]);
      BN_copy(P, temp);
    }
    for (i=0; i<numServers; i++) {
      BN_add(temp, Q, Qs[i]);
      BN_copy(Q, temp);
    }
    BN_mul(testN, P, Q, genCtx->bnCtx);
    
    /* Print results */
    txt = BN_bn2hex(P_share);
    UTL_Log("GEN_GenerateModulus", LOG_TEST, "P share = %s", txt);
    free(txt);
    
    txt = BN_bn2hex(Q_share);
    UTL_Log("GEN_GenerateModulus", LOG_TEST, "Q share = %s", txt);
    free(txt);

    txt = BN_bn2hex(P);
    UTL_Log("GEN_GenerateModulus", LOG_TEST, "P = %s  (P has %d bits)", txt, BN_num_bits(P));
    free(txt);
    if (BN_is_prime(P, 10, NULL, GEN_BNContext(genCtx), NULL)) {
      UTL_Log("GEN_GenerateModulus", LOG_TEST, "P is probably prime.");
    } else {
      UTL_Log("GEN_GenerateModulus", LOG_WARNING, "P is NOT prime.");
      goto redo;
      UTL_CloseAllConnections(streams);
      goto error;
    }      

    txt = BN_bn2hex(Q);
    UTL_Log("GEN_GenerateModulus", LOG_TEST, "Q = %s  (Q has %d bits)", txt, BN_num_bits(Q));
    free(txt);
    if (BN_is_prime(Q, 10, NULL, GEN_BNContext(genCtx), NULL)) {
      UTL_Log("GEN_GenerateModulus", LOG_TEST, "Q is probably prime.");
    } else {
      UTL_Log("GEN_GenerateModulus", LOG_WARNING, "Q is NOT prime.");
      goto redo;
      UTL_CloseAllConnections(streams);
      goto error;
    }      

    if (BN_cmp(N, testN) == 0) {
      UTL_Log("GEN_GenerateModulus", LOG_NOTIFY, "*** The modulus N was computed correctly.");
    } else {
      UTL_Log("GEN_GenerateModulus", LOG_FATAL, "*** There was an error in the computation of N!");
      UTL_CloseAllConnections(streams);
      goto error;
    }

  }


  /*--------------------------------------------------------------------*/
  /*  Return N, P_share, and Q_share.                                   */
  /*--------------------------------------------------------------------*/

  err = UTL_RevertHeap(state, heap);
  if (err != RC_OK) goto foundError;
  global_N = N;
  global_P_share = P_share;
  global_Q_share = Q_share;
  globalGenCtx = genCtx;

  return RC_OK;


 error:
 foundError:
  err = RC_ERROR;
  goto cleanUp;

 notFind:
  err = RC_OK;
  goto cleanUp;

 cleanUp:
  UTL_CloseAllConnections(GEN_Streams(genCtx));
  UTL_FreeStats(GEN_Stat(genCtx));
  UTL_RevertHeap(state, heap);
  if (genCtx) GEN_CloseGenCtx(genCtx);
  return err;


}









