/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/***************************************************************/
/*                                                             */
/* GEN_GenerateSharedKey.c                                     */
/*                                                             */
/* Generate a t-out-of-n sharing of the private key            */
/*                                                             */
/***************************************************************/

#include "ITTC.h"


/* This structure is used to construct the vector of shares */
/* Each node in the list points to the start of the list. */
/* The node at the start of the list keeps count of the total */
/* number of nodes in the list */
typedef struct gen_list {
  BIGNUM *share;               /* A server's share */
  struct gen_list *next;       /* The next node in linked-list */
  struct gen_list *start;      /* The head of the linked-list */
  int total;                   /* Total number of shares in list */
} GEN_LIST;


/* Pretty function to print debugging info */
void GEN_PrintInfo(int n, int k, int start, int done) {
  int i;
  char txt[1024];
  char t[1024];

  sprintf(txt, "[");
  for (i=0; i<3; i++) if ( (i>=start) && (i<=start+n-1) ) {
    sprintf(t, " %2d", i);
    strcat(txt, t);
  } else {
    sprintf(t, "   ");
    strcat(txt, t);
  }
  sprintf(t, " ]    ");
  strcat(txt, t);

  sprintf(t, "Splitting %2d out of %2d", k, n);
  strcat(txt, t);
  if (done) {
    sprintf(t, "  ***");
    strcat(txt, t);
  }

  UTL_Log("GEN_GenerateSharedKey", LOG_TEST, "%s", txt);
}


/* Pretty function to print a list of shares */
int GEN_PrintList(GEN_LIST *list, int importance) {
  GEN_LIST *node;
  char *txt;

  node = list;
  while (node) {
    txt = BN_bn2hex(node->share);
    UTL_Log("GEN_GenerateSharedKey", importance, "  share: %s", txt);
    free(txt);
    node = node->next;
  }

  return RC_OK;
}




/*----------------------------------------*/
/* Recursively free the whole linked-list */
/*----------------------------------------*/
int GEN_FreeList(GEN_LIST *list) {

  if (list->next) GEN_FreeList(list->next);
  free(list);

  return RC_OK;
}


/*-----------------------------------------------------------*/
/* Add a share to a certain computer's linked-list of shares */
/*-----------------------------------------------------------*/
int GEN_AddShare(GEN_LIST **vector, int machine, BIGNUM *share) {

  GEN_LIST *node = NULL;

  if (vector[machine] == NULL) {

    vector[machine] = (GEN_LIST*)malloc(sizeof(GEN_LIST));
    if (vector[machine] == NULL) goto error;
    node = vector[machine];
    node->share = NULL;
    node->share = BN_dup(share);
    if (node->share == NULL) goto error;
    node->next = NULL;
    node->start = node;
    node->total = 1;

  } else {
    
    node = vector[machine];
    node->next = (GEN_LIST*)malloc(sizeof(GEN_LIST));
    if (vector[machine] == NULL) goto error;
    node->next->share = NULL;
    node->next->share = BN_dup(share);
    if (node->next->share == NULL) {
      node = node->next;
      goto error;
    }
    node->next->next = NULL;
    node->next->start = node->start;
    vector[machine] = node->next;
    node->total = 1;
    (node->start->total)++;

  }

  return RC_OK;

 error:
  if (node->share) BN_clear_free(node->share);
  if (vector[machine]) free(vector[machine]);
  return RC_ERROR;

}


/*--------------------------------------------------------------------*/
/* Count the number of 1's from bit positions start to end, inclusive */
/*--------------------------------------------------------------------*/
int GEN_Ones(int coal, int start, int end) {

  int total = 0;
  int i;

  for (i=start; i<=end; i++) {
    if (coal & (1<<i)) total++;
  }

  return total;
}


/*-----------------------------------------------------*/
/* Recursive function to generate a k-out-of-n sharing */
/* - All paremeters refer to the specific splitting    */
/*   that's occuring, not total numbers.               */
/*                                                     */
/* n - number of servers                               */
/* k - number of servers required for a decryption     */
/* start - first server involved in this splitting     */
/* total - the total share to be split                 */
/* success - boolean, tells if it's possible that      */
/*           the desired share is somewhere in the     */
/*           current splitting                         */
/* coalition - bit-vector giving the desired coalition */
/* machine - the machine in coalition for which        */
/*           you want the share                        */
/* vector - a linked-list of shares for each server    */
/*          (note: vector IS global)                   */
/*                                                     */
/*-----------------------------------------------------*/
int GEN_Split(int n, int k, BIGNUM *modulo, int start, BIGNUM *total, int success, 
	  int coalition, int machine, GEN_LIST **vector) {

  int i;                /* Loop variable */
  int count = 0;        /* Total number of shares generated */
  int tc;               /* Temporary count */
  int suc;              /* Used to compute the "success" variable to pass for recursion */
  BIGNUM *temp = NULL;         /* Temporary variable used to computer random numbers */
  BIGNUM *tot = NULL;          /* Another temporary variable */
  BIGNUM *tmp = NULL;          /* Another temporary variable */
  BIGNUM *t = NULL;          /* Another temporary variable */
  BN_CTX *bnCtx = NULL;

  int smallHalf;        /* Number of shares in the small partition */
  int bigHalf;          /* Number of shares in the big partition */

  BIGNUM *smallShare = NULL;   /* The total share for the small partition */
  BIGNUM *bigShare = NULL;     /* The total share for the big partition */

  int smallTarget;      /* Number of 1's in the target coalition */
  int bigTarget;        /*   corresponding to each partition */

  /*char *txt = NULL;*/
  /*char t1[1024];*/


  /*------------------*/
  /* Check for errors */
  /*------------------*/

  if ( (k > n) || (n < 1) ) {
    UTL_Log("GEN_GenerateSharedKey", LOG_INFO, "%d out of %d not possible!", k, n);
    goto error;;
  }


  /*------------*/
  /* Initialize */
  /*------------*/

  temp = BN_new();
  if (temp == NULL) goto error;

  tot = BN_new();
  if (tot == NULL) goto error;

  tmp = BN_new();
  if (tmp == NULL) goto error;

  t = BN_new();
  if (t == NULL) goto error;

  smallShare = BN_new();
  if (smallShare == NULL) goto error;

  bigShare = BN_new();
  if (bigShare == NULL) goto error;

  bnCtx = BN_CTX_new();
  if (bnCtx == NULL) goto error;


  /*------------------------------------*/
  /* Check for recursion end conditions */
  /*------------------------------------*/

  /* An equal sharing */
  if (k == n) {

    /* Make shares for all but the first in the group */
    if (!BN_set_word(tot, 0)) goto error;
    for (i=1; i<n; i++) {

      /* Make a random number from [-modulo, modulo), add to list */
      if (!BN_lshift(t, modulo, 1)) goto error;
      if (!BN_rand(temp, BN_num_bits(modulo)+12, 0, 0)) goto error;
      if (!BN_mod(tmp, temp, t, bnCtx)) goto error;
      if (!BN_sub(temp, tmp, modulo)) goto error;
      if (GEN_AddShare(vector, start+i, temp) != RC_OK) goto error;

      /* Keep a running count of total */
      if (!BN_add(tmp, tot, temp)) goto error;
      if (!BN_copy(tot, tmp)) goto error;
    }

    /* Now make the share for the first server so the total is correct */
    if (!BN_sub(temp, total, tot)) goto error;
    if (GEN_AddShare(vector, start, temp) != RC_OK) goto error;

    /* Print pretty stuff */
    /*GEN_PrintInfo(n, k, start, 1);
    for (i=start; i<=start+n-1; i++) {
      sprintf(t1, "                :: Server %d is sharing ", i);
      txt = BN_bn2hex(total);
      if (txt) {
	strcat(t1, txt);
	free(txt);
      }
      UTL_Log("GEN_GenerateSharedKey", LOG_TEST, "%s", t1);
    }
    */

    return n;
  }

  /* Each server gets the whole share */
  if (k == 1) {

    /* Give all servers the same share */
    for (i=0; i<n; i++) {
      if (GEN_AddShare(vector, start+i, total) != RC_OK) goto error;
    }

    /* Print pretty stuff */
    /*
    GEN_PrintInfo(n, k, start, 1);
    for (i=start; i<=start+n-1; i++) {
      sprintf(t1, "                :: Server %d has share ", i);
      txt = BN_bn2hex(total);
      if (txt) {
	strcat(t1, txt);
	free(txt);
      }
      UTL_Log("GEN_GenerateSharedKey", LOG_TEST, "%s", t1);
    }
    */

    return n;
  }


  /*-----------------------*/
  /* Print out information */
  /*-----------------------*/

  /*GEN_PrintInfo(n, k, start, 0);*/


  /*----------------------------------*/
  /* Figure out how to split n in two */
  /*----------------------------------*/

  smallHalf = n/2;
  bigHalf = smallHalf;
  if (n&1) bigHalf++;

  smallTarget = GEN_Ones(coalition, start+bigHalf-1, start+n);
  bigTarget = GEN_Ones(coalition, start, start+bigHalf-1);


  /*-----------------------*/
  /* Now recursively split */
  /*-----------------------*/

  for (i=k; i>=0; i--) {
    if ( (i>=0) && (i<=bigHalf) && (k-i>=0) && (k-i<=smallHalf) ) {

      /* Check if one branch is 0 out of n, then recurse */
      suc = success;
      if (i == 0) {
	if (k-i != smallTarget) suc = 0;
	tc = GEN_Split(smallHalf, k-i, modulo, start + bigHalf, total, suc, coalition, machine, vector);
	if (tc < 0) goto error;
	count += tc;
      } else if (k-i == 0) {
	if (i != bigTarget) suc = 0;
	tc = GEN_Split(bigHalf, i, modulo, start, total, suc, coalition, machine, vector);
	if (tc < 0) goto error;
	count += tc;
      } else {

	/* Make bigShare a random number from [-modulo, modulo) */
	if (!BN_lshift(t, modulo, 1)) goto error;
	if (!BN_rand(temp, BN_num_bits(modulo)+12, 0, 0)) goto error;
	if (!BN_mod(tmp, temp, t, bnCtx)) goto error;
	if (!BN_sub(bigShare, tmp, modulo)) goto error;

	/* Set smallShare to make the total correct */
	IT_sub(smallShare, total, bigShare);

	if (i != bigTarget) suc = 0;
	tc = GEN_Split(bigHalf, i, modulo, start, bigShare, suc, coalition, machine, vector);
	if (tc < 0) goto error;
	count += tc;

	tc = GEN_Split(smallHalf, k-i, modulo, start + bigHalf, smallShare, suc, coalition, machine, vector);
	if (tc < 0) goto error;
	count += tc;
      }

    }
  }


 done:

  /* Finish */
  if (temp) BN_clear_free(temp);
  if (tot) BN_clear_free(tot);
  if (tmp) BN_clear_free(tmp);
  if (t) BN_clear_free(t);
  if (smallShare) BN_clear_free(smallShare);
  if (bigShare) BN_clear_free(bigShare);
  if (bnCtx) BN_CTX_free(bnCtx);

  return count;


 error:
  count = -1;
  goto done;
}



/*---------------------------------------------------*/
/* Generates a k-out-of-n sharing of the private key */
/*---------------------------------------------------*/

int GEN_GenerateSharedKey(RSA_SHARE *key, COM_STREAM *admStream, GEN_CTX *genCtx) {

  int k = 0, n = 0;
  int i;
  int err;
  int count=0;
  int me;
  int sentOK = FALSE;
  /*char *txt = NULL;*/
  BIGNUM *ptr = NULL;
  BIGNUM *total = NULL;
  BIGNUM *modulo = NULL;
  BIGNUM *temp = NULL;
  BIGNUM *tmp = NULL;
  BIGNUM *g = NULL;
  BIGNUM *N = NULL;
  BIGNUM *e = NULL;
  BIGNUM *gshare = NULL;
  GEN_LIST *node = NULL;
  GEN_LIST **vector = NULL;    /* Each server gets a linked-list of shares */
  UTL_STREAMS *streams = NULL;


  /*------------*/
  /* Initialize */
  /*------------*/

  N = KEY_RSA_share_n(key);
  if (N == NULL) goto error;

  e = KEY_RSA_share_e(key);
  if (e == NULL) goto error;

  n = KEY_RSA_share_nparties(key);
  if (n < 0) goto error;

  k = KEY_RSA_share_threshold(key);
  if (k < 0) goto error;

  total = KEY_RSA_share_d(key, 0);
  if (total == NULL) goto error;

  modulo = KEY_RSA_share_n(key);
  if (modulo == NULL) goto error;

  vector = (GEN_LIST**)malloc(n*sizeof(GEN_LIST*));
  if (vector == NULL) goto error;
  for (i=0; i<n; i++) vector[i] = NULL;

  streams = GEN_Streams(genCtx);

  temp = BN_new();
  if (temp == NULL) goto error;

  tmp = BN_new();
  if (tmp == NULL) goto error;

  g = BN_new();
  if (g == NULL) goto error;

  gshare = BN_new();
  if (gshare == NULL) goto error;

  me = GEN_MyID(genCtx);
  
  UTL_monitor(GEN_UTLContext(genCtx), 0, "Generating %d out of %d sharing.", k, n);
  /*
  txt = BN_bn2hex(total);
  if (txt) {
    UTL_Log("GEN_GenerateSharedKey", LOG_TEST, "total: %s", txt);
    free(txt);
  }
  */


  /*-----------------------------*/
  /* Recursively generate shares */
  /*-----------------------------*/

  count = GEN_Split(n, k, modulo, 0, total, 0, 0, 0, vector);
  if (count < 0) goto error;


  /*---------------------------------*/
  /* Print out debugging information */
  /*---------------------------------*/

  /*
  UTL_Log("GEN_GenerateSharedKey", LOG_TEST, "Total number of shares: %d", count);
  for (i=0; i<n; i++) {
    vector[i] = vector[i]->start;
    UTL_Log("GEN_GenerateSharedKey", LOG_INFO, "%d shares for server %d", vector[i]->total, i);
    GEN_PrintList(vector[i], LOG_TEST);
  }
  */

  /*--------------------------------------------------*/
  /* Decide opon a choice for g, put it in RSA key    */
  /*--------------------------------------------------*/

  for (i=0; i<n; i++) {
    vector[i] = vector[i]->start;
  }

  /* Choose a random number in [1, N] */
  if (BN_rand(temp, BN_num_bits(N)+12, 0, 0) < 0) {
    UTL_Log("GEN_GenerateSharedKey", LOG_FATAL, "Error generating random number.");
    goto error;
  }
  if (!BN_mod(tmp, temp, N, GEN_BNContext(genCtx))) goto error;

  /* Send that number to all servers */
  if (UTL_SendSameToAll(tmp, GEN_Streams(genCtx)) != RC_OK) {
    UTL_Log("GEN_GenerateSharedKey", LOG_INFO, "Error sending same to all.");
    goto error;
  }

  /* Get a similar number from all servers, set g to the sum mod N */
  if (UTL_GetSumFromAll(g, tmp, N, GEN_BNContext(genCtx), GEN_Streams(genCtx), 
			GEN_Heap(genCtx), GEN_Stat(genCtx)) != RC_OK) {
    UTL_Log("GEN_GenerateSharedKey", LOG_INFO, "Error getting sum from all.");
    goto error;
  }

  /* set d0 to be g */
  ptr = KEY_RSA_share_d(key, 0);
  BN_copy(ptr, g);

  /* Print out the result */
  /*
  txt = BN_bn2hex(g);
  if (txt) {
    UTL_Log("GEN_GenerateSharedKey", LOG_TEST, "g was chosen to be %s", txt);
    free(txt);
  }
  */


  /*----------------------------------*/
  /* Send shares to all other servers */
  /*----------------------------------*/

  for (i=0; i<n; i++) {
    node = vector[i]->start;
    if ( (i != me) && (UTL_Stream(i, streams) != NULL) ) {
      UTL_Log("GEN_GenerateSharedKey", LOG_TEST, "Sending %d shares to server %d", node->total, i);
      while(node) {
	if (COM_write_bn(UTL_Stream(i, streams), node->share) <= 0) {
	  UTL_Log("GEN_GenerateSharedKey", LOG_INFO, "Error sending BIGNUM.");
	  goto error;
	}
	node = node->next;
      }
    }
  }


  /*--------------------------------------------------*/
  /* Get shares from all other servers, add up shares */
  /*--------------------------------------------------*/

  for (i=0; i<n; i++) {
    node = vector[me]->start;
    if ( (i != GEN_MyID(genCtx)) && (UTL_Stream(i, streams) != NULL) ) {
      UTL_Log("GEN_GenerateSharedKey", LOG_TEST, "Getting %d shares from machine %d", vector[me]->total, i);
      while(node) {
	
	/* Get a BIGNUM */
	if (UTL_GetFromOne(temp, streams, i, GEN_Heap(genCtx), NULL) != RC_OK) {
	  UTL_Log("GEN_GenerateSharedKey", LOG_INFO, "Error receiving BIGNUM.");
	  goto error;
	}

	/* Add it to the correct share */
	if (!BN_add(tmp, node->share, temp)) goto error;
	if (!BN_copy(node->share, tmp)) goto error;
	node = node->next;

      }
    }
  }
    

  /*---------------------------------*/
  /* Print out debugging information */
  /*---------------------------------*/

  vector[me] = vector[me]->start;
  //UTL_Log("GEN_GenerateSharedKey", LOG_TEST, "%d final shares for server %d:", vector[me]->total, me);
  //GEN_PrintList(vector[me], LOG_NOTIFY);


  /*-------------------------------------*/
  /* Send OK code to ADM                 */
  /*-------------------------------------*/

  if (admStream != NULL) {
    if (COM_write_string(admStream, "OK") <= 0) {
      UTL_Log("GEN_GenerateSharedKey", LOG_INFO, "Error sending OK code.");
      sentOK = TRUE;
      goto error;
    }
  }

  sentOK = TRUE;


  /*--------------------------------------------------*/
  /* Send N, e, g to ADM                              */
  /*--------------------------------------------------*/

  /* Start sending the public key data */
  if (admStream != NULL) {
    if (COM_write_bn(admStream, N) <= 0) {
      UTL_Log("GEN_GenerateSharedKey", LOG_INFO, "Error sending N to ADM.");
      goto error;
    }
    if (COM_write_bn(admStream, e) <= 0) {
      UTL_Log("GEN_GenerateSharedKey", LOG_INFO, "Error sending e to ADM.");
      goto error;
    }
    if (COM_write_bn(admStream, g) <= 0) {
      UTL_Log("GEN_GenerateSharedKey", LOG_INFO, "Error sending share to ADM.");
      goto error;
    }
  }


  /*-------------------------------------*/
  /* Add shares to RSA key, send to ADM  */
  /*-------------------------------------*/

  node = vector[me]->start;

  /* Send the total number of shares to ADM */
  if (!BN_set_word(tmp, node->total)) goto error;
  if (admStream != NULL) {
    if (COM_write_bn(admStream, tmp) <= 0) {
      UTL_Log("GEN_GenerateSharedKey", LOG_INFO, "Error sending share to ADM.");
      goto error;
    }
  }

  /* Now save the shares, send g^share to ADM */
  while (node) {

    /* Save share in RSA key */
    KEY_RSA_share_add_d(key, node->share);

    /* Calculate g^share, send to admin */
    if (!IT_mod_exp_NEG(gshare, g, node->share, N, GEN_BNContext(genCtx))) goto error;
    if (admStream != NULL) {
      if (COM_write_bn(admStream, gshare) <= 0) {
	UTL_Log("GEN_GenerateSharedKey", LOG_INFO, "Error sending share to ADM.");
	goto error;
      }
    }
    
    /* Advance in list */
    node = node->next;
  }


  /*---------------------*/
  /* Free memory, return */
  /*---------------------*/

  err = RC_OK;

 done:

  if (vector) {
    for (i=0; i<n; i++) {
      if (vector[i]) GEN_FreeList(vector[i]);
    }
    free(vector);
  }
  if (temp) BN_clear_free(temp);
  if (tmp) BN_clear_free(tmp);
  if (g) BN_clear_free(g);
  if (gshare) BN_clear_free(gshare);

  return err;


 error:
  if ( (!sentOK) && (admStream != NULL) ) 
    COM_write_string(admStream, "ERROR: Unable to complete key generation");
  err = RC_ERROR;
  goto done;
}


