/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*----------------------------------------------------------*/
/*--                                                      --*/
/*--  GEN_Init.c                                          --*/
/*--                                                      --*/
/*--  Functions to handle initialize GEN_CTX              --*/
/*--  structures.                                         --*/
/*--                                                      --*/
/*----------------------------------------------------------*/

#include "ITTC.h"


int GEN_PrepareSieve(GEN_CTX *genCtx) {
  BIGNUM *c, *d;
  BIGNUM *sixty;
  BIGNUM *M;
  BIGNUM *temp, *tmp;
  int sieveP;
  int i;
  int bits;


  /*------------*/
  /* Initialize */
  /*------------*/

  sieveP = GEN_SieveP(genCtx);
  M = GEN_M(genCtx);
  tmp = UTL_GetBN(GEN_Heap(genCtx));
  temp = UTL_GetBN(GEN_Heap(genCtx));
  sixty = UTL_GetBN(GEN_Heap(genCtx));


  /*------------------------*/
  /* Do this if not sieving */
  /*------------------------*/

  if (!sieveP) {
    genCtx->numSmallPrimes = 0;
    UTL_Log("GEN_PrepareSieve", LOG_INFO, "No sieving.");
    return RC_OK;
  }


  /*-----------------------*/
  /* Do this if sieving    */
  /*-----------------------*/

  /* set M = product of small primes */ 
  /* Automatically set M to be the right size */
  if (!BN_set_word(sixty, 60)) {
      UTL_Log("GEN_PrepareSieve", LOG_INFO, "Error.");
      goto error;
  }
  if (!BN_set_word(M, 1)) goto error;
  if (!BN_set_word(temp, 1)) goto error;
  bits = GEN_ShareBits(genCtx);
  i = 3;

  while (BN_num_bits(temp) + 6 + RANDOM_BUFFER + BN_num_bits_word(GEN_NumServers(genCtx))-1 < bits) {

    /* Multiply M by another prime */
    if (!BN_set_word(temp, UTL_Prime(i))) goto error;
    if (!BN_mul(tmp, temp, M, genCtx->bnCtx)) goto error;
    if (!BN_copy(M, tmp)) goto error;

    /* Figure out what it'll be if one more multiplication */
    if (!BN_set_word(tmp, UTL_Prime(i))) goto error;
    if (!BN_mul(temp, tmp, M, genCtx->bnCtx)) goto error;
    i++;
  }


  genCtx->numSmallPrimes = i - 3;
  UTL_Log("GEN_PrepareSieve", LOG_INFO, "%d sieving primes, ending with %d.", i-3, UTL_Prime(i-1));
  /*UTL_Log("GEN_PrepareSieve", LOG_TEST, "M = %ld", UTL_IntBN(M));*/


  /* Initialize sieve*/
  genCtx->range = 2*BN_num_bits_word((BN_ULONG)UTL_Prime(i));
  genCtx->sieve = (char*)malloc(GEN_SieveRange(genCtx)*sizeof(char));
  if (genCtx->sieve == NULL) goto error;

  /* Generate CRT coefficients */
  if (!BN_mul(GEN_SixtyM(genCtx), sixty, GEN_M(genCtx), genCtx->bnCtx)) {
      UTL_Log("GEN_PrepareSieve", LOG_INFO, "Error.");
      goto error;
  }

  c = BN_mod_inverse(NULL, sixty, M, GEN_BNContext(genCtx));
  if (c == NULL) goto error;

  d = BN_mod_inverse(NULL, M, sixty, GEN_BNContext(genCtx));
  if (c == NULL) goto error;


  if (!BN_mod_mul(GEN_CRTa(genCtx), M, d, GEN_SixtyM(genCtx), GEN_BNContext(genCtx))) {
      UTL_Log("GEN_PrepareSieve", LOG_INFO, "Error in CRTa.");
      goto error;
  }
  if (!BN_mod_mul(GEN_CRTb(genCtx), sixty, c, GEN_SixtyM(genCtx), GEN_BNContext(genCtx))) {
      UTL_Log("GEN_PrepareSieve", LOG_INFO, "Error in CRTb.");
      goto error;
  }


  /*-----------------*/
  /* Free and return */
  /*-----------------*/
  BN_clear_free(c);
  BN_clear_free(d);
  return RC_OK;


 error:
  if (c) BN_clear_free(c);
  if (d) BN_clear_free(d);
  free(genCtx->sieve);
  return RC_ERROR;

}



int GEN_MakeR(GEN_CTX *genCtx) {

  int i;
  int bits;
  int heapState = -1;
  BIGNUM *bigTemp = NULL;
  BIGNUM *R = NULL;

  bits = GEN_RBits(genCtx);
  R = GEN_R(genCtx);
  heapState = UTL_GetHeapState(GEN_Heap(genCtx));
  if (heapState == -1) goto error;

  bigTemp = UTL_GetBN(GEN_Heap(genCtx));
  if (bigTemp == NULL) goto error;

  if (!BN_set_word(R, 1)) goto error;
  for (i=0; i<bits-1; i++) {
    if (!BN_lshift1(bigTemp, R)) {
      UTL_Log("GEN_MakeR", LOG_INFO, "Error generating R - shift.");
      goto error;
    }
    if (!BN_copy(R, bigTemp)) {
      UTL_Log("GEN_MakeR", LOG_INFO, "Error generating R - copy.");
      goto error;
    }
    if (!BN_set_bit(R, 0)) {
      UTL_Log("GEN_MakeR", LOG_INFO, "Error generating R - set_bit.");
      goto error;
    }
  }

  UTL_RevertHeap(heapState, GEN_Heap(genCtx));
  return RC_OK;

 error:
  if (heapState >= 0) UTL_RevertHeap(heapState, GEN_Heap(genCtx));
  return RC_ERROR;
}




int GEN_ModBin(GEN_CTX *genCtx) {

  int *mb = NULL, *mbp = NULL;   /* mb holds pascal's triangle.  mbp is a temporary array */
  int heapState;
  int i, j;
  int numServers;
  int err = RC_OK;
  BIGNUM *tmp = NULL;

  /* Initialize */
  tmp = UTL_GetBN(GEN_Heap(genCtx));
  if (tmp == NULL) goto error;

  numServers = GEN_NumServers(genCtx);

  heapState = UTL_GetHeapState(GEN_Heap(genCtx));
  if (heapState < 0) goto error;

  if ((mbp = (int*)malloc((numServers+3)*sizeof(int))) == NULL) goto error;
  if (mbp == NULL) goto error;

  if ((mb = (int*)malloc((numServers+3)*sizeof(int))) == NULL) goto error;
  if (mb == NULL) goto error;
  
  /* Initialize arrays for Pascal's Triangle */
  for (i=0; i<numServers+3; i++) {
    mb[i] = 0;
    mbp[i] = 0;
  }
  mb[1] = 1;

  /* Generate Pascal's Triagle */
  for (i=0; i<numServers; i++) {
    for (j=1; j<numServers+2; j++) mbp[j] = mb[j] + mb[j - 1];
    for (j=1; j<numServers+2; j++) mb[j] = mbp[j];
  }

  /* Convert Pascal's Triangle into BIGNUM's */
  for (i=0; i<numServers; i++) {
    if (!BN_set_word(GEN_Binomials(genCtx)[i], mb[i+2])) goto error;
  }

  /*
  for (i=0; i<numServers; i++) {
    UTL_Log("GEN_ModBin", LOG_TEST, "Binomial[%d] = %ld", i, UTL_IntBN(GEN_Binomials(genCtx)[i]));  
  }
  */

  err = RC_OK;


 done:

  /* Free and return */
  if (mb) free(mb);
  if (mbp) free(mbp);
  if (heapState >= 0) UTL_RevertHeap(heapState, GEN_Heap(genCtx));
  return err;


 error:
  err = RC_ERROR;
  goto done;
}



int GEN_MakeFactors(GEN_CTX *genCtx) {

  unsigned long int max;
  long unsigned int *factors = NULL;
  int numServers;
  int i, num;
  int start;
  int err = RC_OK;
  int end;
  int me;
  int *raw = NULL;
  int trialDivEnd;
  int done = 0;
  int numPrimes;


  /*------------*/
  /* Initialize */
  /*------------*/

  trialDivEnd = GEN_TrialDivBound(genCtx);
  if (GEN_NumSievePrimes(genCtx) == 0) {
    start = 1;   /* No sieving */
  } else {
    start = GEN_NumSievePrimes(genCtx)+3;   /* sieving */
  }

  /*---------------------------------------*/
  /* Find the last prime to be in a factor */
  /*---------------------------------------*/

  /* Check if trial division will be used */
  if ( (trialDivEnd > UTL_Prime(start)) ) {
    /* If so, count the number of primes involved */
    for (num=start; UTL_Prime(num) <= trialDivEnd; num++) ;
    end = num - 1;
  } else {
    genCtx->numFactors = 0;
    UTL_Log("GEN_MakeFactors", LOG_WARNING, "Trial division bound too small to do trial division."); 
    return RC_OK;
  }

  numPrimes = end - start;
  if (numPrimes == 0) {
    genCtx->numFactors = 0;
    UTL_Log("GEN_MakeFactors", LOG_WARNING, "No trial division factors."); 
    return RC_OK;
  }


  /*---------------------*/
  /* More initialization */
  /*---------------------*/

  factors = (long unsigned int*)malloc(numPrimes*sizeof(long unsigned int));
  if (factors == NULL) goto error;

  raw = (int*)malloc((numPrimes+start+1)*sizeof(int));
  if (raw == NULL) goto error;

  genCtx->numFactors = 0;
  genCtx->factors = factors;
  me = GEN_MyID(genCtx);
  numServers = GEN_NumServers(genCtx);
  for (i=0; i<numPrimes+start+1; i++) raw[i] = 0;


  /*------------------------------------*/
  /* Decide what size factors should be */
  /*------------------------------------*/

  max = 0;
  switch (GEN_WordBits(genCtx)) {
  case 32:
    for (i=0; i<32; i++) {
      max <<= 1;
      max |= 1;
    }
    break;
  case 64:
    for (i=0; i<64; i++) {
      max <<= 1;
      max |= 1;
    }
    break;
  default:
    UTL_Log("GEN_MakeFactors", LOG_INFO, "Problem with configuration file.");
    goto error;
  }

  /*--------------------------*/
  /* Start generating factors */
  /*--------------------------*/

  for (i=0; !done; i++) {

    /* Make the factor fit in a word */
    factors[i] = 1;
    for (num=start; raw[num] == 1; num++) {         /* Find the starting factor */
      if (num>=104729) {
	UTL_Log("GEN_MakeFactors", LOG_INFO, "Too many factors requested. (1)");
	goto error;
      }
    }
    while ( (num < end) && (factors[i] <= max/UTL_Prime(num)) ) {
      /*UTL_Log("GEN_MakeFactors", LOG_TEST, "Using %ld in trial division.", UTL_Prime(num));*/
      factors[i] *= UTL_Prime(num);
      raw[num] = 1;
      num += numServers;
      if (num>=104729) {
	UTL_Log("GEN_MakeFactors", LOG_INFO, "Too many factors requested. (2)");
	goto error;
      }
    }
    if (factors[i] == 1) {
      done = 1;
    } else {
      genCtx->numFactors++;
      /*UTL_Log("GEN_MakeFactors", LOG_TEST, "Factor[%d] = %lu", i, factors[i]);*/
    }
  }
  
  UTL_Log("GEN_MakeFactors", LOG_INFO, "Trial Division factors cover %d primes, ending with %ld.", 
	  GEN_NumTrialDivFactors(genCtx), UTL_Prime(num));

  free(raw);
  err = RC_OK;

 done:
  return err;


 error:
  if (factors) free(factors);
  if (raw) free(raw);
  err = RC_ERROR;
  goto done;
}
