/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/***************************************************************/
/*                                                             */
/* GEN_PickShares.c                                            */
/*                                                             */
/* Pick a random P and Q.                                      */
/*                                                             */
/***************************************************************/

#include "ITTC.h"


/* A wrapper for BN_rand() */
int GEN_Rands(BIGNUM *rnd, int bits, int top, int bottom) {
  int x;

  //fprintf(stderr, "#a");
  x = BN_rand(rnd, bits, top, bottom);
  //fprintf(stderr, "#b");

  return x;
}


int GEN_DoSieve(BIGNUM *n, BIGNUM *start, int range, int *found, GEN_CTX *genCtx) {
  int i, j;
  int a;
  int inc;
  char *sieve;
  int heapState;
  int numSmallPrimes;
  BIGNUM *temp = NULL;
  BIGNUM *tmp = NULL;
  BN_CTX *bnCtx = NULL;


  /*------------*/
  /* Initialize */
  /*------------*/

  heapState = UTL_GetHeapState(GEN_Heap(genCtx));

  temp = UTL_GetBN(GEN_Heap(genCtx));
  if (temp == NULL) goto error;

  tmp = UTL_GetBN(GEN_Heap(genCtx));
  if (tmp == NULL) goto error;

  numSmallPrimes = GEN_NumSievePrimes(genCtx);

  bnCtx = GEN_BNContext(genCtx);
  if (bnCtx == NULL) goto error;

  sieve = GEN_Sieve(genCtx);
  if (sieve == NULL) goto error;

  for (i=0; i<range; i++) sieve[i] = 0;


  /*-------------------*/
  /* Perform the sieve */
  /*-------------------*/

  for (i=3; i<numSmallPrimes+3; i++) {

    /* Figure the starting point for this prime */
    a = BN_mod_word(start, (unsigned long)UTL_Prime(i));
    if (a!=0) a = UTL_Prime(i) - a;
    
    /* Sieve this prime */
    inc = UTL_Prime(i);
    for (j=a; j<range; j+=inc) sieve[j] = 1;

  }
  

  /*----------------------------------------*/
  /* Find the first one that was not sieved */
  /*----------------------------------------*/

  for (i=0; i<range; i++) {
    if (sieve[i] == 0) {
      if (!BN_set_word(temp, i)) {
	UTL_Log("GEN_Sieve", LOG_INFO, "Error in BN_set_word (2).");
	goto error;
      }
      if (!BN_add(n, temp, start)) {
	UTL_Log("GEN_Sieve", LOG_INFO, "Error in BN_add.");
	goto error;
      }

      *found = 1;
      goto done;
    }
  }

  UTL_Log("GEN_Sieve", LOG_TEST, "No sieve value found.");
  *found = 0;

 done:
  UTL_RevertHeap(heapState, GEN_Heap(genCtx));
  return RC_OK;

 error:
  UTL_Log("GEN_Sieve", LOG_TEST, "Error");
  UTL_RevertHeap(heapState, GEN_Heap(genCtx));
  return RC_ERROR;

}


int GEN_MakeBlumShare(BIGNUM *share, int bits, int modSixty, GEN_CTX *genCtx) {

  BIGNUM *tmp = NULL;       /* Temporary bignum */
  int heapState;

  /* Initialize */
  heapState = UTL_GetHeapState(GEN_Heap(genCtx));
  tmp = UTL_GetBN(GEN_Heap(genCtx));
  if (tmp == NULL) goto error;

  /* Choose share */
  if (!GEN_Rands(tmp, bits - 2 - (BN_num_bits_word(GEN_NumServers(genCtx))-1), 0, 0)) {
    UTL_Log("GEN_MakeBlumShare", LOG_INFO, "Error picking share.");
    goto error;
  }
  if (!BN_lshift(share, tmp, 2)) {
    UTL_Log("GEN_MakeBlumShare", LOG_INFO, "Error in left-shift.");
    goto error;
  }

  /* Make it a Blum integer */
  if (modSixty == 3) {
    
    /* Add 3 to share */
    if (!BN_set_bit(share, 0)) {
      UTL_Log("GEN_MakeBlumShare", LOG_INFO, "Error in BN_set_bit().");
      goto error;
    }
    if (!BN_set_bit(share, 1)) {
      UTL_Log("GEN_MakeBlumShare", LOG_INFO, "Error in BN_set_bit().");
      goto error;
    }

  }

  UTL_RevertHeap(heapState, GEN_Heap(genCtx));
  return RC_OK;
  
 error:
  UTL_Log("GEN_MakeBlumShare", LOG_INFO, "Error");
  UTL_RevertHeap(heapState, GEN_Heap(genCtx));
  return RC_ERROR;

}


int GEN_MakeSieveShare(BIGNUM *share, int bits, int modSixty, GEN_CTX *genCtx) {

  int me;
  int err = RC_OK;
  int i;
  int numServers;
  int found;
  int heapState;
  int randBits;
  BIGNUM *tmp = NULL;       /* Temporary bignum */
  BIGNUM *temp = NULL;      /* Temporary bignum */
  BIGNUM *t1 = NULL, *t2 = NULL;
  BIGNUM *M = NULL;
  BIGNUM *a = NULL;
  BIGNUM *b = NULL;
  BIGNUM *test = NULL;
  BN_CTX *bnCtx = NULL;


  /*----------------------------------------*/
  /* Initialize                             */
  /*----------------------------------------*/

  heapState = UTL_GetHeapState(GEN_Heap(genCtx));
  me = GEN_MyID(genCtx);
  numServers = GEN_NumServers(genCtx);

  M = GEN_M(genCtx);
  if (M == NULL) goto error;

  bnCtx = GEN_BNContext(genCtx);
  if (bnCtx == NULL) goto error;

  test = UTL_GetBN(GEN_Heap(genCtx));
  if (test == NULL) goto error;

  tmp = UTL_GetBN(GEN_Heap(genCtx));
  if (tmp == NULL) goto error;

  temp = UTL_GetBN(GEN_Heap(genCtx));
  if (temp == NULL) goto error;

  a = UTL_GetBN(GEN_Heap(genCtx));
  if (a == NULL) goto error;

  b = UTL_GetBN(GEN_Heap(genCtx));
  if (b == NULL) goto error;

  t1 = UTL_GetBN(GEN_Heap(genCtx));
  if (t1 == NULL) goto error;

  t2 = UTL_GetBN(GEN_Heap(genCtx));
  if (t2 == NULL) goto error;


  /*---------------------------------------------------------*/
  /* Do the sieving to find a number relatively prime to M.  */
  /*---------------------------------------------------------*/

  found = 0;
  while (!found) {

    /* Choose starting point for sieve */
    if (UTL_Random(temp, M, bnCtx, GEN_Heap(genCtx)) != RC_OK) {
      UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Error picking start.");
      goto error;
    }

    /* Perform the sieve */
    if (GEN_DoSieve(tmp, temp, GEN_SieveRange(genCtx), &found, genCtx) != RC_OK) {
      UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Error sieving.");
      goto error;
    }
  }
  


  /*----------------------------------------*/
  /* Test Code                              */
  /*----------------------------------------*/

  if (GEN_Testing(genCtx)) {

    if (!BN_gcd(temp, tmp, M, bnCtx)) {
      UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Error in BN_GCD.");
      goto error;
    } 
    if (!BN_is_one(temp)) {
      UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Initial share is NOT relatively prime to M. (1)");
      UTL_Log("GEN_MakeSieveShare", LOG_INFO, "GCD(share, M) = %ld", UTL_IntBN(tmp));
      goto error;
    }

    err = UTL_SendSameToAll(tmp, GEN_Streams(genCtx));
    if (err != RC_OK) {
      UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Error in UTL_SendSameToAll().");
      goto error;
    }

    err = UTL_GetProductFromAll(test, tmp, M, GEN_BNContext(genCtx), GEN_Streams(genCtx), GEN_Heap(genCtx), GEN_Stat(genCtx));
    if (err != RC_OK) {
      UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Error in UTL_GetProductFromAll().");
      goto error;
    }

    if (!BN_gcd(temp, test, M, bnCtx)) {
      UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Error in BN_GCD.");
      goto error;
    } 
    if (!BN_is_one(temp)) {
      UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Modulus is NOT relatively prime to M. (1')");
      goto error;
    }

  }


  /*---------------------------------------------------------------*/
  /* The product of the servers' sieve results is relatively prime */
  /* to M, so right now we have a multiplicitive sharing of a      */
  /* number relatively prime to M.  We use the BGW function to     */
  /* convert this multiplicative sharing into an additive sharing. */
  /*---------------------------------------------------------------*/

  /* First do servers 0 and 1 */    /* !!! Use scalar-scalar BGW here */
  if (me == 0) {
    if (!BN_copy(a, tmp)) goto error;
    if (!BN_zero(b)) goto error;
  } else if (me == 1) {
    if (!BN_zero(a)) goto error;
    if (!BN_copy(b, tmp)) goto error;
  } else {
    if (!BN_zero(a)) goto error;
    if (!BN_zero(b)) goto error;
  }
  UTL_AddCountStat(2, GEN_Stat(genCtx));
  UTL_StartStat(6, GEN_Stat(genCtx));
  if (GEN_BGW(share, a, b, M, BGW_ADDITIVE_SHARE, 0, 0, genCtx) != RC_OK) {
    UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Error in first sharing.");
    goto error;
  }
  UTL_FinishStat(6, GEN_Stat(genCtx));
  
  /* Now do the rest */     /* !!! Use scalar-sum BGW here */
  for (i=2; i<numServers; i++) {

    if (!BN_copy(a, share)) goto error;
    if (i == me) {
      if (!BN_copy(b, tmp)) goto error;
    } else {
      if (!BN_zero(b)) goto error;
    }
    UTL_AddCountStat(2, GEN_Stat(genCtx));
    UTL_StartStat(6, GEN_Stat(genCtx));
    if (GEN_BGW(share, a, b, M, BGW_ADDITIVE_SHARE, 0, 0, genCtx) != RC_OK) {
      UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Error in sharing number %d.", i);
      goto error;
    }
    UTL_FinishStat(6, GEN_Stat(genCtx));

  }


  /*----------------------------------------*/
  /* Test Code                              */
  /*----------------------------------------*/

  if (GEN_Testing(genCtx)) {

    if (UTL_SendSameToAll(share, GEN_Streams(genCtx)) != RC_OK) goto error;
    if (UTL_GetSumFromAll(t2, share, M, GEN_BNContext(genCtx), GEN_Streams(genCtx), 
			  GEN_Heap(genCtx), GEN_Stat(genCtx)) != RC_OK) goto error;

    if (!BN_mod(temp, t2, M, bnCtx)) goto error;

    if (BN_cmp(test, temp)) {
      UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Value modulo M has changed! (2)");
      goto error;
    } 

  }


  /*----------------------------------------*/
  /* Now do CRT on share                    */
  /*----------------------------------------*/

  /* 
     if server 0 then       p =  8 mod 15
                            p =  3 mod  4
                            p =  p mod  M

                            q = 11 mod 15
                            q =  3 mod  4
                            q =  q mod  M

     if server != 0 then    p =  0 mod 15
                            p =  0 mod  4
                            p =  p mod  M

                            q =  0 mod 15
                            q =  0 mod  4
                            q =  q mod  M
  */

  if (!BN_set_word(temp, modSixty)) {
    UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Error in CRT. (1b)");
    goto error;
  } 
  if (!BN_mod_mul(t1, GEN_CRTa(genCtx), temp, GEN_SixtyM(genCtx), bnCtx)) {
    UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Error in CRT. (2b)");
    goto error;
  }
  if (!BN_mod_mul(t2, GEN_CRTb(genCtx), share, GEN_SixtyM(genCtx), bnCtx)) {
    UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Error in CRT. (3b)");
    goto error;
  } 
  if (!BN_add(tmp, t1, t2)) {
    UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Error in CRT. (4b)");
    goto error;
  } 
  if (!BN_mod(share, tmp, GEN_SixtyM(genCtx), bnCtx)) {
    UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Error in CRT. (5b)");
    goto error;
  } 


  /*----------------------------------------*/
  /* Test Code                              */
  /*----------------------------------------*/

  if (GEN_Testing(genCtx)) {

    if (UTL_SendSameToAll(share, GEN_Streams(genCtx)) != RC_OK) goto error;
    if (UTL_GetSumFromAll(t2, share, M, GEN_BNContext(genCtx), GEN_Streams(genCtx), 
			  GEN_Heap(genCtx), GEN_Stat(genCtx)) != RC_OK) goto error;

    if (!BN_mod(temp, t2, M, bnCtx)) goto error;

    if (BN_cmp(test, temp)) {
      UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Value modulo M has changed! (3)");
      goto error;
    } 

  }


  /*----------------------------------------*/
  /* Now modify the share into "bits" bits  */
  /*----------------------------------------*/

  /* Now modify share to be a "bits"-bit number */
  randBits = bits - BN_num_bits(M) - 6 - (BN_num_bits_word(GEN_NumServers(genCtx))-1);
  if (randBits < RANDOM_BUFFER) {
    UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Sieving by too many small primes.");
    goto error;
  }
  if (!GEN_Rands(tmp, randBits, 1, 0)) {
    UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Error picking random number.");
    goto error;
  }
  if (!BN_mul(temp, GEN_SixtyM(genCtx), tmp, genCtx->bnCtx)) {
    UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Error in BN_mod_mul.");
    goto error;
  }
  if (!BN_add(tmp, share, temp)) {
    UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Error in BN_add.");
    goto error;
  }
  if (!BN_copy(share, tmp)) {
    UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Error in BN_copy.");
    goto error;
  }


  /*----------------------------------------*/
  /* Test Code                              */
  /*----------------------------------------*/

  if (GEN_Testing(genCtx)) {
    
    if (UTL_SendSameToAll(share, GEN_Streams(genCtx)) != RC_OK) goto error;
    if (UTL_GetSumFromAll(t2, share, M, GEN_BNContext(genCtx), GEN_Streams(genCtx), 
			  GEN_Heap(genCtx), GEN_Stat(genCtx)) != RC_OK) goto error;

    if (!BN_mod(temp, t2, M, bnCtx)) goto error;

    if (BN_cmp(test, temp)) {
      UTL_Log("GEN_MakeSieveShare", LOG_INFO, "Value modulo M has changed! (4)");
      goto error;
    } 

  }


  /*----------------------------------------*/
  /* Finish                                 */
  /*----------------------------------------*/

  UTL_RevertHeap(heapState, GEN_Heap(genCtx));
  return RC_OK;

 error:
  UTL_RevertHeap(heapState, GEN_Heap(genCtx));
  return RC_ERROR;
  
}






int GEN_PickShares(BIGNUM *P_share, BIGNUM *Q_share, GEN_CTX *genCtx) {

  BIGNUM* tmp = NULL;       /* Temporary bignum */
  BIGNUM* temp = NULL;      /* Temporary bignum */
  int bits;
  int heapState;


  /*----------------------------------------*/
  /* Initialize                             */
  /*----------------------------------------*/

  heapState = UTL_GetHeapState(GEN_Heap(genCtx));

  tmp = UTL_GetBN(GEN_Heap(genCtx));
  if (tmp == NULL) goto error;

  temp = UTL_GetBN(GEN_Heap(genCtx));
  if (temp == NULL) goto error;

  bits = GEN_ShareBits(genCtx);


  /*----------------------------------------*/
  /* Make the shares                        */
  /*----------------------------------------*/

  if (GEN_SieveP(genCtx) == 0) {

    if (GEN_MyID(genCtx) == 0) {
      if (GEN_MakeBlumShare(P_share, bits, 3, genCtx) != RC_OK) {
	UTL_Log("GEN_PickShares", LOG_INFO, "Error making share.");
	goto error;
      }
      if (GEN_MakeBlumShare(Q_share, bits, 3, genCtx) != RC_OK) {
	UTL_Log("GEN_PickShares", LOG_INFO, "Error making share.");
	goto error;
      }
    } else {
      if (GEN_MakeBlumShare(P_share, bits, 0, genCtx) != RC_OK) {
	UTL_Log("GEN_PickShares", LOG_INFO, "Error making share.");
	goto error;
      }
      if (GEN_MakeBlumShare(Q_share, bits, 0, genCtx) != RC_OK) {
	UTL_Log("GEN_PickShares", LOG_INFO, "Error making share.");
	goto error;
      }
    }  

  } else {

    if (GEN_MyID(genCtx) == 0) {
      if (GEN_MakeSieveShare(P_share, bits, 23, genCtx) != RC_OK) {
	UTL_Log("GEN_PickShares", LOG_INFO, "Error making share.");
	goto error;
      }
      if (GEN_MakeSieveShare(Q_share, bits, 11, genCtx) != RC_OK) {
	UTL_Log("GEN_PickShares", LOG_INFO, "Error making share.");
	goto error;
      }
    } else {
      if (GEN_MakeSieveShare(P_share, bits, 0, genCtx) != RC_OK) {
	UTL_Log("GEN_PickShares", LOG_INFO, "Error making share.");
	goto error;
      }
      if (GEN_MakeSieveShare(Q_share, bits, 0, genCtx) != RC_OK) {
	UTL_Log("GEN_PickShares", LOG_INFO, "Error making share.");
	goto error;
      }
    }  

  }

  /*----------------------------------------*/
  /* Report the shares                      */
  /*----------------------------------------*/
  
  /*
  UTL_Log("GEN_PickShares", LOG_INFO, "P share = %ld", IntBN(P_share));
  UTL_Log("GEN_PickShares", LOG_INFO, "P share has %d bits", BN_num_bits(P_share));

  UTL_Log("GEN_PickShares", LOG_INFO, "Q share = %ld", IntBN(Q_share));
  UTL_Log("GEN_PickShares", LOG_INFO, "Q share has %d bits", BN_num_bits(Q_share));
  */


  /*----------------------------------------*/
  /* Return the error code                  */
  /*----------------------------------------*/

  UTL_RevertHeap(heapState, GEN_Heap(genCtx));
  return RC_OK;

 error:
  UTL_RevertHeap(heapState, GEN_Heap(genCtx));
  return RC_ERROR;
}

