/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/***************************************************************/
/*                                                             */
/* GEN_PrimeTest.c                                             */
/*                                                             */
/* Interact with other servers to see if the modulus N         */
/* is prime.                                                   */
/*                                                             */
/***************************************************************/

#include "ITTC.h"


int GEN_PrimeTest(int *failed, BIGNUM *N, BIGNUM *P_share, 
		  BIGNUM *Q_share, int id, GEN_CTX *genCtx) {

  int heapState;
  int numServers;     /* The total number of servers */
  BN_CTX *bnCtx = NULL;      /* The bignum context for temporary bignum variables */
  BIGNUM *exp = NULL;        /* Used in exponents */
  BIGNUM *myProduct = NULL;  /* 2^(p+q) */
  BIGNUM *product = NULL;    /* The running product of 2^(pi+qi) */
  BIGNUM *bigTemp = NULL;    /* A temporary BIGNUM */
  BIGNUM *temp = NULL;       /* A temporary BIGNUM */
  BIGNUM *two = NULL;        /* The constant 2 */
  BIGNUM *one = NULL;        /* The constant 1 */
  BIGNUM *R = NULL;
  UTL_HEAP *heap = NULL;
  UTL_STREAMS *streams = NULL;
  void *ret = NULL;
  int err;
  COM_STREAM *str = NULL;


  /*------------*/
  /* Initialize */
  /*------------*/

  *failed = 0;
  numServers = GEN_NumServers(genCtx);
  heapState = UTL_GetHeapState(GEN_Heap(genCtx));

  bnCtx = GEN_BNContext(genCtx);
  if (bnCtx == NULL) goto error;

  R = GEN_R(genCtx);
  if (R == NULL) goto error;

  one = GEN_One(genCtx);
  if (one == NULL) goto error;

  two = GEN_Two(genCtx);
  if (two == NULL) goto error;

  heap = GEN_Heap(genCtx);
  if (heap == NULL) goto error;

  streams = GEN_Streams(genCtx);
  if (streams == NULL) goto error;

  bigTemp = UTL_GetBN(heap);
  if (bigTemp == NULL) goto error;

  temp = UTL_GetBN(heap);
  if (temp == NULL) goto error;

  exp = UTL_GetBN(heap);
  if (exp == NULL) goto error;

  myProduct = UTL_GetBN(heap);
  if (myProduct == NULL) goto error;

  product = UTL_GetBN(heap);
  if (product == NULL) goto error;



  /*---------------------------------------------*/
  /* Compute primality test value: 2^{p+q} mod N */
  /*---------------------------------------------*/

  if (!BN_add(exp, P_share, Q_share)) {
    UTL_Log("GEN_PrimeTest", LOG_INFO, "Error computing primality test values.");
    goto error;
  }
  
  if (!BN_mod_exp(myProduct, two, exp, N, bnCtx)) {
    UTL_Log("GEN_PrimeTest", LOG_INFO, "Error computing primality test values.");
    goto error;
  }



  /*--------------------------------------------*/
  /* Server  id  does this                      */
  /*--------------------------------------------*/

  if (id == GEN_MyID(genCtx)) {

    /* Get the numbers from all of the other servers. */
    
    if (UTL_GetProductFromAll(temp, myProduct, N, GEN_BNContext(genCtx), streams, heap, GEN_Stat(genCtx)) != RC_OK) {
      UTL_Log("GEN_PrimeTest", LOG_INFO, "Error getting from all.");
      goto error;
    }
    if (!BN_copy(myProduct, temp)) {
      UTL_Log("GEN_PrimeTest", LOG_INFO, "Error computing primality test values.");
      goto error;
    }

    /* Compute total test value: 2^{N+1} mod N */

    if (!BN_add(exp, N, one)) {
      UTL_Log("GEN_PrimeTest", LOG_INFO, "Error computing primality test values.");
      goto error;
    }
    if (!BN_mod_exp(product, two, exp, N, bnCtx)) {
      UTL_Log("GEN_PrimeTest", LOG_INFO, "Error computing primality test values.");
      goto error;
    }

    /* Now check if the real product and expected product are the same */

    if (BN_cmp(product, myProduct) == 0) {
      
      *failed = 0;
      if (UTL_SendSameToAll(GEN_Zero(genCtx), streams) != RC_OK) {
	UTL_Log("GEN_PrimeTest", LOG_INFO, "Error sending to all.");
	goto error;
      }
      
    } else {
      
      *failed = 1;
      if (UTL_SendSameToAll(one, streams) != RC_OK) {
	UTL_Log("GEN_PrimeTest", LOG_INFO, "Error sending to all.");
	goto error;
      }

    }

  } else {


    /*-----------------------------------------*/
    /* The other servers do this               */
    /*-----------------------------------------*/

    /* Send myProduct to server  id */

    if (COM_write_bn(GEN_Stream(id, genCtx), myProduct) <= 0) {
      UTL_Log("GEN_PrimeTest", LOG_INFO, "Error sending BIGNUM.");
      goto error;
    }

    /* Get back result from server  id */

    str = GEN_Stream(id, genCtx);
    err = COM_select_v(GEN_COMContext(genCtx), &ret, &str, 1,
		       COM_SELECT_READ | COM_SELECT_CLOSE | COM_SELECT_TIMEOUT);

    if (err == COM_SELECTED_CLOSE) {
      UTL_Log("GEN_PrimeTest", LOG_INFO, "Stream closed.");
      goto error;
    }

    if (err == COM_SELECTED_TIMEOUT) {
      UTL_Log("GEN_PrimeTest", LOG_INFO, "Stream timed out.");
      goto error;
    }

    if (err != COM_SELECTED_READ) {
      UTL_Log("GEN_PrimeTest", LOG_INFO, "Select error.");
      goto error;
    }    

    if (COM_read_bn(GEN_Stream(id, genCtx), product) <= 0) {
      UTL_Log("GEN_PrimeTest", LOG_INFO, "Error reading BIGNUM.");
      goto error;
    }

    if (!BN_cmp(product, one)) {
      *failed = 1;
    } else {
      *failed = 0;
    }

  }    
    
    

  /*-----------------------*/
  /* Return the error code */
  /*-----------------------*/

  UTL_RevertHeap(heapState, GEN_Heap(genCtx));
  return RC_OK;

 error:
  UTL_RevertHeap(heapState, GEN_Heap(genCtx));
  return RC_ERROR;

}

