/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/****************************************************************/
/*								*/
/*	GEN_PrivateKey.c					*/
/*								*/
/*	Generate shares of  d  for the given public exponent e	*/
/*								*/
/*	Returns  GEN_INVALID_PUBLIC_EXP  if gcd(e, phi(N)) > 1	*/
/*								*/
/****************************************************************/

#include "ITTC.h"

void PBN(BIGNUM *bn, char* title) {
  char *txt = NULL;

  txt = BN_bn2hex(bn);
  if (txt) {
    UTL_Log("GEN_PrivateKey", LOG_NOTIFY, "%s = %s", title, txt);
    free(txt);
  }
}


/*--------------------------------------------------------------*/
/*								*/
/*	GEN_CreateAdditiveSharing				*/
/*								*/
/*	Generate shares of a given input modulo a given modulus	*/
/*								*/
/*--------------------------------------------------------------*/

int GEN_CreateAdditiveSharing(BIGNUM **output, BIGNUM *share, 
			      BIGNUM *modulus, GEN_CTX *genCtx) {

  int heapState;
  int numServers, j;
  BIGNUM *sum = NULL;
  BIGNUM *temp = NULL;
  BIGNUM *zero = NULL;


  heapState = UTL_GetHeapState(GEN_Heap(genCtx));
  numServers = GEN_NumServers(genCtx);

  zero = GEN_Zero(genCtx);

  sum = UTL_GetBN(GEN_Heap(genCtx));
  if (sum == NULL) goto error;

  temp = UTL_GetBN(GEN_Heap(genCtx));
  if (temp == NULL) goto error;


  for(j=1; j<numServers; j++) {
    if (UTL_Random(output[j], modulus, GEN_BNContext(genCtx), GEN_Heap(genCtx)) != RC_OK) goto error;
    if (!BN_add(sum, sum, output[j])) goto error;
  }

  if (!BN_mod(sum, sum, modulus, GEN_BNContext(genCtx))) goto error;
  if (!BN_sub(temp, share, sum)) goto error;
  if (BN_cmp(temp, zero) < 0) BN_add(temp, modulus, temp);
  if (!BN_copy(output[0], temp)) goto error;

  if (GEN_Testing(genCtx)) {
    if (!BN_zero(sum)) goto error;
    for(j=0; j<numServers; j++) {
      if (!BN_add(sum, sum, output[j])) goto error;
    }
    if (!BN_mod(sum, sum, modulus, GEN_BNContext(genCtx))) goto error;
    if (BN_cmp(share, sum)) {
      UTL_Log("GEN_CreateAdditiveSharing", LOG_FATAL, "Additive sharing computed incorrectly.");
      UTL_Log("GEN_CreateAdditiveSharing", LOG_ERROR, "share = %ld", UTL_IntBN(share));
      UTL_Log("GEN_CreateAdditiveSharing", LOG_ERROR, "sum = %ld", UTL_IntBN(sum));
      goto error;
    }
  }

  UTL_RevertHeap(heapState, GEN_Heap(genCtx));
  return RC_OK;


 error:
  UTL_RevertHeap(heapState, GEN_Heap(genCtx));
  return RC_ERROR;
}

/*--------------------------------------------------------------*/
/*	GEN_PrivateKey.c					*/
/*--------------------------------------------------------------*/

int GEN_PrivateKey(RSA_SHARE *key, BIGNUM *N, BIGNUM *P_Share, BIGNUM *Q_Share, GEN_CTX *genCtx) {

  int me, numServers, i;
  int small_e;
  int err = RC_OK;
  int heapState;
  int x = 0;
  char *txt;
  BIGNUM *d_me = NULL, *phi_me = NULL, *temp = NULL;
  BIGNUM *temp1 = NULL, *temp2 = NULL, *temp3 = NULL, *tmp = NULL;
  BIGNUM *d = NULL, *P = NULL, *Q = NULL, *phi = NULL, *ethree = NULL;
  BIGNUM *cipher = NULL, *myPlaintext = NULL;
  BIGNUM *zero = NULL, *one = NULL, *two = NULL, *three = NULL;
  BIGNUM **fs = NULL;
  BIGNUM *product = NULL;
  BIGNUM *e = NULL, *phi_mod_e = NULL, *phi_me_mod_e = NULL;
  BIGNUM *N_mod_e = NULL, *sum = NULL, *eta = NULL;
  BIGNUM *gamma = NULL;
  BN_CTX *bnCtx = NULL;         /* Bignum context */
  UTL_HEAP *heap = NULL;

  heapState = UTL_GetHeapState(GEN_Heap(genCtx));
  me = (BN_ULONG) GEN_MyID(genCtx);
  numServers = GEN_NumServers(genCtx);

  bnCtx = GEN_BNContext(genCtx);
  if (bnCtx == NULL) goto error;

  heap = GEN_Heap(genCtx);
  if (heap == NULL) goto error;

  product = UTL_GetBN(heap);
  if (product == NULL) goto error;

  d       = UTL_GetBN(heap);
  if (d == NULL) goto error;

  P       = UTL_GetBN(heap);
  if (P == NULL) goto error;

  Q       = UTL_GetBN(heap);
  if (Q == NULL) goto error;

  phi     = UTL_GetBN(heap);
  if (phi == NULL) goto error;

  ethree  = UTL_GetBN(heap);
  if (ethree == NULL) goto error;

  d_me    = UTL_GetBN(heap);
  if (d_me == NULL) goto error;

  phi_me  = UTL_GetBN(heap);
  if (phi_me == NULL) goto error;

  sum     = UTL_GetBN(heap);
  if (sum == NULL) goto error;

  cipher  = UTL_GetBN(heap);
  if (cipher == NULL) goto error;

  tmp    = UTL_GetBN(heap);
  if (tmp == NULL) goto error;

  temp    = UTL_GetBN(heap);
  if (temp == NULL) goto error;

  temp1   = UTL_GetBN(heap);
  if (temp1 == NULL) goto error;

  temp2   = UTL_GetBN(heap);
  if (temp2 == NULL) goto error;

  temp3   = UTL_GetBN(heap);
  if (temp3 == NULL) goto error;

  myPlaintext = UTL_GetBN(heap);
  if (myPlaintext == NULL) goto error;

  phi_mod_e    = UTL_GetBN(heap);
  if (phi_mod_e == NULL) goto error;

  phi_me_mod_e = UTL_GetBN(heap);
  if (phi_me_mod_e == NULL) goto error;

  N_mod_e      = UTL_GetBN(heap);
  if (N_mod_e == NULL) goto error;

  zero  = GEN_Zero(genCtx);
  if (zero == NULL) goto error;

  one   = GEN_One(genCtx);
  if (one == NULL) goto error;

  two   = GEN_Two(genCtx);
  if (two == NULL) goto error;

  three = GEN_Three(genCtx);
  if (three == NULL) goto error;


  fs = UTL_GetBNArray(numServers, heap);
  if (fs == NULL) goto error;



  /*------------------------------------------------------*/
  /* Start, check for errors  				  */
  /*------------------------------------------------------*/

  UTL_monitor(GEN_UTLContext(genCtx), 0, "Generating keys.");

  if ( (e = key->e = BN_new()) == NULL ) {
    UTL_Log("GEN_PrivateKey", LOG_INFO, "Failed to allocate memory.");
    goto error;
  }


  /*------------------------------------------------------*/
  /* Get public exponent value				  */
  /*------------------------------------------------------*/

  small_e = UTL_GetIntConfigParam(GEN_UTLContext(genCtx), "Public_Key", 
				  GEN_DEFAULT_PUBLIC_KEY);

  if (!BN_set_word(e, small_e)) {
    UTL_Log("GEN_PrivateKey", LOG_INFO, "Can't set word.");
    goto error;
  }


  /*------------------------------------------------------*/
  /* Compute  phi_me  such that  phi_me = p_me + q_me     */
  /*------------------------------------------------------*/

  if (!BN_add(phi_me, P_Share, Q_Share)) goto error;


  /*--------------------------*/
  /* Compute  N+1  mod  e     */
  /*--------------------------*/

  if (!BN_copy(tmp, N)) goto error;
  if (!BN_add_word(tmp, 1)) goto error;
  if (!BN_mod(N_mod_e, tmp, e, bnCtx)) goto error;


  /*-----------------------------------------------------*/
  /* Compute  phi(N)  mod e   using Benaloh's protocol   */
  /*-----------------------------------------------------*/

  if (!BN_mod(phi_me_mod_e, phi_me, e, bnCtx)) goto error;

  /* Write phi_me_mod_e as sum of  numServers  random numbers */

  if (GEN_CreateAdditiveSharing(fs, phi_me_mod_e, e, genCtx) != RC_OK) {
    UTL_Log("GEN_PrivateKey", LOG_INFO, "Can't create additive sharing.");
    goto error;
  }

  /* Send shares of  phi_me_mod_e  to everyone */

  if (UTL_SendToAll(fs, GEN_Streams(genCtx)) != RC_OK) {
    UTL_Log("GEN_PrivateKey", LOG_INFO, "Error sending to all.");
    goto error;
  }

  /* Compute sum modulo  e  from everyone  */

  if (UTL_GetSumFromAll(sum, fs[me], e, GEN_BNContext(genCtx), GEN_Streams(genCtx), heap, GEN_Stat(genCtx)) != RC_OK) {
    UTL_Log("GEN_PrivateKey", LOG_INFO, "Error getting sum from all.");
    goto error;
  }

  /* Broadcast my sum to everyone */

  if (UTL_SendSameToAll(sum, GEN_Streams(genCtx)) != RC_OK) {
    UTL_Log("GEN_PrivateKey", LOG_INFO, "Error sending to all.");
    goto error;
  }

  /* Compute sum modulo  e  from everyone  */

  if (UTL_GetSumFromAll(temp, sum, e, GEN_BNContext(genCtx), GEN_Streams(genCtx), heap, GEN_Stat(genCtx)) != RC_OK) {
    UTL_Log("GEN_PrivateKey", LOG_INFO, "Error getting sum from all.");
    goto error;
  }

  if (GEN_Testing(genCtx)) {
    if (UTL_SendSameToAll(phi_me_mod_e, GEN_Streams(genCtx)) != RC_OK) {
      UTL_Log("GEN_PrivateKey", LOG_INFO, "Failed in SendSameToAll.");
      goto error;
    }
    if (UTL_GetSumFromAll(temp3, phi_me_mod_e, e, GEN_BNContext(genCtx), GEN_Streams(genCtx), heap, GEN_Stat(genCtx)) != RC_OK) {
      UTL_Log("GEN_PrivateKey", LOG_INFO, "Failed in GetSumFromAll.");
      goto error;
    }
    if (BN_cmp(temp, temp3)) {
      UTL_Log("GEN_PrivateKey", LOG_INFO, "phi_mod_e corrupted during sharing.");
      goto error;
    }
  }

  /* Add  N+1 mod e  so  phi_mod_e = N+1-P-Q (mod e) = phi (mod e) */

  if (!BN_sub(phi_mod_e, N_mod_e, temp)) goto error;
  if(!BN_mod(phi_mod_e, phi_mod_e, e, bnCtx)) goto error;
  while (BN_cmp(phi_mod_e, zero) < 1) {
    if (!BN_add(tmp, phi_mod_e, e)) goto error;
    if (!BN_copy(phi_mod_e, tmp)) goto error;
  }

  
  /*----------------------------------------*/
  /* Compute  eta = phi_mod_e^{-1} mod e    */
  /*----------------------------------------*/

  eta = BN_mod_inverse(NULL, phi_mod_e, e, bnCtx);

  if (!eta) {
    UTL_Log("GEN_PrivateKey", LOG_ERROR, "Invalid public exponent.");
    return GEN_INVALID_PUBLIC_EXP;
  }
 
  /*--------------------------------------------*/
  /* Compute  d_i = floor( (eta*phi_i)/e )      */
  /*    Note:  d = (1+eta*(P+Q-N-1))/e          */
  /*--------------------------------------------*/

  if (!BN_mul(temp, phi_me, eta, bnCtx)) goto error;
  if (!BN_div(d_me, temp1, temp, e, bnCtx)) goto error;

  /* Server #0 subtracts floor( (eta)*(N+1)/e )  from its own share d_1  */

  if (me == 0) {
    if (!BN_add(temp, N, one)) goto error;
    if (!BN_mul(tmp, temp, eta, bnCtx)) goto error;
    if (!BN_div(temp1, temp2, tmp, e, bnCtx)) goto error;	// temp1 = (eta)*(N+1)/e
    if (!BN_sub(temp, d_me, temp1)) goto error;
    if (!BN_sub(d_me, temp, one)) goto error;
  }

  BN_clear_free(eta);
  eta = NULL;


  /*--------------------------------*/
  /* Compute adjustment constant    */
  /*--------------------------------*/

  /* Build  temp = (3^e)^(d_me) mod N */

  if (!BN_mod_exp(ethree, three, e, N, bnCtx)) goto error;
  gamma = BN_mod_inverse(NULL, ethree, N, bnCtx);  
  if (gamma == NULL) goto error;
  if (!IT_mod_exp_NEG(myPlaintext, ethree, d_me, N, bnCtx)) goto error;


  /* Send  myPlaintext  to server 0 */
  if (me != 0) {
    if (COM_write_bn(GEN_Stream(0, genCtx), myPlaintext) <= 0) {
      UTL_Log("GEN_PrivateKey", LOG_INFO, "Error sending BIGNUM.");
      goto error;
    }
  }
  
  if (me == 0) {
    
    if (UTL_GetProductFromAll(product, myPlaintext, N, GEN_BNContext(genCtx), 
			      GEN_Streams(genCtx), heap, GEN_Stat(genCtx)) != RC_OK) {
      UTL_Log("GEN_PrivateKey", LOG_INFO, "Error getting product from all.");
      goto error;
    }
    
    /* Keep multiplying by ciphertext until plaintext is obtained */
    for(i=0; i <= numServers + 1; i++) {
      UTL_Log("GEN_PrivateKey", LOG_TEST, "adjusting %d", i);
      /* Loop comparison */
      if (!BN_cmp(product, three)) break;

      /* Update product */
      if (!BN_mod_mul(temp, product, ethree, N, bnCtx)) goto error;
      if (!BN_copy(product, temp)) goto error;

      /* Increment d_me */
      if (!BN_add(tmp, d_me, one)) goto error;
      if (!BN_copy(d_me, tmp)) goto error;

    }

    x = 0;
    if (BN_cmp(product, three)) {   /* Error! */
      UTL_Log("GEN_PrivateKey", LOG_INFO, "Error adjusting server zero's share of d.");
      x = 1;
      goto error;
    }

  }

  BN_clear_free(gamma);
  gamma = NULL;


  /*--------------------------------*/
  /* Final Test                     */
  /*--------------------------------*/

  if (GEN_Testing(genCtx) || UTL_GetBoolConfigParam(GEN_UTLContext(genCtx), "Final_Test", FALSE)) {

    if (UTL_SendSameToAll(d_me, GEN_Streams(genCtx)) != RC_OK) {
      UTL_Log("GEN_PrivateKey", LOG_INFO, "Failed in SendSameToAll.");
      goto error;
    }
    if (UTL_SendSameToAll(P_Share, GEN_Streams(genCtx)) != RC_OK) {
      UTL_Log("GEN_PrivateKey", LOG_INFO, "Failed in SendSameToAll.");
      goto error;
    }
    if (UTL_SendSameToAll(Q_Share, GEN_Streams(genCtx)) != RC_OK) {
      UTL_Log("GEN_PrivateKey", LOG_INFO, "Failed in SendSameToAll.");
      goto error;
    }

    /* d */
    if (UTL_GetSumFromAll(d, d_me, NULL, GEN_BNContext(genCtx), GEN_Streams(genCtx), heap, GEN_Stat(genCtx)) != RC_OK) {
      UTL_Log("GEN_PrivateKey", LOG_INFO, "Failed in GetSumFromAll.");
      goto error;
    }

    /* P */
    if (UTL_GetSumFromAll(P, P_Share, NULL, GEN_BNContext(genCtx), GEN_Streams(genCtx), heap, GEN_Stat(genCtx)) != RC_OK) {
      UTL_Log("GEN_PrivateKey", LOG_INFO, "Failed in GetSumFromAll.");
      goto error;
    }
    /* Q */
    if (UTL_GetSumFromAll(Q, Q_Share, NULL, GEN_BNContext(genCtx), GEN_Streams(genCtx), heap, GEN_Stat(genCtx)) != RC_OK) {
      UTL_Log("GEN_PrivateKey", LOG_INFO, "Failed in GetSumFromAll.");
      goto error;
    }
    
    if (!BN_mod_exp(temp, ethree, d, N, bnCtx)) goto error;

    /* P = P-1, Q = Q-1 */
    if (!BN_sub(tmp, P, one)) goto error;
    if (!BN_copy(P, tmp)) goto error;
    if (!BN_sub(tmp, Q, one)) goto error;
    if (!BN_copy(Q, tmp)) goto error;

    /* phi = (P-1)*(Q-1) */
    if (!BN_mul(phi, P, Q, bnCtx)) goto error;

    if (!BN_mod(phi_mod_e, phi, e, bnCtx)) goto error;

    if (BN_cmp(d, zero) < 0) {
      if (!BN_add(tmp, d, phi)) goto error;
    }


    if (!BN_mod_mul(temp, e, d, phi, GEN_BNContext(genCtx))) goto error;

    while (BN_cmp(temp, zero) < 0) {
      if (!BN_add(tmp, temp, phi)) goto error;
      if (!BN_copy(temp, tmp)) goto error;
    }
    if (x == 1) goto error;
    if (BN_cmp(temp, one)) {
      UTL_Log("GEN_PrivateKey", LOG_INFO, "e and d not inverses mod phi.");
      goto error;
    }

    txt = BN_bn2hex(d);
    if (txt) {
      UTL_Log("GEN_PrivateKey", LOG_NOTIFY, "d = %s  (d has %d bits)", txt, BN_num_bits(d));
      free(txt);
    }

  }


  /*----------------------------------------*/
  /* Add the share to the key structure	    */
  /*----------------------------------------*/

  KEY_RSA_share_add_d(key, d_me);

  err = RC_OK;


 done:
  if (eta) BN_clear_free(eta);
  if (gamma) BN_clear_free(gamma);
  UTL_RevertHeap(heapState, GEN_Heap(genCtx));
  return err;


 error:
  err = RC_ERROR;
  goto done;


  return RC_ERROR;
}                   


