/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	KEY_Init.c							--*/
/*--									--*/
/*--	Initialize key manager.						--*/
/*--	Load share of secret key into memory.				--*/
/*--	Also reads and checks Administrator's password.			--*/
/*------------------------------------------------------------------------*/

#include <sys/types.h>
#include "ITTC.h"
#ifndef WIN32
#include <dirent.h>
#endif

KEY_CTX *
KEY_Init(utl_ctx, id)
     UTL_CTX * utl_ctx;
     int id;
{
  KEY_CTX * ret;
  char * dirname, *filename;
  char dirbuf[256];
  int len;
#ifdef WIN32
  HANDLE dir;
  WIN32_FIND_DATA dp;
#else
  DIR * dir;  
  struct dirent * dp;
#endif
  

  // Load all keys from "key_dir"

  if((dirname = UTL_GetPathServerParam(utl_ctx, "Share_Dir", id, NULL)) == NULL)
    return NULL;

#ifdef WIN32
  strcpy(dirbuf, dirname);
  strcat(dirbuf, "\\*.*");
  if((dir = FindFirstFile(dirbuf, &dp)) == INVALID_HANDLE_VALUE)
	return NULL;
#else
  if((dir = opendir(dirname)) == NULL)
	return NULL;
#endif
    
  if((ret = (KEY_CTX *) malloc(sizeof(KEY_CTX))) == NULL)
    return NULL;

  if((ret->keys = (KEY_CTX_ENT *) malloc(MAX_KEYS * sizeof(KEY_CTX_ENT))) == NULL)
    return NULL;

  ret->dirname = dirname;

  ret->count = 0;
  while (ret->count < MAX_KEYS) {

#ifdef WIN32
	filename = dp.cFileName;
#else
    if((dp = readdir(dir)) == NULL)
      break;
	filename = dp->d_name;
#endif

    len = strlen(filename);
    if(len > 9 /* && strcmp(filename + (len - 4), ".pem") == 0 */) {
      sprintf(dirbuf, "%s/%s", dirname, filename);
      if((ret->keys[ret->count].share = KEY_RSA_share_read(dirbuf)) != NULL) {
        ret->keys[ret->count].id = KEY_RSA_get_ID(ret->keys[ret->count].share);
        ret->keys[ret->count].filename = strdup(dirbuf);
	UTL_Log("KEY_Init", LOG_INFO, "ID=0x%X \t version %d \t  -- %3d keys loaded", 
		ret->keys[ret->count].id, 
		KEY_RSA_share_version(ret->keys[ret->count].share), 
		ret->count+1);
        ++ret->count;
      } else UTL_Log("KEY_Init", LOG_INFO, "%s  -- Rejected.", filename);
    }

#ifdef WIN32
	if(!(FindNextFile(dir, &dp)))
	  break;
#endif

  }

#ifdef WIN32
  FindClose(dir);
#else
  closedir(dir);
#endif

  return ret;
}

int
KEY_CTX_add(ctx, share)
     KEY_CTX * ctx;
     RSA_SHARE * share;
{
  char namebuf[256];
  int i;


  if(ctx->count < MAX_KEYS) {
    i = ctx->count;
    ++ctx->count;
  }
  else {
    i = 0;
    while(i < ctx->count)
      if(ctx->keys[i].share == NULL)
	break;
      else
	++i;
    if(i >= ctx->count)
      return RC_ERROR;
  }

  ctx->keys[i].share = share;
  ctx->keys[i].id = KEY_RSA_get_ID(share);
  sprintf(namebuf, "%s/%08lX.%d", ctx->dirname, ctx->keys[i].id,
	  KEY_RSA_share_version(ctx->keys[i].share));
  ctx->keys[i].filename = strdup(namebuf);

  if(KEY_RSA_share_write(share, namebuf, NULL, NULL, 0) < 0) {
    return RC_ERROR;
  } else {
    return RC_OK;
  }

}

int
KEY_CTX_flushidver(ctx, id, ver)
  KEY_CTX * ctx;
  BN_ULONG id;
  int ver;
{
  int i;

  for(i = 0; i < ctx->count; ++i)
    if(ctx->keys[i].id == id && ctx->keys[i].share != NULL &&
       KEY_RSA_share_version(ctx->keys[i].share) == ver) {
      if(KEY_RSA_share_write(ctx->keys[i].share, ctx->keys[i].filename, NULL, NULL, 0) < 0)
	return RC_ERROR;
      else
	return RC_OK;
    }
  return RC_ERROR;
}

int
KEY_CTX_getversions(ctx, id, pmin, pmax)
     KEY_CTX * ctx;
     BN_ULONG id;
     int * pmin;
     int * pmax;
{
  int min = -1, max = -1;
  int i, v;

  for(i = 0; i < ctx->count; ++i)
    if(ctx->keys[i].id == id && ctx->keys[i].share != NULL) {
      v = KEY_RSA_share_version(ctx->keys[i].share);
      if(min < 0 || v < min)
	min = v;
      if(max < 0 || v > max)
	max = v;
    }
  if(pmin && min >= 0)
    *pmin = min;
  if(pmax && max >= 0)
    *pmax = max;
  return (min >= 0 && max >= 0);
}

RSA_SHARE *
KEY_CTX_getbyid(ctx, id)
     KEY_CTX * ctx;
     BN_ULONG id;
{
  int i;
  RSA_SHARE * ret = NULL;

  for(i = 0; i < ctx->count; ++i) {
    if(ctx->keys[i].id == id && ctx->keys[i].share != NULL) {
      if(ret == NULL || KEY_RSA_share_version(ctx->keys[i].share) >
	 KEY_RSA_share_version(ret))
	ret = ctx->keys[i].share;
    }
  }

  return ret;
}

RSA_SHARE *
KEY_CTX_getbyidver(ctx, id, ver)
     KEY_CTX * ctx;
     BN_ULONG id;
     int ver;
{
  int i;

  for(i = 0; i < ctx->count; ++i)
    if(ctx->keys[i].id == id && ctx->keys[i].share != NULL &&
       KEY_RSA_share_version(ctx->keys[i].share) == ver)
      return ctx->keys[i].share;
  return NULL;
}

RSA_SHARE *
KEY_CTX_getbyindex(ctx, idx)
     KEY_CTX * ctx;
     int idx;
{
  if(idx < ctx->count && idx >= 0)
    return ctx->keys[idx].share;
  else
    return NULL;
}

int
KEY_CTX_deletebyindex(ctx, idx)
     KEY_CTX * ctx;
     int idx;
{
  if(ctx->keys[idx].share != NULL) {
    KEY_RSA_share_free(ctx->keys[idx].share);
    ctx->keys[idx].id = 0;
    unlink(ctx->keys[idx].filename);
    free(ctx->keys[idx].filename);
    ctx->keys[idx].filename = NULL;
    ctx->keys[idx].share = NULL;	// Mark as free
    return RC_OK;
  }
  else
    return RC_ERROR;
}

int
KEY_CTX_deletebyidver(ctx, id, ver)
     KEY_CTX * ctx;
     BN_ULONG id;
     int ver;
{
  int i;

  for(i = 0; i < ctx->count; ++i)
    if(ctx->keys[i].share != NULL && ctx->keys[i].id == id &&
       KEY_RSA_share_version(ctx->keys[i].share) == ver) {
      KEY_CTX_deletebyindex(ctx, i);
      return RC_OK;
    }
  return RC_ERROR;
}

int
KEY_CTX_collapse_idver(ctx, id, ver)
     KEY_CTX * ctx;
     BN_ULONG id;
     int ver;
{
  int i;

  for(i = 0; i < ctx->count; ++i)
    if(ctx->keys[i].share != NULL && ctx->keys[i].id == id &&
       KEY_RSA_share_version(ctx->keys[i].share) != ver)
      KEY_CTX_deletebyindex(ctx, i);
  return RC_OK;
}

void
KEY_CTX_close(ctx)
     KEY_CTX * ctx;
{
  int i;

  for(i = 0; i < ctx->count; ++i)
    if(ctx->keys[i].share != NULL)
      KEY_RSA_share_free(ctx->keys[i].share);
  free(ctx->keys);
  free(ctx);
}
