/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/***************************************************************/
/*                                                             */
/* KEY_get_public_share.c                                      */
/*                                                             */
/* Given a coalition, returns the share of the private key     */
/* necessary to do an encryption with that coalition.          */
/*                                                             */
/***************************************************************/

#include "ITTC.h"


/* This structure is used to construct the vector of shares */
/* Each node in the list points to the start of the list. */
/* The node at the start of the list keeps count of the total */
/* number of nodes in the list */
typedef struct key_list {
  BIGNUM *share;               /* A server's share */
  struct key_list *next;       /* The next node in linked-list */
  struct key_list *start;      /* The head of the linked-list */
  int total;                   /* Total number of shares in list */
} KEY_LIST;


/* Pretty function to print debugging info */
void KEY_PrintInfo(int n, int k, int start, int done) {
  int i;
  char txt[1024];
  char t[1024];

  sprintf(txt, "[");
  for (i=0; i<3; i++) if ( (i>=start) && (i<=start+n-1) ) {
    sprintf(t, " %2d", i);
    strcat(txt, t);
  } else {
    sprintf(t, "   ");
    strcat(txt, t);
  }
  sprintf(t, " ]    ");
  strcat(txt, t);

  sprintf(t, "Splitting %2d out of %2d", k, n);
  strcat(txt, t);
  if (done) strcat(txt, "  ***");

  UTL_Log("KEY_PrintInfo", LOG_TEST, "%s", txt);
}



/*--------------------------------------------------------------------*/
/* Count the number of 1's from bit positions start to end, inclusive */
/*--------------------------------------------------------------------*/
int KEY_Ones(int coal, int start, int end) {

  int total = 0;
  int i;

  for (i=start; i<=end; i++) {
    if (coal & (1<<i)) total++;
  }

  return total;
}


/*-----------------------------------------------------*/
/* Recursive function to return the k-out-of-n share   */
/* - All paremeters refer to the specific splitting    */
/*   that's occuring, not total numbers.               */
/* - Returns 1 for success, 0 for failure,             */
/* - num is the number of the desired share.           */
/*   Note that if success is impossible, num returns   */
/*   the total number of shares.                       */
/*   num must be initialized to zero before splitting  */
/*                                                     */
/* n - number of servers                               */
/* k - number of servers required for a decryption     */
/* start - first server involved in this splitting     */
/* total - the total share to be split                 */
/* success - boolean, tells if it's possible that      */
/*           the desired share is somewhere in the     */
/*           current splitting                         */
/* coalition - bit-vector giving the desired coalition */
/* machine - the machine in coalition for which        */
/*           you want the share                        */
/* vector - a linked-list of shares for each server    */
/*          (note: vector IS global)                   */
/*                                                     */
/*-----------------------------------------------------*/
int KEY_Split(int *num, int n, int k, int start, int success, int coalition, int machine) {

  int i;                /* Loop variable */
  int suc;              /* Used to compute the "success" variable to pass for recursion */
  int tmp;

  int smallHalf;        /* Number of shares in the small partition */
  int bigHalf;          /* Number of shares in the big partition */

  int smallTarget;      /* Number of 1's in the target coalition */
  int bigTarget;        /*   corresponding to each partition */


  /*------------------*/
  /* Check for errors */
  /*------------------*/

  if ( (k > n) || (n < 1) ) {
    UTL_Log("KEY_Split", LOG_FATAL, "%d out of %d not possible!", k, n);
    return -1;
  }


  /*------------------------------------*/
  /* Check for recursion end conditions */
  /*------------------------------------*/

  /* An equal sharing or each server gets the whole share */
  if ( (k == n) || (k == 1) ) {
    if ( (machine >= start) && (machine <= start+n-1) ) {
      /*KEY_PrintInfo(n, k, start, 1);*/
      *num = *num + 1;
      /*if (success) UTL_Log("KEY_Split", LOG_TEST, "Found desired share.");*/
      return success;
    } else {
      /*KEY_PrintInfo(n, k, start, 0);*/
    }
    return 0;
  }


  /*-----------------------*/
  /* Print out information */
  /*-----------------------*/

  /*KEY_PrintInfo(n, k, start, 0);*/


  /*----------------------------------*/
  /* Figure out how to split n in two */
  /*----------------------------------*/

  smallHalf = n/2;
  bigHalf = smallHalf;
  if (n&1) bigHalf++;

  smallTarget = KEY_Ones(coalition, start+bigHalf, start+n-1);
  bigTarget = KEY_Ones(coalition, start, start+bigHalf-1);


  /*-----------------------*/
  /* Now recursively split */
  /*-----------------------*/

  for (i=k; i>=0; i--) {
    if ( (i>=0) && (i<=bigHalf) && (k-i>=0) && (k-i<=smallHalf) ) {

      /* Check if one branch is 0 out of n, then recurse */
      suc = success;
      if (i == 0) {
	if (k-i != smallTarget) suc = 0;
	tmp = KEY_Split(num, smallHalf, k-i, start + bigHalf, suc, coalition, machine);
	if (tmp != 0) return tmp;
      } else if (k-i == 0) {
	if (i != bigTarget) suc = 0;
	tmp = KEY_Split(num, bigHalf, i, start, suc, coalition, machine);
	if (tmp != 0) return tmp;
      } else {
	if (i != bigTarget) suc = 0;
	tmp = KEY_Split(num, bigHalf, i, start, suc, coalition, machine);
	if (tmp != 0) return tmp;
	tmp = KEY_Split(num, smallHalf, k-i, start + bigHalf, suc, coalition, machine);
	if (tmp != 0) return tmp;
      }

    }
  }

  /* Finish */
  return 0;
}


/*---------------------------------------------------*/
/* Returns a g^d share of a public key               */
/*                                                   */
/* Coalition is a bit-vector of servers to include.  */
/* Server is the server whose share it will find     */
/*                                                   */
/* Returns a pointer to the share of the key that    */
/* should be used.                                   */
/*---------------------------------------------------*/
BIGNUM* KEY_RSA_shared_key_get_gd(RSA_SHARE *key, int coalition, int server) {

  int k, n;
  int i;
  BIGNUM *share = NULL;
  int shareNum = 0;
  int me;
  int result;
  char str[40];

  /*------------*/
  /* Initialize */
  /*------------*/

  n = KEY_RSA_shared_key_nparties(key);
  if (n <= 0) return NULL;

  k = KEY_RSA_shared_key_threshold(key);
  if (k <= 0) return NULL;

  me = server;


  /*------------------*/
  /* Check for errors */
  /*------------------*/

  if (KEY_Ones(coalition, 0, 30) != k) {
    UTL_Log("KEY_RSA_shared_key_get_gd", LOG_FATAL, "Can't get coalition %d with %d-out-of-%d sharing.", coalition, k, n);
    return NULL;
  }

  if ((coalition & (1<<me)) == 0) {
    UTL_Log("KEY_RSA_shared_key_get_gd", LOG_FATAL, "ERROR: Server %d is not in coalition %d.", me, coalition);
    return NULL;
  }

  
  /*---------------------------------------------------*/
  /* Go past all the shares belonging to other servers */
  /*---------------------------------------------------*/

  for (i=0; i<server; i++) {
    result = KEY_Split(&shareNum, n, k, 0, 0, 0, i);
    if (result < 0) {
      UTL_Log("KEY_RSA_shared_key_get_gd", LOG_FATAL, "Problem with recursive splitting");
      return NULL;
    }
  }

  /*---------------------------------------------------*/
  /* Recursively find the share for the correct server */
  /*---------------------------------------------------*/

  result =  KEY_Split(&shareNum, n, k, 0, 1, coalition, me);
  shareNum--;       /* Compensate for g^d's starting at 0, whereas d's start at 1 */

  if (result < 0) {
    UTL_Log("KEY_RSA_shared_key_get_gd", LOG_FATAL, "Problem with recursive splitting");
    return NULL;
  } else if (result == 0) {
    UTL_Bit2Str(str, coalition, n);
    UTL_Log("KEY_RSA_shared_key_get_gd", LOG_FATAL, "No g-share found for coalition %s.", str);
    return NULL;
  } else {
    share = KEY_RSA_shared_key_gd(key, shareNum);
    UTL_Bit2Str(str, coalition, n);
    UTL_Log("KEY_RSA_shared_key_get_gd", LOG_INFO, "For coalition %s using g-share %d.", str, shareNum);
  }


  /*---------------------*/
  /* Return              */
  /*---------------------*/

  return share;
}


/*---------------------------------------------------*/
/* Returns a d share of a private key                */
/*                                                   */
/* Coalition is a bit-vector of servers to include.  */
/* Server is the server whose share it will find     */
/*                                                   */
/* Returns a pointer to the share of the key that    */
/* should be used.                                   */
/*---------------------------------------------------*/
BIGNUM* KEY_RSA_share_get_d(RSA_SHARE *key, int coalition, int server) {

  int k, n;
  BIGNUM *share = NULL;
  int shareNum = 0;
  int me;
  int result;
  char str[40];

  /*------------*/
  /* Initialize */
  /*------------*/

  n = KEY_RSA_share_nparties(key);
  if (n <= 0) return NULL;

  k = KEY_RSA_share_threshold(key);
  if (k <= 0) return NULL;

  me = server;


  /*------------------*/
  /* Check for errors */
  /*------------------*/

  if (KEY_Ones(coalition, 0, 30) != k) {
    UTL_Log("KEY_RSA_share_get_d", LOG_FATAL, "Can't get coalition %d with %d-out-of-%d sharing.", coalition, k, n);
    return NULL;
  }

  if ((coalition & (1<<me)) == 0) {
    UTL_Log("KEY_RSA_share_get_d", LOG_FATAL, "ERROR: Server %d is not in coalition %d.", me, coalition);
    return NULL;
  }

  
  /*---------------------------------------------------*/
  /* Recursively find the share for the correct server */
  /*---------------------------------------------------*/

  result =  KEY_Split(&shareNum, n, k, 0, 1, coalition, me);

  if (result < 0) {
    UTL_Log("KEY_RSA_share_get_d", LOG_FATAL, "Problem with recursive splitting");
    return NULL;
  } else if (result == 0) {
    UTL_Bit2Str(str, coalition, n);
    UTL_Log("KEY_RSA_share_get_d", LOG_FATAL, "No d-share found for coalition %s.", str);
    return NULL;
  } else {
    share = KEY_RSA_share_d(key, shareNum);
    UTL_Bit2Str(str, coalition, n);
    UTL_Log("KEY_RSA_share_get_d", LOG_INFO, "For coalition %s using d-share %d.", str, shareNum);
  }


  /*---------------------*/
  /* Return              */
  /*---------------------*/

  return share;
}

