/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	NTD_serve.c							--*/
/*--									--*/
/*--	Service client.							--*/
/*------------------------------------------------------------------------*/

#include "ITTC.h"

int NTD_serve(CTL_CTX *ctx, int op, COM_STREAM *stream)
{

  BIGNUM * m = NULL;
  BIGNUM * r = NULL;
  int coalition, offline, corrupt;
  char comment[256];
  char txt[256];
  BN_ULONG key_id;
  int rc;
  int temp;

  /* Initialize */
  if ( (m = BN_new()) == NULL) goto error;
  if ( (r = BN_new()) == NULL) goto error;

  /* Read parameters for decryption/signing */
  if (UTL_WaitRead(stream) != RC_OK) goto error;
  if(COM_read_bn(stream, m) < 0) {
    UTL_Log("NTD_Serve", LOG_INFO, "Error reading BIGNUM");
    goto error;
  }

  if (UTL_WaitRead(stream) != RC_OK) goto error;
  if(COM_read_int(stream, &coalition) < 0) {
    UTL_Log("NTD_Serve", LOG_INFO, "Error reading coalition");
    goto error;
  }

  if (UTL_WaitRead(stream) != RC_OK) goto error;
  if(COM_read_int(stream, &temp) < 0) {
    UTL_Log("NTD_Serve", LOG_INFO, "Error reading key ID");
    goto error;
  }
  key_id = (BN_ULONG)temp;

  if (UTL_WaitRead(stream) != RC_OK) goto error;
  if(COM_read_int(stream, &corrupt) < 0) {
    UTL_Log("NTD_Serve", LOG_INFO, "Error reading corrupt mask");
    goto error;
  }

  if (UTL_WaitRead(stream) != RC_OK) goto error;
  if(COM_read_int(stream, &offline) < 0) {
    UTL_Log("NTD_Serve", LOG_INFO, "Error reading offline mask");
    goto error;
  }


  /* Do the operation */
  UTL_Bit2Str(txt, coalition, 6);
  rc = RC_ERROR;
  switch (op & TLB_OP_MASK) {

  case TLB_OP_APPLY_DECRYPT:
    sprintf(comment, "[DECRYPT]  coalition = ");
    strcat(comment, txt);
    UTL_Log("NTD_serve", LOG_INFO, comment);
    UTL_monitor(ctx->utlctx, 0, comment);
    rc = NTD_decrypt_normal(r, ctx->keyctx, m, coalition, ctx->serverID, key_id, corrupt, offline);
    break;

  case TLB_OP_APPLY_SIG:
    sprintf(comment, "[SIGN]  coalition = ");
    strcat(comment, txt);
    UTL_Log("NTD_serve", LOG_INFO, comment);
    UTL_monitor(ctx->utlctx, 0, comment);
    rc = NTD_sign_normal(r, ctx->keyctx, m, coalition, ctx->serverID, key_id, corrupt, offline);
    break;

  case TLB_OP_APPLY_HASHED:
    sprintf(comment, "[HASH-SIGN]  coalition = ");
    strcat(comment, txt);
    UTL_Log("NTD_serve", LOG_INFO, comment);
    UTL_monitor(ctx->utlctx, 0, comment);
    rc = NTD_decrypt_hashed(r, ctx->keyctx, m, coalition, ctx->serverID, key_id, corrupt, offline);
    break;

  default:
    goto error;
  }


  if (rc == RC_NO_KEY) {


    UTL_Log("NTD_Serve", LOG_NOTIFY, "Key does not exist");

    /* Begin code */
    if (COM_write_string(stream, "ERROR: Key does not exist") < 0) {
      UTL_Log("NTD_Serve", LOG_NOTIFY, "Error sending error code to ADM");
      goto noKey;
    }

    goto noKey;


  } else if (rc == RC_OK) {


    /* Check if answer should be corrupted */
    if (ctx->status == CTL_CORRUPT) {
      if (!BN_lshift1(m, r)) goto error;
      if (!BN_copy(r, m)) goto error;
    }

    /* Begin code */
    if (COM_write_string(stream, "OK") < 0) {
      UTL_Log("NTD_Serve", LOG_INFO, "Unable to send begin code");
      goto error;
    }

    /* Send decryption */
    if(COM_write_bn(stream, r) < 0) {
      UTL_Log("NTD_Serve", LOG_INFO, "Error writing result of NTD operation");
      goto error;
    }


  } else {

    UTL_Log("NTD_Serve", LOG_INFO, "Signature failed");
    COM_write_string(stream, "ERROR: Unable to perform NTD operation");
    goto error;


  }

  BN_clear_free(m);
  BN_clear_free(r);
  return RC_OK;


 error:
  BN_clear_free(m);
  BN_clear_free(r);
  return RC_ERROR;


 noKey:
  BN_clear_free(m);
  BN_clear_free(r);
  return RC_NO_KEY;

}



