/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	NTD_sig.c							--*/
/*--									--*/
/*--	Apply key share on share server.				--*/
/*------------------------------------------------------------------------*/

#include "ITTC.h"


/*------------------------------------------------------------------------*/
/*--	NTD_decrypt_normal						--*/
/*------------------------------------------------------------------------*/

int
NTD_decrypt_normal(BIGNUM * ret, KEY_CTX * kctx, BIGNUM * msg, 
		   int coalition, int myID,
		   BN_ULONG keyID, int corrupt, int offline)
{
  RSA_SHARE * share = NULL;
  BIGNUM * d_share = NULL;
  BN_CTX * bnctx = NULL;
  BIGNUM *temp = NULL;
  /*char *txt;*/

  /* Find the key based on the key ID */
  share = KEY_CTX_getbyid(kctx, keyID);
  if (share == NULL) {
    UTL_Log("NTD_decrypt_normal", LOG_INFO, "Key ID 0x%X not found", keyID);
    return RC_NO_KEY;
  }

  /* Check that the key is not corrupted */
  if (KEY_RSA_share_k(share) < 0) {
    UTL_Log("NTD_decrypt_normal", LOG_INFO, "Malformed RSA share");
    return RC_ERROR;
  }

  /* Get the correct share of the private key */
  if ( (bnctx = BN_CTX_new()) == NULL) goto error;
  if ( (d_share = BN_new()) == NULL) goto error;
  temp = KEY_RSA_share_get_d(share, coalition, myID);
  if (temp == NULL) {
    UTL_Log("NTD_decrypt_normal", LOG_INFO, "Unable to load d");
    goto error;
  }
  if (!BN_copy(d_share, temp)) goto error;

  /* Do the RSA decryption */
  if (!IT_mod_exp_NEG(ret, msg, d_share, KEY_RSA_share_n(share), bnctx)) goto error;

  /* Free memory and return */
  BN_clear_free(d_share);
  BN_CTX_free(bnctx);
  return RC_OK;

 error:
  BN_clear_free(d_share);
  BN_CTX_free(bnctx);
  return RC_ERROR;

}

/*------------------------------------------------------------------------*/
/*--	NTD_sign_normal							--*/
/*------------------------------------------------------------------------*/

int NTD_sign_normal(BIGNUM * ret, KEY_CTX * kctx, BIGNUM * msg, 
		    int coalition, int myID,
		    BN_ULONG keyID, int corrupt, int offline)
{
  unsigned char * buf = NULL;
  unsigned char * p = NULL;
  int i, len;

  len = BN_num_bytes(msg);
  if((buf = malloc(len)) == NULL)
    return RC_ERROR;
  BN_bn2bin(msg, buf);

  /*----   Check that message is properly formated by (old) PKCS 1   ----*/

  p = buf;
  if(*p++ != 0x01) {	// Pad must start with '01'
    free(buf);
    return RC_ERROR;
  }

  for(i = 0; i < len - 2; ++i, ++p)
    if(*p != 0xFF) {
      if(*p == 0)
	break;
      else {
	free(buf);
	return RC_ERROR;	// Needs a '0' separator
      }
    }

  free(buf);

  if(i == len - 2 || i < 8)	// No message, or not enough pad
    return RC_ERROR;

  /*----   Apply key share to generate signature   ----*/

  return NTD_decrypt_normal(ret, kctx, msg, coalition, myID, keyID, 
			    corrupt, offline);
}

/*------------------------------------------------------------------------*/
/*--	NTD_decrypt_hashed						--*/
/*------------------------------------------------------------------------*/

int NTD_decrypt_hashed(BIGNUM * ret, KEY_CTX * kctx, BIGNUM * msg, 
		       int coalition, int myID,
		       BN_ULONG keyID, int corrupt, int offline)
{
  int err;

  /* Decrypt the message */
  err = NTD_decrypt_normal(ret, kctx, msg, coalition, myID, keyID, 
			   corrupt, offline);
  if (err != RC_OK) {
    UTL_Log("NTD_decrypt_hashed", LOG_INFO, "Normal decryption failed");
    return err;
  }

  /* Now hash the decrypted message */
  if (UTL_Hash(ret, ret) != RC_OK) {
    UTL_Log("NTD_decrypt_hashed", LOG_INFO, "Hash failed.");
    return RC_ERROR;
  }

  /* Return */
  return RC_OK;
}
