/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	REF_do.c							--*/
/*--									--*/
/*--	Do refresh.     						--*/
/*--									--*/
/*------------------------------------------------------------------------*/

#include "ITTC.h"


GEN_CTX* REF_InitializeGenCtx(COM_CTX *context, UTL_CTX *utlCtx) {

  GEN_CTX* genCtx = NULL;

  /* Initialize */
  genCtx = (GEN_CTX*)calloc(1, sizeof(GEN_CTX));
  if (genCtx == NULL) return NULL;
  genCtx->statistics = NULL;

  /* Initialize contexts */
  genCtx->bnCtx = BN_CTX_new();
  if (genCtx->bnCtx == NULL) {
    free(genCtx);
    return NULL;
  }
  genCtx->utlCtx = utlCtx;
  if (genCtx->utlCtx == NULL) {
    free(genCtx->bnCtx);
    free(genCtx);
    return NULL;
  }

  /* Initialize Heap */
  if ( (genCtx->heap = UTL_InitHeap(UTL_GetIntConfigParam(utlCtx, "Heap_Size", 1000))) == NULL) {
    UTL_Log("REF_InitializeGenCtx", LOG_FATAL, "Error initializing heap.");
    free(genCtx->bnCtx);
    free(genCtx);
    return NULL;
  }  
    
  genCtx->streams = NULL;

  /* Return */
  return genCtx;
}




int REF_Refresh(BN_ULONG id, int coalition, BIGNUM *oldN, BIGNUM *olde, int oldNumParties, 
		int oldThreshold, int version, KEY_CTX *keyCtx, COM_CTX *comCtx, UTL_CTX *utlCtx) {

  RSA_SHARE *oldShare = NULL;
  RSA_SHARE *share = NULL;
  GEN_CTX *genCtx = NULL;
  BIGNUM *goodShare = NULL;
  int me;
  COM_INFO info;
  int err = RC_ERROR;
  int active;

  BIGNUM *N = NULL;
  BIGNUM *pubExp = NULL;
  int numServers;
  int threshold;
  int oldVersion;


  /*--------------------------------------------*/
  /* Initialize GEN context                     */
  /*--------------------------------------------*/

  genCtx = REF_InitializeGenCtx(comCtx, utlCtx);
  if (genCtx == NULL) {
    UTL_Log("REF_Refresh", LOG_FATAL, "Failed initialize context.");
    return RC_ERROR;
  }


  /*----------------------------------*/
  /* Find the key based on the key ID */
  /*----------------------------------*/

  oldShare = KEY_CTX_getbyid(keyCtx, id);


  /*------------------*/
  /* Make the new key */
  /*------------------*/

  if(oldShare == NULL) {


    UTL_Log("REF_Refresh", LOG_FATAL, "Share ID 0x%X not found, generating new share", id);

    if ( (N = BN_dup(oldN)) == NULL) {
      UTL_Log("REF_Refresh", LOG_FATAL, "Error in BN_dup()  [1]");
      goto error;
    }
    
    if ( (pubExp = BN_dup(olde)) == NULL) {
      UTL_Log("REF_Refresh", LOG_FATAL, "Error in BN_dup()  [2]");
      goto error;
    }
    
    numServers = oldNumParties;
    if (numServers <= 0) {
      UTL_Log("REF_Refresh", LOG_FATAL, "Error getting the number of servers");
      goto error;
    }
    
    threshold = oldThreshold;
    if (threshold <= 0) {
      UTL_Log("REF_Refresh", LOG_FATAL, "Error getting threshold");
      goto error;
    }
    
    oldVersion = version;
    if (oldVersion < 0) {
      UTL_Log("REF_Refresh", LOG_FATAL, "Error getting version");
      goto error;
    }
    
    if ((share = KEY_RSA_share_create(N, pubExp)) == NULL) {
      UTL_Log("REF_Refresh", LOG_FATAL, "Unable to create RSA key.");
      goto error;
    }
    
    KEY_RSA_share_set_threshold_params(share, numServers, threshold);
    KEY_RSA_share_set_version(share, oldVersion+1);  

    
  } else {


    if ( (N = BN_dup(KEY_RSA_share_n(oldShare))) == NULL) {
      UTL_Log("REF_Refresh", LOG_FATAL, "Error in BN_dup()  [1]");
      goto error;
    }
    if (BN_cmp(N, oldN)) {
      UTL_Log("REF_Refresh", LOG_FATAL, "N from ADM does not match N in share");
      goto error;
    }

    if ( (pubExp = BN_dup(KEY_RSA_share_e(oldShare))) == NULL) {
      UTL_Log("REF_Refresh", LOG_FATAL, "Error in BN_dup()  [2]");
      goto error;
    }
    if (BN_cmp(pubExp, olde)) {
      UTL_Log("REF_Refresh", LOG_FATAL, "e from ADM does not match e in share");
      goto error;
    }
    
    numServers = KEY_RSA_share_nparties(oldShare);
    if (numServers <= 0) {
      UTL_Log("REF_Refresh", LOG_FATAL, "Error getting the number of servers");
      goto error;
    }
    if (numServers != oldNumParties) {
      UTL_Log("REF_Refresh", LOG_FATAL, "Number of servers from ADM does not match number of servers in share");
      goto error;
    }
    
    threshold = KEY_RSA_share_threshold(oldShare);
    if (threshold <= 0) {
      UTL_Log("REF_Refresh", LOG_FATAL, "Error getting threshold");
      goto error;
    }
    if (threshold != oldThreshold) {
      UTL_Log("REF_Refresh", LOG_FATAL, "Threshold from ADM does not match threshold in share");
      goto error;
    }
    
    oldVersion = KEY_RSA_share_version(oldShare);
    if (oldVersion != version) {
      UTL_Log("REF_Refresh", LOG_FATAL, "ADM is requesting version that does not exist  (old = %d, adm = %d)", oldVersion, version);
      goto error;
    }
    if (oldVersion < 0) {
      UTL_Log("REF_Refresh", LOG_FATAL, "Error getting version");
      goto error;
    }
    
    if ((share = KEY_RSA_share_create(N, pubExp)) == NULL) {
      UTL_Log("REF_Refresh", LOG_FATAL, "Unable to create RSA key.");
      goto error;
    }
    
    KEY_RSA_share_set_threshold_params(share, numServers, threshold);
    KEY_RSA_share_set_version(share, oldVersion+1);  


  }


  /*--------------------------------------------*/
  /* Set up connections                         */
  /*--------------------------------------------*/

  /* Get my ID number */
  info = COM_ctx_get_info(comCtx);
  me = COM_INFO_GET_NUM(info);

  /* Initialize the streams */
  genCtx->streams = UTL_InitStreams(numServers, me, comCtx);
  if (genCtx->streams == NULL) {
    UTL_Log("REF_Refresh", LOG_FATAL, "Error initializing streams.");
    goto error;
  }

  me = GEN_MyID(genCtx);
  if (UTL_SetupConnections(~0, GEN_Streams(genCtx), GEN_Heap(genCtx), 0) != RC_OK) {
    UTL_Log("REF_Refresh", LOG_FATAL, "Failed to setup connections.");
    goto error;
  }


  /*--------------------------------------------------------*/
  /* Decide what this server contributes to the new sharing */
  /*--------------------------------------------------------*/

  active = ( coalition & (1<<me) );
  if (active) {
    goodShare = BN_dup(KEY_RSA_share_get_d(oldShare, coalition, me));
    if (goodShare == NULL)  {
      UTL_Log("REF_Refresh", LOG_FATAL, "Error in BN_dup()");
      goto error;
    }
  } else {
    goodShare = BN_new();
    if (goodShare == NULL)  {
      UTL_Log("REF_Refresh", LOG_FATAL, "Error in BN_new()");
      goto error;
    }
    BN_set_word(goodShare, 0);
  }

  /*
  txt = BN_bn2hex(goodShare);
  if (txt) {
    UTL_Log("REF_Refresh", LOG_TEST, "       Contributing: %s", txt);
    free(txt);
  }
  */


  /*--------------------------------------------------*/
  /* Make a new sharing, do NOTsend g^d values to ADM */
  /* A recover will follow the refresh and will send  */
  /* the public key to the ADM.                       */
  /*--------------------------------------------------*/

  /* Set goodShare to be share 0 of the key, to */
  /* pass it into GEN_GenerateSharedKey */
  KEY_RSA_share_add_d(share, goodShare);

  err = GEN_GenerateSharedKey(share, NULL, genCtx);  
  if (err != RC_OK) {
    UTL_Log("REF_Refresh", LOG_ERROR, "Error sharing key.");
    goto error;
  }


  /*------------------------------*/
  /* Save the new sharing to disk */
  /*------------------------------*/

  err = KEY_CTX_add(keyCtx, share);
  if (err != RC_OK) {
    UTL_Log("REF_Refresh", LOG_ERROR, "Cannot save new key to disk.");
    goto error;
   }


  /*-----------------*/
  /* Free and return */
  /*-----------------*/

  err = RC_OK;


 done:

  if (genCtx) {
    if (GEN_Streams(genCtx)) UTL_CloseAllConnections(GEN_Streams(genCtx));
    if (genCtx->bnCtx) BN_CTX_free(genCtx->bnCtx);
    if (genCtx->heap) UTL_FreeHeap(genCtx->heap);
    free(genCtx);
  }

  if (N) BN_clear_free(N);
  if (pubExp) BN_clear_free(pubExp);
  if (goodShare) BN_clear_free(goodShare);

  return err;


 error:
  if (err != RC_NO_KEY) err = RC_ERROR;
  goto done;


  return RC_ERROR;
}
