/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	TLB_ApplyKey.c							--*/
/*--									--*/
/*--	Contact share servers and apply private key.			--*/
/*--									--*/
/*------------------------------------------------------------------------*/

#include "ITTC.h"



int TLB_CountBits(int x) {
  int i, count;

  for (i=0, count=0; i<32; i++) {
    if ( x & (1<<i) ) count++;
  }

  return count;
}



/*------------------------------------------------------------------------*/
/*--		TLB_DoApplyKey						--*/
/*------------------------------------------------------------------------*/

static int
TLB_DoApplyKey(BIGNUM *sig, BIGNUM *msg, RSA_SHARED_KEY *pubkey, 
	       int flags, int *inCoalition, TLB_INTERNAL_CTX *ctx)
{
  long int offline, corrupt, coalition;
  int heapState;
  int numServers, t;
  int i, j, pos, rand;
  int count, rc, found;
  int maximum;
  BIGNUM **sigs = NULL;
  BIGNUM *iBig = NULL, *randBig = NULL;
  char str[40];
  int mode;
  /*char *txt;*/


  /*----------------------------------------------------*/
  /*	Initialize					*/
  /*----------------------------------------------------*/

  /* Init local variables */
  heapState = UTL_GetHeapState(ctx->heap);

  numServers = KEY_RSA_shared_key_nparties(pubkey);
  if (numServers <= 0) goto error;

  t = KEY_RSA_shared_key_threshold(pubkey);
  if (t <= 0) goto error;

  if ( (iBig = UTL_GetBN(ctx->heap)) == NULL) goto error;
  if ( (randBig = UTL_GetBN(ctx->heap)) == NULL) goto error;
  maximum = UTL_GetIntConfigParam(ctx->UTLctx, "Zero_Knowledge", TLB_MAX_ATTEMPTS);

  /* Check threshold */
  if (t > numServers) {
    UTL_Log("TLB_DoApplyKey", LOG_ERROR, "Threshold too large.");
    goto error;
  }

  /* Set the mode */
  if (inCoalition == NULL) {
    mode = 0;
  } else if (*inCoalition == 0) {
    mode = 1;
  } else if (TLB_CountBits(*inCoalition) == t) {
    mode = 2;
  } else {
    UTL_Log("TLB_DoApplyKey", LOG_ERROR, "Passed inappropriate coalition.");
    goto error;
  }    

  /* Set-up `sigs' as array of BIGNUM's */
  sigs = UTL_GetBNArray(numServers, ctx->heap);
  if (sigs == NULL) {
    UTL_Log("TLB_DoApplyKey", LOG_ERROR, "Out of space on heap.");
    goto error;
  }
  
  /* offline and corrupt sets are bit vectors indicating state	*/
  /* of shares servers.  Initially empty.			*/
  offline = 0;
  corrupt = 0;

  /*----------------------------------------------------*/
  /*	Loop until valid signature is generated		*/
  /*----------------------------------------------------*/

  UTL_Log("TLB_ApplyKey", LOG_INFO, "Applying key 0x%08X ...", KEY_RSA_get_ID(pubkey));
  UTL_monitor(ctx->UTLctx, 0, "Applying key 0x%08X ...", KEY_RSA_get_ID(pubkey));

  do {

    /* Make sure TLB doesn't re-use streams that */
    /* could still have data on them */

    UTL_DeactivateAll(ctx->streams);
    for (i=0; i<numServers; i++) {
      if (UTL_Stream(i, ctx->streams) != NULL) {
	COM_close(UTL_Stream(i, ctx->streams));
	UTL_SetStream(i, NULL, ctx->streams);
      }
    }

    /*--------------------------------------------------*/
    /* Mode 0 or 1:                                     */
    /*	 Build coalition by picking random set of	*/
    /*   t  servers not offline or corrupt.		*/
    /* Mode 2:                                          */
    /*   Coalition is given and need not be computed    */
    /*--------------------------------------------------*/

    if (mode == 2) {

      coalition = *inCoalition;

    } else {

      /* Set the coalition bit-vector to be all that aren't */
      /* either corrupt or offline */
      count = coalition = 0;
      for (i=0; i<numServers; ++i) {
	if ( !(corrupt & (1<<i)) && !(offline & (1<<i))) {
	  coalition |= (1<<i);
	  count++;
	}
      }

      /* Check if there are too few active servers */
      if (count < t) {
	UTL_Log("TLB_DoApplyKey", LOG_ERROR, "Too many bad servers.");
	goto error;
      }

      /* Delete random elements until only t are left */
      for (i=count; i>t; --i) {
	BN_set_word(iBig, i);
	UTL_Random(randBig, iBig, ctx->BNctx, ctx->heap);
	rand = BN_get_word(randBig);
	
	for (pos=0,j=0; pos<numServers; pos++) {
	  if (coalition & (1<<pos)) {
	    if (j == rand) break;
	    j++;
	  }
	}
	coalition &= ~(1<<pos);
      }

      if (mode == 1) *inCoalition = coalition;

    }

    UTL_Bit2Str(str, coalition, numServers);
    UTL_Log("TLB_DoApplyKey", LOG_NOTIFY, "Using coalition %s", str);
    UTL_monitor(ctx->UTLctx, 0, "Using coalition %s.", str);


    /*--------------------------------------------------*/
    /*	Ask servers in coalition to apply their shares	*/
    /*--------------------------------------------------*/

    rc = TLB_UseKey(sigs, msg, coalition, pubkey, &offline, corrupt, flags, ctx);

    /* Offline servers were detected */
    if (rc == TLB_OFFLINE) {
      UTL_Bit2Str(str, offline, numServers);
      UTL_Log("TLB_DoApplyKey", LOG_NOTIFY, "Offline servers:  %s", str);
      UTL_monitor(ctx->UTLctx, 0, "Offline servers detected: %s", str);
      /* LOG_ERROR if given coalition and could not compute signature */
      if (mode == 2) goto error;
      continue;
    }

    /* Check for internal LOG_ERROR */
    if (rc != RC_OK) {
      UTL_Log("TLB_DoApplyKey", LOG_NOTIFY, "Key not used correctly.");
      goto error;
    }


    /*--------------------------------------------------*/
    /*	Check signature					*/
    /*--------------------------------------------------*/

    rc = TLB_CheckSig(sig, sigs, msg, pubkey, coalition, ctx);

    /* Proper signature was generated */
    if ( (rc == RC_OK) && (flags == TLB_TEST_SERVERS) )
      rc = TLB_CORRUPT;

    /* Proper signature was generated */
    if (rc == RC_OK)
      goto done;

    /* LOG_ERROR if the given coalition is not able to compute signature */
    if (mode == 2) goto error;

    /* Check for internal LOG_ERROR */
    if (rc != TLB_CORRUPT) {
      UTL_Log("TLB_DoApplyKey", LOG_ERROR, "Error in TLB_CheckSig.");
      goto error;
    }

    UTL_Log("TLB_DoApplyKey", LOG_NOTIFY, "Checking for corrupt servers.");


    /*--------------------------------------------------*/
    /*	Find corrupt servers				*/
    /*--------------------------------------------------*/

    found = 0;

    for(count=0; (count < maximum) && (!found) ; ++count) {

      for( i=0; i<numServers; ++i) {

	/* Ignore servers not in coalition */
	if (!(coalition & (1<<i)))
	  continue;

	/* Verify a server */
	UTL_Log("TLB_DoApplyKey", LOG_NOTIFY, "Validating server %d.", i);
	rc = TLB_ValidateServer( msg, sigs[i], pubkey, i, coalition, ctx->streams, ctx);

	/* Continue if the server is OK */
	if (rc == RC_OK)
	  continue;
	
	/* Remember if the server went offline */
	if (rc == TLB_OFFLINE) {
	  found = 1;
	  offline |= (1<<i);
	  continue;
	}

	/* Check for internal LOG_ERROR */
	if (rc != TLB_CORRUPT) {
	  UTL_Log("TLB_DoApplyKey", LOG_ERROR, "Error in TLB_ValidateServer.");
	  goto error;
	}

	/* Corrupt server was found */
	found = 1;
	corrupt |= (1<<i);
	if (flags == TLB_TEST_SERVERS) goto corrupt;

      }

    }

    if (!found) {

      UTL_Log("TLB_DoApplyKey", LOG_INFO, "Did not find any corrupt servers.");
      if (flags == TLB_TEST_SERVERS) goto done;
      goto error;

    } else {

      UTL_Bit2Str(str, offline, numServers);
      UTL_Log("TLB_DoApplyKey", LOG_NOTIFY, "Offline servers:  %s", str);
      UTL_monitor(ctx->UTLctx, 0, "Offline servers detected: %s", str);

      UTL_Bit2Str(str, corrupt, numServers);
      UTL_Log("TLB_DoApplyKey", LOG_NOTIFY, "Corrupt servers:  %s", str);
      UTL_monitor(ctx->UTLctx, 0, "Corrupt servers detected: %s", str);

    }

  } while(1);


 done:
  UTL_RevertHeap(heapState, ctx->heap);
  return RC_OK;

 error:
  UTL_RevertHeap(heapState, ctx->heap);
  return RC_ERROR;

 corrupt:
  UTL_Bit2Str(str, corrupt, numServers);
  UTL_Log("TLB_DoApplyKey", LOG_NOTIFY, "Corrupt servers:  %s", str);
  UTL_monitor(ctx->UTLctx, 0, "Corrupt servers detected: %s", str);
  UTL_RevertHeap(heapState, ctx->heap);
  return RC_CORRUPT;

}

/*------------------------------------------------------------------------*/
/*--		TLB_ApplyKey						--*/
/*------------------------------------------------------------------------*/

int
TLB_ApplyKey(BIGNUM *sig, BIGNUM *msg, int flags, int *coalition, RSA_SHARED_KEY *pubkey)
{
  int rc, i;
  TLB_CTX *tlbCtx = NULL;
  TLB_INTERNAL_CTX *ctx = NULL;


  /*----------------------------------------------------*/
  /*	Initialize					*/
  /*----------------------------------------------------*/

  /* Get context from key */
  tlbCtx = KEY_RSA_shared_key_ctx(pubkey);
  if (tlbCtx == NULL) {
    fprintf(stderr, "TLB_ApplyKey: Uninitialized context.\n");
    return RC_ERROR;
  }

  /* Make an internal context */
  ctx = TLB_INTERNAL_CTX_new(tlbCtx);
  if (ctx == NULL) {
    fprintf(stderr, "TLB_ApplyKey: Uninitialized internal context.\n");
    return RC_ERROR;
  }

  /* Get the number of servers from the key */
  ctx->numServers = KEY_RSA_shared_key_nparties(pubkey);

  /* Initialize streams */
  /* TJW: I'm not a server, so me == -1 */
  if ((ctx->streams = UTL_InitStreams(ctx->numServers, -1, ctx->COMctx)) == NULL) {
    fprintf(stderr, "TLB_ApplyKey: can't init streams.\n");
    BN_CTX_free(ctx->BNctx);
    UTL_FreeHeap(ctx->heap);
    free(ctx);
    return NULL;
  }

  /* Free all temporary variables on heap */
  UTL_ReleaseHeap(ctx->heap);

  /* Init streams */
  UTL_DeactivateAll(ctx->streams);


  /*----------------------------------------------------*/
  /*	Do opearation					*/
  /*----------------------------------------------------*/

  rc = TLB_DoApplyKey(sig, msg, pubkey, flags, coalition, ctx);


  /*----------------------------------------------------*/
  /*	Close internal context (leaving global TLB_CTX) */
  /*----------------------------------------------------*/

  UTL_DeactivateAll(ctx->streams);
  for (i=0; i<ctx->numServers; i++) {
    if (UTL_Stream(i, ctx->streams) != NULL) {
      COM_close(UTL_Stream(i, ctx->streams));
      UTL_SetStream(i, NULL, ctx->streams);
    }
  }


  /*----------------------------------------------------*/
  /*	Close internal context (leaving global TLB_CTX) */
  /*----------------------------------------------------*/

  TLB_INTERNAL_CTX_free(ctx);


  return rc;
}


