/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	TLB_UseKey.c							--*/
/*--									--*/
/*--	Asks servers to generate shares of signature.                   --*/
/*--	Returns  TLB_OFFLINE  if some servers are not responding.	--*/
/*--									--*/
/*------------------------------------------------------------------------*/

#include "ITTC.h"


int TLB_UseKey(BIGNUM *sigs[], BIGNUM *msg, long int coalition, 
	       RSA_SHARED_KEY *pubkey, long int *offline, long int corrupt, 
	       int flags, TLB_INTERNAL_CTX *ctx)
{
  
  int numServers;
  int i;
  int rc = RC_OK;
  int timeout;
  int initialOffline;
  char resp[256];
  COM_STREAM * s = NULL;
  UTL_STREAMS *streams = NULL;
  int op, opcode;


  /*----------------------------------------------------*/
  /* Initialize					        */
  /*----------------------------------------------------*/

  numServers = KEY_RSA_shared_key_nparties(pubkey);
  if (numServers <= 0) return RC_ERROR;
  timeout = UTL_GetIntConfigParam(ctx->UTLctx, "Client_Timeout", 5000);
  streams = ctx->streams;
  if (streams == NULL) return RC_ERROR;
  op = (flags & TLB_SIGN_MESSAGE) ? TLB_OP_APPLY_SIG : TLB_OP_APPLY_DECRYPT;
  initialOffline = *offline;


  /*----------------------------------------------------*/
  /* Open streams to servers, send requests		*/
  /*----------------------------------------------------*/

  for (i=0; i<numServers; ++i) {

    /* Ignore servers not in coalition */
    if ( !(coalition & (1<<i)))
      continue;

    /* Connect to server if not already connected */
    s = UTL_Stream(i, streams);
    if (s == NULL) {
      s = COM_connect(ctx->COMctx, COM_TYPE_SHARE_SERVER, i, ITTC_SHARE_SERVER_VPORT);
      UTL_SetStream(i, s, streams);
    }

    /* Check for connection problems */
    if (s == NULL) {
      UTL_Log("TLB_UseKey", LOG_NOTIFY, "Can't open connection to server %d.", i);
      *offline |= 1<<i;
      return TLB_OFFLINE;
    }

    /* Send request to server */
    rc = TLB_SendRequest(s, pubkey, op, msg, coalition, *offline, corrupt, ctx);
    if (rc != RC_OK) {
      UTL_Log("TLB_UseKey", LOG_INFO, "Error sending to server.");
      *offline |= 1<<i;
      return TLB_OFFLINE;
    }

    /* Set timeout for responses */
    COM_set_timeout(UTL_Stream(i, streams), timeout);

  }

  COM_flush(s);


  /*----------------------------------------------------*/
  /*	Collect reponses				*/
  /*----------------------------------------------------*/

  for (i=0; i<numServers; ++i) {

    /* Initialize signature to zero */
    if (!BN_set_word(sigs[i], 0)) return RC_ERROR;

    /* Ignore servers not in coalition */
    if ( !(coalition & (1<<i)))
      continue;

    /* Get the stream to this server */
    s = UTL_Stream(i, streams);
    if (s == NULL) {
      UTL_Log("TLB_UseKey", LOG_INFO, "Server %d offline.", i);
      *offline |= (1<<i);
      continue;
    }

    /* Get header */
    if (UTL_WaitRead(s) != RC_OK) {
      UTL_Log("TLB_UseKey", LOG_INFO, "No response");
      *offline |= (1<<i);
      continue;
    }
    if(COM_read_string(s, resp) < 0) {
      UTL_Log("TLB_UseKey", LOG_INFO, "Unable to read header");
      *offline |= (1<<i);
      continue;
    }
    if (strcmp(resp, "RESPONSE") != 0) {
      UTL_Log("TLB_UseKey", LOG_INFO, "Incorrect header");
      *offline |= (1<<i);
      continue;
    }

    /* Get opcode */
    if (UTL_WaitRead(s) != RC_OK) {
      UTL_Log("TLB_UseKey", LOG_INFO, "No response");
      *offline |= (1<<i);
      continue;
    }
    if(COM_read_int(s, &opcode) < 0) {
      UTL_Log("TLB_UseKey", LOG_INFO, "Unable to read opcode");
      *offline |= (1<<i);
      continue;
    }
    if (opcode != op) {
      UTL_Log("TLB_UseKey", LOG_INFO, "Incorrect opcode");
      *offline |= (1<<i);
      continue;
    }

    /* Get OK code */
    if (UTL_WaitRead(s) != RC_OK) {
      UTL_Log("TLB_UseKey", LOG_INFO, "No response");
      *offline |= (1<<i);
      continue;
    }
    if(COM_read_string(s, resp) < 0) {
      UTL_Log("TLB_UseKey", LOG_INFO, "Unable to read begin code");
      *offline |= (1<<i);
      continue;
    }
    if (strcmp(resp, "OK") != 0) {
      UTL_Log("TLB_UseKey", LOG_INFO, "Begin code not OK:");
      UTL_Log("TLB_UseKey", LOG_INFO, "%s", resp);
      *offline |= (1<<i);
      continue;
    }

    /* Get signature (decryption) */
    if (UTL_WaitRead(s) != RC_OK) {
      UTL_Log("TLB_UseKey", LOG_INFO, "No response");
      *offline |= (1<<i);
      continue;
    }
    if(COM_read_bn(s, sigs[i]) < 0) {
      UTL_Log("TLB_UseKey", LOG_INFO, "Unable to read result of NTD operation");
      *offline |= (1<<i);
      continue;
    }
    if (sigs[i]->neg) {
      UTL_Log("TLB_UseKey", LOG_INFO, "Server %d returned a negative signature.", i);
      *offline |= (1<<i);
      continue;
    }

  }

  if (*offline != initialOffline) return TLB_OFFLINE;

  return RC_OK;

}
