/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	TLB_ValidateServer.c						--*/
/*--									--*/
/*--	Determine if server's share is corrupt				--*/
/*--									--*/
/*------------------------------------------------------------------------*/

#include "ITTC.h"

int
TLB_ValidateServer(BIGNUM * msg, BIGNUM * sig, RSA_SHARED_KEY *pubkey,
		   int server, long int coalition, UTL_STREAMS *streams, 
		   TLB_INTERNAL_CTX * ctx)
{
  BIGNUM *g = NULL, *gd = NULL;
  BIGNUM *z = NULL, *N = NULL;
  BIGNUM *k = NULL, *l = NULL;
  BIGNUM *ret = NULL;
  BIGNUM *tmp1 = NULL, *tmp2 = NULL;
  BN_CTX *bnCtx = NULL;
  COM_STREAM *s = NULL;
  int heapState;
  int rc = RC_ERROR;
  char resp[256];
  int op;
  /*char *txt = NULL;*/


  /*------------------------------*/
  /* Initialize                   */
  /*------------------------------*/

  heapState = UTL_GetHeapState(ctx->heap);
  bnCtx = ctx->BNctx;
  if (bnCtx == NULL) {
    rc = RC_ERROR;
    goto error;
  }
  s = UTL_Stream(server, streams);
  if (s == NULL) {
    rc = RC_ERROR;
    goto error;
  }

  g = KEY_RSA_shared_key_g(pubkey);
  if (g == NULL) {
    UTL_Log("TLB_ValidateServer", LOG_INFO, "Malformed public key (g).");
    rc = RC_ERROR;
    goto error;
  }

  N = KEY_RSA_shared_key_n(pubkey);

  if ( (z = UTL_GetBN(ctx->heap)) == NULL) {
    rc = RC_ERROR;
    goto error;
  }

  if ( (k = UTL_GetBN(ctx->heap)) == NULL) {
    rc = RC_ERROR;
    goto error;
  }

  if ( (l = UTL_GetBN(ctx->heap)) == NULL) {
    rc = RC_ERROR;
    goto error;
  }

  if ( (tmp1 = UTL_GetBN(ctx->heap)) == NULL) {
    rc = RC_ERROR;
    goto error;
  }

  if ( (tmp2 = UTL_GetBN(ctx->heap)) == NULL) {
    rc = RC_ERROR;
    goto error;
  }

  /*------------------------------*/
  /* Construct challange          */
  /*------------------------------*/

  do {
    rc = UTL_Random(k, N, bnCtx, ctx->heap);
    if (rc != RC_OK) {
      UTL_Log("TLB_ValidateServer", LOG_INFO, "Error computing random k");
      rc = RC_ERROR;
      goto error;
    }
  } while BN_is_zero(k);

  do {
    rc = UTL_Random(l, N, bnCtx, ctx->heap);
    if (rc != RC_OK) {
      UTL_Log("TLB_ValidateServer", LOG_INFO, "Error computing random l");
      rc = RC_ERROR;
      goto error;
    }
  } while BN_is_zero(l);

  if (!BN_set_word(k, 2))  {
    rc = RC_ERROR;
    goto error;
  }

  if (!BN_set_word(l, 3))  {
    rc = RC_ERROR;
    goto error;
  }

  if (!IT_mod_exp_NEG(tmp1, msg, k, N, bnCtx))  {
    rc = RC_ERROR;
    goto error;
  }

  if (!IT_mod_exp_NEG(tmp2, g, l, N, bnCtx))  {
    rc = RC_ERROR;
    goto error;
  }

  if (!BN_mod_mul(z, tmp1, tmp2, N, bnCtx))  {
    rc = RC_ERROR;
    goto error;
  }


  /*------------------------------*/
  /* Send request to server       */
  /*------------------------------*/

  rc = TLB_SendRequest( s, pubkey, TLB_OP_APPLY_HASHED, z, coalition, 0, 0, ctx);

  if (rc != RC_OK) {
    UTL_Log("TLB_ValidateServer", LOG_INFO, "Error sending to server.");
    goto offline;
  }


  /*------------------------------*/
  /* Get response from server     */
  /*------------------------------*/

  /* Get header */
  if ( (rc = UTL_WaitRead(s)) != RC_OK) goto offline;
  if(COM_read_string(s, resp) < 0) {
    UTL_Log("TLB_ValidateServer", LOG_INFO, "Unable to read header");
    goto offline;
  }
  if (strcmp(resp, "RESPONSE") != 0) {
    UTL_Log("TLB_ValidateServer", LOG_INFO, "Incorrect header");
    goto offline;
  }

  /* Get opcode */
  if ( (rc = UTL_WaitRead(s)) != RC_OK) goto offline;
  if(COM_read_int(s, &op) < 0) {
    UTL_Log("TLB_ValidateServer", LOG_INFO, "Unable to read opcode");
    goto offline;
  }
  if (op != TLB_OP_APPLY_HASHED) {
    UTL_Log("TLB_ValidateServer", LOG_INFO, "Incorrect opcode");
    goto offline;
  }

  /* Get OK code */
  if ( (rc = UTL_WaitRead(s)) != RC_OK) goto offline;
  if(COM_read_string(s, resp) < 0) {
    UTL_Log("TLB_ValidateServer", LOG_INFO, "Unable to read begin code");
    goto offline;
  }
  if (strcmp(resp, "OK") != 0) {
    UTL_Log("TLB_ValidateServer", LOG_INFO, "Begin code not OK:");
    UTL_Log("TLB_ValidateServer", LOG_INFO, "%s", resp);
    goto offline;
  }

  /* Get response to challenge */
  ret = UTL_GetBN(ctx->heap);
  if (ret == NULL) goto error;
  rc = UTL_GetFromOneSimple(ret, s);
  if (rc != RC_OK) {
    UTL_Log("TLB_ValidateServer", LOG_INFO, "Server %d error getting signature.", server);
    goto offline;
  }


  /*--------------------------------------*/
  /* Get appropriate g^d from public key  */
  /*--------------------------------------*/

  gd = KEY_RSA_shared_key_get_gd(pubkey, coalition, server);

  if (gd == NULL) {
    UTL_Log("TLB_ValidateServer", LOG_INFO, "Malformed public key (gd).");
     goto error;
  }


  /*------------------------------*/
  /* Check response to challange  */
  /*------------------------------*/

  if (!IT_mod_exp_NEG(tmp1, sig, k, N, bnCtx)) goto error;
  if (!IT_mod_exp_NEG(tmp2, gd, l, N, bnCtx)) goto error;
  if (!BN_mod_mul(z, tmp1, tmp2, N, bnCtx)) goto error;

  rc = UTL_Hash(z, z);
  if (rc != RC_OK) {
    UTL_Log("TLB_ValidateServer", LOG_INFO, "Error hashing response.");
    goto error;
  }

  if (BN_cmp(ret, z)) {
    UTL_Log("TLB_ValidateServer", LOG_WARNING, "Server %d is corrupt!", server);
    rc = TLB_CORRUPT;
  } else {
    UTL_Log("TLB_ValidateServer", LOG_INFO, "Server %d is not corrupt.", server);
    rc = RC_OK;
  }
    

  /*------------------------------*/
  /* Free memory and return       */
  /*------------------------------*/

  UTL_RevertHeap(heapState, ctx->heap);
  return rc;


 error:
  UTL_RevertHeap(heapState, ctx->heap);
  return RC_ERROR;


 offline:
  UTL_RevertHeap(heapState, ctx->heap);
  return TLB_OFFLINE;
}


