/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	TLB_init.c							--*/
/*--									--*/
/*--	Initialize TLB context						--*/
/*--									--*/
/*------------------------------------------------------------------------*/

#include "ITTC.h"

TLB_CTX *
TLB_CTX_new(const char *configfile)
{
  TLB_CTX *ctx = NULL;
  int myID;
  COM_TYPE myType;

  if(configfile == NULL)
    return NULL;

  /*--------------------------------------------*/
  /*	Initialize context structure		*/
  /*--------------------------------------------*/

  if ((ctx = malloc(sizeof(TLB_CTX))) == NULL) {
    fprintf(stderr, "TLB_init: Out of memory.\n");
    return NULL;
  }

  ctx->UTLctx = UTL_init_ctx();
  if (ctx->UTLctx == NULL) {
    free(ctx);
    return NULL;
  }
  if (UTL_LoadConfigFile(ctx->UTLctx, configfile) != RC_OK) { 
    fprintf(stderr, "TLB_init: Bad config file.\n");
    free(ctx);
    UTL_free_ctx(ctx->UTLctx);
    return NULL;
  }

  myID = UTL_GetIntConfigParam(ctx->UTLctx, "Client_ID", -1);
  if(myID < 0) {
    myID = UTL_GetIntConfigParam(ctx->UTLctx, "Admin_ID", -1);
    if(myID < 0) {
      free(ctx);
      UTL_free_ctx(ctx->UTLctx);
      return NULL;
    } else {
      myType = COM_TYPE_ADMIN;
    }
  } else {
    myType = COM_TYPE_CLIENT;
  }

  UTL_MonitorInit(ctx->UTLctx, myType, myID);

  if (UTL_OpenLogFile(ctx->UTLctx, myType, myID) != RC_OK) {
    fprintf(stderr, "TLB_init: UTL_OpenLogFile returned error code.\n");
  }


  /*--------------------------------------------*/
  /*	Initialize COM				*/
  /*--------------------------------------------*/

  if ((ctx->COMctx = COM_init(ctx->UTLctx, myType, myID)) == NULL) {
    fprintf(stderr, "TLB_init: NULL COM context.\n");
    free(ctx);
    UTL_free_ctx(ctx->UTLctx);
    return NULL;
  }

  /*COM_ctx_set_flags(ctx->COMctx, COM_FLAG_NONBLOCK);*/
  /*COM_ctx_set_flags(ctx->COMctx, COM_FLAG_BLOCKING_CONNECT);*/

  return ctx;
} 


TLB_INTERNAL_CTX *
TLB_INTERNAL_CTX_new(TLB_CTX *tlbCtx)
{
  TLB_INTERNAL_CTX *ctx = NULL;

  /* Malloc the internal context */
  if ((ctx = (TLB_INTERNAL_CTX*)malloc(sizeof(TLB_INTERNAL_CTX))) == NULL) {
    fprintf(stderr, "TLB_INTERNAL_CTX_new: Out of memory.\n");
    return NULL;
  }

  /* Set TLB context, others that are mirrored from there */
  ctx->tlbCtx = tlbCtx;
  ctx->COMctx = tlbCtx->COMctx;
  ctx->UTLctx = tlbCtx->UTLctx;

  /* Initialize heap */
  if ((ctx->heap = UTL_InitHeap(30)) == NULL) {
    fprintf(stderr, "TLB_INTERNAL_CTX_new: can't init heap.\n");
    free(ctx);
    return NULL;
  }

  /* Initialize bignum context */
  if ((ctx->BNctx = BN_CTX_new()) == NULL) {
    fprintf(stderr, "TLB_INTERNAL_CTX_new: can't init BN.\n");
    UTL_FreeHeap(ctx->heap);
    free(ctx);
    return NULL;
  }

  /* Return */
  return ctx;
}  


void
TLB_CTX_free(ctx)
     TLB_CTX * ctx;
{
  if (ctx) {
    if (ctx->COMctx) COM_ctx_close(ctx->COMctx);
    if (ctx->UTLctx) UTL_free_ctx(ctx->UTLctx);
  }
}


/* Doesn't free TLB_CTX */
void
TLB_INTERNAL_CTX_free(ctx)
     TLB_INTERNAL_CTX * ctx;
{
  if (ctx) {
    if (ctx->BNctx) BN_CTX_free(ctx->BNctx);
    if (ctx->heap) UTL_FreeHeap(ctx->heap);
    if (ctx->streams) UTL_CloseAllConnections(ctx->streams);
  }
}

