/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	UTL_ConfigParams.c						--*/
/*--									--*/
/*--	Utility for handling configuration paramters			--*/
/*--									--*/
/*--	Configuration file structure:					--*/
/*--		A list of	type: value	pairs.  One per line.	--*/
/*--		Comments denoted by  ';'  as first char on line.	--*/
/*--		#INCLUDE filename	includes config values from	--*/
/*--					another file.			--*/
/*--									--*/
/*------------------------------------------------------------------------*/

#include "ITTC.h"
#include <stdio.h>
#include <ctype.h>

#ifdef WIN32
#define strcasecmp(X,Y) _stricmp(X,Y)
#endif

/*------------------------------------------------------------------------*/
/*--	UTL_GetConfigParam						--*/
/*--									--*/
/*--	Returns value of config parameter or default if parameter       --*/
/*--    is not defined.							--*/
/*------------------------------------------------------------------------*/

char *UTL_GetConfigParam(const UTL_CTX *ctx, const char *name, char *deflt) 
{
  int i;

  if (ctx->NumLoaded == 0) {		/* Is config file loaded?	*/
    UTL_Log("UTL_GetConfigParam", LOG_FATAL, 
	  "Config file not loaded!  Accessing: `%s'", name);
    return deflt;
  }

  for(i=0; i<ctx->NumParam; i++)
    if ( strcmp(ctx->ConfigParam[i].type, name) == 0)
      return ctx->ConfigParam[i].value;

  return deflt;
}

/*------------------------------------------------------------------------*/
/*--	UTL_GetPathConfigParam						--*/
/*--									--*/
/*--	Returns value of config parameter or default if parameter       --*/
/*--    is not defined.	The value is interpreted as a pathname.		--*/
/*------------------------------------------------------------------------*/

char *UTL_GetPathConfigParam(UTL_CTX *ctx, const char *name, char *deflt) 
{
  int i;

  if (ctx->NumLoaded == 0) {		/* Is config file loaded?	*/
    UTL_Log("UTL_GetConfigParam", LOG_FATAL, 
	  "Config file not loaded!  Accessing: `%s'", name);
    return deflt;
  }

  for(i=0; i<ctx->NumParam; i++)
    if ( strcmp(ctx->ConfigParam[i].type, name) == 0) {
      if(ctx->ConfigParam[i].path)
	return ctx->ConfigParam[i].path;
      else
	return(ctx->ConfigParam[i].path =
	  UTL_RelativeToFile(NULL, ctx->ConfigParam[i].context, ctx->ConfigParam[i].value));
    }

  return deflt;
}

/*------------------------------------------------------------------------*/
/*--	UTL_GetIntConfigParam						--*/
/*--									--*/
/*--	Returns value of int config parameter or default if parameter   --*/
/*--    is not defined.							--*/
/*------------------------------------------------------------------------*/

int   UTL_GetIntConfigParam(const UTL_CTX *ctx, const char *name, 
			    const int deflt) 
{
  int i;

  if (ctx->NumLoaded == 0) {		/* Is config file loaded?	*/
    UTL_Log("UTL_GetConfigParam", LOG_FATAL, 
	  "Config file not loaded!  Accessing: `%s'", name);
    return deflt;
  }

  for(i=0; i<ctx->NumParam; i++)
    if ( strcmp(ctx->ConfigParam[i].type, name) == 0)
      return atoi(ctx->ConfigParam[i].value);

  return deflt;
}

/*------------------------------------------------------------------------*/
/*--	UTL_GetBoolConfigParam						--*/
/*--									--*/
/*--	Returns value of bool config parameter or default if parameter  --*/
/*--    is not defined.							--*/
/*--	Parameter is considered 'true' if it's value is one of		--*/
/*--   	   '\0'   ||  1   ||  true  ||  on  || yes   (case insensative)	--*/
/*------------------------------------------------------------------------*/

int   UTL_GetBoolConfigParam(const UTL_CTX *ctx, const char *name, 
			     const int deflt) 
{
  int i;

  if (ctx->NumLoaded == 0) {		/* Is config file loaded?	*/
    UTL_Log("UTL_GetConfigParam", LOG_FATAL, 
	  "Config file not loaded!  Accessing: `%s'", name);
    return deflt;
  }

  for(i=0; i<ctx->NumParam; i++) {
    if ( strcmp(ctx->ConfigParam[i].type, name) == 0) {
      if ((*ctx->ConfigParam[i].value == '\0')  ||      
          (*ctx->ConfigParam[i].value == '1')   ||      
          (*ctx->ConfigParam[i].value == 't')   ||      
          (*ctx->ConfigParam[i].value == 'T')   ||      
          (*ctx->ConfigParam[i].value == 'y')   ||      
          (*ctx->ConfigParam[i].value == 'Y')   ||      
          (*(ctx->ConfigParam[i].value+1) == 'n') ||
          (*(ctx->ConfigParam[i].value+1) == 'N'))
	return TRUE;
      else
	return FALSE;
    }
  }

  /*UTL_Log("UTL_GetConfigParam", INFO, 
    "Using default value for parameter `%s'", name);*/

  return deflt;
}

/*------------------------------------------------------------------------*/
/*--	UTL_LoadConfigFile						--*/
/*--									--*/
/*--	Load configuration file.  Must be called before configuration	--*/
/*--    access functions can be used.					--*/
/*------------------------------------------------------------------------*/

PRIVATE int UTL_LoadConfigFile_r(UTL_CTX *ctx, const char *filename)
{
  FILE *fp;
  char line[UTL_MAXLINE], *c, *c1, *p, *path, *context;

  if(ctx->NumLoaded >= UTL_MAX_INCLUDES) {
    fprintf(stderr, "Too many includes!\n");
    return RC_ERROR;
  }

  /*--  Open configuration file  --*/
  
  fp = fopen(filename, "r");

  if (!fp) {
    fprintf(stderr, "Unable to open config file %s\n", filename);
    return RC_ERROR;
  }

  if((context = malloc(strlen(filename) + 1)) == NULL) {
    fprintf(stderr, "Unable to allocate space for %s\n", filename);
    return RC_ERROR;
  }
  strcpy(context, filename);
  ctx->ConfigFileTable[ctx->NumLoaded++] = context;

  while ( (fgets(line, UTL_MAXLINE, fp)) ) {

    if (line[0] == ';')		    // Ignore comments.
      continue;

    for(c=line; isspace(*c) && *c; ++c);

    /*--  Check for preprocessor directive  --*/

    if ((*c) == '#') {

      for(c1=c; !isspace(*c1) && *c1; ++c1);
      p = c1+1;
      if (*c1 == '\0')
	p = NULL;
      else
        *c1 = '\0';

      if (p) {				// Read directive parameter.
	for(; isspace(*p) && *p; ++p);	// Skip over blanks.
	if (*p == '\0')
	  p = NULL;
	else {
	  for(c1=p; !isspace(*c1) && *c1; ++c1);  // Skip to end of parameter.
	  *c1 = '\0';
	}
      }

      /*--  Process directives  --*/

      if (strcasecmp(c, INCLUDE_DIRECTIVE) == 0) {
        if (p == NULL) { 		// No filename is given.
	  fprintf(stderr, "Empty #INCLUDE directive in file %s\n",
			filename);
	  return RC_ERROR;
	}
	path = UTL_RelativeToFile(NULL, filename, p);
	if (UTL_LoadConfigFile_r(ctx, path)) {
	  free(path);
	  fprintf(stderr, "An error occured while loading config file %s\n", p);
	  return RC_ERROR;
	}
	free(path);
      }

      continue;
    }

    /*--  Process standard config param  --*/

    if ( (c1 = (char *)strchr(c,':')) == NULL)
      continue;
    *c1 = '\0';

    if (ctx->NumParam >= UTL_MAX_PARAM) {
      fprintf(stderr, "Too many parameters in configuration file.\n");
      return RC_ERROR;
    }

    ctx->ConfigParam[ctx->NumParam].type = (char *)malloc( strlen(c)+1 );
    strcpy(ctx->ConfigParam[ctx->NumParam].type, c);

    for(c=c1+1; isspace(*c) && *c; ++c);
    
    //  Look for new line.
    for(c1=c; *c1 && (*c1 != '\n'); ++c1);

    if (*c1 != '\n') {
      fprintf(stderr, "Parameter %s too long in file %s\n", 
		ctx->ConfigParam[ctx->NumParam].type, filename);
      return RC_ERROR;
    }

    *c1 = '\0';

    ctx->ConfigParam[ctx->NumParam].value = (char *)malloc( strlen(c)+1 );
    strcpy(ctx->ConfigParam[ctx->NumParam].value, c);

    ctx->ConfigParam[ctx->NumParam].context = context;
    ctx->ConfigParam[ctx->NumParam].path = NULL;

    ctx->NumParam ++ ;
  }

  fclose(fp);

  return RC_OK;
}

int UTL_LoadConfigFile(UTL_CTX *ctx, const char *filename)
{

  /*--  Test if file already loaded  --*/

  /*--  Recursively load configuration files  --*/
  if (UTL_LoadConfigFile_r(ctx, filename)) {
    fprintf(stderr, "An error occured while loading config file %s\n", 
		filename);
    return RC_ERROR;
  }

  /*--  Print config params  --*/
  /*
  printf("\nConfiguration parameters: \n");
  printf("----------------------------------------\n");
  for(i=0; i<ctx->NumParam; ++i)
    printf("%s: %s \n", ctx->ConfigParam[i].type, ctx->ConfigParam[i].value);
  printf("----------------------------------------\n\n");
  */
  return RC_OK;
}

/*------------------------------------------------------------------------*/
/*--	UTL_UnLoadConfigFile						--*/
/*--									--*/
/*--	Unload config file and enable an alternate file to be loaded	--*/
/*------------------------------------------------------------------------*/

int UTL_UnLoadConfigFile(UTL_CTX *ctx)
{
  int i;

  if (ctx->NumLoaded == 0)
    return RC_OK;

  for(i = 0; i < ctx->NumLoaded; ++i)
    free(ctx->ConfigFileTable[i]);
  ctx->NumLoaded = 0;

  for(i=0; i<ctx->NumParam; i++) {
    free(ctx->ConfigParam[i].type);
    free(ctx->ConfigParam[i].value);
    if(ctx->ConfigParam[i].path)
      free(ctx->ConfigParam[i].path);
  }

  ctx->NumParam = 0;

  return RC_OK;
}

/*------------------------------------------------------------------------*/
/*--	UTL_GetServerParam						--*/
/*--									--*/
/*--	Helper functions for getting server configuration parameters.  --*/
/*------------------------------------------------------------------------*/

/* TJW 10/17/98 */

PRIVATE void *
UTL_get_param_helper(ctx, f, name, number, dflt)
     const UTL_CTX *ctx;
     void * ((*f)());
     const char * name;
     int number;
     void * dflt;
{
  char buf[128];
  void * res;

  res = (*f)(ctx, name, dflt);
  sprintf(buf, "Server_%s", name);
  res = (*f)(ctx, buf, res);
  sprintf(buf, "Server_%s_%d", name, number);
  return (*f)(ctx, buf, res);
}

PACKAGE char *
UTL_GetServerParam(ctx, name, number, dflt)
     const UTL_CTX *ctx;
     const char * name;
     int number;
     char * dflt;
{
  return (char *) UTL_get_param_helper(ctx, UTL_GetConfigParam,
				       name, number, (void *) dflt);
}

PACKAGE char *
UTL_GetPathServerParam(ctx, name, number, dflt)
     UTL_CTX *ctx;
     const char * name;
     int number;
     char * dflt;
{
  return (char *) UTL_get_param_helper(ctx, UTL_GetPathConfigParam,
				       name, number, (void *) dflt);
}

PACKAGE int
UTL_GetIntServerParam(ctx, name, number, dflt)
     const UTL_CTX *ctx;
     const char * name;
     int number;
     int dflt;
{
  return (int) UTL_get_param_helper(ctx, UTL_GetIntConfigParam,
				    name, number, (void *) dflt);
}

PACKAGE int
UTL_GetBoolServerParam(ctx, name, number, dflt)
     const UTL_CTX *ctx;
     const char * name;
     int number;
     int dflt;
{
  return (int) UTL_get_param_helper(ctx, UTL_GetBoolConfigParam,
				    name, number, (void *) dflt);
}

PRIVATE void *
UTL_get_adm_param_helper(ctx, f, name, number, dflt)
     const UTL_CTX *ctx;
     void * ((*f)());
     const char * name;
     int number;
     void * dflt;
{
  char buf[128];
  void * res;

  res = (*f)(ctx, name, dflt);
  sprintf(buf, "Admin_%s", name);
  res = (*f)(ctx, buf, res);
  sprintf(buf, "Admin_%s_%d", name, number);
  return (*f)(ctx, buf, res);
}

PACKAGE char *
UTL_GetPathAdminParam(ctx, name, number, dflt)
     UTL_CTX *ctx;
     const char * name;
     int number;
     char * dflt;
{
  return (char *) UTL_get_adm_param_helper(ctx, UTL_GetPathConfigParam,
					   name, number, (void *) dflt);
}

PRIVATE void *
UTL_get_cli_param_helper(ctx, f, name, number, dflt)
     const UTL_CTX *ctx;
     void * ((*f)());
     const char * name;
     int number;
     void * dflt;
{
  char buf[128];
  void * res;

  res = (*f)(ctx, name, dflt);
  sprintf(buf, "Client_%s", name);
  res = (*f)(ctx, buf, res);
  sprintf(buf, "Client_%s_%d", name, number);
  return (*f)(ctx, buf, res);
}

PACKAGE char *
UTL_GetPathClientParam(ctx, name, number, dflt)
     UTL_CTX *ctx;
     const char * name;
     int number;
     char * dflt;
{
  return (char *) UTL_get_cli_param_helper(ctx, UTL_GetPathConfigParam,
					   name, number, (void *) dflt);
}
