/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/***************************************************************/
/*                                                             */
/* UTL_Connection.c                                            */
/*                                                             */
/* Set up and close connections between all of the servers,    */
/* allocate and free stream_array.                             */
/*                                                             */
/***************************************************************/

#include "ITTC.h"

/* Internal function for UTL_SetupConnections */
int IncludeP(int i, int vector) {
  return ( (vector & (1<<i)) > 0);
}


/* Initialize connections.  Open connections between the share servers. */
/* Thread controls which virtual port is used by COM */
/* vectors is a bit vector telling which servers to connect to.  */
/* It will skip the server number returned by UTL_MyID, but only if this */
/* number is greater than zero */
/* Hint: to connect to all server, make vector = ~0 */
int UTL_SetupConnections(int vector, UTL_STREAMS *streams, UTL_HEAP *heap, int thread) {

  int i;               /* Loop variable */
  int me = 0;          /* This server's number */
  int numServers = 3;  /* The total number of servers */
  int toGo = 1;        /* If there are any servers left */
  int ret;             /* Return value from COM_SELCT */
  COM_STREAM *temp = NULL;    /* A temporary pointer to a COM_STREAM */
  COM_STREAM *tmp = NULL;     /* A temporary pointer to a COM_STREAM */
  void *vd = NULL;            /* A void pointer */
  int peerID;
  BIGNUM **bnArray = NULL;    /* An array for SendToAll and GetFromAll */
  COM_CTX *ctx = NULL;
  char *status = NULL;
  int start;
  int vect;
  int heapState;
  int err = RC_OK;
  int vport = ITTC_SHARE_SERVER_VPORT+1;

  /* Initialize everything */
  heapState = UTL_GetHeapState(heap);
  numServers = UTL_NumServers(streams);
  me = UTL_MyID(streams);
  bnArray = UTL_GetBNArray(numServers, heap);
  if (bnArray == NULL) goto error;
  ctx = UTL_COMContext(streams);
  if (ctx == NULL) goto error;
 
  if ( (status = (char*)malloc(numServers*sizeof(char))) == NULL) {
    UTL_Log("UTL_SetupConnection", LOG_TEST, "Error in malloc.");
    return RC_ERROR;
  }
  for (i=0; i<numServers; i++) status[i] = 0;
  status[me] = 1;

  /* Don't attempt to connect where connections are already established */
  vect = vector;
  for (i=0; i<numServers; i++) {
    if (UTL_Stream(i, streams)!=NULL) vect &= ~(1<<i);
  }

  /* Send out connection requests to all servers with ID > this server */
  if (me>=0) {
    start = me + 1;
  } else {
    start = 0;
  }
  for (i=start; i<numServers; i++) {
    if (IncludeP(i, vect)) {
      UTL_Log("UTL_SetupConnection", LOG_TEST, "Requesting connection from server %d, port %d", i, vport+thread);
      tmp = COM_connect(ctx, COM_TYPE_SHARE_SERVER, i, (vport_t)vport+thread);
      if (UTL_SetStream(i, tmp, streams) != RC_OK) {   /* UTL_SetStream checks for tmp == NULL */
	UTL_Log("UTL_SetupConnection", LOG_INFO, "Error in connect", i);
	err = RC_ERROR;
	goto error;
      }
      status[i] = 1;
    }
  }

  /* Accept connection requests from servers with ID < this server */
  if (me == 0) {
    toGo = 0;
  } else {
    toGo = 1;
  }
  while (toGo) {

    UTL_Log("UTL_SetupConnection", LOG_TEST, "Waiting for connection attempt.");

    /* Handle one new connection */
    ret = COM_select(ctx, &vd, COM_SELECT_ACCEPT | COM_SELECT_CLOSE | 
		     COM_SELECT_TIMEOUT | COM_SELECT_PORT(vport+thread) );
    UTL_Log("UTL_SetupConnection", LOG_TEST, "Made connection on port %d.", vport+thread);
    temp = (COM_STREAM*)vd;
    if (ret == COM_SELECTED_TIMEOUT) {                     /* Check for timeout */
      UTL_Log("UTL_SetupConnection", LOG_INFO, "Select timed out.");
      err = RC_TIMEOUT;
      goto error;
    }
    if (ret == COM_SELECTED_CLOSE) {                     /* Check for timeout */
      UTL_Log("UTL_SetupConnection", LOG_INFO, "Select closed.");
      err = RC_CLOSE;
      goto error;
    }
    if (ret != COM_SELECTED_ACCEPT) {                       /* Check for errors */
      UTL_Log("UTL_SetupConnection", LOG_INFO, "Select failed.");
      	err = RC_ERROR;
	goto error;
    }

    /* Get the peer's ID number */
    if ((peerID = UTL_StreamID(temp)) < 0) {
      UTL_Log("UTL_SetupConnection", LOG_INFO, "Error getting peer's ID.");
      	err = RC_ERROR;
	goto error;
    }

    /* Check if the connection is OK */
    if (!IncludeP(peerID, vect)) {
      UTL_Log("UTL_SetupConnection", LOG_INFO, "Unwanted server attempted a connect.");
      	err = RC_ERROR;
	goto error;
    }
    if (status[peerID] == 1) {
      UTL_Log("UTL_SetupConnection", LOG_INFO, "Server %d sent multiple connects.", peerID);
      	err = RC_ERROR;
	goto error;

    }
    if (UTL_SetStream(peerID, temp, streams) != RC_OK) {
      UTL_Log("UTL_SetupConnection", LOG_INFO, "Error in SetStream", i);
      	err = RC_ERROR;
	goto error;
    }
    status[peerID] = 1;

    /* See if there are any connections left */
    toGo = 0;
    for (i=0; i<me; i++) if ( (status[i] == 0) && IncludeP(i, vect) ) {
      toGo = 1;
      i = me;
    }

    UTL_Log("UTL_SetupConnection", LOG_TEST, "Connection accepted from server %d on port %d.", peerID, vport+thread);
  }


  /************************/
  /* Now do two-way ACK's */
  /************************/

  /* Prepare to send ACK's */
  UTL_Log("UTL_SetupConnection", LOG_TEST, "Starting ACK's.");
  for (i=0; i<numServers; i++) {
    if (BN_set_word(bnArray[i], i) == 0) {
      UTL_Log("UTL_SetupConnection", LOG_INFO, "Error generating array of ACK's.");
      	err = RC_ERROR;
	goto error;
    }
  }

  /* Send ACK's */
  if (UTL_SendToAll(bnArray, streams) != RC_OK) {
    UTL_Log("UTL_SetupConnection", LOG_INFO, "Error sending ACK's.");
    	err = RC_ERROR;
	goto error;
  }

  /* Receive ACK's */
  if (UTL_GetFromAll(bnArray, streams, heap, NULL) != RC_OK) {
    UTL_Log("UTL_SetupConnection", LOG_INFO, "Error receiving ACK's.");
    	err = RC_ERROR;
	goto error;
  }

  /* Check ACK's */
  for (i=0; i<numServers; i++) {
    if ( (int)BN_get_word(bnArray[i]) != me ) {
      UTL_Log("UTL_SetupConnection", LOG_INFO, "ACK's incorrect.");
      	err = RC_ERROR;
	goto error;
    }
  }  
  
  UTL_Log("UTL_SetupConnection", LOG_TEST, "Done with ACK's.");
  err = RC_OK;

 error:
  UTL_RevertHeap(heapState, heap);
  if (status) free(status);
  return err;
}



/* Close connections with all of the servers, free stream_array */
int UTL_CloseAllConnections(UTL_STREAMS *streams) {

  int i;               /* Loop variable */
  int numServers = 3;  /* The total number of servers */

  /* Close all connections, close everything */
  if (streams == NULL) return RC_OK;
  numServers = UTL_NumServers(streams);
  for (i=0; i<numServers; i++) {
    if (UTL_Stream(i, streams) != NULL) {
      UTL_Log("UTL_CloseConnection", LOG_TEST, "Connection %d closing.", i);
      COM_close(UTL_Stream(i, streams));
      UTL_Log("UTL_CloseConnection", LOG_TEST, "Connection %d closed.", i);
    }
  }
  free(streams->active);
  free(streams->streams);
  free(streams);

  /* Return the error code */
  return RC_OK;
}




