/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */



/*----------------------------------------------------------*/
/*--                                                      --*/
/*--  UTL_GetSend.c                                       --*/
/*--                                                      --*/
/*--  GetFromAll and SendToAll functions                  --*/
/*--                                                      --*/
/*----------------------------------------------------------*/

#include "ITTC.h"


/* Does a select until there is input to read, the */
/* stream closes, or the stream times out */
int UTL_WaitRead(COM_STREAM* stream) {
  int err = RC_ERROR;
  COM_INFO info;
  void *vt;
  COM_STREAM *tStream = NULL;

  err = COM_select_v(stream->ctx, &vt, &stream, 1, 
		     COM_SELECT_READ | COM_SELECT_TIMEOUT | COM_SELECT_CLOSE);

  if (err == COM_SELECTED_TIMEOUT) {
    COM_get_peer_info(stream, &info);
    UTL_Log("UTL_WaitRead", LOG_FATAL, "Timeout in select.  (server %d)", COM_INFO_GET_NUM(info));
    return RC_TIMEOUT;
  }

  if (err == COM_SELECTED_CLOSE) {
    COM_get_peer_info(stream, &info);
    UTL_Log("UTL_WaitRead", LOG_FATAL, "Close in select.  (server %d)", COM_INFO_GET_NUM(info));
    return RC_CLOSE;
  }

  if (err != COM_SELECTED_READ) {
    COM_get_peer_info(stream, &info);
    UTL_Log("UTL_WaitRead", LOG_FATAL, "Error in select.  (server %d)", COM_INFO_GET_NUM(info));
    return RC_ERROR;
  }

  tStream = (COM_STREAM*)vt;
  if(tStream != stream) {
    UTL_Log("UTL_WaitRead", LOG_WARNING, "Received event from wrong stream (%08x != %08x)", tStream, stream);
    return RC_ERROR;
  }

  return RC_OK;
}


/* send is an array of BIGNUM*'s, and each bignum is sent to the  */
/* corresponding server. */
/* If a server's COM_STREAM is NULL, nothing is sent to that server */
/* It tries not to send to itself, but that only works if UTL_MyID is */
/* between 0 and numServers-1 */
int UTL_SendToAll(BIGNUM** send, UTL_STREAMS *streams) {

  int numServers;
  int me;
  int i;

  me = UTL_MyID(streams);
  numServers = UTL_NumServers(streams);

  for (i=0; i<numServers; i++) {
    if ( (i != me) && (UTL_Stream(i, streams) != NULL) ) {
      if (COM_write_bn(UTL_Stream(i, streams), send[i]) <= 0) {
	UTL_Log("UTL_SendToAll", LOG_INFO, "Error sending BIGNUM.");
	return RC_ERROR;
      }
    }
  }

  return RC_OK;
}


/* Like UTL_SendToAll, but sends an identical BIGNUM to each server */
int UTL_SendSameToAll(BIGNUM* send, UTL_STREAMS *streams) {

  int numServers;
  int me;
  int i;

  me = UTL_MyID(streams);
  numServers = UTL_NumServers(streams);

  for (i=0; i<numServers; i++) {
    if ( (i != me) && (UTL_Stream(i, streams) != NULL) ) {
      if (COM_write_bn(UTL_Stream(i, streams), send) <= 0) {
	UTL_Log("UTL_SendSameToAll", LOG_INFO, "Error sending BIGNUM.");
	return RC_ERROR;
      }
    }
  }

  return RC_OK;
}


/* get is an array of BIGNUM's.  The function gets a BIGNUM from each */
/* server and puts it into the corresponding spot in each the array.  */
int UTL_GetFromAll(BIGNUM **get, UTL_STREAMS *streams, UTL_HEAP *heap, UTL_STATS* stats) {

  int err = RC_OK;
  int numServers;
  int me;
  int i;
  int code = RC_OK;
  void *vt;
  COM_STREAM *tStream = NULL;
  BIGNUM *bigTemp = NULL;
  int num;
  int numStreams;   /* The number of streams to get from */
  int heapState;

  /* Initializations */
  heapState = UTL_GetHeapState(heap);
  me = UTL_MyID(streams);
  numServers = UTL_NumServers(streams);
  if (UTL_ActivateAll(streams) != RC_OK) goto error;
  bigTemp = UTL_GetBN(heap);
  if (bigTemp == NULL) goto error;

  /* Count the number of streams */
  numStreams = 0;
  for (i=0; i<numServers; i++) {
    if ( (UTL_Stream(i, streams) != NULL) && (i != me) ) numStreams++;
  }
  if (numStreams < 1) {
    UTL_Log("UTL_GetFromAll", LOG_INFO, "No streams to get from.");
    goto error;
  }

  /* A loop to get a BIGNUM from each server */
  for (i=0; i<numStreams; i++) {

    /* Select for input */
    if (stats) {
      UTL_AddCountStat(3, stats);    
      UTL_StartStat(8, stats);
    }
    err = COM_select_v(UTL_COMContext(streams), &vt, UTL_Active(streams), numServers, 
    		       COM_SELECT_READ | COM_SELECT_CLOSE | COM_SELECT_TIMEOUT);
    if (stats) UTL_FinishStat(8, stats);

    if (err == COM_SELECTED_TIMEOUT) {
      UTL_Log("UTL_GetFromAll", LOG_NOTIFY, "Timeout in select.");
      code = RC_TIMEOUT;
      goto error;
    }
    if ( (err != COM_SELECTED_READ) && (err != COM_SELECTED_CLOSE) ) {
      UTL_Log("UTL_GetFromAll", LOG_INFO, "Done with select.");
      goto error;
    }
    tStream = (COM_STREAM*)vt;

    /* Get the peer's ID number */
    if ((num = UTL_StreamID(tStream)) < 0) {
      UTL_Log("UTL_GetFromAll", LOG_FATAL, "Error getting peer's ID.");
      goto error;
    }

    /* Check is the stream actually closed instead of sending a bignum */
    if (err == COM_SELECTED_CLOSE) {
      UTL_Log("UTL_GetFromAll", LOG_NOTIFY, "Stream %d closed.", num);
      if (UTL_DeactivateAll(streams) != RC_OK) goto error;
      if (UTL_Activate(num, streams) != RC_OK) goto error;
      code = RC_CLOSE;
      goto error;
    }

    /* Remember not to select on this stream again */
    if (UTL_Deactivate(num, streams) != RC_OK) goto error;
    
    /* Read the BIGNUM from the stream */
    if (COM_read_bn(tStream, bigTemp) <= 0) {
      UTL_Log("UTL_GetFromAll", LOG_INFO, "Error reading BIGNUM.");
      goto error;
    }
    if (!BN_copy(get[num], bigTemp)) goto error;
    
  }
  
  err = RC_OK;

 done:
  UTL_RevertHeap(heapState, heap);
  return err;


 error:
  UTL_Log("UTL_GetFromAll", LOG_INFO, "Error.");
  if (code == RC_OK) {
    err = RC_ERROR;
  } else {
    err = code;
  }
  goto done;


  return RC_ERROR;
}


/* The function gets a single BIGNUM from server number "server" */
int UTL_GetFromOne(BIGNUM *get, UTL_STREAMS *streams, int server, UTL_HEAP *heap, UTL_STATS* stats) {

  int err = 0;
  int numServers;
  int me;
  void *vt;
  COM_STREAM *tStream = NULL;
  BIGNUM *bigTemp = NULL;
  int heapState;

  /* Initializations */
  heapState = UTL_GetHeapState(heap);
  me = UTL_MyID(streams);
  numServers = UTL_NumServers(streams);
  if (UTL_DeactivateAll(streams) != RC_OK) goto error;
  if( UTL_Activate(server, streams) != RC_OK) goto error;
  bigTemp = UTL_GetBN(heap);
  if (bigTemp == NULL) goto error;

  if ( (UTL_Stream(server, streams) == NULL) || (server == me) ) {
    UTL_Log("UTL_GetFromOne", LOG_INFO, "Invalid stream.");
    goto error;
  }

  /* Select for input */
  if (stats) {
    UTL_AddCountStat(3, stats);    
    UTL_StartStat(8, stats);
  }
  err = COM_select_v(UTL_COMContext(streams), &vt, UTL_Active(streams), numServers, 
		     COM_SELECT_READ | COM_SELECT_TIMEOUT | COM_SELECT_CLOSE);
  if (stats) UTL_FinishStat(8, stats);

  if (err == COM_SELECTED_TIMEOUT) {
    UTL_Log("UTL_GetFromOne", LOG_INFO, "Timeout in select.");
    UTL_RevertHeap(heapState, heap);
    return RC_TIMEOUT;
  }
  if ( (err != COM_SELECTED_READ) && (err != COM_SELECTED_CLOSE) ) {
    UTL_Log("UTL_GetFromOne", LOG_INFO, "Error in select.");
    goto error;
  }
  tStream = (COM_STREAM*)vt;

  /* Check is the stream actually closed instead of sending a bignum */
  if (err == COM_SELECTED_CLOSE) {
    UTL_Log("UTL_GetFromOne", LOG_NOTIFY, "Stream %d closed.", server);
    UTL_RevertHeap(heapState, heap);
    return RC_CLOSE;
  }

  /* Read the BIGNUM from the stream */
  if (COM_read_bn(tStream, bigTemp) <= 0) {
    UTL_Log("UTL_GetFromOne", LOG_INFO, "Error reading BIGNUM.");
    goto error;
  }
  if (!BN_copy(get, bigTemp)) goto error;
  
  UTL_RevertHeap(heapState, heap);
  return RC_OK;


 error:
  UTL_RevertHeap(heapState, heap);
  return RC_ERROR;

}






/*  Same as UTL_GetFromAll, but multiplies all numbers together as it gets them, and */
/*  returns the value of the product.  myNum is used as the contribution of the */
/*  server calling the function. */
/*  Math is done mod M */
int UTL_GetProductFromAll(BIGNUM *product, BIGNUM *myNum, BIGNUM *M, BN_CTX *bnCtx, 
			  UTL_STREAMS *streams, UTL_HEAP *heap, UTL_STATS* stats) {
   return UTL_GetFNFromAll(product, myNum, M, GET_PRODUCT, bnCtx, streams, heap, stats);
}


/*  Same as UTL_GetFromAll, but adds all numbers together as it gets them, and */
/*  returns the value of the product.  myNum is used as the contribution of the */
/*  server calling the function. */
int UTL_GetSumFromAll(BIGNUM *sum, BIGNUM *myNum, BIGNUM *M, BN_CTX *bnCtx, 
		      UTL_STREAMS *streams, UTL_HEAP *heap, UTL_STATS* stats) {
   return UTL_GetFNFromAll(sum, myNum, M, GET_SUM, bnCtx, streams, heap, stats);
}


/* Does a GetFromAll and performs some running calculation on all of the received numbers. */
/* Math is done mod M where for multiplies, not adds */
int UTL_GetFNFromAll(BIGNUM *result, BIGNUM *myNum, BIGNUM *M, int function, BN_CTX *bnCtx, 
		     UTL_STREAMS *streams, UTL_HEAP *heap, UTL_STATS* stats) {

  int err = 0;
  int num;
  int numServers;
  int me;
  int i;
  void *vt;
  COM_STREAM *tStream = NULL;
  BIGNUM *bigTemp = NULL;
  BIGNUM *input = NULL;
  int numStreams;
  int heapState;


  /* Initializations */
  heapState = UTL_GetHeapState(heap);
  if (!BN_copy(result, myNum)) goto error;
  me = UTL_MyID(streams);
  numServers = UTL_NumServers(streams);
  if (UTL_ActivateAll(streams) != RC_OK) goto error;
  if ( (bigTemp = UTL_GetBN(heap)) == NULL) goto error;
  if ( (input = UTL_GetBN(heap)) == NULL) goto error;

  /* Count the number of streams */
  numStreams = 0;
  for (i=0; i<numServers; i++) {
    if ( (UTL_Stream(i, streams) != NULL) && (i != me) ) numStreams++;
  }
  if (numStreams < 1) {
    UTL_Log("UTL_GetFNFromAll", LOG_INFO, "No streams to get from.");
    goto error;
  }

  /* A loop to get a BIGNUM from each server */
  for (i=0; i<numStreams; i++) {

    /* Select for input */
    if (stats) {
      UTL_AddCountStat(3, stats);    
      UTL_StartStat(8, stats);
    }
    err = COM_select_v(UTL_COMContext(streams), &vt, UTL_Active(streams), numServers,
		       COM_SELECT_READ | COM_SELECT_TIMEOUT | COM_SELECT_CLOSE);
    if (stats) UTL_FinishStat(8, stats);

    if (err == COM_SELECTED_TIMEOUT) {
      UTL_Log("UTL_GetFNFromAll", LOG_INFO, "Timeout in select.");
      UTL_RevertHeap(heapState, heap);
      return RC_TIMEOUT;
    }

    if ( (err != COM_SELECTED_READ) && (err != COM_SELECTED_CLOSE) ) {
      UTL_Log("UTL_GetFNFromAll", LOG_INFO, "Error in select.");
      goto error;
    }

    if (err == COM_SELECTED_DONE) {
      UTL_Log("UTL_GetFNFromAll", LOG_INFO, "Connections closed prematurely.");
      goto error;
    }
    tStream = (COM_STREAM*)vt;

    /* Get the peer's ID number */
    if ((num = UTL_StreamID(tStream)) < 0) {
      UTL_Log("UTL_GetFNFromAll", LOG_INFO, "Error getting peer's ID.");
      goto error;
    }

    /* Check is the stream actually closed instead of sending a bignum */
    if (err == COM_SELECTED_CLOSE) {
      UTL_Log("UTL_GetFNFromAll", LOG_NOTIFY, "Stream %d closed.", num);
      if (UTL_DeactivateAll(streams) != RC_OK) goto error;
      if (UTL_Activate(num, streams) != RC_OK) goto error;
      UTL_RevertHeap(heapState, heap);
      return RC_CLOSE;
    }

    /* Remember not to select on this stream again */
    if (UTL_Deactivate(num, streams) != RC_OK) goto error;
    
    /* Read the BIGNUM from the stream */
    if (COM_read_bn(tStream, input) <= 0) {
      UTL_Log("UTL_GetFNFromAll", LOG_INFO, "Error reading BIGNUM.");
      goto error;
    }
    /*UTL_Log("UTL_GetFNFromAll", LOG_INFO, "Received %ld from server %d", UTL_IntBN(input), num);*/

    /* Do the Math */
    switch (function) {
    case GET_PRODUCT:
      if (M == NULL) {
	if (!BN_mul(bigTemp, result, input, bnCtx)) goto error;
	if (!BN_copy(result, bigTemp)) goto error;
      } else {
	if (!BN_mod_mul(bigTemp, result, input, M, bnCtx)) goto error;
	if (!BN_copy(result, bigTemp)) goto error;
      }
      break;
    case GET_SUM:
      if (!BN_add(bigTemp, result, input)) goto error;
      if (!BN_copy(result, bigTemp)) goto error;
      break;
    }
  }

  /* Finish the math */
  switch (function) {
  case GET_PRODUCT:
    break;
  case GET_SUM:
    if (M != NULL) if (!BN_mod(result, result, M, bnCtx)) goto error;
    break;
  }

  UTL_RevertHeap(heapState, heap);
  return RC_OK;


 error:
  UTL_RevertHeap(heapState, heap);
  return RC_ERROR;

}




/* The function gets a single BIGNUM from server number "server" */
int UTL_GetFromOneSimple(BIGNUM *get, COM_STREAM *stream) {

  int rc;
  COM_STREAM *tStream = NULL;
  void *vt = NULL;

  rc = COM_select_v(stream->ctx, &vt, &stream, 1, 
		     COM_SELECT_READ | COM_SELECT_TIMEOUT | COM_SELECT_CLOSE);

  if (rc == COM_SELECTED_TIMEOUT) {
    UTL_Log("UTL_CheckResponse", LOG_NOTIFY, "Timeout.");
    return RC_TIMEOUT;
  }

  if (rc == COM_SELECTED_CLOSE) {
    UTL_Log("UTL_CheckResponse", LOG_NOTIFY, "Stream closed.");
    return RC_CLOSE;
  }

  if (rc != COM_SELECTED_READ) {
    UTL_Log("UTL_CheckResponse", LOG_INFO, "Error in select.");
    return RC_ERROR;
  }

  tStream = (COM_STREAM*)vt;

  if (COM_read_bn(tStream, get) < 0) {
    UTL_Log("UTL_CheckResponse", LOG_INFO, "Error reading response bignum");
    return RC_ERROR;
  }

  return RC_OK;

}
