/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	UTL_Heap.c							--*/
/*--									--*/
/*--	Allocation and de-allocation of bignums on the UTL heap 	--*/
/*--									--*/
/*------------------------------------------------------------------------*/

#include "ITTC.h"


/* Initialize the heap of BIGNUM's */
UTL_HEAP *UTL_InitHeap(int maxSize) {
  UTL_HEAP *heap;
  int i;

  heap = (UTL_HEAP*)malloc(sizeof(UTL_HEAP));
  if (heap == NULL) return NULL;
  heap->top = 0;
  heap->maxSize = maxSize;

  heap->array = (BIGNUM**)malloc(maxSize*sizeof(BIGNUM*));
  if (heap->array == NULL) {
    free(heap);
    return NULL;
  }
  for (i=0; i<maxSize; i++) heap->array[i] = NULL;

  return heap;
}


int UTL_GetHeapState(UTL_HEAP *heap) {
  if (heap) return heap->top;
  return -1;
}


/* Return a BIGNUM that the calling procedure can use */
BIGNUM *UTL_GetBN(UTL_HEAP *heap) {
  BIGNUM *temp;

  if (heap->top >= heap->maxSize) {
    UTL_Log("UTL_GetBN", LOG_FATAL, "Trying to access BIGNUM out of heap.");
    return NULL;
  }
  temp = heap->array[heap->top];
  if (temp == NULL) temp = BN_new();
  if (temp == NULL) return NULL;
  
  BN_zero(temp);
  heap->top++;
  return temp;
}


/* Return a BIGNUM array that the calling procedure can use */
BIGNUM **UTL_GetBNArray(int num, UTL_HEAP *heap) {
  BIGNUM **temp;
  int i;

  if (heap->top + num - 1 >= heap->maxSize) {
    UTL_Log("UTL_GetBNArray", LOG_FATAL, "Trying to access BIGNUM out of heap.");
    exit(-11);
  }

  for (i=0; i<num; i++) {
    if (heap->array[heap->top+i] == NULL) heap->array[heap->top+i] = BN_new();
    if (heap->array[heap->top+i] == NULL) return NULL;
    BN_zero(heap->array[heap->top+i]);
  }
  temp = &(heap->array[heap->top]);
  heap->top += num;

  return temp;
}


/* Release all heap BIGNUM's for re-use */
int UTL_ReleaseHeap(UTL_HEAP *heap) {
  heap->top = 0;
  return RC_OK;
}


int UTL_RevertHeap(int state, UTL_HEAP *heap) {
  heap->top = state;
  return RC_OK;
}


/* Free's all BIGNUM's in the heap */
int UTL_FreeHeap(UTL_HEAP *heap) {
  int i;

  for (i=0; i<heap->maxSize; i++) {
    if (heap->array[i]) BN_clear_free(heap->array[i]);
  }
  free(heap->array);
  free(heap);
  return RC_OK;
}






