/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	UTL_Log.c							--*/
/*--									--*/
/*--	Logging utility							--*/
/*------------------------------------------------------------------------*/

#include "ITTC.h"
#include <time.h>
#include <stdio.h>
#include <stdarg.h>

static FILE *logfile = NULL;
static int logLevel;
static int outputLevel;

/*------------------------------------------------------------------------*/
/*--	UTL_OpenLogFile()						--*/
/*------------------------------------------------------------------------*/

int UTL_OpenLogFile(UTL_CTX *ctx, COM_TYPE ty, int num)
{
  char *name = NULL, *sname = NULL;
  UTL_TIME temp;
  char timestr[HMSMDY_SIZE + 1];

  if (logfile != NULL)
    fclose(logfile);

  /*
  UTL_TimeGetCurrent(&temp);
  UTL_TimeToHMSMDY(&temp, timestr);
  */

  if(ty == COM_TYPE_CLIENT)
    sname = UTL_GetPathClientParam(ctx, "Log_File", num, "log");
  else if(ty == COM_TYPE_ADMIN)
    sname = UTL_GetPathAdminParam(ctx, "Log_File", num, "log");
  else
    sname = UTL_GetPathServerParam(ctx, "Log_File", num, "log");

  /*
  name  = (char *)malloc( strlen(sname)+30 );
  if (name == NULL) return RC_ERROR;

  sprintf(name, "%s/%s.log", sname, timestr);
  */
  name = sname;

  logfile = fopen(name, "a");

  if (logfile == NULL) {
    fprintf(stderr, "ittcd: cannot open log file %s \n", name);
    return RC_ERROR;
  }

  /*printf("Log file set to: %s\n\n", name);*/


  /*--------------------------*/
  /* Decide on logging levels */
  /*--------------------------*/

  switch ( (UTL_GetConfigParam(ctx, "Log_Level", "Info"))[0] ) {
  case 'B':
  case 'b':
    logLevel = LOG_BLANK;
    break;
  case 'F':
  case 'f':
    logLevel = LOG_FATAL;
    break;
  case 'E':
  case 'e':
    logLevel = LOG_ERROR;
    break;
  case 'W':
  case 'w':
    logLevel = LOG_WARNING;
    break;
  case 'N':
  case 'n':
    logLevel = LOG_NOTIFY;
    break;
  case 'I':
  case 'i':
    logLevel = LOG_INFO;
    break;
  case 'D':
  case 'd':
    logLevel = LOG_DEBUG;
    break;
  case 'T':
  case 't':
    logLevel = LOG_TEST;
    break;
  default:
      fprintf(stderr, "UTL_Log:  unknown log message level in %s.\n", name);
      fclose(logfile);
      abort();
  }      
    

  /*-------------------------*/
  /* Decide on output levels */
  /*-------------------------*/

  switch ( (UTL_GetConfigParam(ctx, "Output_Level", "Test"))[0] ) {
  case 'B':
  case 'b':
    outputLevel = LOG_BLANK;
    break;
  case 'F':
  case 'f':
    outputLevel = LOG_FATAL;
    break;
  case 'E':
  case 'e':
    outputLevel = LOG_ERROR;
    break;
  case 'W':
  case 'w':
    outputLevel = LOG_WARNING;
    break;
  case 'N':
  case 'n':
    outputLevel = LOG_NOTIFY;
    break;
  case 'I':
  case 'i':
    outputLevel = LOG_INFO;
    break;
  case 'D':
  case 'd':
    outputLevel = LOG_DEBUG;
    break;
  case 'T':
  case 't':
    outputLevel = LOG_TEST;
    break;
  default:
      fprintf(stderr, "UTL_Log:  unknown output message level in %s.\n", name);
      fclose(logfile);
      abort();
  }      
    

  return RC_OK;
}


/*------------------------------------------------------------------------*/
/*--	UTL_Log()							--*/
/*--									--*/
/*--	loc - location of error  (function name).			--*/
/*--	severity -  see UTL.h for a list of these			--*/
/*--	msg - Log message (printf string).				--*/
/*--									--*/
/*--	Looks in the config file to see which level of message		--*/
/*--	to print to the screen.						--*/
/*--									--*/
/*------------------------------------------------------------------------*/

void UTL_Log(char *loc, int severity, char *msg, ...)
{
  char timestr[UTL_TIME_STRING_LENGTH];
  char *type = NULL;
  UTL_TIME temp;
  va_list args;

  if (!logfile)
    return;


  /*-------------------------------------*/
  /* Decide the level of the log message */
  /*-------------------------------------*/

  switch (severity) {
    case LOG_BLANK:    type = "BLANK:";    break;
    case LOG_FATAL:    type = "FATAL:";    break;
    case LOG_ERROR:    type = "ERROR:";    break;
    case LOG_WARNING:  type = "WARNING:";  break;
    case LOG_NOTIFY:   type = "NOTIFY:";   break;
    case LOG_INFO:     type = "INFO: ";    break;
    case LOG_DEBUG:    type = "DEBUG:";    break;
    case LOG_TEST:     type = "TEST:";     break;
    default:  
      fprintf(stderr, "UTL_Log:  unknown log message type in %s. \n", loc);
      return;
  }      


  /*-----------------------*/
  /* Prapare for the log   */
  /*-----------------------*/

  UTL_TimeGetCurrent(&temp);
  UTL_TimeToString(&temp, timestr);

  va_start(args, msg);


  /*--------------------------------------------------*/
  /* Output to the log file.                          */
  /*--------------------------------------------------*/

  if (severity <= logLevel) {
    fprintf(logfile, "%s %s   %s: ", type, timestr, loc);
    vfprintf(logfile, msg, args);
    fprintf(logfile, "\n");
    
    fflush(logfile);
  }


  /*--------------------------------------------------*/
  /* Print to the screen.                             */
  /*--------------------------------------------------*/

  if (severity <= outputLevel) {
    fprintf(stderr, "%5u %26s: \t", UTL_thread_current_id(), loc);
    vfprintf(stderr, msg, args);
    fprintf(stderr, "\n");
  }

}

/*------------------------------------------------------------------------*/
/*--	UTL_CloseLogFile()						--*/
/*------------------------------------------------------------------------*/

int UTL_CloseLogFile()
{
	if(logfile) {
		fclose(logfile);
		logfile = NULL;
	}

  return RC_OK;
}

/*------------------------------------------------------------------------*/
/*--	UTL_FlushLogFile()						--*/
/*------------------------------------------------------------------------*/

int UTL_FlushLogFile()
{
  fflush(logfile);

  return RC_OK;
}

