/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	UTL_Monitor.c							--*/
/*--									--*/
/*--	Send information to monitor server				--*/
/*--									--*/
/*------------------------------------------------------------------------*/

#include "ITTC.h"
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/types.h>
#ifndef WIN32
#include <sys/socket.h>
#endif
/*#include <sys/time.h>*/
#ifndef WIN32
#include <netinet/in.h>
#include <netdb.h>
#endif
#include <fcntl.h>
#include <stdlib.h>
#ifndef WIN32
#include <strings.h>
#include <arpa/inet.h>
#endif

#ifdef WIN32
#define vsnprintf _vsnprintf
#endif

#ifndef INADDR_NONE
#define INADDR_NONE     0xffffffff      /* should be in <netinet/in.h> */
#endif

/*----------------------------------------------------------------------*/
/*	UTL_monitor							*/
/*		Send information to monitor server			*/
/*		Total string message len must be < 80 characters	*/
/*----------------------------------------------------------------------*/

int UTL_monitor(UTL_CTX *ctx, int msgattr, char *msg, ...)
{

  va_list args;
  int len;
  UTL_TIME ut;
  unsigned char buf[UTL_MONITOR_MAXMSG + 26];	/* max message = 240 bytes */
  unsigned char * p = NULL;
  u_long t;
  
  /*-- Message format:
	peerId:   32 bit int;
	time:	  14 byte string;
	msgattr:  32 bit int;
	msglen:   32 bit int;
        msg:      char *
	total size = 26 + message length
  --*/

  if (ctx->socket == 0 || ctx->saddr == NULL)
    return RC_OK;		/*  Monitor program is inactive  */

  va_start(args, msg);
  vsnprintf(ctx->space, UTL_MONITOR_MAXMSG, msg, args);
  len = strlen(ctx->space);	/* could add 1 to null-terminate */

  UTL_TimeGetCurrent(&ut);

  p = buf;

  /* TJW: SPARCs hate this trick, so we can't use it here :-( */
  /* *((u_long *) p) = htonl(ctx->me); */
  t = htonl(ctx->me);
  memcpy(p, &t, sizeof(u_long));
  p += sizeof(u_long);

  UTL_TimeToHMSMDY(&ut, p);
  p += HMSMDY_SIZE;

  /* *((u_long *) p) = htonl(msgattr); */
  t = htonl(msgattr);
  memcpy(p, &t, sizeof(u_long));
  p += sizeof(u_long);

  /* *((u_long *) p) = htonl(len); */
  t = htonl(len);
  memcpy(p, &t, sizeof(u_long));
  p += sizeof(u_long);

  memcpy(p, ctx->space, len);
  p += len;

  if(sendto(ctx->socket, (void *) buf, p - buf, 0, ctx->saddr, ctx->saddr_len) < 0) {
    perror("UTL_monitor");
    UTL_Log("UTL_monitor", LOG_ERROR, "Error sending msg");
    return RC_ERROR;
  }

  return RC_OK;
}

/*----------------------------------------------------------------------*/
/*	UTL_ConnectToMonitor						*/
/*		Connect to monitor server				*/
/*----------------------------------------------------------------------*/

PRIVATE int UTL_ConnectToMonitor(UTL_CTX *ctx) {
  int port;
  char *host;
  struct sockaddr_in * saddr;
  struct hostent hent;
#ifdef WIN32
  struct hostent * hptr;
#else
  int err;
#endif
  unsigned char buf[256];
  unsigned long inAddr;

  host = UTL_GetConfigParam   (ctx, "Monitor_IP_Addr", "127.0.0.1");
  port = UTL_GetIntConfigParam(ctx, "Monitor_IP_Port", 7000);

  ctx->saddr_len = sizeof(struct sockaddr_in);
  if((saddr = malloc(ctx->saddr_len)) == NULL)
    return RC_ERROR;
  ctx->saddr = (struct sockaddr *) saddr;

  memset((char *) saddr, 0, ctx->saddr_len);
  saddr->sin_family = AF_INET;
  saddr->sin_port = htons(port);

  if((inAddr = inet_addr(host)) == INADDR_NONE) {
#ifdef WIN32
    if((hptr = gethostbyname(host)) == NULL) {
      UTL_Log("UTL_ConnectToMonitor", LOG_FATAL, "Unable to resolve hostname");
      return RC_ERROR;
    }
    else
      memcpy((char *) &saddr->sin_addr, hptr->h_addr, hptr->h_length);
#else
    if(gethostbyname_r(host, &hent, buf, sizeof(buf), &err) == NULL) {
      UTL_Log("UTL_ConnectToMonitor", LOG_FATAL, "Unable to resolve hostname");
      return RC_ERROR;
    }
    else
      memcpy((char *) &saddr->sin_addr, hent.h_addr, hent.h_length);
#endif
  }
  else
    saddr->sin_addr.s_addr = inAddr;

  if ((ctx->socket = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    UTL_Log("UTL_ConnectToMonitor", LOG_FATAL, "Can't open socket");
    return RC_ERROR;
  }

  /*
  if(connect(ctx->socket, (struct sockaddr *) &saddr, sizeof(saddr)) < 0) {
    UTL_Log("UTL_ConnectToMonitor", LOG_FATAL, "Failed to connect to monitor");
    return RC_ERROR;
  }
  */

  return RC_OK;
}

/*----------------------------------------------------------------------*/
/*	UTL_MonitorInit							*/
/*		Initialize monitor service				*/
/*									*/
/*	mytype, myID are the COM values for my server			*/
/*	type (COM_CLIENT, COM_SERVER) and number.			*/
/*----------------------------------------------------------------------*/

int UTL_MonitorInit(UTL_CTX *ctx, COM_TYPE mytype, int myID) {
  if (!UTL_GetBoolConfigParam(ctx, "Monitor", FALSE)) {
    ctx->socket = 0;
    return RC_OK;
  }

  if ((ctx->space = malloc(UTL_MONITOR_MAXMSG)) == NULL) {
    UTL_Log("UTL_MonitorInit", LOG_FATAL, "Can't malloc space");
    return RC_ERROR;
  }

  ctx->saddr = NULL;

  ctx->me = COM_MAKE_INFO(mytype, myID, 0);

  if (UTL_ConnectToMonitor(ctx))
    return RC_ERROR;

  return RC_OK;
}


/*----------------------------------------------------------------------*/
/*	UTL_MonitorClose						*/
/*		Close connection to monitor service			*/
/*----------------------------------------------------------------------*/

int UTL_MonitorClose(UTL_CTX *ctx) {
  if (ctx->space)
    free(ctx->space);

  if (ctx->socket == 0)
    return RC_OK;

#ifdef WIN32
  closesocket(ctx->socket);
#else
  close(ctx->socket);
#endif
  ctx->socket = 0;

  if (ctx->saddr)
    free(ctx->saddr);

  return RC_OK;
}

