/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/***************************************************************/
/*                                                             */
/* UTL_Notify.c                                                */
/*                                                             */
/* Notify others of success or failure, get status from        */
/* others, decide failure.                                     */
/*                                                             */
/***************************************************************/

#include "ITTC.h"


int UTL_Notify(int *failed, UTL_STREAMS *streams, UTL_HEAP *heap) {

  int numServers;
  int me;
  int i;
  BIGNUM **bf = NULL;   /* BIGNUM versions of the variable "failed" */
  int heapState;


  /***************************/
  /* Initialize              */
  /***************************/

  me = UTL_MyID(streams);
  numServers = UTL_NumServers(streams);
  heapState = UTL_GetHeapState(heap);
  bf = UTL_GetBNArray(numServers, heap);
  if (bf == NULL) {
    UTL_RevertHeap(heapState, heap);
    return RC_ERROR;
  }
  if (!BN_set_word(bf[me], *failed)) {
    UTL_Log("UTL_Notify", LOG_INFO, "Error with BIGNUM.");
    UTL_RevertHeap(heapState, heap);
    return RC_ERROR;
  }


  /***************************/
  /* Send status to others   */
  /***************************/

  if (UTL_SendSameToAll(bf[me], streams) != RC_OK) {
    UTL_Log("UTL_Notify", LOG_INFO, "Error with UTL_SendSameToAll.");
    UTL_RevertHeap(heapState, heap);
    return RC_ERROR;
  }


  /***************************/
  /* Get status from others  */
  /***************************/

  if (UTL_GetFromAll(bf, streams, heap, NULL) != RC_OK) {
    UTL_Log("UTL_Notify", LOG_INFO, "Error with UTL_GetFromAll.");
    UTL_RevertHeap(heapState, heap);
    return RC_ERROR;
  }


  /**************************/
  /* Decide if it passed    */
  /**************************/

  if (*failed == 1) {
    UTL_Log("UTL_Notify", LOG_TEST, "I sent a \"1\".");
    UTL_RevertHeap(heapState, heap);
    return RC_OK;
  }
  for (i=0; i<numServers; i++) {
    if (BN_is_one(bf[i])) {
      *failed = 1;
      UTL_Log("UTL_Notify", LOG_TEST, "Another server sent a \"1\".");
      UTL_RevertHeap(heapState, heap);
      return RC_OK;
    }
  }


  /***************************/
  /* Free and return         */
  /***************************/

  UTL_RevertHeap(heapState, heap);
  return RC_OK;
}
