/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*----------------------------------------------------------*/
/*--                                                      --*/
/*--  UTL_Stats.c                                         --*/
/*--                                                      --*/
/*--  Functions to handle statistics.                     --*/
/*--                                                      --*/
/*----------------------------------------------------------*/

#include "ITTC.h"


/************************/
/* Statistics functions */
/************************/

/* Not much error-checking in the statistics code */


/* Initialize the statistics structure */
int UTL_InitStats(int count, int times, UTL_STATS* stats) {

  int i;

  if (stats == NULL) return RC_ERROR;

  stats->numCount = count;
  stats->numTimes = times;

  stats->count = (int*)malloc(count*sizeof(int));
  if (stats->count == NULL) return RC_ERROR;

  stats->times = (unsigned long int*)malloc(times*sizeof(unsigned long int));
  if (stats->times == NULL) {
    free(stats->count);
    return RC_ERROR;
  }

  stats->start = (UTL_TIME*)malloc(times*sizeof(UTL_TIME));
  if (stats->start == NULL) {
    free(stats->times);
    free(stats->count);
    return RC_ERROR;
  }

  stats->countNames = (char**)malloc( (count) * sizeof(char*) );
  if (stats->countNames == NULL) {
    free(stats->start);
    free(stats->times);
    free(stats->count);
    return RC_ERROR;
  }

  stats->timeNames = (char**)malloc( (times) * sizeof(char*) );
  if (stats->timeNames == NULL) {
    free(stats->countNames);
    free(stats->start);
    free(stats->times);
    free(stats->count);
    return RC_ERROR;
  }

  stats->pread = 0;
  stats->pwrite = 0;
  stats->bread = 0;
  stats->bwrite = 0;
 
  for (i=0; i<count; i++) stats->count[i] = 0;
  for (i=0; i<times; i++) stats->times[i] = 0;
  for (i=0; i<times; i++) UTL_TimeGetCurrent(&(stats->start[i]));
  for (i=0; i<count; i++) stats->countNames[i] = NULL;
  for (i=0; i<times; i++) stats->timeNames[i] = NULL;
 
  return RC_OK;
}


/* Free everything in the statistics structure */
int UTL_FreeStats(UTL_STATS* stats) {

  int i;

  if (stats == NULL) return RC_ERROR;

  for (i=0; i<stats->numCount; i++) if (stats->countNames[i] != NULL) free(stats->countNames[i]);
  for (i=0; i<stats->numTimes; i++) if (stats->timeNames[i] != NULL) free(stats->timeNames[i]);

  free(stats->count);
  free(stats->times);
  free(stats->start);
  free(stats->countNames);
  free(stats->timeNames);
   
  return RC_OK;
}


/* Start timing on a certain statistic */
int UTL_StartStat(int stat, UTL_STATS* stats) {

  UTL_TimeGetCurrent(&(stats->start[stat]));

  return RC_OK;
}


/* Finish timing on a certain statistic, add time to total time */
int UTL_FinishStat(int stat, UTL_STATS* stats) {
  UTL_TIME temp;

  UTL_TimeGetCurrent(&temp);
  stats->times[stat] += (unsigned long)UTL_TimeSubtractMsec(&temp, &(stats->start[stat]));

  return RC_OK;
}


/* Increment a counter statistic */
int UTL_AddCountStat(int stat, UTL_STATS* stats) {
  stats->count[stat] += 1;
  return RC_OK;
}


/* Give a name to a certain count statistic */
int UTL_NameCountStat(int stat, char* name, UTL_STATS* stats) {
  stats->countNames[stat] = strdup(name);
  return RC_OK;
}


/* Give a name to a certain time statistic */
int UTL_NameTimeStat(int stat, char* name, UTL_STATS* stats) {
  stats->timeNames[stat] = strdup(name);
  return RC_OK;
}


/* Print out the statistics */
int UTL_PrintStats(UTL_STATS* stats) {
  int i;

  UTL_Log("PrintStats", LOG_NOTIFY, "---------------------------------------------------");
  for (i=0; i<stats->numCount; i++) {
    UTL_Log("PrintStats", LOG_NOTIFY, "%30s:  %10d", stats->countNames[i], stats->count[i]);
  }
  UTL_Log("PrintStats", LOG_NOTIFY, "---------------------------------------------------");
  for (i=0; i<stats->numTimes; i++) {
    UTL_Log("PrintStats", LOG_NOTIFY, "%30s:  %10lu", stats->timeNames[i], stats->times[i]);
  }
  UTL_Log("PrintStats", LOG_NOTIFY, "---------------------------------------------------");
  
  if (stats->count[0])
    UTL_Log("PrintStats", LOG_NOTIFY, "%30s:  %10.0f", "Avg. Pick Candidates time", (float)stats->times[0]/stats->count[0]);

  if (stats->count[0])
    UTL_Log("PrintStats", LOG_NOTIFY, "%30s:  %10.0f", "Avg. Iteration time", (float)stats->times[7]/stats->count[0]);

  if (stats->count[0])
    UTL_Log("PrintStats", LOG_NOTIFY, "%30s:  %10.0f", "Avg. Trial Division time", (float)stats->times[2]/stats->count[0]);

  if (stats->count[1])
    UTL_Log("PrintStats", LOG_NOTIFY, "%30s:  %10.0f", "Avg. Prime Test time", (float)stats->times[3]/stats->count[1]);

  if (stats->count[2])
    UTL_Log("PrintStats", LOG_NOTIFY, "%30s:  %10.0f", "Avg. BGW time", (float)stats->times[6]/stats->count[2]);

  UTL_Log("PrintStats", LOG_NOTIFY, "---------------------------------------------------");

  return RC_OK;
}  


unsigned long int sumTime(int num, int iterations, UTL_STATS** array) {
  int i;
  unsigned long int sum;

  sum = 0;
  for (i=0; i<iterations; i++) {
    sum += array[i]->times[num];
  }

  return sum;
}


int sumCount(int num, int iterations, UTL_STATS** array) {
  int i, sum;

  sum = 0;
  for (i=0; i<iterations; i++) {
    sum += array[i]->count[num];
  }

  return sum;
}




/* Print out the statistics */
int UTL_PrintAverageStats(int iterations, UTL_STATS** stats) {
  int i, j;
  float tmp;
  int temp;

  UTL_Log("", LOG_NOTIFY, "");
  UTL_Log("", LOG_NOTIFY, "");
  UTL_Log("", LOG_NOTIFY, "");
  UTL_Log("PrintAverageStats", LOG_NOTIFY, "Average statistics over %d iterations.", iterations);
  UTL_Log("PrintAverageStats", LOG_NOTIFY, "---------------------------------------------------");
  for (i=0; i<stats[0]->numCount; i++) {
    for (j=0, tmp=0; j<iterations; j++) tmp += (float)stats[j]->count[i];
    UTL_Log("PrintAverageStats", LOG_NOTIFY, "%30s:  %10.2f", stats[0]->countNames[i], tmp/iterations);
  }
  UTL_Log("PrintAverageStats", LOG_NOTIFY, "---------------------------------------------------");
  for (i=0; i<stats[0]->numTimes; i++) {
    for (j=0, tmp=0; j<iterations; j++) tmp += stats[j]->times[i];
    UTL_Log("PrintAverageStats", LOG_NOTIFY, "%30s:  %10.0f", stats[0]->timeNames[i], tmp/iterations);
  }
  UTL_Log("PrintAverageStats", LOG_NOTIFY, "---------------------------------------------------");
  UTL_Log("PrintAverageStats", LOG_NOTIFY, "%30s:  %10.0f", "Avg. Pick Candidates time", 
	  (float)sumTime(0, iterations, stats)/sumCount(0, iterations, stats)/iterations);

  UTL_Log("PrintAverageStats", LOG_NOTIFY, "%30s:  %10.0f", "Avg. Iteration time", 
	  (float)sumTime(7, iterations, stats)/sumCount(0, iterations, stats)/iterations);

  UTL_Log("PrintAverageStats", LOG_NOTIFY, "%30s:  %10.0f", "Avg. Trial Division time", 
	  (float)sumTime(2, iterations, stats)/sumCount(0, iterations, stats)/iterations);

  UTL_Log("PrintAverageStats", LOG_NOTIFY, "%30s:  %10.0f", "Avg. Prime Test time", 
	  (float)sumTime(3, iterations, stats)/sumCount(1, iterations, stats)/iterations);

  UTL_Log("PrintAverageStats", LOG_NOTIFY, "%30s:  %10.0f", "Avg. BGW time", 
	  (float)sumTime(6, iterations, stats)/sumCount(2, iterations, stats)/iterations);

  UTL_Log("PrintAverageStats", LOG_NOTIFY, "---------------------------------------------------");

  for (i=0, temp=0; i<iterations; i++) temp += stats[i]->pread;
  UTL_Log("PrintAverageStats", LOG_NOTIFY, "%30s:  %10d", "Avg. Packets Read", temp);

  for (i=0, temp=0; i<iterations; i++) temp += stats[i]->pwrite;
  UTL_Log("PrintAverageStats", LOG_NOTIFY, "%30s:  %10d", "Avg. Packets Written", temp);

  for (i=0, temp=0; i<iterations; i++) temp += stats[i]->bread;
  UTL_Log("PrintAverageStats", LOG_NOTIFY, "%30s:  %10d", "Avg. Bytes Read", temp);

  for (i=0, temp=0; i<iterations; i++) temp += stats[i]->bwrite;
  UTL_Log("PrintAverageStats", LOG_NOTIFY, "%30s:  %10d", "Avg. Bytes Written", temp);

  UTL_Log("PrintAverageStats", LOG_NOTIFY, "---------------------------------------------------");

  return RC_OK;
}  


/* Return a specific count statistic */
int UTL_CountStat(int stat, UTL_STATS* stats) {
  return stats->count[stat];
}


/* Return a specific time statistic */
int UTL_TimeStat(int stat, UTL_STATS* stats) {
  return stats->times[stat];
}

