/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	UTL_Stream.c							--*/
/*--									--*/
/*--	Handles steam data structure    			        --*/
/*------------------------------------------------------------------------*/

#include "ITTC.h"


/***********************/
/* Access functions    */
/***********************/

int UTL_Clear(UTL_STREAMS *streams) {
  int i;
  COM_STREAM *stream = NULL;
  COM_PKT_T type;
  BIGNUM *bn = NULL;
  char str[256];
  int nt;

  for (i=0; i<streams->numServers; i++) {
    stream = UTL_Stream(i, streams);
    if (stream) {

      while (COM_packet_pending(stream, &type, NULL)) {
	switch (type) {
	case COM_PKT_INT:
	  COM_read_int(stream, &nt);
	  UTL_Log("UTL_Clear", LOG_NOTIFY, "Read int [%d] from server %d", nt, i);
	  break;
	case COM_PKT_STRING:
	  COM_read_string(stream, str);
	  UTL_Log("UTL_Clear", LOG_NOTIFY, "Read string \"%s\" from server %d", str, i);
	  break;
	case COM_PKT_BIGNUM:
	  bn = BN_new();
	  if (bn == NULL) return RC_ERROR;
	  COM_read_bn(stream, bn);
	  BN_clear_free(bn);
	  UTL_Log("UTL_Clear", LOG_NOTIFY, "Read BN from server %d", i);
	  break;
	default:
	  UTL_Log("UTL_Clear", LOG_NOTIFY, "Unknown type from server %d", i);
	  return RC_ERROR;
	  break;
	}
      }

    }
  }
  return RC_OK;
}


int UTL_StreamID(COM_STREAM *stream) {
  COM_INFO info;

  /* Get the peer's ID number */
  if ( (stream) && (COM_get_peer_info(stream, &info) < 0) ) {
    UTL_Log("StreamID", LOG_FATAL, "Error in COM_get_peer_info().");
    return -1;
  }
  return COM_INFO_GET_NUM(info);
}


COM_CTX* UTL_COMContext(UTL_STREAMS *streams) {
  if (streams) return streams->context;
  return NULL;
}


int UTL_NumServers(UTL_STREAMS *streams) {
  if (streams) return streams->numServers;
  return -1;
}

int UTL_MyID(UTL_STREAMS *streams) {
  if (streams) return streams->myID;
  return -1;
}

COM_STREAM* UTL_Stream(int num, UTL_STREAMS *streams) {
  if (streams) return streams->streams[num];
  return NULL;
}

COM_CTX* UTL_Context(UTL_STREAMS *streams) {
  if (streams) return streams->context;
  return NULL;
}  

int UTL_SetStream(int num, COM_STREAM* stream, UTL_STREAMS *streams) {
  if (streams == NULL)  {
    UTL_Log("SetStream", LOG_INFO, "No stream structure.");
    return RC_ERROR;
  }
  if (streams->streams == NULL)  {
    UTL_Log("SetStream", LOG_INFO, "No streams in stream structure.");
    return RC_ERROR;
  }
  streams->streams[num] = stream;
  return RC_OK;
}
  

/* Activate a stream */
int UTL_Activate(int num, UTL_STREAMS *streams) {

  if (num == streams->myID) {
    UTL_Log("Activate", LOG_INFO, "Can't activate self.", num);
    return RC_ERROR;
  }

  if ( (streams) && (streams->active) ) {
    streams->active[num] = UTL_Stream(num, streams);
  } else return RC_ERROR;

  return RC_OK;
}


/* De-activate a stream */
int UTL_Deactivate(int num, UTL_STREAMS *streams) {

  if ( (streams) && (streams->active) ) {
    streams->active[num] = NULL;
  } else return RC_ERROR;

  return RC_OK;
}


/* Activate all streams */
int UTL_ActivateAll(UTL_STREAMS *streams) {
  int i;

  if (streams == NULL) return RC_ERROR;

  if ( (streams->myID > -1) && (streams->myID < streams->numServers) ) 
    UTL_Deactivate(streams->myID, streams);

  for (i=0; i<streams->numServers; i++) {
    if (i != streams->myID) {
      UTL_Activate(i, streams);
    }
  }

  return RC_OK;
}


/* De-activate all streams */
int UTL_DeactivateAll(UTL_STREAMS *streams) {
  int i;

  if (streams == NULL) return RC_ERROR;
  for (i=0; i<streams->numServers; i++) {
    UTL_Deactivate(i, streams);
  }

  return RC_OK;
}


/* Returns a bit vector where 1 signifies an active stream, 0 */
/* signifies an inactive stream */
int UTL_ActiveV(UTL_STREAMS* streams) {
  int i;
  int vect = 0;
  int numServers;

  if (streams == NULL) return RC_ERROR;
  numServers = UTL_NumServers(streams);
  for (i=0; i<numServers; i++) {
    if (streams->active[i] != NULL) vect |= (1<<i);
  }

  return vect;
}


/* Returns an array of active streams */
COM_STREAM** UTL_Active(UTL_STREAMS* streams) {

  if (streams == NULL) return NULL;
  return streams->active;
}


UTL_STREAMS *UTL_InitStreams(int numServers, int me, COM_CTX *context) {
  UTL_STREAMS *streams;
  int i;
  
  if (context == NULL) return NULL;

  /* Initialize Streams */
  streams = (UTL_STREAMS*)malloc(sizeof(UTL_STREAMS));
  if (streams == NULL) return NULL;

  streams->streams = (COM_STREAM**)malloc(numServers*sizeof(COM_STREAM*));
  if (streams->streams == NULL) {
    free(streams);
    return NULL;
  }

  streams->numServers = numServers;
  streams->myID = me;
  for (i=0; i<numServers; i++) UTL_SetStream(i, NULL, streams);

  /* Make array of active streams */
  if ((streams->active = (COM_STREAM**)malloc(numServers*sizeof(COM_STREAM*))) == NULL) {
    free(streams);
    free(streams->streams);
    UTL_Log("UTL_InitStream", LOG_INFO, "Error in malloc.");
    return NULL;
  }
  for (i=0; i<numServers; i++) streams->active[i] = NULL;

  /* Set the COM context */
  streams->context = context;

  return streams;
}


