/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

#ifndef _COM_H_
#define _COM_H_

#include <sys/types.h>
#ifdef WIN32
typedef short pid_t;
#endif

#include <stdio.h>
#include "openssl/bn.h"
#include "rbuffer.h"

#include "UTL_ctx.h"
#include "UTL_mutex.h"
#include "UTL_cond.h"
#include "UTL_thread.h"

#include "UTL_Time.h"

/* Tweakables for win32 */
#ifndef PUBLIC
#define PUBLIC
#define PACKAGE
#define PRIVATE static
#endif

#define COM_DEFAULT_PORT 8500

#define COM_CTX_DEFAULT_TIMEOUTMS	20000
#define FUDGE_MS	8	/* Fudge factor (in ms) for 60Hz clocks */

/* These flags are used for both COM_init() and COM_connect(). */

/* Bitmasks */
#define COM_TYPE_CLIENT		0x00
#define COM_TYPE_SHARE_SERVER	0x01
#define COM_TYPE_ADMIN		0x02

typedef unsigned char COM_TYPE;
typedef unsigned short vport_t;

#define COM_INFO_TYPE_MASK	0x7f000000	/* 24-30 */
#define COM_INFO_NUM_MASK	0x00ff0000	/* 16-23 */
#define COM_INFO_PORT_MASK	0x0000ffff	/* Lower 16 */
#define COM_INFO_INVALID_MASK	0x80000000	/* 31st bit */

#define COM_INFO_GET_TYPE(i)	(((i) & COM_INFO_TYPE_MASK) >> 24)
#define COM_INFO_GET_NUM(i)	(((i) & COM_INFO_NUM_MASK) >> 16)
#define COM_INFO_GET_PORT(i)	((i) & COM_INFO_PORT_MASK)
#define COM_INFO_GET_VALID(i)	(((i) & COM_INFO_INVALID_MASK) == 0)

#define COM_MAKE_INFO(t,n,p)	((t) << 24 | (n) << 16 | (p))

typedef int COM_INFO;

typedef struct com_cb_st {
  int fd;
  struct com_cb_st * next;
} COM_INPUT;

/*** COM_TRANSPORT ***/

/*
 * The interface between the COM package and the underlying transport.
 * COM still handles the networking - the transport only deals with
 * the resulting stream(s).
 */
typedef void TRANS_CTX;
typedef void TRANS_STREAM;
typedef void TRANS_MAGIC;

typedef struct com_trans_st {
  TRANS_CTX * (* new_server_ctx_method)(TRANS_MAGIC *, int);
  TRANS_CTX * (* new_client_ctx_method)(TRANS_MAGIC *, int);
  void (* close_ctx_method)(TRANS_CTX *);
  TRANS_STREAM * (* new_server_stream_method)(TRANS_CTX *, int);
  TRANS_STREAM * (* new_client_stream_method)(TRANS_CTX *, int);
  int (* handshake_stat_method)(TRANS_STREAM *);
  int (* do_handshake_method)(TRANS_STREAM *);
  char * (* get_id_method)(TRANS_STREAM *);
  int (* read_method)(TRANS_STREAM *, unsigned char *, int);
  int (* write_method)(TRANS_STREAM *, unsigned char *, int);
  void (* close_stream_method)(TRANS_STREAM *, int);
  void (* error_method)(char *);

  TRANS_MAGIC * magic;
} COM_TRANSPORT;

extern PACKAGE TRANS_CTX * COM_new_client_transport_ctx(COM_TRANSPORT *, TRANS_MAGIC *, int);
extern PACKAGE TRANS_CTX * COM_new_server_transport_ctx(COM_TRANSPORT *, TRANS_MAGIC *, int);
extern PACKAGE void COM_close_transport_ctx(COM_TRANSPORT *, TRANS_CTX *);
extern PACKAGE TRANS_STREAM *
  COM_new_transport_client_stream(COM_TRANSPORT *, TRANS_CTX *, int);
extern PACKAGE TRANS_STREAM *
  COM_new_transport_server_stream(COM_TRANSPORT *, TRANS_CTX *, int);
extern PACKAGE int COM_transport_get_fd(COM_TRANSPORT *, TRANS_STREAM *);
extern PACKAGE int COM_transport_handshake_status(COM_TRANSPORT *,
						  TRANS_STREAM *);
extern PACKAGE int COM_transport_do_handshake(COM_TRANSPORT *,
					      TRANS_STREAM *);
extern PACKAGE char * COM_transport_get_id(COM_TRANSPORT *, TRANS_STREAM *);
extern PACKAGE int COM_transport_read(COM_TRANSPORT *, TRANS_STREAM *,
				      unsigned char *, int);
extern PACKAGE int COM_transport_write(COM_TRANSPORT *, TRANS_STREAM *,
				       unsigned char *, int);
extern PACKAGE void COM_transport_close_stream(COM_TRANSPORT *, TRANS_STREAM *, int shutdown);
extern PACKAGE void COM_transport_handle_error(COM_TRANSPORT *, char *);

/* The only two transports currently implemented */
extern PUBLIC COM_TRANSPORT * COM_SSL_transport();
extern PUBLIC COM_TRANSPORT * COM_clear_transport();

extern PUBLIC COM_TRANSPORT * COM_transport_lookup();

/*** COM_SEQUENCE ***/

#define COM_SEQ_MAX_LOCKS	32
#define COM_SEQ_INVALID_LOCK	-1

#define COM_SEQ_DIRTY	0x01
#define COM_SEQ_LOCKED	0x02

typedef struct com_seq_waiter_st {
  void (* waiter)();
  void * arg;
  struct com_seq_waiter_st * next;
} COM_SEQ_WAITER;

typedef struct com_seq_st {
  struct com_seq_ent_st {
    COM_INFO inf;
    unsigned int seq;
    int locked;
    unsigned int lock_owner;
    COM_SEQ_WAITER * wlist;
    struct com_seq_st * parent;
    struct com_seq_ent_st * next;
  } * head;
#ifdef WIN32
#ifdef W32_LOCK
  HANDLE hFile;
  DWORD dwTop;
#endif
#endif
  FILE * sfp;
  unsigned char sv_lock;
  UTL_MUTEX up_lock;
  UTL_COND up_cond;	/* TODO: consider making this part of the entry */
  int flags;
  time_t mtime;
  int nlocks;
} COM_SEQUENCE;

typedef struct com_seq_ent_st COM_SEQUENCE_ENTRY;

/*** COM_CTX / COM_STREAM ***/

typedef enum {
  COM_STATUS_NOCONNECTION, COM_STATUS_INPROGRESS, COM_STATUS_INHANDSHAKE,
  COM_STATUS_WAITID, COM_STATUS_SEQ_WAIT, COM_STATUS_SEQ_A, COM_STATUS_SEQ_B,
  COM_STATUS_CONNECTED
} COM_STATUS;

#define COM_AST_NOT_SIGNALED	0
#define COM_AST_SIGNALED	1
#define COM_AST_ACCEPTED	2

/* True iff underlying transport can read/write */
#define COM_ioready(s)	((s)->state > COM_STATUS_INHANDSHAKE)

/* True iff COM_STREAM can be read/written publicly */
#define COM_usermode(s)	((s)->state == COM_STATUS_CONNECTED)

#ifdef STATS
typedef struct com_stat_st {
  unsigned long b_read;		// Bytes read
  unsigned long p_read;		// Packets read
  unsigned long b_write;	// Bytes written
  unsigned long p_write;	// Packets written
  UTL_TIME t_wait;	// Not used
} COM_STAT;
#endif

/*
 * One COM_CTX exists for each client or server process, and it keeps
 * state common to a group of connections.  If the COM_CTX is intended
 * for server-side use, then COM_init will set up an SSL server-side
 * context, and it will also attempt to bind to the appropriate port.
 * The COM_CTX also contains a linked list of all the connections
 * associated with it.
 */
typedef struct com_ctx_st {
  COM_TRANSPORT * trans;
  TRANS_CTX * s_ctx;
  TRANS_CTX * c_ctx;
  UTL_CTX * utl_ctx;
  int accept_socket;
  int flags;
  int timeoutms;

  COM_INFO myid;
  COM_SEQUENCE * sequence;
  COM_INPUT * inputs;
  int rseq;

  UTL_MUTEX llock;
  UTL_MUTEX slock;

  UTL_COND condnew;

  UTL_THREAD incubator;

  /*
   * Each COM_STREAM has a read buffer and a write buffer associated with
   * it.  Data sent or received along a COM_STREAM is packetized according
   * to data type.
   */
  struct com_stream_st {
    COM_TRANSPORT * trans;
    struct com_ctx_st * ctx;
    TRANS_STREAM * stream;
    UTL_MUTEX tlock;

    BUFFER * wbuf;
    BUFFER * rbuf;
    BUFFER * tbuf;

    /*char * peerid;*/
    COM_INFO peerid;

    int fd;

    COM_STATUS state;
    int flags;
    int timeoutms;
    int rseq;
    pid_t owner_pid;
    int connecting;
    int accept_state;
    int incr;
    int owned;
    unsigned int iowner;

    UTL_TIME marker;
    int usemarker;

    COM_SEQUENCE_ENTRY * s_ent;

#ifdef STATS
    COM_STAT stats;
#endif

    struct com_stream_st * next;
  } * streams;
} COM_CTX;

typedef struct com_stream_st COM_STREAM;

typedef enum {
  COM_PKT_BYTE,		/* 8 bits */
  COM_PKT_SHORT,	/* 16 bits */
  COM_PKT_INT,		/* 32 bits */
  COM_PKT_BYTES,	/* variable-length */
  COM_PKT_BIGNUM,	/* variable-length */
  COM_PKT_STRING,	/* variable-length */
  COM_PKT_OTHER
} COM_PKT_T;

/* Not used yet */
typedef struct com_pkt_hdr_st {
  unsigned char hbyte;
  COM_PKT_T ptype;
  int len;	/* optional */
} COM_PKT_HDR;

#define HEADER_BYTE 0xff

/* If any threading is used, we want INCUBATOR enabled */

#define INCUBATOR 1

/*
 * COM_init() sets up various system parameters according to the
 * contents of the (previously-read) configuration file.
 * This must be called before COM_connect(), COM_accept(), or
 * any of the other I/O routines are called.
 */
extern PUBLIC COM_CTX * COM_init(UTL_CTX * uctx, COM_TYPE type, int num);

/*
 * COM_set_error_stream() directs all error messages to the given
 * stream.  The default error stream is stderr.
 */
extern PUBLIC void COM_set_error_stream(FILE * ep);

/*
 * Sets the timeout, in milliseconds.
 */
extern PUBLIC void COM_set_timeout(COM_STREAM * str, unsigned long millisec);
extern PUBLIC void COM_ctx_set_timeout(COM_CTX * ctx, unsigned long millisec);

/*
 * COM_connect() opens a secure connection of the specified type
 * to the numbered host, as specified by the configuration files.
 * Typically, server_num is an integer between 0 and n-1, where
 * n is the total number of servers participating in the threshold
 * scheme.
 *
 * The "type" and "num" arguments correspond exactly to the
 * arguments used in COM_init().
 *
 * The "vport" argument tells COM_connect which virtual port to use
 * and hence which COM_accept() on the server side will receive this
 * connection.
 */
extern PUBLIC COM_STREAM * COM_connect(COM_CTX * ctx,
				       COM_TYPE type, int num, vport_t vport);

/*
 * COM_accept() listens for an COM connection on the specified virtual
 * port and returns a COM_STREAM when a connection is established.
 * This blocks until success, or until a fatal error occurs, in which
 * case NULL is returned.
 */
extern PUBLIC COM_STREAM * COM_accept(COM_CTX * ctx, vport_t vport);

/*
 * Gets the COM_CTX associated with a given COM_STREAM.
 */
extern COM_CTX * COM_get_CTX(COM_STREAM * s);

/*
 * Forks off a child process, which will execute the function
 * `func' with arguments `s' and `magic'.  The stream `s' is set
 * up so that the parent process can COM_close() it immediately
 * after COM_fork() returns without interfering with the child
 * process.
 */
typedef void (* forkfunc)(COM_STREAM *, void *);
extern PUBLIC void COM_fork(forkfunc func, COM_STREAM * s, void * magic);

/*
 * Changes ownership of the given stream to (or away from) the current process.
 */
extern PUBLIC void COM_grab(COM_STREAM * s);
extern PUBLIC void COM_ungrab(COM_STREAM * s);

/*
 * Obtains information about the peer on the other end of the
 * stream s.  Returns 0 upon success, -1 otherwise.  If successful,
 * the structure pointed to by "ret" will be filled with information
 * about the peer.
 */
extern PUBLIC int COM_get_peer_info(COM_STREAM * s, COM_INFO * ret);

/*
 * Returns information about the COM_ctx itself.
 */
extern PUBLIC COM_INFO COM_ctx_get_info(COM_CTX * ctx);

/*
 * COM_ctx_{set,unset}_flags() specifies a bitwise or-ed list of
 * flags to {set,unset} on the given context.  By default, all
 * flags are unset.
 * COM_{set,unset}_flags() specifies a bitwise or-ed list
 * of flags to {set,unset} on the given stream.  By default, all
 * flags are inherited from the parent context.
 *
 * COM_FLAG_BUFFERED causes writes to occur only when COM_flush() is
 * called.  If it is not set, writes to the socket occur after
 * every write_{int,string,..} call.

 * COM_FLAG_NEED_{CLIENT,SERVER}_CERT specify the level of certification
 * required by the client or server.  The COM_connect() and
 * COM_accept() routines will automatically try to present a
 * certificate to the other party, if one is available.  If one is not
 * presented, and the other party requires one (as dictated by
 * COM_set_verify()), the connection is refused.  Normally, this
 * flag need not be changed, since the configuration file should
 * specify a default setting.
 *
 * COM_FLAG_BLOCKING_CONNECT will override COM_FLAG_NONBLOCK for
 * COM_connect() only.  This is used in buggy multithreaded platforms.
 */
#define COM_FLAG_BUFFERED		0x01
#define COM_FLAG_NONBLOCK		0x02
#define COM_FLAG_NEED_CLIENT_CERT	0x04
#define COM_FLAG_NEED_SERVER_CERT	0x08
#define COM_FLAG_BLOCKING_CONNECT	0x10

extern PUBLIC int COM_ctx_get_flags(COM_CTX * ctx);
extern PUBLIC void COM_ctx_set_flags(COM_CTX * ctx, int flags);
extern PUBLIC void COM_ctx_unset_flags(COM_CTX * ctx, int flags);
extern PUBLIC int COM_get_flags(COM_STREAM * s);
extern PUBLIC void COM_set_flags(COM_STREAM * s, int flags);
extern PUBLIC void COM_unset_flags(COM_STREAM * s, int flags);

/*
 * Adds fd to the list of file descriptors selected by COM_select.
 */
extern PUBLIC void COM_ctx_add_input(COM_CTX * ctx, int fd);

/*
 * COM_select() can wait for any or all of the following conditions
 * on any of the COM_STREAMS associated with it:
 *
 * COM_SELECT_ACCEPT:  A new connection was accepted by the server.
 * COM_SELECT_READ:  A stream has a packet waiting to be read.
 * COM_SELECT_WRITE:  A stream emptied its write buffer successfully.
 * COM_SELECT_INPUT:  A user-specified file descriptor has input ready.
 * COM_SELECT_CLOSE:  A stream was closed because of a network error/timeout.
 * COM_SELECT_TIMEOUT:  Nothing happened within a specified time interval.
 *
 * The flags can be bitwise or-ed to wait for multiple conditions.
 * For ACCEPT, READ, and WRITE, a pointer to the appropriate COM_STREAM
 * is returned through the `ret' parameter.  For INPUT, a pointer to
 * a COM_INPUT is returned this way.  The return value, one of
 * COM_SELECTED_{ERROR,TIMEOUT,ACCEPT,READ,WRITE} indicates what
 * caused COM_select to return.
 *
 * COM_SELECT_ACCEPT requires an argument, which specifies the 'virtual
 * port' to wait on, or COM_ACCEPT_ANY, which specifies any virtual port.
 */
#define COM_SELECT_ACCEPT	0x010000
#define COM_SELECT_READ		0x020000
#define COM_SELECT_WRITE	0x040000
#define COM_SELECT_INPUT	0x080000
#define COM_SELECT_CLOSE	0x100000
#define COM_SELECT_TIMEOUT	0x200000

#define COM_SELECT_PORT(p)	(COM_SELECT_ACCEPT | (p))
#define COM_ACCEPT_ANY	0
#define COM_ACCEPT_MASK	0xffff

#define COM_SELECTED_ERROR	-1
#define COM_SELECTED_UNKNOWN	-1
#define COM_SELECTED_TIMEOUT	0
#define COM_SELECTED_ACCEPT	1
#define COM_SELECTED_READ	2
#define COM_SELECTED_WRITE	3
#define COM_SELECTED_INPUT	4
#define COM_SELECTED_CLOSE	5
#define COM_SELECTED_DONE	6	/* Nothing left */

extern PUBLIC int COM_select(COM_CTX * ctx, void ** ret, int flags);
extern PUBLIC int COM_select_v(COM_CTX * ctx, void ** ret,
			       COM_STREAM ** streams, int n, int flags);

/*
 * If a packet is waiting to be read in s, returns the length
 * (in bytes) of its data representation, else 0.  The data type
 * of the packet will be returned through type_ret if successful.
 */
extern PUBLIC int COM_packet_pending(COM_STREAM * s,
				     COM_PKT_T * type_ret, int * len_ret);

/*
 * Returns the number of bytes yet to be written on the given
 * COM_STREAM.
 */
extern PUBLIC int COM_write_pending(COM_STREAM * s);

/*
 * Flushes all pending writes on the given COM_STREAM.
 */
extern PUBLIC int COM_flush(COM_STREAM * s);

/*
 * Convenience routines for transmitting various data types
 */
extern PUBLIC int COM_read_int(COM_STREAM * s, int * np);
extern PUBLIC int COM_write_int(COM_STREAM * s, int n);

extern PUBLIC int COM_read_string(COM_STREAM * s, char * str);
extern PUBLIC int COM_write_string(COM_STREAM * s, char * str);

extern PUBLIC int COM_read_bytes(COM_STREAM * s, unsigned char * bytes);
extern PUBLIC int COM_write_bytes(COM_STREAM * s, unsigned char * bytes, int len);

extern PUBLIC int COM_read_bn(COM_STREAM * s, BIGNUM * num);
extern PUBLIC int COM_write_bn(COM_STREAM * s, BIGNUM * num);

/*
 * COM_close() closes a previously-opened COM_STREAM.
 */
extern PUBLIC void COM_close(COM_STREAM * s);

extern PUBLIC void COM_print_stats(COM_STREAM * s);
#ifdef STATS
extern PUBLIC COM_STAT * COM_get_stats(COM_STREAM * s);
#endif

/*
 * COM_ctx_close() closes down the entire context, along with any live
 * streams that are still open.  A good way to end your program.
 */
extern PUBLIC void COM_ctx_close(COM_CTX * ctx);

#endif
