/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	GEN.h								--*/
/*--									--*/
/*--	Header file for secret share management package			--*/
/*------------------------------------------------------------------------*/

#ifndef GEN_H
#define GEN_H

#include "GEN_Context.h"


/*--------------------------*/
/*--  Security Constants  --*/
/*--------------------------*/

#define GEN_MAX_SHARES_PER_SERVER     20
#define GEN_DEFAULT_PUBLIC_KEY     65537

#define STRONG_R_BITS        2063
#define NORMAL_R_BITS        1039
#define WEAK_R_BITS           521
#define TEST_R_BITS           521

#define STRONG_COEFF_BITS    2413 
#define NORMAL_COEFF_BITS    1289 
#define WEAK_COEFF_BITS       531
#define TEST_COEFF_BITS         5

#define STRONG_SHARE_BITS    1024
#define NORMAL_SHARE_BITS     512
#define WEAK_SHARE_BITS       256
#define TEST_SHARE_BITS        64

#define RANDOM_BUFFER           3

#define STRONG_KEY              3
#define NORMAL_KEY              2
#define WEAK_KEY                1
#define TEST_KEY                0


/*------------------------*/
/*--  Naming Constants  --*/
/*------------------------*/

#define BGW_INTERPOLATE         0
#define BGW_ADDITIVE_SHARE      1

#define GEN_INVALID_PUBLIC_EXP  -4

#define GEN_BASE_VPORT          255


/*--------------------------------*/
/*--  Main function prototypes  --*/
/*--------------------------------*/

/*  Generate the RSA key */
/*     Returns a UTL statistic under *stats.  If null, it's ignored. */
/*     myServer is the server number */
/*     keySize is either   STRONG_KEY, NORMAL_KEY, WEAK_KEY, or TEST_KEY  */
/*     keyType is unused */
/*     numServers is the number of servers */
/*     threshold is the number of servers needed for a decryption */
int GEN_GenKey(int myServer, COM_STREAM *admStream, int keySize, int keyType, 
	       int numServers, int threshold, UTL_STATS **stats, 
	       UTL_CTX *utlCtx, KEY_CTX *keyCtx, COM_CTX *comCtx);

/*  Pick a P and Q for all servers */
int GEN_PickShares(BIGNUM *P_share, BIGNUM *Q_share, GEN_CTX *genCtx);     

/* Servers interact to find an RSA modulus */
int GEN_GenerateModulus(int numServers, int me, int thread, int keySize, int keyType, 
			UTL_CTX *utlCtx, COM_CTX *comCtx);

/*  Do trial division on the modulus to throw out bad values */
int GEN_TrialDivision(int *failed, BIGNUM *N, GEN_CTX *genCtx);   

/*  Test if the modulus is prime */
int GEN_PrimeTest(int *failed, BIGNUM *N, BIGNUM *P_share, 
		  BIGNUM *Q_share, int id, GEN_CTX *genCtx);   

/*  Generate a public/private key pair from a modulus */
int GEN_PrivateKey(RSA_SHARE *key, BIGNUM *N, BIGNUM *P_Share, BIGNUM *Q_Share, 
		   GEN_CTX *genCtx);

/* Generate a t-out-of-n sharing of the key */
int GEN_GenerateSharedKey(RSA_SHARE *key, COM_STREAM *admStream, GEN_CTX *genCtx);

/*  Save the modulus to disk */
int GEN_SaveKeyToDisk(KEY_CTX *keyCtx, RSA_SHARE *key);


/*------------------------------------------*/
/*--  Initialization function prototypes  --*/
/*------------------------------------------*/

/*  Generate the modified binomials for use in laGrange interpolation*/
int GEN_ModBin(GEN_CTX *genCtx);  

/*  Make R, the modulus for calculating N */
int GEN_MakeR(GEN_CTX *genCtx);

/* Make the factors that will be used in trial division. */
/* These factors are products of the small primes. */
int GEN_MakeFactors(GEN_CTX *genCtx);

/* Calculates M, the product of small primes, and pre-computes */
/* the CRT coefficients. */
int GEN_PrepareSieve(GEN_CTX *genCtx);


/*----------------------------------*/
/*--  BIGNUM function prototypes  --*/
/*----------------------------------*/

/* Takes a BIGNUM and an int.  Returns 1 if they're relatively prime, */
/* a 0 if they're not, and a -1 on an error. */
int GEN_RelPrime(BIGNUM *bn, int n);


/*---------------------------------*/
/*--  Other function prototypes  --*/
/*---------------------------------*/

/* Take a pair of bignums, x and y, and communicate with other servers */
/* to calculate (x1+x2+...+xn)*(y1+y2+...+yn).  If share = 1, GEN_BGW */
/* returns an additive sharing of N.  Otherwise it returns the value of N. */
/* All math is done mod R.  xType and yType are not used at this time. */
int GEN_BGW(BIGNUM *Z, BIGNUM *X, BIGNUM *Y, BIGNUM *modulus, 
	    int share, int xType, int yType, GEN_CTX *genCtx);

/* Set n to be the first number from start to start+range that is relatively */
/* prime to all primes <= 101.  Found is set to 1 if number is found, 0 otherwise. */
int GEN_DoSieve(BIGNUM *n, BIGNUM *start, int range, int *found, GEN_CTX *genCtx);

/* Make number that is relatively prime to M, the product of small primes <= 101 */
int GEN_MakeSieveShare(BIGNUM *share, int bits, int modSixty, GEN_CTX *genCtx);

/* Make number that is 3 mod 4 */
int GEN_MakeBlumShare(BIGNUM *share, int bits, int modFour, GEN_CTX *genCtx);


#endif


