/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	GEN_Context.h							--*/
/*--									--*/
/*--	Header file for secret share management package			--*/
/*------------------------------------------------------------------------*/

#ifndef GEN_CONTEXT_H
#define GEN_CONTEXT_H


#define TESTING 1
#define RUNNING 0


/* This structure keeps all relevant GEN information, and also keeps */
/* lots of pre-defined BIGNUM's */
typedef struct gen_ctx {

  /* Streams */
  UTL_STREAMS *streams;

  /* General Security */
  int security;                // The strength of security
  int wordBits;                // The number of bits in a word
  int rBits;                   // The number of bits in R
  int coeffBits;               // The number of bits in BGW polynomial coefficients
  int shareBits;               // The number of bits in P and Q
  int threshold;               // The degree of BGW polynomials
  BIGNUM *R;                   // The modulus for working with the shares

  /* Sieving */
  char *sieve;                 // The sieve itself
  int sieveP;                  // Is sieving in use
  int numSmallPrimes;          // The number of sieving primes
  int range;                   // The size of the sieve
  BIGNUM *sixtyM;              // Sixty times M
  BIGNUM *M;                   // The product of small primes for sieving

  /* Trial Division */
  int trialDivEnd;             // The maximum trial division prime
  int numFactors;              // The number of factors for trial division
  long unsigned int *factors;                // The factors for trial division

  /* Constants */
  BIGNUM** binomials;          // The binomials
  BIGNUM *zero;                // The BIGNUM representation of 0
  BIGNUM *one;                 // The BIGNUM representation of 1
  BIGNUM *two;                 // The BIGNUM representation of 2
  BIGNUM *three;               // The BIGNUM representation of 3
  BIGNUM *CRTa;                // Used in sieving to set a number mod 60
  BIGNUM *CRTb;                // Used in sieving to set a number mod 60

  /* Other Constants */
  int testMode;               

  /* The heap of BIGNUM's */
  UTL_HEAP *heap;

  /* BIGNUM context */
  BN_CTX *bnCtx;

  /* UTL Statistics */
  UTL_STATS* statistics;    /* IMPORTANT: This is not free()'d when GEN_CloseGenCtx() is called. */

  /* UTL Context */
  UTL_CTX *utlCtx;

  /* Thread number */
  int thread;

} GEN_CTX;


/*----------------------------------*/
/*-- Context Access Functions     --*/
/*----------------------------------*/

/* Initialize GEN_CTX */
int GEN_InitializeGenCtx(COM_CTX *context, int keySize, int keyType, int numServers,
			 UTL_CTX *utlCtx, GEN_CTX **genCtx);

/* Close GEN_CTX.  Frees everything inside GenCtx except for COM_CTX and UTL_STATS */
/* Note: The calling function must call free(genCtx) to free the context.  This just */
/*       frees stuff inside of the context. */
int GEN_CloseGenCtx(GEN_CTX *genCtx);

/*  0 = test, 1 = weak, 2 = normal, 3 = strong, -1 = error (doesn't use RC_OK or RC_ERROR) */
int GEN_InitSecurity(int keySize, int keyType, GEN_CTX* genCtx);

/* Set the thread number */
int GEN_SetThread(int num, GEN_CTX* genCtx);

/* Return the thread number */
int GEN_Thread(GEN_CTX* genCtx);

/* Set the context statistics */
int GEN_SetStats(UTL_STATS* stats, GEN_CTX* genCtx);

/* Return the stream context */
UTL_STREAMS *GEN_Streams(GEN_CTX *genCtx);

/* Returns the heap */
UTL_HEAP* GEN_Heap(GEN_CTX *genCtx);

/* Returns current ID */
int GEN_MyID(GEN_CTX *genCtx);       

/* Return the number of servers */
int GEN_NumServers(GEN_CTX *genCtx);            

/* Returns stream number num */
COM_STREAM* GEN_Stream(int num, GEN_CTX *genCtx);

/* Set a stream to the given stream */
int GEN_SetStream(int num, COM_STREAM* stream, GEN_CTX *genCtx);    

/* Returns COM context */
COM_CTX* GEN_COMContext(GEN_CTX *genCtx);    

/* Returns UTL context */
UTL_CTX* GEN_UTLContext(GEN_CTX *genCtx);    

/*  0 = test, 1 = weak, 2 = normal, 3 = strong */
int GEN_Security(GEN_CTX *ctx);

/* Returns 1 if testing, 0 if running */
int GEN_Testing(GEN_CTX *genCtx);

/* The number of bits in a word on this system */
int GEN_WordBits(GEN_CTX *genCtx);

/* The number of bits in R */
int GEN_RBits(GEN_CTX *genCtx);

/* The number of bits in a BGW polynomial coefficient */
int GEN_CoeffBits(GEN_CTX *genCtx);

/* The number of bits in P and Q */
int GEN_ShareBits(GEN_CTX *genCtx);

/* The degree of BGW polynomials */
int GEN_Threshold(GEN_CTX *genCtx);

/* Returns R */
BIGNUM* GEN_R(GEN_CTX *genCtx);

/* Returns the sieve */
char* GEN_Sieve(GEN_CTX *genCtx);

/* Returns 1 if the sieve is active, 0 otherwise */
int GEN_SieveP(GEN_CTX *genCtx);

/* The number of primes used in sieving */
int GEN_NumSievePrimes(GEN_CTX *genCtx);

/* The range of the sieve */
int GEN_SieveRange(GEN_CTX *genCtx);

/* Returns M */
BIGNUM* GEN_M(GEN_CTX *genCtx);

/* The upper bound for trial division primes */
int GEN_TrialDivBound(GEN_CTX *genCtx);

/* The number of factors used in trial division */
int GEN_NumTrialDivFactors(GEN_CTX *genCtx);

/* returns an array of the trial division factors */
long unsigned int* GEN_TrialDivFactors(GEN_CTX *genCtx);

/* Returns the binomials for LaGrange interpolation */
BIGNUM** GEN_Binomials(GEN_CTX *genCtx);

/* Returns 0 if the first binoimial is negative in interpolation, */
/* 1 if the second is negative */
int GEN_ModBinFirstNeg(GEN_CTX *genCtx);

/* Returns the RSA encryption exponent, e */
int GEN_ERSA(GEN_CTX *genCtx);

/* Return the BIGNUM representation of these numbers */
BIGNUM* GEN_Zero(GEN_CTX *genCtx);
BIGNUM* GEN_One(GEN_CTX *genCtx);
BIGNUM* GEN_Two(GEN_CTX *genCtx);
BIGNUM* GEN_Three(GEN_CTX *genCtx);

/* Return the BIGNUM's used to set a number mod 60 */
BIGNUM* GEN_CRTa(GEN_CTX *genCtx);
BIGNUM* GEN_CRTb(GEN_CTX *genCtx);

/* Return sixty times M */
BIGNUM* GEN_SixtyM(GEN_CTX *genCtx);

/* Return the BIGNUM context */ 
BN_CTX* GEN_BNContext(GEN_CTX *genCtx);

/* Return the UTL statistics */
UTL_STATS* GEN_Stat(GEN_CTX *genCtx);

#endif


