/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

#ifndef _KEY_H_
#define _KEY_H_

#include "COM.h"
#include "UTL.h"
#include "TLB_ctx.h"

#include <stdio.h>
#include "openssl/pem.h"

#define MAX_RSA_SHARING 20

/*
 * An RSA_SHARE * is identical to any other RSA *, except that its
 * app_data field points to an RSA_SHARE_APP_DATA structure and its
 * meth field contains a different set of function pointers.  These
 * functions handle RSA_SHAREs, which can be handed off to SSLeay
 * functions that accept RSAs just fine.  We use the RSA_SHARE *
 * data type to indicate that we mean an RSA * that has this
 * special form.
 */

/* Private (per-server) RSA share */
typedef struct rsa_share_app_st {
  /* Maybe need a pad here? */
  int nparties;
  int threshold;
  int ver;	/* Version number */
  int k;
  BIGNUM * d[MAX_RSA_SHARING];
} RSA_SHARE_APP_DATA;

/* RSA_SHARE is a "derived class" of RSA. */
typedef RSA RSA_SHARE;

/*
 * We use the same polymorphism for an RSA_SHARED_KEY as we
 * did for RSA_SHAREs.
 */

/* Public (gateway) key */
typedef struct rsa_shared_key_app_st {
  /* Maybe need a pad here? */
  int nparties;
  int threshold;
  BIGNUM * g;
  int k;
  BIGNUM * gd[MAX_RSA_SHARING];	    /* TJW: may need to increase array size */
  TLB_CTX * ctx;
} RSA_SHARED_KEY_APP_DATA;

/* RSA_SHARED_KEY is a "derived class" of RSA */
typedef RSA RSA_SHARED_KEY;

#define PEM_STRING_RSA_SHARE		"RSA SHARE"
#define PEM_STRING_RSA_SHARED_KEY	"RSA SHARED KEY"


/* These are Mike's function to retrieve a share, given a coalition and a server */
extern BIGNUM * KEY_RSA_share_get_d(RSA_SHARE * share, int coalition, int server);
extern BIGNUM * KEY_RSA_shared_key_get_gd(RSA_SHARED_KEY * key, int coalition, int server);

extern void KEY_RSA_share_initialize();		/* Call me first */
extern RSA_METHOD * KEY_RSA_share_meth_ssleay();
extern RSA_SHARE * KEY_RSA_share_new();
extern RSA_SHARE * KEY_RSA_share_create(BIGNUM *, BIGNUM *);
extern void KEY_RSA_share_set_threshold_params(RSA_SHARE *, int, int);
extern void KEY_RSA_share_set_version(RSA_SHARE *, int);
extern void KEY_RSA_share_add_d(RSA_SHARE *, BIGNUM *);
extern void KEY_RSA_share_free(RSA_SHARE *);

extern RSA_SHARE * KEY_RSA_share_read(char *);
extern int KEY_RSA_share_write(RSA_SHARE *, const char *, const char * enc_alg,
			       unsigned char * key, int keylen);
extern int KEY_RSA_share_write_cipher(RSA_SHARE *, const char *, EVP_CIPHER *,
				      unsigned char * key, int keylen);

extern BIGNUM * KEY_RSA_share_n(RSA_SHARE *);
extern BIGNUM * KEY_RSA_share_e(RSA_SHARE *);
extern int KEY_RSA_share_nparties(RSA_SHARE *);
extern int KEY_RSA_share_threshold(RSA_SHARE *);
extern int KEY_RSA_share_version(RSA_SHARE *);
extern int KEY_RSA_share_k(RSA_SHARE *);
extern BIGNUM * KEY_RSA_share_d(RSA_SHARE *, int);
extern void KEY_RSA_share_print_all_d(RSA_SHARE *);   /* Mike's function */

/* Set default passphrase */
extern void KEY_set_passphrase(const char *);
extern void KEY_set_cipher_name(const char *);

/* For (package-)internal use only */
PACKAGE char * KEY_get_cipher_name();
PACKAGE int KEY_use_callback();
PACKAGE int KEY_pp_callback();

PACKAGE RSA_SHARE_APP_DATA * KEY_RSA_share_get_data(RSA_SHARE *);

/* SSLeay-level functions */
RSA_SHARE * PEM_read_RSA_share(FILE *, RSA_SHARE **, int (*)(), void *);
RSA_SHARE * PEM_read_bio_RSA_share(BIO *, RSA_SHARE **, int (*)(), void *);
int PEM_write_RSA_share(FILE *, RSA_SHARE *,
			EVP_CIPHER *, unsigned char *, int, int (*)(), void *);
int PEM_write_bio_RSA_share(BIO *, RSA_SHARE *,
			    EVP_CIPHER *, unsigned char *, int, int (*)(), void *);

extern void KEY_RSA_shared_key_initialize();		/* Call me first */
extern RSA_METHOD * KEY_RSA_shared_key_meth_ssleay();
extern RSA_SHARED_KEY * KEY_RSA_shared_key_new();
extern RSA_SHARED_KEY * KEY_RSA_shared_key_create(BIGNUM *, BIGNUM *);
extern void KEY_RSA_shared_key_set_threshold_params(RSA_SHARED_KEY *, int, int);
extern void KEY_RSA_shared_key_set_ctx(RSA_SHARED_KEY *, TLB_CTX *);
extern void KEY_RSA_shared_key_set_g(RSA_SHARED_KEY *, BIGNUM *);
extern void KEY_RSA_shared_key_add_gd(RSA_SHARED_KEY *, BIGNUM *);
extern void KEY_RSA_shared_key_free(RSA_SHARED_KEY *);

extern RSA_SHARED_KEY * KEY_RSA_shared_key_read(char *);
extern int KEY_RSA_shared_key_write(RSA_SHARED_KEY *, char *);

extern BIGNUM * KEY_RSA_shared_key_n(RSA_SHARED_KEY *);
extern BIGNUM * KEY_RSA_shared_key_e(RSA_SHARED_KEY *);
extern int KEY_RSA_shared_key_nparties(RSA_SHARED_KEY *);
extern int KEY_RSA_shared_key_threshold(RSA_SHARED_KEY *);
extern BIGNUM * KEY_RSA_shared_key_g(RSA_SHARED_KEY *);
extern int KEY_RSA_shared_key_k(RSA_SHARED_KEY *);
extern TLB_CTX * KEY_RSA_shared_key_ctx(RSA_SHARED_KEY *);
extern BIGNUM * KEY_RSA_shared_key_gd(RSA_SHARED_KEY *, int);

PACKAGE RSA_SHARED_KEY_APP_DATA * KEY_RSA_shared_key_get_data(RSA_SHARED_KEY *);

/* SSLeay-level functions */
RSA_SHARED_KEY * PEM_read_RSA_shared_key(FILE *, RSA_SHARED_KEY **, int (*)(), void *);
RSA_SHARED_KEY * PEM_read_bio_RSA_shared_key(BIO *, RSA_SHARED_KEY **, int (*)(), void *);
int PEM_write_RSA_shared_key(FILE *, RSA_SHARED_KEY *,
			     EVP_CIPHER *, unsigned char *, int, int (*)(), void *);
int PEM_write_bio_RSA_shared_key(BIO *, RSA_SHARED_KEY *,
				 EVP_CIPHER *, unsigned char *, int, int (*)(), void *);

extern BN_ULONG KEY_RSA_get_ID(RSA *);



#define MAX_KEYS	256

typedef struct key_share_st {
  BN_ULONG id;		/* For efficiency */
  char * filename;
  RSA_SHARE * share;
} KEY_CTX_ENT;

typedef struct key_ctx_st {
  char * dirname;
  int count;
  KEY_CTX_ENT * keys;
} KEY_CTX;

extern KEY_CTX * KEY_Init(UTL_CTX * utl_ctx, int entity_id);
extern int KEY_CTX_getversions(KEY_CTX * ctx, BN_ULONG id, int * pmin, int * pmax);	// gets lowest, highest ver. numbers
extern RSA_SHARE * KEY_CTX_getbyid(KEY_CTX * ctx, BN_ULONG id);	// most recent
extern RSA_SHARE * KEY_CTX_getbyidver(KEY_CTX * ctx, BN_ULONG id, int ver);
extern RSA_SHARE * KEY_CTX_getbyindex(KEY_CTX * ctx, int index);
extern int KEY_CTX_add(KEY_CTX * ctx, RSA_SHARE * share);	// Add to ctx, save to disk
extern int KEY_CTX_flushidver(KEY_CTX * ctx, BN_ULONG id, int ver);  // Flush to disk
extern int KEY_CTX_deletebyindex(KEY_CTX * ctx, int index);
extern int KEY_CTX_deletebyidver(KEY_CTX * ctx, BN_ULONG id, int ver);	// delete specific version of share
extern int KEY_CTX_collapse_idver(KEY_CTX * ctx, BN_ULONG id, int ver);  // delete all but specified version of share
extern void KEY_CTX_close(KEY_CTX * ctx);

#endif
