/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	TLB.h								--*/
/*--									--*/
/*--	Header file for threshold library				--*/
/*------------------------------------------------------------------------*/

#ifndef TLB_H
#define TLB_H

#include "COM.h"
#include "UTL.h"
#include "KEY.h"
#include "openssl/rsa.h"

#include "TLB_ctx.h"


typedef struct TLB_internal_ctx_struct {

  // The following are only in the internal context
  BN_CTX	*BNctx;
  UTL_HEAP	*heap;
  UTL_STREAMS	*streams;
  TLB_CTX       *tlbCtx;

  // The following are mirrored from tlbCtx
  COM_CTX	*COMctx;
  UTL_CTX	*UTLctx;

  int numServers;

} TLB_INTERNAL_CTX;


/*------------------------*/
/*--  Naming Constants  --*/
/*------------------------*/

/* Corrupt of Offline servers were discovered during sig gen */
#define TLB_OFFLINE             2001
#define TLB_CORRUPT	        2002
#define TLB_ERROR               2003

/* TJW 9/12/98: Just a guess to get this to compile... */
#define TLB_MAX_ATTEMPTS	10

/*
 * Protocol opcodes (lowest byte only)
 */
#define TLB_OP_DONE		0x00
#define TLB_OP_APPLY_DECRYPT	0x01
#define TLB_OP_APPLY_SIG	0x02
#define TLB_OP_APPLY_HASHED	0x03
#define TLB_OP_MASK		0xFF

/*
 * Internal flag settings
 */
#define TLB_DECRYPT_MESSAGE	0x00
#define TLB_SIGN_MESSAGE	0x01
#define TLB_TEST_SERVERS	0x02


/*--------------------------------*/
/*--  function prototypes	--*/
/*--------------------------------*/

PUBLIC TLB_CTX * TLB_CTX_new(const char *configfile);
PUBLIC void TLB_CTX_free(TLB_CTX *);


TLB_INTERNAL_CTX *TLB_INTERNAL_CTX_new(TLB_CTX *tlbCtx);
void TLB_INTERNAL_CTX_free(TLB_INTERNAL_CTX * ctx);     // Does not free the TLB_CTX


PUBLIC int TLB_SignWithKey(BIGNUM * signature, BIGNUM * msg, 
			   RSA_SHARED_KEY * key);
PUBLIC int TLB_DecryptWithKey(BIGNUM * result, BIGNUM * msg, 
			      RSA_SHARED_KEY * key);


/*----------------------------------------*/
/*--  Replacement functions for SSLeay  --*/
/*----------------------------------------*/

int TLB_private_encrypt(int flen, unsigned char * from, unsigned char * to,
			RSA * rsa, int padding);
int TLB_private_decrypt(int flen, unsigned char * from, unsigned char * to,
			RSA * rsa, int padding);


/*----------------------------------------*/
/*--  TLB Internal functions.		--*/
/*----------------------------------------*/

/* TLB_ApplyKey: If coalition is NULL, a random coalition will be chosen.
                 If coalition is a pointer to 0 (zero), a random coalition will
		 be chosen and will be returned in the pointer.
		 If coalition is a pointer to a number > 0, that coalition
		 will be used in the decryption.  If that coalition fails, an
		 error will be returned and the decryption will NOT be attempted
		 with another coalition. */

PACKAGE int TLB_ApplyKey(BIGNUM *sig, BIGNUM *msg, int flags, int *coalition,
			 RSA_SHARED_KEY *pubkey);

extern int TLB_UseKey(BIGNUM *sigs[], BIGNUM *m, long int coalition, 
		      RSA_SHARED_KEY *pubkey, long int *offline, 
		      long int corrupt, int flags, TLB_INTERNAL_CTX *ctx);

extern int TLB_CheckSig(BIGNUM * sig, BIGNUM ** sigs, BIGNUM * msg,
			RSA_SHARED_KEY * pubkey, long coalition, 
			TLB_INTERNAL_CTX * ctx);

extern int TLB_ValidateServer(BIGNUM * msg, BIGNUM * sig, 
			      RSA_SHARED_KEY *pubkey,
		   int server, long int coalition, UTL_STREAMS *streams, 
		   TLB_INTERNAL_CTX * ctx);

extern int TLB_SendRequest(COM_STREAM *s, RSA_SHARED_KEY *pubkey, 
		     int op, BIGNUM *msg, long int coalition, 
		     long int offline, long int corrupt, TLB_INTERNAL_CTX *ctx);


/* Test the servers "Zero_Knowledge" times to check for corrupt servers
   or corrupt shares */
int TLB_TestServers(RSA_SHARED_KEY * key);

#endif


