/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*------------------------------------------------------------------------*/
/*--									--*/
/*--	UTL.h								--*/
/*--									--*/
/*--	Header file for utilities component.				--*/
/*--									--*/
/*------------------------------------------------------------------------*/

#ifndef UTL_h
#define UTL_h


/*-----------------*/
/*--  Constants  --*/
/*-----------------*/

#define UTL_MAXLINE      80
#define UTL_MAX_BN     1000

#define UTL_MONITOR_MAXMSG	230

/* Preprocessor directives in configuration file */
#define INCLUDE_DIRECTIVE "#INCLUDE"

#define GET_PRODUCT   0
#define GET_SUM       1

#define UTL_TYPE_STRING    17
#define UTL_TYPE_INT       19


/*-------------------------*/
/*--  Log message types  --*/
/*-------------------------*/

#define LOG_BLANK     -1
#define LOG_FATAL      0
#define LOG_ERROR      1
#define LOG_WARNING    2
#define LOG_NOTIFY     3
#define LOG_INFO       4
#define LOG_DEBUG      5
#define LOG_TEST       6


/*-------------------------------*/
/*---  Structure definitions  ---*/
/*-------------------------------*/

/* UTL Context */

#include "UTL_ctx.h"

/* Used for communications */
typedef struct UTL_streams {
  COM_STREAM **active;         // An array containing pointers to the active streams
  COM_STREAM **streams;        // An array of COM_STREAMs
  COM_CTX *context;            // COM context
  int myID;                    // My ID number
  int numServers;              // The total number of servers
} UTL_STREAMS;


/* Used for the heap of BIGNUM's */
typedef struct UTL_heap {
  int maxSize;
  int top;
  BIGNUM **array;
} UTL_HEAP;


/* Used to take statistics */
typedef struct UTL_stats {

  int numCount;    /* How many count statistics there are */
  int numTimes;    /* How many times statistics there are */

  int *count;      /* How many times a function have been run */

  unsigned long int *times;     /* Times taken by functions */
  UTL_TIME *start;     /* The starting time of a function */

  char **countNames;    /* Text names for the statistics */
  char **timeNames;     /* Text names for the statistics */

  int pread;     /* Used to count communications overhead */
  int pwrite;
  int bread;
  int bwrite;

} UTL_STATS;



/*---------------------------*/
/*--  Function prototypes  --*/
/*---------------------------*/


/*** Config File 
     -----------
     
     LoadConfigFile:  Load the config file at the path given by filename.
		UTL context must already be allocated (using UTL_init_ctx).

     UnLoadConfigFile:  Unload config file. Enables a new file to be loaded.

     GetConfigParam:  Returns a pointer to the value of name in the config
         file.  If this value is not found in the config file, deflt is 
	 returned instead.  Note: Do not modify the string that is returned,
	 because it points to a string that is used elsewhere.

     GetIntConfigParam:  Returns an integer config parameter.  Returns
         deflt if the parameter is not in the config file.

     GetBoolConfigParam:  Returns a boolean config parameter.  Returns
         deflt if the parameter is not in the config file.
*/
int   UTL_LoadConfigFile(UTL_CTX *ctx, const char *filename);
int   UTL_UnLoadConfigFile(UTL_CTX *ctx);
char *UTL_GetConfigParam(const UTL_CTX *ctx, const char *name, char *deflt);
char *UTL_GetPathConfigParam(UTL_CTX *ctx, const char *name, char *deflt);
int   UTL_GetIntConfigParam(const UTL_CTX *ctx, const char *name, 
			    const int deflt);
int   UTL_GetBoolConfigParam(const UTL_CTX *ctx, const char *name, 
			     const int deflt);
char *UTL_GetServerParam(const UTL_CTX *ctx, const char * name, int number, 
			 char * deflt);
char *UTL_GetPathServerParam(UTL_CTX *ctx, const char * name, int number, 
			     char * deflt);
int   UTL_GetIntServerParam(const UTL_CTX *ctx, const char * name, int number,
			    int deflt);
int   UTL_GetBoolServerParam(const UTL_CTX *ctx, const char * name, int number,
			     int deflt);
char *UTL_GetPathAdminParam(UTL_CTX *ctx, const char * name, int number, 
			    char * deflt);
char *UTL_GetPathClientParam(UTL_CTX *ctx, const char * name, int number, 
			     char * deflt);


/*** UTL_Context
     -----------

     init_ctx:  Allocates space for a UTL_CTX structure.
		Must be called before UTL_LoadConfigFile is called.

     free_ctx:  Deallocates a UTL_CTX structure.
*/

UTL_CTX *UTL_init_ctx();
int UTL_free_ctx(UTL_CTX *ctx);


/*** Logging
     -------
     
     Log:  Makes a log entry.  Loc is the location that is calling 
         UTL_Log, severity is the importance of the log entry, and 
	 msg is the log message.  Parameters can be added after the message
	 in the same form as a call to printf.  The configuration file 
	 tells what message severity is important enough to print to screen
	 and to put in the log file.
	 * Severity levels, from most important to least:
	   BLANK, FATAL, ERROR, WARNING, NOTIFY, INFO, DEBUG, TEST

     OpenLogFile:  Opens the log file.  The location of this file is given
         in the configuration file.

     CloseLogFile:  Closes the log file.

     FlushLogFile:  Causes all pending log items to put flushed to the log file.
*/
void UTL_Log(char *loc, int severity, char *msg, ...);
int  UTL_OpenLogFile(UTL_CTX *ctx, COM_TYPE ty, int num);
int  UTL_CloseLogFile();
int  UTL_FlushLogFile();


/*** Monitor utility
     ---------------

     MonitorInit:  Establish connection to monitor server.

     monitor:  Sends a message to the monitor server.

     MonitorClose:  Close connection to server.
*/

int UTL_monitor(UTL_CTX *ctx, int msgattr, char *msg, ...);
int UTL_MonitorInit(UTL_CTX *ctx, COM_TYPE mytype, int myID);
int UTL_MonitorClose(UTL_CTX *ctx);


/*** Heap 
     ----

     InitHeap:  Returns a malloc'd heap, that is initialized to hold maxSize
         bignums.

     GetBN:  Returns a BIGNUM* from the heap that can be used by the calling
         program.  The BIGNUM* is initialized to zero.

     GetBNArray:  Returns an array of BIGNUM*'s.  There are num entries in 
         the array, which are initialized to zero.

     GetHeapState:  Returns an integer that summarizes the state of the heap.
         This integer is used with UTL_RevertHeap to free all BIGNUM*'s allocated
	 since this function has been called.

     RevertHeap:  When given the number returned by GetHeapState, this function
         deallocates all BIGNUM*'s allocated since GetHeapState was called.
	 Note: Deallocating a BIGNUM* isn't the same as free'ing it.  It simply 
	 means that it is available to be used by a call to UTL_GetBN.

     ReleaseHeap:  Like UTL_RevertHeap, but it it deallocates ALL BIGNUM*'s in
         the heap.

     FreeHeap:  This function actually free's the heap and all associated 
         BIGNUM*'s.  UTL_InitHeap needs to be called before the heap can be 
	 used again.
*/
UTL_HEAP *UTL_InitHeap(int maxSize);
BIGNUM *UTL_GetBN(UTL_HEAP *heap);
BIGNUM **UTL_GetBNArray(int num, UTL_HEAP *heap);
int   UTL_GetHeapState(UTL_HEAP *heap);
int   UTL_RevertHeap(int state, UTL_HEAP *heap);
int   UTL_ReleaseHeap(UTL_HEAP *heap);
int   UTL_FreeHeap(UTL_HEAP *heap);


/*** Stats
     -----

     # There are two types of statistics: timing-statistics and
       counting-statistics.  Counting-statistics can only be
       incremented.  Timing-statistics can be incremented continuously
       over periods of time.

     InitStats:  This function initializes a UTL_STATS* variable.
         The variable "stats" must already be malloc'd.

     FreeStats:  Free's a UTL_STATS* and all internal structures.

     StartStat:  Starts adding time to timing-statistic number "stat".

     FinishStat:  Finishes adding time to timing-statistic number "stat".

     AddCountStat:  Increments counting-statistic number "stat".

     NameCountStat:  Gives the name for counting-stat number "stat".

     NameTimeStat:  Gives the name for timing-stat number "stat".

     PrintStats:  Prints out statistics.

     PrintAverageStats:  Prints out average statistics.  The array 
         "stats" has "iterations" elements in it.

     CountStat:  Returns the value of the counting-statistic "stat".

     TimeStat:  Returns the value of the timing-statistic "stat".
*/
int UTL_InitStats(int count, int times, UTL_STATS* stats);
int UTL_FreeStats(UTL_STATS* stats);
int UTL_StartStat(int stat, UTL_STATS* stats);
int UTL_FinishStat(int stat, UTL_STATS* stats);
int UTL_AddCountStat(int stat, UTL_STATS* stats);
int UTL_NameCountStat(int stat, char* name, UTL_STATS* stats);
int UTL_NameTimeStat(int stat, char* name, UTL_STATS* stats);
int UTL_PrintStats(UTL_STATS* stats);
int UTL_PrintAverageStats(int iterations, UTL_STATS** stats);
int UTL_CountStat(int stat, UTL_STATS* stats);
int UTL_TimeStat(int stat, UTL_STATS* stats);


/*** GetFrom and SendTo functions 
     ----------------------------

     WaitRead: Takes as a parameter a stream, and returns when the
         stream has data to read.  Also returns if the stream closes
	 or times out.
     SendToAll: "send" is an array of BIGNUM*'s.  Each element in the
         array is sent to the corresponding server.  For example,
         array element 3 is sent to server number 3.  There are two
         exceptions to this rule: 1) a server never sends a BIGNUM to
         itself, and 2) if the UTL_Stream(x, servers)==NULL then no
         BIGNUM is sent to server x.  Note that the first exception is
         meaningless if UTL_MyID is greater than numServers or less
         than zero.  For example, if UTL_MyID==-1, then a BIGNUM will
         be sent to all servers, except those excluded by exception 2.
         Note: this function is non-blocking.

     SendSameToAll: This is just like UTL_SendToAll, but it sends the
         same BIGNUM to each server.

     GetFromAll: "get" is an array of BIGNUM's.  This function
         receives a BIGNUM from each server and puts it into the spot
         of the array corresponding to the server number.  It will not
         get BIGNUM's from itself, or from any server where
         UTL_Stream(x, streams)==NULL.  If UTL_MyID==-1, for example,
         it will get a BIGNUM from every server except those where
         UTL_Stream(x, streams)==NULL.  This is a blocking function,
         and it operates asynchronously.  That is, it can receive the
         BIGNUM's in any order from the other servers.

	 If RC_TIMEOUT or RC_CLOSE is returned, then the UTL_STREAMS
	 structure will have it's active array modified so only the 
	 streams that have timed out, or the stream that closed, will
	 be active.

	 If the stats parameter is NULL no statistics will be taken.
	 Otherwise, statistics will be taken on the amount of time
	 that the procedure spends in a select.

     GetFromOne: Gets a single BIGNUM from a server number "server".
         This is a convenience function in that it doesn't do anything 
	 complicated, but it does all of the error-checking and 
	 stream-handling for you.
	 
     GetFromOneSimple: Just like GetFromOne(), but it requires less
         complicated parameters, just a stream and a bignum.
	 
     GetProductFromAll: This is just like GetFromAll, but it returns
         the product of all received BIGNUM's.  This product is also
         multiplied by "myNum", which is the current server's
         contribution to the product.  Additionally, if "M" is
         non-NULL, all multiplications will be performed mod "M".  

     GetSumFromAll: This is just like GetFromAll, but it returns the
         sum of all received BIGNUM's.  "myNum" is added as the
         current server's contribution to the sum.  Additionally, if
         "M" is non-NULL, the final sum will be reduced mod "M".  

     GetFNFromAll: UTL_GetProductFromAll and GetSumFromAll are
         actually wrappers for this function.  The possible values for
         "function" are GET_PRODUCT and GET_SUM.

     Notify: Notifies all other servers of a failure.  Sends the value
         of "failed" to all other servers (except itself or servers
         with UTL_Stream==NULL) and sets "failed" to be 1 if it was
         originally 1 or if this function received a 1 from any other
         server.  
*/

int UTL_WaitRead(COM_STREAM* stream);
int UTL_SendToAll(BIGNUM** send, UTL_STREAMS *streams);  
int UTL_SendSameToAll(BIGNUM* send, UTL_STREAMS *streams);
int UTL_GetFromAll(BIGNUM **get, UTL_STREAMS *streams, UTL_HEAP *heap, UTL_STATS* stats);
int UTL_GetFromOne(BIGNUM *get, UTL_STREAMS *streams, int server, UTL_HEAP *heap, UTL_STATS* stats);
int UTL_GetFromOneSimple(BIGNUM *get, COM_STREAM *stream);
int UTL_GetProductFromAll(BIGNUM *product, BIGNUM *myNum, BIGNUM *M, BN_CTX *bnCtx, 
			  UTL_STREAMS *streams, UTL_HEAP *heap, UTL_STATS* stats);
int UTL_GetSumFromAll(BIGNUM *sum, BIGNUM *myNum, BIGNUM *M, BN_CTX *bnCtx, 
		      UTL_STREAMS *streams, UTL_HEAP *heap, UTL_STATS* stats);
int UTL_GetFNFromAll(BIGNUM *result, BIGNUM *myNum, BIGNUM *M, int function, BN_CTX *bnCtx, 
		     UTL_STREAMS *streams, UTL_HEAP *heap, UTL_STATS* stats);
int UTL_Notify(int *failed, UTL_STREAMS *streams, UTL_HEAP *heap);


/*** Check functions
     ---------------

     CheckResponseStr: Gets a string from a server.  Returns RC_OK, 
         RC_ERROR, RC_CLOSE, and RC_TIMEOUT.

     CheckResponseInt: This is the same as CheckResponseStr except
         that it takes an int from each server.  

     CheckResponse: The previous two functions are wrappers for this
         function.
*/

int UTL_CheckResponseStr(COM_STREAM *stream, char *msg);
int UTL_CheckResponseInt(COM_STREAM *stream, int goal);
int UTL_CheckResponse(COM_STREAM *stream, void *data, int flag);


/*** Connection functions
     --------------------

     * These functions connect to share servers, and are primarily 
       used in the GEN module.

     InitStreams: Returns a malloc'd, initialized UTL_STREAMS
         structure.

     SetupConnections: "vector" is a bit vector.  A connection is
         established with all servers where the corresponding bit of
         "vector" is 1, unless that corresponds to the server calling
         the function.  All bits in "vector" corresponding to servers
         greater than numServers-1 are ignored.  A convenient way to
         connect to all servers is to set vector=~0.  The connections
         are set up asynchronously.  This function can be called
         multiple times to establish more connections.  If a
         connection already exists, a new connection will _not_ be
         established.  Thread tells COM which virtual port to use.

     CloseAllConnections: Closes all connections and free's the
         UTL_STREAMS structure.

     OpenConnection: Opens a single connection to a specific server.
	 If it is connecting to a server where the server's ID >
	 UTL_MyID then it is non-blocking.  Otherwise, it is blocking.
	 It can never cause a server to establish a connection with
	 itself.  This function should only be called if it is know
	 that server number "server" is calling this function at the
	 same time to complete the connection.

     CloseConnection: Closes a single connection to server number
         "server".

     MyID: Returns the ID of the server calling this function.

     NumServers: Returns the number of share servers.

     Stream: Returns a pointer to the stream to server number "num".
         If no connection exists to this server, NULL will be returned.

     SetStream: Sets the connection to server number "num" to be
         "stream".  This stream replaces any previous stream, so you
         have to be careful to free whatever was there before, if
         necessary.

     COMContext: Returns the COM_CTX for the server calling the function.

     MakeConnection: Makes a connection to a server.  Does not make it 
         active.
*/
UTL_STREAMS *UTL_InitStreams(int numServers, int me, COM_CTX *context);
COM_STREAM* UTL_Stream(int num, UTL_STREAMS *streams);
int UTL_SetStream(int num, COM_STREAM* stream, UTL_STREAMS *streams);    

int UTL_MyID(UTL_STREAMS *streams);       
int UTL_NumServers(UTL_STREAMS *streams);            
COM_CTX* UTL_COMContext(UTL_STREAMS *streams);    

int UTL_SetupConnections(int vector, UTL_STREAMS *streams, UTL_HEAP *heap, int thread);
int UTL_CloseAllConnections(UTL_STREAMS *streams);   


/* Internal stream functions */
int UTL_Activate(int num, UTL_STREAMS *streams);
int UTL_Deactivate(int num, UTL_STREAMS *streams);
int UTL_ActivateAll(UTL_STREAMS *streams);
int UTL_DeactivateAll(UTL_STREAMS *streams);
int UTL_ActiveV(UTL_STREAMS *streams);             /* Returns a bit vector of active streams */
COM_STREAM** UTL_Active(UTL_STREAMS *streams);     /* Returns an array of active streams */
int UTL_Clear(UTL_STREAMS *streams);    /* reads out all strings, int's, and bignum's */


/*** Other UTL functions
     -------------------

     Random:  Returns "randNum", a number between 0 and "range".

     StreamID:  Returns the ID number of "stream".

     Prime: Returns prime number "num".  For example, UTL_Prime(0)==2,
         UTL_Prime(1)==3, and UTL_Prime(3)==7.

     UTL_IntBN: Returns the long int form of a BIGNUM.  This function
         has problems if "b" is too large tp fit in a long int.

     UTL_RelativeToFile: Interprets a pathname relative to another
         filename, writes result to "out".
	
     UTL_Bit2Str: Converts the bit-vector "bits" into a string, returns
         the string in "str".  "str" must be pre-allocated, and must be
	 at least 33 char's long.  THe bit vector has length "len".
*/
int UTL_Random(BIGNUM *randNum, BIGNUM *range, BN_CTX *bnCtx, UTL_HEAP *heap);
int UTL_StreamID(COM_STREAM *stream);
int UTL_Prime(int num);
long int UTL_IntBN(BIGNUM *b);
char * UTL_RelativeToFile(char * out, const char * filespec, const char * newfile);
int UTL_Hash(BIGNUM *out, BIGNUM *in);
void UTL_Bit2Str(char *str, int bits, int len);

#endif

