/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

/*----------------------------------------------------------------------*/
/*									*/
/*	RSA *ITTC_load_RSA_key(char *filename)				*/
/*									*/
/*	Main (only) interface function to the ITTC system.		*/
/*	Used to load a public RSA shared key  (a .rsk file).		*/
/*	Used by ITTC clients.						*/
/*									*/
/*	The functions returns a standard RSA structure to be used	*/
/*	by SSLeay functions.						*/
/*									*/
/*----------------------------------------------------------------------*/

#include "ITTC.h"
#include <openssl/err.h>
#ifndef WIN32
#include <signal.h>
#endif

int
ITTC_init()
{
	static int initialized = 0;

	if(!initialized) {
#ifdef WIN32
		WSADATA wsadata;
		WSAStartup(MAKEWORD(1, 1), &wsadata);
		COM_set_error_stream(stderr);
#else
		sigignore(SIGPIPE);
#endif
		SSLeay_add_all_algorithms();
		ERR_load_crypto_strings();

		KEY_RSA_share_initialize();
		KEY_RSA_shared_key_initialize();

		initialized = 1;
	}
	return RC_OK;
}

RSA_SHARED_KEY *
ITTC_load_RSA_key(keyconf)
     const char * keyconf;
{
  RSA_SHARED_KEY * key;
  TLB_CTX * tc;
  char * keyfile;

  ITTC_init();

  if((tc = TLB_CTX_new(keyconf)) == NULL) {
    fprintf(stderr, "ITTC_load_RSA_key: error creating TLB_ctx.\n");
    return NULL;
  }

  keyfile = UTL_GetPathConfigParam(tc->UTLctx, "Shared_Key", NULL);
  if(keyfile == NULL) {
    fprintf(stderr, "ITTC_load_RSA_key: Unable to locate shared key file.\n");
    TLB_CTX_free(tc);
    return NULL;
  }

  key = KEY_RSA_shared_key_read(keyfile);

  if (!key) {
    fprintf(stderr, "ITTC_load_RSA_key: Unable to load shared key file %s.\n",
		keyfile);
    
    TLB_CTX_free(tc);
    return NULL;
  }

  KEY_RSA_shared_key_set_ctx(key, tc);

  return key;
}
