/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

#include "ITTC.h"


/* Generates a SHA-1 hash of in, stores result in out */
int UTL_Hash(BIGNUM *out, BIGNUM *in) {

  BIGNUM *temp;
  unsigned char *txt;
  unsigned char hash[SHA_DIGEST_LENGTH + 2];
  int len;
  int rc;
  int neg;

  /* Initialize */
  len = BN_num_bytes(in);
  txt = (unsigned char*)malloc((len+2)*sizeof(unsigned char));

  BN_print_fp(stdout, in);
  printf("\t");

  /* Make a clean form of the input */
  temp = BN_new();
  neg = in->neg;
  in->neg = 0;
  BN_add(temp, in, temp);
  in->neg = neg;

  BN_print_fp(stdout, temp);
  printf("\t");

  /* Convert the input to a string */
  rc = BN_bn2bin(in, txt);
  if (rc <= 0) {
    UTL_Log("UTL_Hash", LOG_ERROR, "Problem converting bignum to string.");
    return RC_ERROR;
  }

  /* Perform the hash */
  SHA1(txt, strlen(txt), hash);

  /* Convert the hash to a bignum */
  if (BN_bin2bn(hash, SHA_DIGEST_LENGTH, out) == NULL) {
    UTL_Log("UTL_Hash", LOG_ERROR, "Problem converting hash to bignum.");
    return RC_ERROR;
  }

  BN_print_fp(stdout, out);
  printf("\n");

  BN_clear_free(temp);
  free(txt);
  return RC_OK;
}



/* Generates a random number from 0 to range */
int Rnd(BIGNUM *randNum, BIGNUM *range) {
  BIGNUM *tmp;
  BN_CTX *bnCtx;

  tmp = BN_new();
  bnCtx = BN_CTX_new();
  
  if (!BN_rand(tmp, BN_num_bits(range)+12, 0, 0)) {
    UTL_Log("KEY_Random", LOG_FATAL, "Error in BN_rand.");
    return RC_ERROR;
  }
  if (!BN_mod(randNum, tmp, range, bnCtx)) {
    UTL_Log("KEY_Random", LOG_FATAL, "Error in BN_mod.");
    return RC_ERROR;
  }


  return RC_OK;
}




void strip(char* txt) {
  int i;
  int len;

  len = strlen(txt);
  
  for (i=0; i<len; i++) {
    txt[i] &= 127;
    if (txt[i] < 'A') txt[i] = '#';
  }

}


void niceprint(char* txt) {
  int i, j;
  int len;

  len = strlen(txt);

  for (i=0; i<len; i++) {
    for (j=7; j>=0; j--) {

      if (txt[i] & (1<<j)) printf("1");
      else printf("0");

    }
    printf(" ");
  }

}



int main() {
  int i;
  BIGNUM *temp, *t1, *t2, *t3;
  BN_CTX *bnCtx;

  // Initialize
  temp = BN_new();
  t1 = BN_new();
  t2 = BN_new();
  t3 = BN_new();
  bnCtx = BN_CTX_new();

  // set t3 = 1000^8
  BN_set_word(t1, 1000*1000);
  BN_set_word(t2, 1000*1000);
  BN_mul(t3, t1, t2, bnCtx);
  BN_copy(t1, t3);
  BN_copy(t2, t3);
  BN_mul(t3, t1, t2, bnCtx);

  // Do the work
  for (i=1; i<=20; i++) {
    Rnd(temp, t3);

    UTL_Hash(t1, temp);

    //txt = (char*)malloc((BN_num_bytes(temp)+3)*sizeof(char*));
    //BN_bn2bin(temp, txt);
    //printf("%3d:  ", i);
    //niceprint(txt);
    //printf("\n");
    //free(txt);
  }

  BN_CTX_free(bnCtx);

  return 0;
}

