/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <stdio.h>
#include <sys/types.h>
#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/time.h>
#endif
#ifdef WIN32
#include "../ADM/getopt.h"
#endif
#include "ITTC.h"

#include "UTL_mutex.h"
#include "UTL_cond.h"
#include "UTL_thread.h"
#include "UTL_Time.h"


static char * progname;

static void
usage()
{
  fprintf(stderr, "Usage: %s [-dqv] [-n num] [-t thr] shared-key\n", progname);
  fprintf(stderr, "\t-d    (debug) Print data received from servers\n");
  fprintf(stderr, "\t-q    (quiet) Don't print stats at end\n");
  fprintf(stderr, "\t-r    (rsa) Treat shared-key as a normal RSA key\n");
  fprintf(stderr, "\t-v    (verbose) Print information on thread creation/deletion\n");
  fprintf(stderr, "\t-n n  (number) Issue a total of n requests\n");
  fprintf(stderr, "\t-t n  (thread) Use a maximum of n threads at any time (default 1)\n");
  exit(1);
}

static double
getelapsed(stimep)
     UTL_TIME * stimep;
{
  UTL_TIME etime;

  UTL_TimeGetCurrent(&etime);
  return UTL_TimeSubtractMsec(&etime, stimep) / 1000.0;
}

typedef struct conn_ctx_st {
  RSA * key;
  int threaded;
} CONN_CTX;

static unsigned long bcount = 0;
static double alltime = 0.0;
static int dflag = 0, vflag = 0;

static UTL_COND gcond;	/* Thread decrement condition var. */
static UTL_MUTEX gmutex;	/* Thread update mutex */
static int thrcount = 0;

/* Wait until there are no more than "n" threads active */
static void
waitcount(n)
     int n;
{
  UTL_mutex_lock(&gmutex);
  while(1) {
    if(thrcount > n)
      UTL_cond_wait(&gcond, &gmutex);
    else {
      UTL_mutex_unlock(&gmutex);
      break;
    }
  }
}

static void
hexprint(s, len)
     unsigned char * s;
     int len;
{
  while(len-- > 0)
    printf("%02X", *s++);
}

#define INPUT_LENGTH 4

static void
hit(ctx)
     CONN_CTX * ctx;
{
  int buflen, r;
  unsigned char * buf;
  unsigned char * buf2;
  unsigned char txt[INPUT_LENGTH];
  UTL_TIME htime;
  double tincr;

  UTL_TimeGetCurrent(&htime);

  buflen = RSA_size(ctx->key);
  if((buf = malloc(buflen)) == NULL || (buf2 = malloc(buflen)) == NULL) {
    fprintf(stderr, "Unable to allocate memory\n");
    exit(3);
  }

  /* Make a random text string */
  RAND_bytes(txt, sizeof(txt));
  if(dflag) {
    printf("%-20s: ", "Original message");
    hexprint(txt, sizeof(txt));
    printf("\n");
  }
    
  /* Uses EAY padding/encryption (local) */
  if((r = RSA_public_encrypt(sizeof(txt), txt, buf, ctx->key, RSA_PKCS1_PADDING)) < 0) {
    fprintf(stderr, "Error encrypting data\n");
    exit(5);
  }
  if(dflag) {
    printf("%-20s: ", "RSA encrypted");
    hexprint(buf, r);
    printf("\n");
  }
    
  /* Performs shared decryption over network */
  if((r = RSA_private_decrypt(r, buf, buf2, ctx->key, RSA_PKCS1_PADDING)) < 0) {
    fprintf(stderr, "Error decrypting data\n");
    exit(6);
  }
  if(dflag) {
    printf("%-20s: ", "RSA decrypted");
    hexprint(buf2, r);
    printf("\n");
  }

  /* Check the result */
  if(memcmp(buf2, txt, sizeof(txt)) == 0)
    bcount += r;
  else {
    fprintf(stderr, "Signature mismatch!\n");
    exit(8);
  }

  free(buf);
  free(buf2);

  tincr = getelapsed(&htime);

  if(ctx->threaded) {
    UTL_mutex_lock(&gmutex);
    alltime += tincr;
    --thrcount;
    if(vflag)
      printf("(-) %d threads\n", thrcount);
    UTL_mutex_unlock(&gmutex);
    UTL_cond_notify(&gcond);
  }
  else
    alltime += tincr;
}

int main(argc, argv)
     int argc;
     char * argv[];
{
  int c;
  int qflag = 0, rflag = 0;
  int num = 1;
  int i;
  int nthreads = 1;
  UTL_TIME stime;
  double elapsed;
  FILE * fp;
  CONN_CTX cctx;
  UTL_THREAD tid;
  int version;

  progname = argv[0];

  while((c = getopt(argc, argv, "dqrvn:p:t:")) != EOF) {
    switch(c) {
    case 'q':
      ++qflag;
      break;
    case 'd':
      ++dflag;
      break;
    case 'r':
      ++rflag;
      break;
    case 'v':
      ++vflag;
      break;
    case 'n':
      num = atoi(optarg);
      break;
    case 't':
      nthreads = atoi(optarg);
      break;
    default:
      usage();
    }
  }

  if(num <= 0 || nthreads < 0 || optind >= argc)
    usage();

  cctx.key = NULL;

  if(rflag) {
    fp = fopen(argv[optind], "r");
    if(fp) {
      cctx.key = PEM_read_RSAPrivateKey(fp, NULL, NULL, NULL);
      fclose(fp);
    }
  }
  else {
    cctx.key = ITTC_load_RSA_key(argv[optind]);
    version = KEY_RSA_share_version(cctx.key);    
    printf("key version: %d\n", version);
  }  

  /* "key" now behaves just like a "normal" RSA public/private key. */
  if(cctx.key == NULL) {
    fprintf(stderr, "%s: Unable to read RSA key '%s'\n", progname, argv[optind]);
    exit(2);
  }
    
  cctx.threaded = (nthreads == 1) ? 0 : 1;

  UTL_cond_init(&gcond);
  UTL_mutex_init(&gmutex);

  UTL_TimeGetCurrent(&stime);
  for(i = 0; i < num; ++i) {
    if(nthreads == 1)
      hit(&cctx);
    else {
      if(nthreads > 0)
	waitcount(nthreads - 1);
      if(UTL_thread_create(&tid, (void*)hit, &cctx) < 0)
	perror("UTL_thread_create");
      else {
	UTL_mutex_lock(&gmutex);
	++thrcount;
	if(vflag)
	  printf("(+) %d threads\n", thrcount);
	UTL_mutex_unlock(&gmutex);
      }
    }
#ifdef ALL_STATS
  if(!qflag) {
  elapsed = getelapsed(&stime);
    printf("   Total bytes processed: %ld\n", bcount);
    printf("Sequential (thread) time: %g s\n", alltime);
    printf("                 Latency: %g s\n", alltime / num);
    printf("    Parallel (real) time: %g s\n", elapsed);
    printf("       Amortized latency: %g s\n", elapsed / num);
    printf("              Throughput: %g ops/s\n", num / elapsed);

  }
#endif
  }
  if(nthreads != 1)
    waitcount(0);

  elapsed = getelapsed(&stime);

  UTL_mutex_destroy(&gmutex);
  UTL_cond_destroy(&gcond);

  KEY_RSA_shared_key_free(cctx.key);

  if(!qflag) {
    printf("\n\n");
    printf("   Total bytes processed: %ld\n", bcount);
    printf("Sequential (thread) time: %g s\n", alltime);
    printf("    Parallel (real) time: %g s\n", elapsed);
    printf("       Amortized latency: %g s\n", elapsed / num);
    printf("\n");
    printf("                 Latency: %g s\n", alltime / num);
    printf("              Throughput: %g ops/s\n", num / elapsed);
    printf("\n");
  }

  return (bcount > 0) ? 0 : 1;
}

