/* Copyright (C) 1999 Stanford University (dabo@cs.stanford.edu)
 * All rights reserved.
 *
 * This package is an implementation of the ITTC system written
 * by Michael Malkin, Tom Wu, and Dan Boneh (dabo@cs.stanford.edu).
 * 
 * This library is free for non-commercial use as long as
 * the following conditions are adhered to.  The following conditions
 * apply to all source code found in the src directory of this distribution.
 * Any source code found in other directories is subject to the copyright 
 * notice within that code.
 * 
 * Copyright belongs to Stanford University and as such any Copyright 
 * notices in the code are not to be removed.  This package may not be 
 * used in a commercial product without the explicit consent of the authors. 
 *
 * Free redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *    "This product includes software written by Mike Malkin, Tom Wu, 
 *     and Dan Boneh."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The license and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution license
 * [including the GNU Public License.]
 */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/time.h>
#include <arpa/inet.h>

#ifdef PTHREADS
#include <pthread.h>
#endif

#include <openssl/err.h>
#include <openssl/ssl.h>

#ifndef INADDR_NONE
#define INADDR_NONE     0xffffffff      /* should be in <netinet/in.h> */
#endif

char http_req[] = "HEAD / HTTP/1.0\r\n\r\n";

static char * progname;

static void
usage()
{
  fprintf(stderr, "Usage: %s [-dqvc] [-n num] [-t thr] [-p port] host\n", progname);
  fprintf(stderr, "\t-d    (debug) Print data received from servers\n");
  fprintf(stderr, "\t-q    (quiet) Don't print stats at end\n");
  fprintf(stderr, "\t-v    (verbose) Print information on thread creation/deletion\n");
  fprintf(stderr, "\t-n n  (number) Issue a total of n requests\n");
  fprintf(stderr, "\t-t n  (thread) Use a maximum of n threads at any time (default 1)\n");
  fprintf(stderr, "\t-p n  (port) Use port n (default 443)\n");
  fprintf(stderr, "\t-c    (clear) Use clear transport (default is SSL)\n");
  exit(1);
}

static double
getelapsed(stimep)
     struct timeval * stimep;
{
  struct timeval etime;

  gettimeofday(&etime, NULL);
  return (etime.tv_sec - stimep->tv_sec) + (etime.tv_usec - stimep->tv_usec) / 1000000.0;
}

typedef struct conn_ctx_st {
  struct sockaddr * c_addr;
  int c_addr_size;
  int threaded;
  int cflag;
  SSL_CTX * c_sslctx;
} CONN_CTX;

static unsigned long bcount = 0;
static double alltime = 0.0;
static int dflag = 0, vflag = 0;

#ifdef PTHREADS
static pthread_cond_t gcond;	/* Thread decrement condition var. */
static pthread_mutex_t gmutex;	/* Thread update mutex */
static int thrcount = 0;

/* Wait until there are no more than "n" threads active */
static void
waitcount(n)
     int n;
{
  pthread_mutex_lock(&gmutex);
  while(1) {
    if(thrcount > n)
      pthread_cond_wait(&gcond, &gmutex);
    else {
      pthread_mutex_unlock(&gmutex);
      break;
    }
  }
}
#endif

static void
hit(ctx)
     CONN_CTX * ctx;
{
  SSL * ssl;
  BIO * bio;
  int s;
  unsigned buflen;
  unsigned char buf[1024];
  struct timeval htime;
  double tincr;

  gettimeofday(&htime, NULL);
  if((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    perror("socket");
    exit(7);
  }

  if(connect(s, ctx->c_addr, ctx->c_addr_size) < 0) {
    perror("connect");
    exit(7);
  }

  if (ctx->cflag) {		// Clear transport
    if (send(s, http_req, strlen(http_req), 0) < 0) {
      perror("send");
      exit(7);
    }
    
    while ((buflen = recv(s, buf, sizeof(buf), 0)) > 0) {
      if(dflag)
	fwrite(buf, 1, buflen, stdout);
      bcount += buflen;
    }

    close(s);
  }
  else {			// SSL transport

    if((ssl = SSL_new(ctx->c_sslctx)) == NULL ||
       (bio = BIO_new_socket(s, BIO_CLOSE)) == NULL) {
      ERR_print_errors_fp(stderr);
      exit(7);
    }

    SSL_set_bio(ssl, bio, bio);
    if (SSL_connect(ssl) <= 0 || 
	SSL_write(ssl, http_req, strlen(http_req)) <= 0) {
      ERR_print_errors_fp(stderr);
      exit(7);
    }

    while((buflen = SSL_read(ssl, buf, sizeof(buf))) > 0) {
      if(dflag)
	fwrite(buf, 1, buflen, stdout);
      bcount += buflen;
    }
      
    SSL_shutdown(ssl);
    BIO_free(bio);
      
  }

  tincr = getelapsed(&htime);

#ifdef PTHREADS
  if(ctx->threaded) {
    pthread_mutex_lock(&gmutex);
    alltime += tincr;
    --thrcount;
    if(vflag)
      printf("(-) %d threads\n", thrcount);
    pthread_mutex_unlock(&gmutex);
    pthread_cond_signal(&gcond);
  }
  else
#endif
    alltime += tincr;
}

int main(argc, argv)
     int argc;
     char * argv[];
{
  char * host;
  int port = 443;	/* Default for secure Web server */
  struct hostent * hptr;
  struct sockaddr_in saddr;
  unsigned long inaddr;
  int c;
  int qflag = 0;
  int cflag = 0;
  int num = 1;
  int i;
  int nthreads = 1;
  struct timeval stime;
  double elapsed;
  SSL_METHOD * meth;
  CONN_CTX cctx;
#ifdef PTHREADS
  pthread_t tid;
#endif

  progname = argv[0];

  while((c = getopt(argc, argv, "dqcvn:p:t:")) != EOF) {
    switch(c) {
    case 'q':
      ++qflag;
      break;
    case 'd':
      ++dflag;
      break;
    case 'v':
      ++vflag;
      break;
    case 'n':
      num = atoi(optarg);
      break;
    case 'p':
      port = atoi(optarg);
      break;
    case 't':
      nthreads = atoi(optarg);
      break;
    case 'c':
      ++cflag;
      break;
    default:
      usage();
    }
  }

  if(port <= 0 || num <= 0 || nthreads < 0 || optind >= argc)
    usage();

  cctx.cflag = cflag;
  if (cflag)
    fprintf(stderr, "Using clear transport.\n");

#ifndef PTHREADS
  if(nthreads != 1) {
    fprintf(stderr, "Multithreading not supported\n");
    exit(2);
  }
#endif

  host = argv[optind];

  ERR_load_crypto_strings();
  SSL_load_error_strings();
  SSLeay_add_ssl_algorithms();

  if((meth = SSLv23_client_method()) == NULL ||
     (cctx.c_sslctx = SSL_CTX_new(meth)) == NULL) {
    ERR_print_errors_fp(stderr);
    exit(2);
  }

  memset((char *) &saddr, 0, sizeof(saddr));
  saddr.sin_family = AF_INET;
  saddr.sin_port = htons(port);

  if((inaddr = inet_addr(host)) == INADDR_NONE) {
    if((hptr = gethostbyname(host)) == NULL) {
      fprintf(stderr, "Unknown hostname %s\n", host);
      return NULL;
    }
    else
      memcpy((char *) &saddr.sin_addr, hptr->h_addr, hptr->h_length);
  }
  else
    saddr.sin_addr.s_addr = inaddr;

  cctx.c_addr = (struct sockaddr *) &saddr;
  cctx.c_addr_size = sizeof(saddr);
#ifdef PTHREADS
  cctx.threaded = (nthreads == 1) ? 0 : 1;
#else
  cctx.threaded = 0;
#endif

#ifdef PTHREADS
  pthread_cond_init(&gcond, NULL);
  pthread_mutex_init(&gmutex, NULL);
#endif

  gettimeofday(&stime, NULL);
  for(i = 0; i < num; ++i) {
    if(nthreads == 1)
      hit(&cctx);
#ifdef PTHREADS
    else {
      if(nthreads > 0)
	waitcount(nthreads - 1);
      if(pthread_create(&tid, NULL, hit, &cctx) < 0)
	perror("pthread_create");
      else {
	pthread_mutex_lock(&gmutex);
	++thrcount;
	if(vflag)
	  printf("(+) %d threads\n", thrcount);
	pthread_mutex_unlock(&gmutex);
      }
    }
#endif
  }
#ifdef PTHREADS
  if(nthreads != 1)
    waitcount(0);
#endif
  elapsed = getelapsed(&stime);

#ifdef PTHREADS
  pthread_mutex_destroy(&gmutex);
  pthread_cond_destroy(&gcond);
#endif

  if(!qflag) {
    printf("        Total bytes read: %ld\n", bcount);
    printf("Sequential (thread) time: %g s\n", alltime);
    printf("                 Latency: %g s\n", alltime / num);
    printf("    Parallel (real) time: %g s\n", elapsed);
    printf("       Amortized latency: %g s\n", elapsed / num);
    printf("              Throughput: %g ops/s\n", num / elapsed);
  }

  return (bcount > 0) ? 0 : 1;
}
